/*
//	Note:  The use of this program shall be restricted to educational,
//	non-commercial use only.  All inquiries regarding licensing this
//	software for commercial use may be addressed to:
//	
//	Georgia Tech Research Corporation
//	Technology Licensing
//	Centennial Research Building - Rm. 275
//	400 Tenth Street, N.W.
//	Atlanta, Georgia  30332-0415
//	
//	*************************************************
//	
//	       Copyright 1996
//	Georgia Tech Research Corporation
//	    All Rights Reserved
//
//
//
*/
/************************************************************************/
/*									*/
/* Rap.c								*/
/*									*/
/*	RAP (Remote Access Protocol) is a protocol layered on top of	*/
/*	ICE that uses the Xt hooks to share client information with	*/
/*	external agents.						*/
/*									*/
/*	This contains the following public entry points:	        */
/*									*/
/*	RapInit(Widget,char *)						*/
/*									*/
/*		Widget 		- The top level shell of the app.	*/
/*		char *		- The ICE network ID's.			*/
/*									*/
/*		Initiate the ICE connection with the external agent.	*/
/*									*/
/*	RapAtom(Display)						*/
/*									*/
/*		Returns the Atom assigned to the RAP protocol name.	*/
/*									*/
/************************************************************************/
#include <stdio.h>
#include <X11/Xmd.h>
#include <X11/Xlib.h>
#include <X11/StringDefs.h>
#include <X11/Intrinsic.h>

#include <X11/ICE/ICElib.h>
#include <X11/ICE/ICEconn.h>
#include <X11/ICE/ICEproto.h>
#include <X11/ICE/ICEmsg.h>
#include "Rap.h"
#include "RapProto.h"

static int rapMajorOpcode=0;
static short CurrentSeqNum = 0;
static IceConn globalIceConn=0;
static short request_types[256];
static short event_types[256];
 

/************************************************************************/
/*									*/
/* RapIceIceErrorHandler						*/
/*									*/
/* 	Called when an unexpected ICE protocol error is encountered.	*/
/*									*/
/************************************************************************/
static
void RapIceErrorHandler(IceConn 	iceConn,
			Bool		swap,
			int		offendingMinorOpcode,
			unsigned long	offendingSequenceNum,
			int		errorClass,
			int		severity,
			IcePointer	values)
{
    printf("RapIceErrorHandler:  offendingMinorOpcode=%d,errorClass=%d,severity=%d\n",
	   offendingMinorOpcode,errorClass,severity);

    /*[[[WDW - Dying agents should be handled differently.]]]*/

} /* RapIceErrorHandler */


/************************************************************************/
/*									*/
/* RapIceIOErrorHandler							*/
/*									*/
/* 	Called when an unexpected ICE protocol error is encountered.	*/
/*									*/
/************************************************************************/
static
void RapIceIOErrorHandler(IceConn iceConn)
{
    printf("RapIceIOErrorHandler\n");

    /*[[[WDW - Dying agents should be handled differently.]]]*/

} /* RapIceIOErrorHandler */


/************************************************************************/
/*									*/
/* RapIOErrorProc							*/
/*									*/
/* 	Called if the ICE connection unexpectedly breaks.  This routine	*/
/*	is established as the IO error handler during the call to	*/
/*	IceRegisterForProtocolSetup.					*/
/*									*/
/************************************************************************/
static
void RapIOErrorProc(IceConn iceConn)
{
    printf("RapIOErrorProc:  The ICE connection broke.\n");
    /*IceCloseConnection(iceConn); */

} /* RapIOErrorProc */


/************************************************************************/
/*									*/
/* RapProcessIceMsg							*/
/*									*/
/* 	Responsible for processing the set of messages that can	        */
/*	be received from ICE.  This routine is set up as the processor	*/
/*	via a call to XtAppAddInput by RapInit, the routine that	*/
/*	establishes the ICE connection with the agent.			*/
/*									*/
/************************************************************************/
static
void RapProcessIceMsg(XtPointer clientData, 
		      int	*source, 
		      XtInputId	*id)
{
    Bool			replyReady;
    IceProcessMessagesStatus	processStatus;
    IceConn			iceConn = (IceConn)clientData;
    
    /* IceProcessMessages ends up calling RapProcessMsg
     * because it is the message handler for the RAP protocol.
     *
     * [[[WDW - I'm not quite sure about what the reply_ready
     *		stuff does, and if we should be taking advantage
     *		of it somehow.]]]
     */
    processStatus = IceProcessMessages(iceConn,
				       NULL,	/* reply_ready_ret */
				       NULL);	/* reply_ready	   */
    

    /* Check the status to see if there were any problems.
     */
    if (processStatus != IceProcessMessagesSuccess) {
#ifdef RAP_DEBUG
	fprintf(stderr,"RapProcessIceMsg:  ");
#endif
	switch(processStatus) {
	case IceProcessMessagesIOError:
	    fprintf(stderr,
		    "IceProcessMessagesIOError -- closing connection\n");
	    IceCloseConnection(iceConn);
	    XtRemoveInput(*id);
	    break;

	case IceProcessMessagesConnectionClosed:
	    fprintf(stderr,"IceProcessMessagesConnectionClosed\n");
	    break;

	default:
	    fprintf(stderr,"Unknown status %1d\n",processStatus);
	    break;
	}
    }
} /* RapProcessIceMsgProc */


/* next need to write this little utility thing ... */
static int ErrMsgSize(error_str)
char *error_str;
{
    /* for now, add macro stuff later ... */	
    return(0);

}


/************************************************************************/
/*									*/
/* RapProcessMsg							*/
/*									*/
/*	Responsible for processing RAP protocol messages.  This routine	*/
/*	was assigned as the process message proc during the call to	*/
/*	IceRegisterForProtocolSetup by RapInit.  It is called as a 	*/
/*	result of RapProcessIceMsgProc calling IceProcessMessages.	*/
/*									*/
/*	The clientData parameter is set to the widget that was passed	*/
/*	to RapInit.							*/
/*									*/
/************************************************************************/
static
void RapProcessMsg(IceConn		iceConn,
		   IcePointer 		clientData,
		   int			opcode,
		   unsigned long	length,
		   Bool			swap,
		   IceReplyWaitInfo	*replyWait,
		   Bool			*replyReadyRet)
{
    Widget		w = (Widget)clientData;
    RapMessageTypes  msg_code = opcode;   
#ifdef RAP_DEBUG
    printf("RapProcessMsg:  %d", opcode);
#endif
  
    switch(msg_code) {
	/* The RapAcknowledge opcode is used by the agent just after
	 * it accepts the ICE connection.  It expects the client to
	 * respond with a RapHello message that contains the ID of the
	 * client's top level window.  This allows the agent to do the
	 * bookkeeping necessary to handle multiple clients.
	 */
    case RapHelloRequest:
	(void) HandleRapHelloRequest(iceConn, rapMajorOpcode, w); 
	break;

	/* The RapQueryTreeRequest is used by the agent to tell 
	 * the client to send its widget hierarchy.
	 *
	 */
    case RapQueryTreeRequest:
	(void) HandleRapQueryTreeRequest(iceConn, rapMajorOpcode, w);	
	break;

    case RapFullQueryTreeRequest:
	(void) HandleRapFullQueryTreeRequest(iceConn, rapMajorOpcode, w);	
	break;

	/* agent notification of callbacks */
    case RapAddNotifyRequest:
	(void) HandleRapAddNotifyRequest(iceConn, rapMajorOpcode, w);
	break;
  	    	
	/* delete agent notification of callbacks */
    case RapRemoveNotifyRequest:
	(void) HandleRapRemoveNotifyRequest(iceConn, rapMajorOpcode);
	break;

        /* agent notification of events */
    case RapSelectEventRequest:
	(void) HandleRapSelectEventRequest(iceConn, rapMajorOpcode, w);
	break;
 
        /* agent notification of requests */
    case RapSelectRequestRequest:
	(void) HandleRapSelectRequestRequest(iceConn, rapMajorOpcode, w);
	break;
 
    case RapGetResourcesRequest:
	(void) HandleRapGetResourcesRequest(iceConn, rapMajorOpcode);
	break;

    case RapGetGCValuesRequest:
	(void) HandleRapGetGCValuesRequest(iceConn, rapMajorOpcode, w);
	break;

    case RapGetValuesRequest:
	(void) HandleRapGetValuesRequest(iceConn, rapMajorOpcode);
	break;

    case RapSetValuesRequest:
	(void) HandleRapSetValuesRequest(iceConn, rapMajorOpcode);
	break;

    case RapLocateObjectRequest:
	(void) HandleRapLocateObjectRequest(iceConn, rapMajorOpcode);
	break;

    case RapObjectToWindowRequest:
	(void) HandleRapObjectToWindowRequest(iceConn, rapMajorOpcode);
	break;

    case RapWindowToObjectRequest:
	(void) HandleRapWindowToObjectRequest(iceConn, rapMajorOpcode);
	break;

    default:
	fprintf(stderr, "Yikes!  Unkown opcode: %d\n",opcode);
	break;
    } /* switch(opcode) */
    
} /* RapProcessMsg */



/************************************************************************/
/*									*/
/* RapInit								*/
/*									*/
/*	Initiate the ICE connection with the agent.  This routine is	*/
/*	called by XteaCheckMessages as a result of the client 		*/
/*	registering the RAP protocol with a call to XteaSetupClient.	*/
/*									*/
/*	In addition to initiating the connection, this routine also 	*/
/*	sets up the callbacks that work with the external agent hooks	*/
/*	built into Xt.  The client data used for the hooks is the ICE 	*/
/*	connection ID.	This allows clients to keep track of multiple 	*/
/*	agents.								*/
/*									*/
/************************************************************************/
void RapInit(Widget	w,
	     char *	networkIds)
{
    IcePoVersionRec		protocolVersions[1];
    IceConn			iceConn;
    char			errorStr[256];
    char			*connectStr;
    IceProtocolSetupStatus	protoSetupStatus;
    int				majorVersion,minorVersion;
    char			*vendorStr,*releaseStr;
    Widget			hooksWidget;
    
#ifdef RAP_DEBUG
    printf("RapInit: ");
#endif

    /* If this is the first time this routine has been called,
     * do the necessary work to let ICE know about RAP and that
     * this client will be able to do a protocol setup for RAP.
     */
    if (!rapMajorOpcode) {
	protocolVersions[0].major_version    = RapMajorVersion;
	protocolVersions[0].minor_version    = RapMinorVersion;
	protocolVersions[0].process_msg_proc = RapProcessMsg;
    
	rapMajorOpcode = IceRegisterForProtocolSetup(RapName,
						     RapVendor,
						     RapRelease,
						     1,protocolVersions,
						     0,NULL,NULL,
						     RapIOErrorProc);

	if (rapMajorOpcode == -1) {
	    fprintf(stderr,"IceRegisterForProtocolSetup failed.\n");
	    rapMajorOpcode = 0;
	    return;
	}
    }
    
#ifdef RAP_DEBUG	
    printf("rapMajorOpcode=%1d\n",rapMajorOpcode);
#endif

    /* Using the ICE network IDs in the ClientMessage sent by the agent
     * (Xtea gets these for us and passes them to RapInit), contact the 
     * agent.
     */
    iceConn = IceOpenConnection(networkIds,
				NULL,
				False,
				rapMajorOpcode,
				255,
				errorStr);

    globalIceConn = iceConn;
    
    if (!iceConn) {
	fprintf(stderr,"IceOpenConnection failed due to %s\n",errorStr);
	return;
    } else {
	connectStr = IceConnectionString(iceConn);
#ifdef RAP_DEBUG
	printf("Connection Opened: ");
	printf("IceConnectionNumber = %1d",IceConnectionNumber(iceConn));
	printf(", IceConnectionString = %s\n",connectStr);
#endif
	free(connectStr);
    }

    /* Set up the protocol and use the widget as the the client data.
     */
    protoSetupStatus = IceProtocolSetup(iceConn,
					rapMajorOpcode,
					w, 		/* client data */
					False,
					&majorVersion,
					&minorVersion,
					&vendorStr,
					&releaseStr,
					255,
					errorStr);

    if (protoSetupStatus != IceProtocolSetupSuccess) {
	fprintf(stderr,"IceProtocolSetup failed due to ");
	
	switch(protoSetupStatus) {
	case IceProtocolSetupFailure:
	    fprintf(stderr,"IceProtocolSetupFailure\n");
	    break;

	case IceProtocolSetupIOError:
	    fprintf(stderr,"IceProtocolSetupIOError\n");
	    break;

	case IceProtocolAlreadyActive:
	    fprintf(stderr,"IceProtocolAlreadyActive\n");
	    break;

	default:
	    fprintf(stderr,":  %s",errorStr);
	}
    }
    
#ifdef RAP_DEBUG
    printf("Protocol Setup:  ");
    printf("Version: %1d.%1d",majorVersion,minorVersion);
    printf(", Vendor:  %s",vendorStr);
    printf(", Release: %s\n",releaseStr);
#endif

    /* need to do some initialization somewhere .. might as well
     be here ...  for now */
    {
        int x;
        for(x=0; x<256; x++)
                event_types[x] = 0;
 
     }

    /* Handle ICE messages using XtAppAddInput.  Use the iceConn 
     * connection number as a way to determine when ICE messages 
     * are being sent on the connection.  In addition, use the
     * iceConn ID as the client data so the RapProcessIceMsg can
     * handle multiple agents.
     */
    XtAppAddInput(XtWidgetToApplicationContext(w),
		  IceConnectionNumber(iceConn),
		  (XtPointer)XtInputReadMask,
		  RapProcessIceMsg,
		  (XtPointer)iceConn);    

    /* Set up the callbacks for the Xt hooks.
     * [[[WDW - Need a way to remove these callbacks if the
     *		agent decides to leave.]]] */
     
    hooksWidget = XtHooksOfDisplay(XtDisplay(w));
    XtAddCallback(hooksWidget,
		  XtNcreateHook,
		  (XtCallbackProc)RapCreateHookProc,
		  iceConn);
    XtAddCallback(hooksWidget,
		  XtNchangeHook,
		  (XtCallbackProc)RapChangeHookProc,
		  iceConn);
    XtAddCallback(hooksWidget,
		  XtNconfigureHook,
		  (XtCallbackProc)RapConfigureHookProc,
		  iceConn);
    XtAddCallback(hooksWidget,
		  XtNgeometryHook,
		  (XtCallbackProc)RapGeometryHookProc,
		  iceConn);
    XtAddCallback(hooksWidget,
		  XtNdestroyHook,
		  (XtCallbackProc)RapDestroyHookProc,
		  iceConn);
    
    /* Set up the handler to handle sudden breaks in the 
     * ICE connection.   [[[WDW - This really isn't such a
     * good thing to do.]]]]
     */
    IceSetErrorHandler(RapIceErrorHandler);
    IceSetIOErrorHandler(RapIceIOErrorHandler);
    
} /* RapInit */

int
RapGetMajorOpcode()
{
	return(rapMajorOpcode);
}

IceConn
RapGetIceConnection()
{
        return(globalIceConn);
}
 
void
RapSetRequestTypes(int request_num, int value)
{
        request_types[request_num] = value;
 
}
 
short *
RapGetRequestTypes()
{
        return(request_types);
 
}
 
void
RapSetEventTypes(int event_num, int value)
{
        event_types[event_num] = value;
 
}
 
short *
RapGetEventTypes()
{
        return(event_types);
 
}
 


/************************************************************************/
/*									*/
/* RapAtom								*/
/*									*/
/* 	Returns the Atom assigned to the rap protocol name.		*/
/*									*/
/************************************************************************/
Atom RapAtom(Display *theDisplay)
{
    return(XInternAtom(theDisplay,
		       RapName,
		       False));
  
} /* RapAtom */
