/*
//	Note:  The use of this program shall be restricted to educational,
//	non-commercial use only.  All inquiries regarding licensing this
//	software for commercial use may be addressed to:
//	
//	Georgia Tech Research Corporation
//	Technology Licensing
//	Centennial Research Building - Rm. 275
//	400 Tenth Street, N.W.
//	Atlanta, Georgia  30332-0415
//	
//	*************************************************
//	
//	       Copyright 1996
//	Georgia Tech Research Corporation
//	    All Rights Reserved
//
//
//
*/
/************************************************************************/
/*									*/
/* Xtea.c								*/
/*									*/
/*	Xt external agent enabling code.  This contains the following	*/
/*	public entry points:						*/
/*									*/
/*	XteaSetupClient(Widget)						*/
/*									*/
/*		Widget	- The top level shell of the client.		*/
/*									*/
/*		Called by the client to install the list of known 	*/
/*		protocols and set up the ClientMessage event handler 	*/
/*		on the top level shell.					*/
/*									*/
/************************************************************************/
#include <X11/IntrinsicP.h>		/* for usual Xt function decls 	*/
#include <X11/Xos.h>			/* for strdup declaration 	*/
#include <X11/Xatom.h>			/* for XA_STRING 		*/
#include "Xtea.h"
/***** DEBUGGING ***/
#include <X11/StringDefs.h>

/************************************************************************/
/*									*/
/* Table to hold list of known protocols and their initialization	*/
/* routines.  This is a linked list -- lookup is through a sequential	*/
/* search.								*/
/*									*/
/************************************************************************/
typedef struct _XteaProtoTableEntry 
{
    Atom 			protoAtom;
    XteaProtoInitializer	initializerProc;
    struct _XteaProtoTableEntry	*next;
} XteaProtoTableEntry, *XteaProtoTable;

/************************************************************************/
/*									*/
/* Local global variables.  						*/
/*									*/
/* knownProtocols is a table of XteaProtoTable with each entry 		*/
/* containing the atomized name of the protocol as well	as the 		*/
/* protocol's initialization procedure.					*/
/*									*/
/* [[[WDW - Is static information bad for multithreaded apps?]]]	*/
/*									*/
/************************************************************************/
static XteaProtoTable  	knownProtocols=NULL;	

/************************************************************************/
/*									*/
/* XteaCheckMessages							*/
/*									*/
/*	XteaCheckMessages handles the incoming ClientMessage events 	*/
/*	from external agents.  This handler is added by a call to	*/
/*	<a href=#XteaSetupClient>XteaSetupClient</a>.			*/
/*									*/
/*	The client message event looks like the following:	        */
/*									*/
/*	cme->type		== ClientMessage			*/
/*	cme->message_type	== Atom ("ExternalAgent")		*/
/*	cme->format		== EXTAGENT_FORMAT (32)			*/
/*	cme->data.l[0]		== protoAtom				*/
/*	cme->data.l[1]		== Atom ("IceNetworkIds")		*/
/*	cme->data.l[2]		== agentWindow				*/
/*									*/
/*	The protoAtom entry is the name of the protocol the agent 	*/
/*	wishes to speak.  The networkIdsAtom is the list of ICE		*/
/*	network ids the client must use to send an ICE protocol setup	*/
/* 	message to the aget.  The agentWindow is the id of the		*/
/*	agent's window that contains the networkIds property.		*/
/*									*/
/************************************************************************/
static
void XteaCheckMessages(Widget		w, 
		       XtPointer 	clientData, 
		       XEvent 		*event, 
		       Boolean 		*cont)
{
    XteaProtoTableEntry	*current = knownProtocols;
    Display 		*theDisplay = XtDisplay(w);
    XClientMessageEvent *cme = (XClientMessageEvent *) event;
    Atom 		protoAtom;
    Atom		networkIdsAtom;
    Window		agentWindow;
    
    /* Make sure this is a ClientMessage event before doing anything.
     */
    if (event->type != ClientMessage) 
	return;

    /* If the message type is not "ExternalAgent,"
     * then we know not to process this event any further.
     */
    if ((cme->message_type != XInternAtom(theDisplay, EXTAGENT_NAME, FALSE)) ||
	(cme->format != EXTAGENT_FORMAT)) 
        return;

    /* Get the information supplied by the ClientMessage event.
     */
    protoAtom      = cme->data.l[0];
    networkIdsAtom = cme->data.l[1];
    agentWindow    = cme->data.l[2];

    /* Look in the table of known protocols for the protocol the
     * agent wishes to speak.  If ths protocol is in the list of
     * known protocols, call the protocol initialization routine
     * with the agents list of ICE network ids.
     */
    while(current)
    {
	if (current->protoAtom == protoAtom)
	{
	    XTextProperty textProp;

	    if(XGetTextProperty(theDisplay,
				agentWindow,
				&textProp,
				networkIdsAtom) != 0)
	    {
		(*(current->initializerProc))(w,(char *)textProp.value);
		XFree(textProp.value);
	    }
	    
	    return;
	}
	current = current->next;
    }

    return;

} /* XteaCheckMessages */

/************************************************************************/
/*									*/
/* XteaAddProtocol							*/
/*									*/
/*	topLevel	- The top level shell of the app.		*/
/*	protoAtom	- The atomized name of the protocol.		*/
/*	initializerProc	- The procedure for initializing the		*/
/*			  protocol.					*/
/*									*/
/*	Add the protocol, designated by the unique name protoAtom, and	*/
/*	its initialization routine to the list of known protocols.	*/
/*	The call to this routine is usually made by a convenience 	*/
/*	routine in the protocol library itself.				*/
/*									*/
/************************************************************************/
static
Boolean 
XteaAddProtocol(Atom 		     	protoAtom, 
		XteaProtoInitializer 	initializerProc)
{
    XteaProtoTableEntry *newEntry, *current, *prev;

    /* The protocol to be added must be have a name and an
     * initialization procedure.
     */
    if (!protoAtom || (initializerProc == NULL))
	return FALSE;

    /* Create and initialize a table node for the new initializer.
     */
    if ((newEntry = XtNew(XteaProtoTableEntry)) == NULL)
	return FALSE;

    newEntry->protoAtom 	= protoAtom;
    newEntry->initializerProc 	= initializerProc;
    newEntry->next 		= NULL;

    /* Install it in the table.
     */
    if (knownProtocols == NULL)
    {
	knownProtocols = newEntry;
    }
    else 
    {
	/* Search the list for a matching node and override it, 
	 * or append at the end of no matching node is found.
	 */
	current = knownProtocols;
	while (current != NULL) 
	{
	    if (current->protoAtom == protoAtom) 
	    {
		current->initializerProc = initializerProc;
		XtFree((char *)newEntry);
		break;
	    } 
	    else 
	    {
		prev = current;
		current = current->next;
	    }
	}

	/* We're at the end of the list and haven't found a match. 
	 */
	if (current == NULL)
	    prev->next = newEntry;
    }

    return TRUE;

} /* XteaAddProtocol */

/************************************************************************/
/*									*/
/* XteaRemoveProtocol							*/
/*									*/
/*	protoAtom	- The atomized name of the protocol.		*/
/*									*/
/*	Remove the protocol from the list of known protocols.		*/
/*									*/
/************************************************************************/
static
void XteaRemoveProtocol(Atom protoAtom)
{
    XteaProtoTableEntry *current = knownProtocols;
    XteaProtoTableEntry *prev = NULL;

    while (current) 
    {
	if (current->protoAtom == protoAtom) 
	{
	    if (prev) 
		prev->next = current->next;
	    else 
	        knownProtocols = current->next;
	    XtFree((char *)current);
	    break;
	}

	prev = current;
	current = current->next;
    }
    return;

} /* XteaRemoveProtocol */

/************************************************************************/
/*									*/
/* XteaSetupClient							*/
/*									*/
/*	topLevel	- The client's top level shell.			*/
/*									*/
/*	Registers the ICE protocols the client speaks and sets up a	*/
/*	ClientMessage event handler.					*/
/*									*/
/*	[[[WDW - For implementations that use static libraries, this	*/
/*	routine needs to be modified for each new protocol.  For 	*/
/*	implementations that use shared libraries, this routine needs	*/
/*	to be modified to use some version of the dynamic loading code	*/
/*	implemented by the loadable working group.]]]			*/
/*									*/
/************************************************************************/
void XteaSetupClient(Widget topLevel)
{
    /* Add the various known protocols 
     */	
/* #ifdef RAP */
    extern void RapInit(Widget,char *);
    extern Atom RapAtom(Display *);
    
    XteaAddProtocol(RapAtom(XtDisplay(topLevel)),RapInit);
/* #endif */

    /* Set up the ClientMessage handler.
     */
    XtAddEventHandler(topLevel,		/* The widget to receive the events */
                      (EventMask) 0,	/* The EventMask		    */
		      TRUE,		/* Call on nonmaskable events	    */
	              XteaCheckMessages,/* The handler.			    */
	              NULL);		/* Client data.			    */

    printf("XteaSetupClient: Added event handler!\n");

    printf("NUM SHELLS = %s\n", XtNnumShells);

} /* XteaSetupClient */

