/*
//	Note:  The use of this program shall be restricted to educational,
//	non-commercial use only.  All inquiries regarding licensing this
//	software for commercial use may be addressed to:
//	
//	Georgia Tech Research Corporation
//	Technology Licensing
//	Centennial Research Building - Rm. 275
//	400 Tenth Street, N.W.
//	Atlanta, Georgia  30332-0415
//	
//	*************************************************
//	
//	       Copyright 1996
//	Georgia Tech Research Corporation
//	    All Rights Reserved
//
//
//
 * Handlers for incoming messages to the client.
 *
 * Right now all of these are pretty simple:  they read in the incoming message
 * (which, for all our current cases, don't really contain much), decide whether
 * or not the message makes sense, and call routines from client-out.c to
 * send the appropriate replies or errors.
 */

#include <stdio.h>
#include <X11/Xmd.h>
#include <X11/IntrinsicP.h>	/* to get inside widget structure opaque types */
#include <X11/ShellP.h>		/* for ApplicationShellWidgetClass */
#include <X11/StringDefs.h>
#include <X11/Xproto.h>		/* needed for select events and requests */

#include <X11/ICE/ICElib.h>
#include <X11/ICE/ICEmsg.h>
#include <X11/ICE/ICEproto.h>
#include <X11/ICE/ICEconn.h>

#include "RapProto.h"
#include "RapLibint.h"
#include "Rap.h"


int
HandleRapHelloRequest(IceConn iceConn, int majorOpcode, Widget w)
{
    RapMessageHeader *request;

#ifdef RAP_DEBUG
    fprintf(stderr, "RECEIVED HELLO REQUEST\n"); fflush(stderr);
#endif
    IceReadSimpleMessage(iceConn,
			 RapMessageHeader,
			 request);

    return RAPHelloReply(iceConn, majorOpcode, w);

}

int 
HandleRapFullQueryTreeRequest(IceConn iceConn, int majorOpcode, Widget w)
{

    RapMessageHeader *request;
 
#ifdef RAP_DEBUG
    fprintf(stderr, "RECEIVED FULL QUERY TREE REQUEST\n"); fflush(stderr);
#endif
    IceReadSimpleMessage(iceConn,
                         RapMessageHeader,
                         request);
 
    return RAPFullQueryTreeReply(iceConn, majorOpcode, w);
}


int
HandleRapQueryTreeRequest(IceConn iceConn, int majorOpcode, Widget w)
{
    RapMessageHeader *request;

#ifdef RAP_DEBUG
    fprintf(stderr, "RECEIVED QUERY TREE REQUEST\n"); fflush(stderr);
#endif
    IceReadSimpleMessage(iceConn,
			 RapMessageHeader,
			 request);

    return RAPQueryTreeReply(iceConn, majorOpcode, w);
}

int
HandleRapGetValuesRequest(IceConn iceConn, int majorOpcode)
{
    RapMessageHeader *request;
    char *pdata;
    CARD32 widget;
    CARD32 numResources;
    char **resources;
    register int i;
    int ret;

#ifdef RAP_DEBUG
    fprintf(stderr, "RECEIVED GET VALUES REQUEST.\n"); fflush(stderr);
#endif
    IceReadCompleteMessage(iceConn,
			   sizeof(RapMessageHeader),
			   RapMessageHeader,
			   request,
			   pdata);

    EXTRACT_CARD32(pdata, FALSE, widget);
#ifdef RAP_DEBUG
    printf("WIDGET is %lx\n", widget);
#endif

    EXTRACT_CARD32(pdata, FALSE, numResources);
#ifdef RAP_DEBUG
    printf("NUM STRINGS is %d\n", numResources);
#endif

    resources = (char **) malloc(sizeof(char *) * numResources);
    for (i=0 ; i<numResources ; i++) {
	char *str;
	EXTRACT_STRING(pdata, FALSE, str);
	resources[i] = str;
#ifdef RAP_DEBUG
	printf("STRING %d is '%s'\n", i, resources[i]);
#endif
    }

    ret = RAPGetValuesReply(iceConn, majorOpcode, widget,
			    numResources, resources);

    IceDisposeCompleteMessage(iceConn, pdata);

    return ret;
}

int
HandleRapGetResourcesRequest(IceConn iceConn, int majorOpcode)
{
    RapMessageHeader *request;
    char *pdata;
    CARD32 numWidgets;
    WidgetList widgets;
    int ret;

#ifdef RAP_DEBUG
    fprintf(stderr, "RECEIVED GET RESOURCES REQUEST\n"); fflush(stderr);
#endif

    IceReadCompleteMessage(iceConn,
			   sizeof(RapMessageHeader),
			   RapMessageHeader,
			   request,
			   pdata);

    EXTRACT_CARD32(pdata, FALSE, numWidgets);
#ifdef RAP_DEBUG
    printf("NUM WIDGETS: %d\n", numWidgets);
#endif

    widgets = (WidgetList) pdata;

    ret = RAPGetResourcesReply(iceConn, majorOpcode, numWidgets, widgets);
    
    IceDisposeCompleteMessage(iceConn, pdata);
    
    return ret;
}

int
HandleRapLocateObjectRequest(IceConn iceConn, int majorOpcode)
{
    RapMessageHeader *request;
    char *pdata;
    CARD32 numWidgets;
    WidgetList widgets;
    int ret;

    IceReadCompleteMessage(iceConn,
			   sizeof(RapMessageHeader),
			   RapMessageHeader,
			   request,
			   pdata);

    EXTRACT_CARD32(pdata, FALSE, numWidgets);
    widgets = (WidgetList) pdata;

    ret = RAPLocateObjectReply(iceConn, majorOpcode, numWidgets, widgets);
    IceDisposeCompleteMessage(iceConn, pdata);
    
    return ret;
}

int
HandleRapGetGCValuesRequest(IceConn iceConn, int majorOpcode, Widget w)
{
    RapMessageHeader *request;
    char *pdata;
    CARD32 numGCs;
    GContext *GCs;
    int ret;
 
#ifdef RAP_DEBUG
    fprintf(stderr, "RECEIVED GET GC VALUES REQUEST\n"); fflush(stderr);
#endif
 
    IceReadCompleteMessage(iceConn,
                           sizeof(RapMessageHeader),
                           RapMessageHeader,
                           request,
                           pdata);
 
    EXTRACT_CARD32(pdata, FALSE, numGCs);

#ifdef RAP_DEBUG
    printf("NUM GCs: %d, first = %lx\n", numGCs, *((GContext *) pdata));
#endif
 
    GCs = (GContext *) pdata;
 
    ret = RAPGetGCValuesReply(iceConn, majorOpcode, numGCs, GCs, w);
 
    IceDisposeCompleteMessage(iceConn, pdata);
 
    return ret;
}
 
 



/************************************************************************/
/*                                                                      */
/* HandleRapAddNotifyRequest                                            */
/*                                                                      */
/*      Client reads in request for installation of various hook        */
/*      callbacks. Called from inside RapProcessMsg - when message      */
/*      of type RapAddNotifyRequest comes in                            */
/*                                                                      */
/************************************************************************/

int HandleRapAddNotifyRequest(IceConn iceConn, int majorOpcode, Widget w)
{
    RapMessageHeader  *notify_request = NULL;
    short             nrequests, *reqs =NULL;
    int               x, request_error = False;
    HookCallbacks     req_code;
    Widget            hooksWidget = XtHooksOfDisplay(XtDisplay(w));

#ifdef RAP_DEBUG
    printf("RECEIVED ADD NOTIFY REQUEST\n");
#endif

    IceReadMessageHeader(iceConn, sizeof(RapMessageHeader),
			 RapMessageHeader, notify_request);
 
    /* next read in number of requests */
    IceReadData16(iceConn, False, sizeof(short), &nrequests);
 
#ifdef RAP_DEBUG
    printf("got %d reqs\n", nrequests);
#endif
 
    for (x=0; x< (nrequests); x++) {
	if (reqs)
	    free(reqs);     
 
	reqs = (short *) malloc(sizeof(short) );
 
	IceReadData16(iceConn, False, 2, reqs);
 
	req_code = *reqs;
 
	switch (req_code) {
	case RapCreateHook:
	    XtAddCallback(hooksWidget, XtNcreateHook,
			  RapCreateHookProc, iceConn);
	    break;
 
	case RapChangeHook:
	    XtAddCallback(hooksWidget, XtNchangeHook,
			  RapChangeHookProc, iceConn);
	    break;
 
	case RapConfigureHook:
	    XtAddCallback(hooksWidget, XtNconfigureHook,
			  RapConfigureHookProc, iceConn);
	    break;
 
	case RapGeometryHook:
	    XtAddCallback(hooksWidget, XtNgeometryHook,
			  RapGeometryHookProc, iceConn);
	    break;

	case  RapDestroyHook:
	    XtAddCallback(hooksWidget, XtNdestroyHook,
			  RapDestroyHookProc, iceConn);
	    break;
 
	default:
	    /* should probably send  an error here .... not sure exactly how ... */
	    request_error = True;
	    break;
 
	}  /* end of local switch  */
    }    /* end of for loop */
 
    if (request_error) { /* send an error msg of sorts ... */
	RAPAddNotifyError(iceConn,majorOpcode, 
			  notify_request->sequenceNumber);
	return 0;
    } else {
	RAPSendAck(iceConn, majorOpcode,
		   notify_request->sequenceNumber);
	return 1;
    }
}


int
HandleRapRemoveNotifyRequest(IceConn iceConn, int majorOpcode)
{
#ifdef RAP_DEBUG
    printf("RECEIVED REMOVE NOTIFY REQUEST\n");
#endif
    return 0;
}


int
HandleRapSelectEventRequest(IceConn iceConn, int majorOpcode, Widget w)
{
    RapMessageHeader  *event_request = NULL;
    short              nrequests, req_event, *event_types;
    int                x;
    Display           *dpy = XtDisplay(w);
 
    event_types = (short *) RapGetEventTypes();
    
    for (x=0; x<256; x++) {
        if (event_types[x]) {
            RapSetEventTypes(x, 0);
            XESetWireToEvent(dpy,x,NULL);
        }
    }
 
    IceReadMessageHeader(iceConn, sizeof(RapMessageHeader),
			 RapMessageHeader, event_request);
 
    /* next read in number of requests */
    IceReadData16(iceConn, False, sizeof(short), &nrequests);
 
    for (x=0; x<nrequests; x++) {
        IceReadData16(iceConn, False, sizeof(short), &req_event);
        XESetWireToEvent(dpy,req_event,RapEventProc);
        RapSetEventTypes(req_event, 1);
 
    }
 
    RAPSendAck(iceConn, majorOpcode, event_request->sequenceNumber);
    return 1;
}
 

int
HandleRapSelectRequestRequest(IceConn iceConn, int majorOpcode, Widget w)
{
    RapMessageHeader  *request = NULL;
    CARD16             nrequests, req_type;
    int                x;
    Display    	      *dpy = XtDisplay(w);
    XExtCodes         *xcodes = XAddExtension(dpy);
    
    IceReadMessageHeader(iceConn, sizeof(RapMessageHeader),
			 RapMessageHeader, request);
 
    /* next read in number of requests */
    IceReadData16(iceConn, False, sizeof(short), &nrequests);
 
    /* now zero out request_type array  - from any previous requests */
    for (x=0; x<256; x++)
        RapSetRequestTypes(x, 0);
 
    for (x=0; x< (nrequests); x++) {
        IceReadData16(iceConn, False, sizeof(short), &req_type);
        RapSetRequestTypes(req_type, 1);
 
        /* check for adding gc text ... */
        if (X_CreateGC == req_type)
	    XESetCreateGC(dpy, xcodes->extension, RapGCFlushProc);
    }
 
    /* now install the flush proc */
    /*    xcodes = XAddExtension(dpy); */
    XESetBeforeFlush(dpy, xcodes->extension, RapFlushProc);
 
    RAPSendAck(iceConn, majorOpcode, request->sequenceNumber);
 
    return 1;
}


int
HandleRapSetValuesRequest(IceConn iceConn, int majorOpcode)
{
    return 0;
}

int
HandleRapWindowToObjectRequest(IceConn iceConn, int majorOpcode)
{
    return 0;
}

int
HandleRapObjectToWindowRequest(IceConn iceConn, int majorOpcode)
{
    return 0;
}

