//
// Alva loadable module.
//
//
// 
// Note:  The use of this program shall be restricted to educational,
// non-commercial use only.  All inquiries regarding licensing this
// software for commercial use may be addressed to:
// 
// Georgia Tech Research Corporation
// Technology Licensing
// Centennial Research Building - Rm. 275
// 400 Tenth Street, N.W.
// Atlanta, Georgia  30332-0415
// 
//        Copyright 1996
// Georgia Tech Research Corporation
//     All Rights Reserved
// 


#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>
#include <stropts.h>
#include <netinet/in.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>

#include "Alva.h"
#include "Loadable.h"
#include "Action.h"
#include "Error.h"
#include "Util.h"
#include "Config.h"
#include "Mercator.h"

extern Mercator *mercator;

LOADABLE_CLASS_DEFN(Alva);


#ifndef MIN
#define MIN(a,b)	((a)<(b) ? (a) : (b))
#endif /* MIN */

const int KEY1_PRESS 	= 0;  	// PROG
const int KEY1_RELEASE 	= 128;  // PROG
const int KEY2_PRESS 	= 1;  	// HOME
const int KEY2_RELEASE 	= 129;  // HOME
const int KEY3_PRESS 	= 2; 	// CURSOR
const int KEY3_RELEASE 	= 130;  // CURSOR
const int KEY4_PRESS 	= 3;  	// UP
const int KEY4_RELEASE 	= 131;  // UP
const int KEY5_PRESS 	= 4;  	// LEFT
const int KEY5_RELEASE 	= 132;  // LEFT
const int KEY6_PRESS 	= 5;  	// RIGHT
const int KEY6_RELEASE 	= 133;  // RIGHT
const int KEY7_PRESS 	= 6;  	// DOWN
const int KEY7_RELEASE 	= 134;  // DOWN
const int KEY8_PRESS    = 7;    // right CURSOR on the 3/80 (This is a guess!)
const int KEY9_PRESS    = 8;    // right HOME on the 3/80 (This is a guess!)
const int KEY10_PRESS   = 9;    // right PROG on the 3/80 (This is a guess!)
const int KEY8_RELEASE  = 135;  // right CURSOR on the 3/80
const int KEY9_RELEASE  = 136;  // right HOME on the 3/80
const int KEY10_RELEASE = 137;  // right PROG on the 3/80
const int TOUCH_CURSOR 	= 114;  // key group identifier


///////////////////////////////////////////////////////////////////////////
//
// This is the capabilities struct which defines the Alva 3-20
//
static Braille::Capabilities alva_3_20(
    20,			// display cells
    3,			// status cells
    1,			// highlight supported?
    1,			// has cursor keys?
    1,			// has prog key?
    1,			// has home key?
    0			// other keys
);

///////////////////////////////////////////////////////////////////////////
//
// This is the capabilities struct which defines the Alva 3-80
//
static Braille::Capabilities alva_3_80(
    80,			// display cells
    5,			// status cells
    1,			// highlight supported?
    1,			// has cursor keys?
    1,			// has prog key?
    1,			// has home key?
    0			// other keys
);


///////////////////////////////////////////////////////////////////////////

Alva::Alva(const char *host, short port) :
    Braille("Alva"),
    text(NULL),
    column(0),
    row(0),
    numRows(0),
    jumpScrollMode(0),
    fd(-1),
    file(NULL)
{
    FieldData *configEntry;
    
    if ((configEntry = Config::GetDefaultByName("alvaModel")) == NULL) {
	Error(EL_WARNING, "Alva::Alva",
	      "No alvaModel specified in config file; defaulting to Alva 3/20.",
	      NULL);
	capabilities = alva_3_20;
    } else if (strcmp(configEntry->String(), "Alva 3/20") == 0) {
	Error(EL_DEBUG, "Alva::Alva", "alvaModel is Alva 3/20", NULL);
	capabilities = alva_3_20;
    } else if (strcmp(configEntry->String(), "Alva 3/80") == 0) {
	Error(EL_DEBUG, "Alva::Alva", "alvaModel is Alva 3/80", NULL);
	capabilities = alva_3_80;
    } else {
	Error(EL_DEBUG, "Alva::Alva", "alvaModel is unknown: ",
	      configEntry->String(),
	      "; expected either Alva 3/20 or Alva 3/80.  ",
	      "Defaulting to Alva 3/20.", NULL);
	capabilities = alva_3_20;
    }

    if (Alva::StartServer() == -1) {
	Error(EL_ERROR, "Alva::Alva",
	      "Could not start server, shutting down...", NULL);
	MarkForCleanup();
    } else {
	if (Alva::ConnectServer(host, port) == -1) {
	    Error(EL_ERROR, "Alva::Alva",
		  "Could not connect to server, shutting down...", NULL);
	    MarkForCleanup();
	}
    }
}

Alva::~Alva()
{
    Error(EL_DEBUG, "Alva::~Alva", "Destroying...", NULL);
    
    for (register int i=0 ; i<numRows ; i++)
	delete [] text[i];

    delete [] text;

    Shutdown(fd);
}

int
Alva::ConnectServer(const char *host, short port)
{
    FieldData *maxTriesData = Config::GetDefaultByName("alvaTimeout");
    int maxTries = 5;
    int i=0;
    
    if (maxTriesData && (maxTriesData->Tag() == INTEGER))
	maxTries = maxTriesData->Integer();
    
    Error(EL_DEBUG, "Alva::ConnectServer",
	  "Connecting to Alva server!", NULL);

    if (host == NULL)
	if ((host = getenv( "ALVAHOST" )) == NULL)
	    host = "localhost";
    
    while(i<maxTries) {
	if ((fd = OpenTCPSocket(host, port)) < 0) {
	    i++;
	    sleep(1);
	    continue;
	}
	else
	    break;
    }

    if (fd == -1) {
	Error(EL_ERROR, "Alva::ConnectServer", "server connection",
	      " timed out", NULL);
	MarkForCleanup();
	return -1;
    } else {
	Error(EL_DEBUG, "Alva::ConnectServer",
	      "connection to server successful", NULL);
	AddFDInterest(fd);
	SetIOMask(fd, ARead);
	if(!(file = fdopen(fd, "w"))) {
	    Error(EL_ERROR, "Alva::Alva", "fdopen ", Perror(), NULL);
	}
	return 0;
    }
}

//
// Return -1 if server was not started, because both the ALVAHOST environment
// variable and the brailleServer string list in the config file are not set.
//
// Return 0 otherwise. (I'm also assuming that if ALVAHOST environment
// variable is only set if the user wants to use a server on a remote machine.
// So essentially setting ALVAHOST to anything will prevent a server from
// being started on the local machine.
//
int
Alva::StartServer()
{
    const char *host;
    
    // first check if an env. var. is set. I'm assuming that ALVAHOST
    // is only set if the user wants to use a remote server
    if((host = getenv("ALVAHOST")) == NULL) {
	if(mercator->ExecList("alvaServer")) {
	    Error(EL_DEBUG, "Alva::StartServer",
		  "alvaServer not found.", NULL);
	    return -1;
	}
	return 0;
    }
    else
	Error(EL_DEBUG, "Alva::StartServer", "Server has been started on ",
	      host, NULL);
    return 0; // assuming the server has been started on another host
}


void
Alva::SetText(const char *t[], int nr)
{
    register int i;
    
    for (i=0 ; i<numRows ; i++)
	delete [] text[i];

    delete [] text;

    numRows = nr;

    text = new char*[numRows];

    for (i=0 ; i<numRows ; i++)
	text[i] = Strdup(t[i]);
    
    column = row = 0;

    DisplayBraille(text[0]);
}

void
Alva::Shutdown(int _fd)
{
    if (_fd != fd) {
	Error(EL_ERROR, "Alva::Shutdown",
	      "Trying to shutdown a descriptor we don't own.", NULL);
	return;
    }

    if (fd == -1)
	return;
    
    RemoveFDInterest(fd);

    if (file && (fclose(file) != 0))
	Error(EL_ERROR, "Alva::Shutdown",
	      "fclosing file pointer failed: ", Perror(), NULL);
    
    if (close(fd) < 0)
	Error(EL_ERROR, "Alva::Shutdown",
	      "Close of Alva server connection failed: ", Perror(), NULL);

    fd = -1;
    
    MarkForCleanup();
}

//
// The following stuff  is based on code provided by Gary.   See copyright
// notice above.
//
int
Alva::HandleActivity(int _fd, short type)
{
    unsigned char key[2];
    int bytes;

    if (fd != _fd)
	Error(EL_ERROR, "Alva::HandleActivity",
	      "Descriptor passed to us is not ours?!", NULL);

    if (type & AWrite) {
	Error(EL_ERROR, "Alva::HandleActivity",
	      "Unexpected writable condition on server connection, "
	      "shutting down...", NULL);
	
	Shutdown(_fd);
	MarkForCleanup();
	return -1;
    }

    if (type & AExcept) {
	Error(EL_ERROR, "Alva::HandleActivity",
	      "Exceptional condition on server connection, ",
	      "shutting down...", NULL);
	
	Shutdown(_fd);
	MarkForCleanup();
	return -1;
    }

    //
    // Check to see how many bytes are available (to ensure that we
    // don't hang...)  We should probably just set nonblocking I/O on
    // this descriptor.
    //
    if (ioctl(_fd, I_NREAD, &bytes) == -1) {
	Error(EL_ERROR, "Alva::HandleActivity",
	      "ioctl failed: ", Perror(), NULL);
	return -1;
    }

    if (bytes < 2) {
	Error(EL_WARNING, "Alva::HandleActivity", "Not enough bytes to read.",
	      NULL);
	return -1;
    }

    if (read(_fd, key, 2) != 2) {
	Error(EL_ERROR, "Alva::HandleActivity",
	      "Read failed: ", Perror(), ", shutting down...", NULL);

	Shutdown(_fd);
	MarkForCleanup();
	return -1;
    }
    
    // Ignore all press codes for now 
    if (key[1] < 128)
	return 0;
    
    // Perform edit cursor routing 
    if (key[0] == TOUCH_CURSOR)
	return 0;

    // Front panel or status cell touch cursor 
    if (key[0] == 113) { 
 
	switch (key[1]) {
	case KEY1_RELEASE:  // PROG key
	case KEY10_RELEASE:  // 3/80 right PROG key
	    Action::Invoke("BrailleProg", 0, NULL);
	    break;
	    
	case KEY2_RELEASE:  // HOME key
	case KEY9_RELEASE:  // 3/80 right HOME key
	    if (text) {
		row = column = 0;
		DisplayBraille(text[row]);
		Action::Invoke("BrailleHome", 0, NULL);
	    }
	    else {
		Error(EL_ERROR, "Alva::HandleActivity", "No text", NULL);
	    }
	    break;
	    
	case KEY3_RELEASE:  // CURSOR key 
	case KEY8_RELEASE:  // 3/80 right CURSOR key
	    Action::Invoke("BrailleCursor", 0, NULL);
	    break;
	    
	case KEY4_RELEASE:  // UP key
	    if (text) {
		if (row > 0) {
		    row--;
		}
		else
		    Action::Invoke("BrailleUpBounds", 0, NULL);

		DisplayBraille(text[row] + column);
		Action::Invoke("BrailleUp", 0, NULL);
	    }
	    else {
		Error(EL_ERROR, "Alva::HandleActivity", "No text", NULL);
	    }
	    break;
	    
	case KEY5_RELEASE:  // LEFT key
	    if (text) {
		if (jumpScrollMode) {
		    if (column >= (capabilities.displayCells)) {
			column -= capabilities.displayCells;
		    }
		    else if (column > 0) { // jump to position 0
			column = 0;
		    }
		    else {
			Action::Invoke("BrailleLeftBounds", 0, NULL);
		    }
		    Action::Invoke("BrailleLeftJumpScroll", 0, NULL);
		}
		else {
		    if (column > 0) {
			column--;
		    }
		    else {
			Action::Invoke("BrailleLeftBounds", 0, NULL);
		    }
		    Action::Invoke("BrailleLeft", 0, NULL);
		}
		
		DisplayBraille(text[row] + column);
	    }
	    else {
		Error(EL_ERROR, "Alva::HandleActivity", "No text", NULL);
	    }
	    break;
	    
	case KEY6_RELEASE:  // RIGHT key
	    if (text) {
		if (jumpScrollMode) {
		    if((strlen(text[row])-column) > capabilities.displayCells){
			column += capabilities.displayCells;
		    }
		    else {
			Action::Invoke("BrailleRightBounds", 0, NULL);
		    }
		    Action::Invoke("BrailleRightJumpScroll", 0, NULL);
		}
		else {
		    if (column < (strlen(text[row]) - 1)) {
			column++;
		    }
		    else {
			Action::Invoke("BrailleRightBounds", 0, NULL);
		    }
		    Action::Invoke("BrailleRight", 0, NULL);
		}
		
		DisplayBraille(text[row] + column);
	    }
	    else {
		Error(EL_ERROR, "Alva::HandleActivity", "No text", NULL);
	    }
	    break;
	    
	case KEY7_RELEASE:   //  DOWN key
	    if (text) {
		if (row < (numRows-1)) { /* text starts at row 0 */
		    row++;
		}
		else
		    Action::Invoke("BrailleDownBounds", 0, NULL);

		DisplayBraille(text[row]+column);
		Action::Invoke("BrailleDown", 0, NULL);
	    }
	    else {
		Error(EL_ERROR, "Alva::HandleActivity", "No text", NULL);
	    }
	    break;
	    
	default:
	    {
		char tmp[20];
		sprintf(tmp, "%d\n", key[1]);
		Error(EL_WARNING, "Alva::HandleActivity",
		      "Unknown keycode.", tmp, NULL);
		break;
	    }
	}
    }
    return 0;
}

void
Alva::DisplayRaw(const char *chars, int rawPosition)
{
    unsigned char translated[128];
    int i;
    unsigned char l, p;
    int length;

    if ((rawPosition < 0) || (rawPosition > capabilities.displayCells +
			      capabilities.statusCells - 1)) {
	Error(EL_ERROR, "Alva::Displayraw",
	      "Position parameter out of range.", NULL);
	return;
    }
    
    if(chars != NULL)
	length = strlen(chars);

    if (length < 1)
	return;
    
    length = MIN(length, capabilities.displayCells +
		 capabilities.statusCells - rawPosition);
    
    p = rawPosition;
    
    l = length;
    
    for(i=0 ; i<length ; i++)
	translated[i] = Translation()[chars[i]];

    //
    // The following two lines have been temporarily inserted to pad the
    // ends of lines with blanks.  This is just for the demo.
    //
    for (i=length ; i<capabilities.displayCells; i++)
	translated[i] = 0; 		    // padding 
    
    length = l = capabilities.displayCells;

    write(fd, "\033B", 2);
    write(fd, &p, 1);
    write(fd, &l, 1);
    write(fd, translated, length);
    write(fd, "\015", 1);

    if(file)
	fflush(file);
    
    return;
}

void
Alva::DisplayBraille(const char *chars, int position)
{
    if ((position < 0) || (position > capabilities.displayCells -1)) {
	Error(EL_ERROR, "Alva::DisplayBraille",
	      "Position parameter out of range.", NULL);
	return;
    }

    if (chars == NULL) {
	Error(EL_ERROR, "Alva::DisplayBraille", "NULL text.", NULL);
	return;
    }

    //
    // Display the text in normal text area
    //
    if(fd > 0) {
	DisplayRaw(chars, capabilities.statusCells);
    }
    else {
	Braille::DisplayRaw(chars, capabilities.statusCells);
    }
    
    return;
}

void
Alva::DisplayStatus(const char *chars)
{
    char *tmpChars;

    if(chars == NULL) {
	Error(EL_ERROR, "Alva::DisplayStatus", "NULL text.", NULL);
	return;
    }
    
    tmpChars = new char[capabilities.statusCells + 1];

    // truncate the chars to the length of the status area
    strncpy(tmpChars, chars, capabilities.statusCells);
    tmpChars[strlen(tmpChars)] = '\0';

    DisplayRaw((const char *)tmpChars, 0);

    if(text)
	DisplayBraille(text[row] + column);

    delete tmpChars;
    return;
}


const Braille::Capabilities&
Alva::GetCapabilities() const
{
    return capabilities;
}

int
Alva::JumpScroll()
{
    return jumpScrollMode;
}

void
Alva::JumpScroll(int js)
{
    jumpScrollMode = js;

    Error(EL_DEBUG, "Alva::JumpScroll", "Jump Scroll Mode ",
	  (jumpScrollMode ? "on" : "off"), NULL);
}

void
Alva::ToggleJumpScroll()
{
    if(jumpScrollMode)
	jumpScrollMode = 0;
    else
	jumpScrollMode = 1;

    Error(EL_DEBUG, "Alva::ToggleJumpScroll", "Jump Scroll Mode ",
	  (jumpScrollMode ? "on" : "off"), NULL);
}


//
// This method should be called just before mercator kills all its
// subprocesses
//
void
Alva::ShutDown()
{
    Error(EL_DEBUG, "Alva::ShutDown", "Shutting down Alva object",
	  NULL);

    // just need to prevent the object from trying to play any sound
    if (fd == -1) {
	close(fd);
	fd = -1;
    }
}

#if 0
int
Alva::OpenAlva(const char *host, short port)
{
    struct sockaddr_in sa; /* IP format socket address */
    struct hostent *hostentry;
    long *hostaddress;
    int fd;
    int status;

    hostentry = gethostbyname(host);
    
    hostaddress = (long *) *(hostentry->h_addr_list);
    
    fd = socket(PF_INET, SOCK_STREAM, 0);
    
    if (fd == -1) {
	Error(EL_ERROR, "Alva::OpenAlva", "socket: ", Perror(), NULL);
	return -1;
    }
    
    memset((char *) &sa, 0, sizeof(sa));

    sa.sin_family = PF_INET;
    sa.sin_port = 1350; /* port number, arbitrarily chosen */
    sa.sin_addr.s_addr = *hostaddress;
    
    status = connect(fd, (struct sockaddr *) &sa, sizeof(sa));
    
    if (status == -1) {
	Error(EL_ERROR, "Alva::OpenAlva", "connect: ", Perror(), NULL);
	return 1;
    }

    return fd;
}
#endif
