//  -*-Mode: C++-*-
//
//
//
//	Note:  The use of this program shall be restricted to educational,
//	non-commercial use only.  All inquiries regarding licensing this
//	software for commercial use may be addressed to:
//	
//	Georgia Tech Research Corporation
//	Technology Licensing
//	Centennial Research Building - Rm. 275
//	400 Tenth Street, N.W.
//	Atlanta, Georgia  30332-0415
//	
//	*************************************************
//	
//	       Copyright 1996
//	Georgia Tech Research Corporation
//	    All Rights Reserved
//
//
//
//
//  $Author: keith $
//  $Source: /net/hm10/keith/proj/mercator/6.0/RCS/AppModelMgr.cc,v $
//  $Revision: 6.15 $
//  $Date: 1995/11/29 21:11:08 $
//
// $Log: AppModelMgr.cc,v $
// Revision 6.15  1995/11/29  21:11:08  keith
// Added FillRectangle support (acts like ClearAreas)
//
// Revision 6.14  1995/11/29  00:12:01  keith
// Got rid of some warnings...
//
// Revision 6.13  1995/10/31  00:09:56  keith
// Made client deletion a little more reliable.
//
// Revision 6.12  1995/10/25  19:24:47  keith
// Fixed memory access problem with SetToNextClient()
//
// Revision 6.11  1995/10/14  21:21:28  keith
// Made traversal of client list optionally circular.
//
// Revision 6.10  1995/10/07  18:14:08  spine
// Removed code which refer to IN3 and (voice) Recongnizer.
//
// Revision 6.9  1995/09/19  21:33:56  spine
// *** empty log message ***
//
// Revision 6.8  1995/08/29  22:15:38  keith
// Made changes about how window IDs, and X, Y, Width, and Height are
// handled (mostly for gadgets).  Some bugs remain though...
//
// Revision 6.7  1995/08/18  22:37:16  keith
// Fixed client deletion bug
//
// Revision 6.6  1995/07/28  20:12:44  spine
// Added copyright notice.
//
// Revision 6.5  1995/07/25  22:45:43  beth
// *** empty log message ***
//
// Revision 6.4  1995/07/01  20:02:19  spine
// Removed all references to LEDA strings and LEDA.
//
// Revision 6.3  1995/06/06  01:02:31  suelong
// added call to get all gcs at startup of application
//
// Revision 6.2  1995/06/05  22:01:08  iansmith
// Experimental support for the new mercator dictionaries.
//
// Removed all LEDA dictionary references in AppModelMgr and replaced
// them with mercator dictionaries.
//
// Added a function to hash strings (needed for dictionaries to have
// strings as part of them). Same for windows.
//
// Made a bunch of simplifications in the destructor as the memory
// cleanup is now pushed into the dictionary rather than falling on the
// AppModelMgr.
//
// Modified GetClient to use a mercator dictionary properly.
//
// Modified CreateXtObject/DeleteXtObject to use a mercator dictionary.
// Delete XtObject's debugging information got significantly simpler by
// removing all the LEDA stuff.
//
// Modified ChangeWindowId to use new dictionaries.
//
// Modified GetXtObjectByLongName to now use a mercator list iterator to
// find things.
//
// Modified DispatchEvent to use new dictionaries.
//
// Changed DeleteEventBind to now use a mercator dictionary. Potential
// problem discovered there. Should we be deleting the storage or just
// removing from the dictionary?
//
//

static const char rcsid[] = "$Header: /net/hm10/keith/proj/mercator/6.0/RCS/AppModelMgr.cc,v 6.15 1995/11/29 21:11:08 keith Exp $";

#include "AppModelMgr.h"
#include "Mercator.h"
#include "XtObject.h"
#include "RAP.h"
#include "Config.h"
#include "Error.h"
#include "Util.h"

#include "RapProto.h"		// for RapUnrealized and RapGadget

// ************************************************************************
// **                     Class Special Functions                    ******
// ************************************************************************

static int inited = 0;

// String hasher for dictionaries.
static unsigned StringHasher(const mString& str) { return str.hash(); }


// hashing function for Window : just use value, its 32 bits
static unsigned WindowHasher(const Window &w) { return (unsigned) w; }

AppModelMgr::AppModelMgr() :
    clientList(new mList<Client>),
    currentClient(NULL),
    clientNames(new mDictValPtr<mString,Client>(StringHasher)),
    xtobjectNames(new mDictValPtr<mString,XtObject>(StringHasher)),
    xtobjectWindows(new mDictValPtr<Window,XtObject>(WindowHasher)),
    eventActions(new mDictPtrPtr<EventStruct,Action>(EventStructHash))
{
    if (inited)
	Error(EL_FATAL, "AppModelMgr::AppModelMgr",
	      "attempt to create a new AppModelMgr.", NULL);

    inited = TRUE;
}

// ************************************************************************
// Just a little closure to  delete all of the clients
class deleteClientWalk : public mListClosure <Client> {
    AppModelMgr *obj;
    
public:
    void startup(void *cdata) {
	obj=(AppModelMgr *)cdata;
    }
    int action(Client *client) {
	obj->DeleteClient(client->GetID());
	return 1; // continue
    }
};

// ************************************************************************

AppModelMgr::~AppModelMgr()
{
    ClientPtr client;
    // dic_item item;
    deleteClientWalk dc;
    mDictPtrPtrIterator<EventStruct,Action> iter(*eventActions);
    EventStruct *es;

    // Delete the clients and free their memory via the closure.
    clientList->walk(&dc,this);
    // blow away the actual list
    delete clientList;
    
    // This maps strings to the clients we just deleted, so we just
    // blow it away knowing that our ValPtr dictionary will delete
    // the strings
    delete clientNames;

    // when we delete clients, it deletes all XtObjects in the
    // clients, so we'll just delete the two dictionaries that
    // map to XtObjects since they are value based.
    delete xtobjectNames;
    delete xtobjectWindows;

    // For actions we'll actually free the memory.   NOTE that this will result
    // in multiple frees (and possible coredumps) if the same action is installed
    // in two event dictionaries.
    // (I ifdef'ed out the item by item delete because more than one
    // events are bound to the same action. So an actions is getting
    // deleted for each event it is bound to. The temporary fix is to not
    // do explicit item by item delete.)
#ifdef X
    while ((es=iter.next())!=0) {
	eventActions->del_item(es);
    }
#endif
    delete eventActions;
}

// ************************************************************************
// **                          Public Functions                      ******
// ************************************************************************

//
// This code is called by RAP to create a new client in the offscreen
// model.
//
// CreateClient does the following:
//	- Create a new client
//	- Put a pointer to it in the RAP instance for the client.
//	- Store it in the AMM's client list.
//	- Sends off the initial messages we're interested in.
//
Client * 
AppModelMgr::CreateClient(RAP *rap)
{
    Client *newClient = new Client(rap);
    int argc = 0;
    char *argv[10];
    
    rap->SetClient(newClient);
    
    Error(EL_DEBUG, "AppModelMgr::CreateClient",
	  "NEW CLIENT BEING CREATED.", NULL);
    
    clientList->append(newClient);

    mString clientName(newClient->GetUniqueName());
    clientNames->insert(clientName, newClient);

    argv[argc++] = Strdup(newClient->ClientName());
    argv[argc++] = Strdup(newClient->GetUniqueName());
    Action::Invoke("ClientCreated", argc, argv);
    while (argc--) delete [] argv[argc];

    // start by sending a request for all your gcs
    // zero for num of gcs means send me all of them
    rap->GetGCValuesRequest(NULL, 0);

    //
    // We send a hello req to confirm the client's toplevel window ID.
    //
    rap->HelloRequest();
    //
    // We send a full query tree req to download the initial app state.
    //
    rap->FullQueryTreeRequest();
    //
    // We're interested in the following events:
    //
    short events[5];
    events[0] = KeyPress;
    events[1] = EnterNotify;
    events[2] = MapNotify;
    events[3] = UnmapNotify;
    events[4] = ButtonPress;
    rap->SelectEventRequest(events, 5);
    //
    // We're interested in the following requests:
    //
    short requests[12];
    requests[0]  = X_CreateGC;
    requests[1]  = X_ChangeGC;
    requests[2]  = X_CopyGC;
    requests[3]  = X_FreeGC;
    requests[4]  = X_ImageText16;
    requests[5]  = X_PolyText16;
    requests[6]  = X_ImageText8;
    requests[7]  = X_PolyText8;
    requests[8]  = X_ClearArea;
    requests[9]  = X_CopyArea;
    requests[10] = X_ReparentWindow;
    requests[11] = X_PolyFillRectangle;
    rap->SelectRequestRequest(requests, 12);

    return newClient;
}

// ************************************************************************

int 
AppModelMgr::DeleteClient(XID id)
{
    Client *client;
    // rewritten by IES 5/22/95 to get rid of the bogus
    // old GetClientListItem LEDAism.

    // Find client's item.
    client = GetClient(id);

     if (client == NULL) {
	char tmp[16];
	sprintf(tmp, "%d", id);
	Error(EL_ERROR, "AppModelMgr::DeleteClient",
	      "couldn't find client, ID = ", tmp, NULL);
	return False;
    }

    // Fire an action before we actually change the current client.
    int argc = 0;
    char *argv[10];

    argv[argc++] = Strdup(client->GetUniqueName());
    Action::Invoke("ClientDeletion", argc, argv);
    while (argc--) delete argv[argc];

    // If current client, move current client to previous client.
    if (currentClient && (id == currentClient->GetID())) {
 	if (client != clientList->head()) {
	    SetCurrentClient(clientList->predecessor(client));
	    // As a hack, call GoTo to set the current location.
	    argv[0] = Strdup(GetCurrentClient()->GetCurrentLocation()->GetUniqueName());
	    Action::Invoke("GoTo", 1, argv);
	    delete argv[0];
	} else {
	    // ok, the client IS the head of the list, try successors
	    if (clientList->successor(client)!=0) {
		// found a successor, so allow that to be the new client
		SetCurrentClient(clientList->successor(client));
		// As a hack, call GoTo to set the current location.
		argv[0] = Strdup(GetCurrentClient()->GetCurrentLocation()->GetUniqueName());
		Action::Invoke("GoTo", 1, argv);
		delete argv[0];
	    } else {
		// we are the last client, so just set the current client to NULL
		Error(EL_ERROR, "AppModelMgr::DeleteClient",
		      "No client to go to!", NULL);
		SetCurrentClient(NULL);
	    }
	}
    }
    // Delete item from list and blow away its storage
    clientList->destroy_item(client);
    
    return True;
}

// ************************************************************************
// Ok, we need a closure that can find a client
class findClientClosure : public mListClosure<Client> {
public:
    Client *foundClient;
    XID xid;
    void startup(void *cdata) { 
	xid=*((XID *)cdata);
	foundClient=0;
    };
    int action(Client *client) {
	if (client->GetID()==xid) {
	    foundClient=client;
	    return 0; // quit, we found
	}
	return 1; // keep looking
    }
};
// ************************************************************************
// Search for and return client by id.

Client* 
AppModelMgr::GetClient(XID id)
{
    findClientClosure fc;

    clientList->walk(&fc,&id);
    return fc.foundClient;
}

// ************************************************************************

Client* 
AppModelMgr::GetClient(mString name)
{
    Client *client;

    if ((client = clientNames->lookup(name)) == NULL) {
        // The LEDA ~ operator is replaced
        // with (char *) operator of mString.
        char * n = (char *)name;
	Error(EL_WARNING, "AppModelMgr::GetClient", 
	      "can't find client by name: ", n, NULL);
	delete [] n;
	return NULL;
    }
 
    return client;
}

// ************************************************************************

void 
AppModelMgr::SetCurrentClient(Client *newClient)
{
    // I changed all this to simplify it when I removed all the
    // old LEDA crap.  [IES , 5/22/95]
    Action    *action;
    Client    *oldClient;

    // validate input
    if (newClient == NULL) {
	Error(EL_ERROR, "AppModelMgr::SetCurrentClient",
	      "called with a NULL client argument.", NULL);
	// May be dangerous...
	currentClient = NULL;
	return;
    }
    
    // give  debug message
    if (currentClient) {
	Error(EL_DEBUG, "AppModelMgr::SetCurrentClient",
	      "CURRENT CLIENT WAS ", currentClient->ClientName(),
	      " NOW ", newClient->ClientName(), NULL);
    } else {
	Error(EL_DEBUG, "AppModelMgr::SetCurrentClient",
	      "CURRENT CLIENT WAS <NONE>, NOW ",
	      newClient->ClientName(), NULL);
    }

    // copy the current client
    oldClient = currentClient; 

    // set the current client to be this new client
    currentClient = newClient;

    // Invoke the ClientChange action
    // char *argv[3];
    // XtObject *obj;
    // if (oldClient) argv[0] = strdup(oldClient->ClientName());
    // else           argv[0] = strdup("NULL");
    // if (currentClient->ClientName() != NULL)
    // argv[1] = strdup(currentClient->ClientName());
    // else argv[1] = strdup("NULL");
    // obj = currentClient->GetCurrentLocation();
    // if (obj) argv[2] = strdup(obj->GetUniqueName());
    // else     argv[2] = strdup("NULL");
    //Action::Invoke("ClientChange",3,argv);
    //free(argv[0]);
    //free(argv[1]);
    //free(argv[2]);
 
}

// ************************************************************************

Client* 
AppModelMgr::SetToNextClient()
{
    // need a copy of the successor sometimes
    Client *tmp;

    // Set the current client item to the next list item if it is
    // not null.  Return client for success or null for no change.

    if (clientList->entries()<2) {
	return NULL;
    }
    // make it circular ... check to see if we have a successor
    if ((tmp=clientList->successor(GetCurrentClient()))!=0) {
	// we do set it to be current and return it
	SetCurrentClient(tmp);
	return tmp;
    } else {
	// Do we loop or don't we loop?
	FieldData *data = Config::GetDefaultByName("loopClients");
	if (data == NULL)
	    return NULL;
	if (data->Tag() != BOOLEAN) {
	    Error(EL_ERROR, "AppModelMgr::SetToNextClient",
		  "loopClients attribute should be boolean", NULL);
	    return NULL;
	}
	if (data->Boolean()) {
	    tmp = clientList->head();
	    SetCurrentClient(tmp);
	    return tmp;
	} else {
	    return NULL;
	}
    }
}

// ************************************************************************

Client* 
AppModelMgr::SetToPrevClient()
{
    // need a copy of the successor sometimes
    Client *tmp;

    // Set the current client item to the previous list item if it is
    // not null.  Return client for success or null for no change.

    if (clientList->entries()<2) {
	return NULL;
    }

    if ((tmp=clientList->predecessor(GetCurrentClient()))!=0) {
	// found a predecessor so make it current
	SetCurrentClient(tmp);
	return tmp;
    } else {
	// Do we loop or don't we loop?
	FieldData *data = Config::GetDefaultByName("loopClients");
	if (data == NULL)
	    return NULL;
	if (data->Tag() != BOOLEAN) {
	    Error(EL_ERROR, "AppModelMgr::SetToPrevClient",
		  "loopClients attribute should be boolean", NULL);
	    return NULL;
	}
	if (data->Boolean()) {
	    tmp = clientList->tail();
	    SetCurrentClient(tmp);
	    return tmp;
	} else {
	    return NULL;
	}
    }
}

// ************************************************************************

void
AppModelMgr::CreateXtObject(XtObject *obj)
{
    // This is called when an XtObject is created.  The function inserts
    // the XtObject into Application's list of ALL the functions.
    // This is a private functions so that it provides "some" protection that
    // only the friend Client::CreateXtObject calls it.  It is assumed
    // that the instance of the XtObject has been created by its Client.
 
    if (obj == NULL) {
	Error(EL_ERROR, "AppModelMgr::CreateXtObject",
	      "called with NULL object argument.", NULL);
	return;
    }
    
    mString xtobjectName(obj->GetUniqueName());
    Window xtobjectWin  = obj->GetWindowId();
 
#ifdef LOTS_O_PRINTFS
    char tmp1[16], tmp2[16];
    sprintf(tmp1, "%d", obj->GetObjID());
    sprintf(tmp2, "%d", xtobjectWin);
    
    if (obj->GetParent() == NULL) {
	Error(EL_DEBUG, "AppModelMgr::CreateXtObject",
	  obj->GetUniqueName(), " (", tmp1, ") ", obj->GetName(), " win=",
	  tmp2, " NO PARENT", NULL);
    } else {
	Error(EL_DEBUG, "AppModelMgr::CreateXtObject",
	  obj->GetUniqueName(), " (", tmp1, ") ", obj->GetName(), " win=",
	  tmp2, " parent = ", obj->GetParent()->GetName(), NULL);
    }
#endif /* LOTS_O_PRINTFS */
 
    // if we don't know about the XtObject already then insert
    // it into the dictionary
    
    if (xtobjectNames->lookup(xtobjectName) == 0) {
	xtobjectNames->insert(xtobjectName, obj);
    }
 
    // if the window is valid then insert it into the dictionary
    if ((xtobjectWin != RapGadget) && (xtobjectWin != RapUnrealized)) {
	if (xtobjectWindows->lookup(xtobjectWin) != 0) {
	    char tmp[16];
	    sprintf(tmp, "%d", xtobjectWin);
	    Error(EL_ERROR, "AppModelMgr::CreateXtObject",
		  "Window for object is already in database!  ",
		  "(obj = ", obj->GetUniqueName(), ", win = ", tmp, ")", NULL);
	} else {
	    xtobjectWindows->insert(xtobjectWin, obj);
	}
    }
}

// ************************************************************************

void
AppModelMgr::DeleteXtObject(XtObject *obj)
{
    XtObject *item, *namep=NULL, *winp=NULL;
    
    // This is called when an XtObject is deleted.  The function deletes
    // the XtObject from its xtobjectName and xtobjectWindows dictionary.
    // This function MUST be called BEFORE the XtObject is deleted.
 
    mString xtobjectName(obj->GetUniqueName());
    Window xtobjectWindow = obj->GetWindowId();
 
    // delete from the names dictionary
    if ((item = xtobjectNames->lookup(xtobjectName)) == 0) {
	Error(EL_ERROR, "AppModelMgr::DeleteXtObject",
	      "Attempting to delete object which was never inserted:  ",
	      obj->GetUniqueName(), NULL);
    } else {
	// Take it out of the dict...
	xtobjectNames->del(xtobjectName);
	namep =  item;
    }
 
    // delete from the window dictionary if not RapGadget or RapUnrealized
    if ((xtobjectWindow != RapGadget) && (xtobjectWindow != RapUnrealized)) {
	if ((item = xtobjectWindows->lookup(xtobjectWindow)) == 0) {
	    Error(EL_ERROR, "AppModelMgr::DeleteXtObject",
		  "Attempting to delete object's window was never inserted:  ",
		  obj->GetUniqueName(), NULL);
	} else {
	    // Take it out of the dict...
	    xtobjectWindows->del(xtobjectWindow);
	    winp = item;
	}
    }

    // double check our dictionary implementation
    if ((namep != NULL) && (winp != NULL) && (namep != winp))
	Error(EL_ERROR,"AppModelMgr::DeleteXtObject",
	      "Inconsistant pointers in the xtobject Name and ",
	      "Window dictionaries!", NULL);

    //
    // Take it out of its client's dict.
    //
    obj->GetClient()->DeleteXtObject(obj->GetObjID());

    delete obj;
}

// ************************************************************************

void
AppModelMgr::ChangeWindowId(XtObject *xtobj, Window newID)
{
    mString xtobjectName(xtobj->GetUniqueName());
    Window xtobjectWin = xtobj->GetWindowId();
 
    char tmp[16];
    sprintf(tmp, "%d", xtobjectWin);
    Error(EL_DEBUG, "AppModelMgr::ChangeWindowId",
	  xtobj->GetUniqueName(), " ==> ", xtobj->GetName(),
	  " (window = ", tmp, ")", NULL);

    // delete the previous window entry
    if (xtobjectWindows->lookup(xtobjectWin) != 0)
	xtobjectWindows->del(xtobjectWin);

    // Update the object itself
    xtobj->SetWindowId(newID);

    // if the window is valid then insert it into the dictionary
    if ((newID != RapGadget) && (newID != RapUnrealized))
	xtobjectWindows->insert(newID, xtobj);
}

// ************************************************************************
#if 0
XtObject* 
AppModelMgr::GetXtObject(Window w)
{
    return xtobjectWindows->lookup(w);
}

// ************************************************************************

XtObject* 
AppModelMgr::GetXtObject(mString name)
{
    return xtobjectNames->lookup(name);
}
#endif
// ************************************************************************

XtObject* 
AppModelMgr::GetXtObjectByLongName(const mString& name)
{
    // Note that GetXtObjectByLongName is *much* more expensive than
    // GetXtObject because our XtObject dictionary is keyed by unique
    //names only.
    mDictValPtrIterator<mString,XtObject> iter(*xtobjectNames);
    XtObject *obj, *ret = NULL;
    char *n=(char *)name;

    while (iter.next()) {
	obj=xtobjectNames->lookup(iter.current());
	if (strcmp(obj->GetLongName(), n) == 0) { // found it!
	    ret = obj;
	    break;
	}
    }
    delete [] n;
    return ret;
}

// ************************************************************************
int
AppModelMgr::DispatchEvent(EventPtr event)
{
    // if an EventBind has been created for the event passed in then
    // invoke the action and return TRUE.  Else return False.

    ActionPtr action;

#ifdef EVENT_DEBUG
    printf("\n**************************************************************\n");
    printf("AppModelMgr::DispatchEvent: event = ");
    switch (event->eventType) {
    case KEYPRESS:   
	printf("Key Press  key=%i, modifier=%X\n", event->key, event->modifier);
	break;
    case KEYRELEASE: 
	printf("Key Release  key=%i, modifier=%X\n", event->key, event->modifier);
	break;
    default:
	printf("UNKNOWN\n");
	break;
    }
#endif

    // look to see if the application is interested in the event
    action = eventActions->lookup(event);
    if (action == NULL) {
        // event binding not found - no action invoked - return false
	return False; 
    }

    // Invoke the action
    action->Invoke(0,NULL);

    Error(EL_DEBUG, "AppModelMgr::DispatchEvent", "invoking action...", NULL);

    return True; // event binding found and invoked - return true
  
}

// ************************************************************************
#if 0
void
AppModelMgr::CreateEventBind(EventPtr event, Action *function)
{
    eventActions->insert(event, function);
}
#endif
// ************************************************************************

int
AppModelMgr::DeleteEventBind(EventPtr event)
{
    if (eventActions->lookup(event) == 0) {
	Error(EL_ERROR, "AppModelMgr::DeleteEventBind"
	      "event was never inserted.", NULL);
	return False;
    } else {
	// Note: possible XXX: Should this be doing a deletion
	// of the storage or just dictionary removal?
	// I'm currently not sure enough to know... IES
	eventActions->del(event);
	return True;
    }
}

// ************************************************************************
