// -*-Mode: C++-*-
//
//
//
//	Note:  The use of this program shall be restricted to educational,
//	non-commercial use only.  All inquiries regarding licensing this
//	software for commercial use may be addressed to:
//	
//	Georgia Tech Research Corporation
//	Technology Licensing
//	Centennial Research Building - Rm. 275
//	400 Tenth Street, N.W.
//	Atlanta, Georgia  30332-0415
//	
//	*************************************************
//	
//	       Copyright 1996
//	Georgia Tech Research Corporation
//	    All Rights Reserved
//
//
//
//
//
// This is my implementation of the generic audio interface for the
// AudioFIle subsystem. 
//

// $Author: spine $
// $Source: /net/hm10/keith/proj/mercator/6.0/RCS/AudioFile.cc,v $
// $Date: 1995/07/28 20:16:36 $
/* 
 * $Log: AudioFile.cc,v $
// Revision 1.2  1995/07/28  20:16:36  spine
// Added copyright notice.
//
// Revision 1.1  1995/07/25  23:02:26  spine
// Initial revision
//
 */

#include "AudioFile.h"
#include "Error.h"
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/uio.h>
#include <unistd.h>
#include <stdlib.h>
#include "Util.h"
#include "Loadable.h"

// Keith's Loadable Magic to get a well-known symbol
LOADABLE_CLASS_DEFN(AudioFile);

// Amount of internal buffering each thread should do
#define BUFFER_LENGTH_IN_SECONDS (0.05)
// This is number of initial work crew members
#define WORK_CREW_START_SIZE 10

/*
 * Initialize the Work Crew Member 
 */
WorkCrewMember::WorkCrewMember(unsigned samplingRate,
			       unsigned bytesPerSample,
			       unsigned channels):
    Thread()
{
    /* just set up the audio specific things here....wait until thread start
       time for the rest */
    sampsPerBuffer=((double)BUFFER_LENGTH_IN_SECONDS)*((double)samplingRate);
    bufSize=sampsPerBuffer*bytesPerSample*channels;
    audioBuffer=new unsigned char [bufSize];
    currentTime=0;
    active=0;
    // call up audio file
    if ((aud=AFOpenAudioConn(""))==NULL) {
	fprintf(stderr,"Error Opening Audio Connection: Check DISPLAY variable\n");
	Error(EL_FATAL,"AudioFile::AudioFile","Error opening connection to AudioFile",
	      NULL);
    }
    // init attributes
    attributes.preempt = Mix;
    attributes.play_gain = 0;
    attributes.rec_gain =  0;
    attributes.channels=1;
    attributes.type = MU255;

    // create an audio context
    ac=AFCreateAC(aud,0,
		  (ACPlayGain | ACEncodingType | ACChannels | 
		   ACRecordGain | ACPreemption),
		  &attributes);
    
    AFSync(aud, ATrue);	// Make sure our device is right 
    // Now allocate the conversion contexts
    sixteen2eightCtx=AConvertAllocInit(LIN16,8000,1,MU255,8000,1);
    eight2sixteenCtx=AConvertAllocInit(MU255,8000,1,LIN16,8000,1);
}

void WorkCrewMember::Initialize()
{
    /* set up the condition vars and stuff */
    MUTEX_init(&mutex);
    COND_init(&condvar,&mutex);
}

/* Main execution path of the Work Crew Members */
void WorkCrewMember::Main()
{
    int ret,convertedSize,i;
    char tmpbuf[80];
    short *convertedData,sixteenBitValue;
    unsigned s1,s2;

    while (1) {
	/* look for an active signal */
	MUTEX_lock(&mutex);
	while (!active) {
	    COND_wait(&condvar,&mutex);
	}
	// At this point, we are locked again ... so unlock to allow others to tell
	// us to stop
	MUTEX_unlock(&mutex);
	// check on the volume flag
	if (volumeChanged) {
	    // rest
	    volumeChanged=0;
	    attributes.play_gain=((volume/20)-2)*6;
	    AFChangeACAttributes(ac,ACPlayGain,&attributes);
	}
 	// read in some bytes
	ret=read(fd,audioBuffer,bufSize);
	if (ret==0) {
	    // check to see if we are "looped" 
	    if (looped) {
		// try to lseek back to start
		if (lseek(fd,0L,SEEK_SET)<0) {
		    // can't reset 
		    Error(EL_ERROR,"WorkCrewMember::Main",
			  "Error In read during audio file I/O (lseek)",
			  Perror(),NULL);
		    active=0;
		    continue;
		}
		// still active
		Error(EL_DEBUG,"AudioFile::Main","looping sound",NULL);
		continue;
	    } else {
		Error(EL_DEBUG,"AudioFile::Main","done with sound",NULL);
		// EOF , we are done
		active=0;
		close(fd);
		// This alerts others that we are done
		fd=-1;
		continue;
	    }
	}
	// error?
	if (ret==-1) {
	    /* error condition */
	    Error(EL_ERROR,"WorkCrewMember::Main",
		  "Error In read during audio file I/O (read)",
		  Perror(),NULL);
	    active=0;
	    continue;
	}
	// do we need to muffle? Don't bother if its just a few bytes... not worth the
	// the computational load
	if (muffle && ret>10) {
	    // we do need to , perform conversion
	    convertedSize=AConvert(eight2sixteenCtx,ret,audioBuffer,
				   &convertedData);
	    // ok, here's the filter:
	    // y[n]=(x[n]+x[n-1])/2
	    // its a cheap hack but it does give a LPF with a cuttoff at 1/2 Nyquist freq.
	    
	    // we ignore the first byte... ugh.
	    s1=convertedData[0];
	    for (i=1; i<convertedSize; ++i) {
		s2=convertedData[i];
		s2=(s2+s1)>>1;
		s1=convertedData[i];
		sixteenBitValue=(short)s2;
		convertedData[i]=sixteenBitValue;
	    }
	    // ok, convert it back
	    convertedSize=AConvert(sixteen2eightCtx,convertedSize,convertedData,
				   &audioBuffer);
	    // size may have changed due to rounding, so reassign
	    ret=convertedSize;
	}
 	// successful read ... shove the bytes out 
 	// first: see what time the server says it is
	if (currentTime<AFGetTime(ac)) {
	    // we are behind the server, so assume this is a "new" file to be
	    // played and update current time... note: we check again since
	    // we know we are behind we want the most up to date version...
	    currentTime=AFGetTime(ac);
	}
	sprintf(tmpbuf,"scheduling some bytes (%d) for time %d\n",ret,currentTime);
	Error(EL_DEBUG,"AudioFile::Main",tmpbuf,NULL);
        // now we schedule the bytes we just read in
	AFPlaySamples(ac,currentTime,ret,audioBuffer);
	currentTime+=sampsPerBuffer;
	AFSync(aud,ATrue);
    }
}

// Return the activity status
int WorkCrewMember::IsActive() const
{
    return active;
}

// This function runs on the callers thread, which will be the main (not the thread
// associated with this object). Since we really shouldn't make audiofile calls from
// this thread, we'll just set  a flag and let the other thread (running in main)
// pick up the flag and make the change
void WorkCrewMember::SetVolume(int newVolume)
{
    volumeChanged=1;
    volume=newVolume;
}


// This is where the audio file object calls into our code and starts things going
// NOTE: Execution of this function body is running on the CALLER's thread, thus
// we have to synch on the active flag
void WorkCrewMember::StartSound(int _fd,int _looped, int _volume, int _muffle)
{
    fd=_fd;
    looped=_looped;
    muffle=_muffle;
    // lock then set the active flag
    MUTEX_lock(&mutex);
    Error(EL_DEBUG,"WorkCrewMember::StartSound","setting active to 1",NULL);
    active=1;
    MUTEX_unlock(&mutex);
    // change the volume if necessary
    if (volume!=_volume) {
	SetVolume(_volume);
    }
    // wake him up from sleeping
    COND_signal(&condvar);
}

// This is where the AudioFile object calls to stop a sound
// NOTE: Execution of this function body is running on the CALLER's thread, thus
// we have to synch on the active flag

void WorkCrewMember::StopSound()
{
    // lock then clear the active flag
    MUTEX_lock(&mutex);
    active=0;
    MUTEX_unlock(&mutex);
    // If he is currently playing, he'll notice that the active is now zero and
    // stop, but he won't close the FD...if he's not playing this is a NOOP
    // since we don't signal him.
    // does he have an open FD?
    if (fd!=-1) {
	close(fd);
    }
}


// This copy constructor is here to make sure that we get an error if someone
// tries to invoke this from within this class
WorkCrewMember::WorkCrewMember(const WorkCrewMember &w)
{
    /* this is not allowed */
    Error(EL_ABORT,"WorkCrewMember::WorkCrewMember",
	  "copy constructor is not allowed!",NULL);
}
// This operator is here to make sure that we get an error if someone
// tries to invoke this from within this class
WorkCrewMember & WorkCrewMember::operator=(const WorkCrewMember &)
{
    /* this is not allowed */
    Error(EL_ABORT,"WorkCrewMember::operator=",
	  "operator = is not allowed!",NULL);
    /* to make compiler happy */
    return *this;
}
// This operator is here to make sure that we get an error if someone
// tries to invoke this from within this class
int WorkCrewMember::operator<(const WorkCrewMember &)
{
    /* this is not allowed */
    Error(EL_ABORT,"WorkCrewMember::operator=",
	  "operator = is not allowed!",NULL);
    /* to make compiler happy */
    return 0;
}

// This is here to allow the rogue wave stuff to find duplicates
int WorkCrewMember::operator==(const WorkCrewMember &w) const
{
    // we define equality to be the same thread
    return (THREAD_equality(ThreadId(),w.ThreadId()));
}

// This is the "hashing function" for unsigned integers ... just return them
static unsigned identity(const unsigned &x)
{
    return x;
} 

// Construct an AudioFile
AudioFile::AudioFile(const char *hostname)
{
    WorkCrewMember *crewmember;

    // counter is initially one (zero is the error value)
    soundId=1;

    // initialize the dictionary
    SoundIdDictionary=new RWTValHashDictionary<unsigned,unsigned>(identity);

    // init the threads system
    THREAD_init();

    // set up the threads 
    for (int i=0; i<WORK_CREW_START_SIZE; ++i) {
	crewmember=new WorkCrewMember(8000,1,1);
	WorkCrew.insert(crewmember);
	crewmember->Start(Thread::HighPriority);
    }
}

AudioFile::~AudioFile(void)
{
    // stop all sounds
    for (int i=0; i<WorkCrew.entries(); ++i) {
	// stop it if its going
	if (WorkCrew[i]->IsActive()) {
	    WorkCrew[i]->StopSound();
	}
    }
    // kill all the threads
    for (i=0; i<WorkCrew.entries(); ++i) {
	THREAD_kill(WorkCrew[i]->ThreadId());
    }
}
    
Audio::SoundID	AudioFile::PlaySound(const char *sound_file, float *rate, 
				  int volume, int muffle, int looped)
{
    int member=-1,fd;
    WorkCrewMember *newWorker;
    unsigned newId;

    Error(EL_DEBUG,"AudioFile::PlaySound","Starting Sound (",
	  sound_file,")",NULL);
    for (int i=0; i<WorkCrew.entries(); ++i) {
	if (!WorkCrew[i]->IsActive()) {
	    // got one
	    member=i;
	    break;
	}
    }
    // member is our guy?
    if (member==-1) {
	// everybody is busy
	newWorker=new WorkCrewMember(8000,1,1);
	WorkCrew.insert(newWorker);
	member=WorkCrew.index(newWorker);
	if (member==-1) {
	    Error(EL_FATAL,"AudioFile::PlaySound","unable to create a new worker!",
		  NULL);
	}
	Error(EL_DEBUG,"AudioFile::PlaySound","creating new worker!",
	      NULL);
    }
    // wake up the player 
    fd=open(sound_file,O_RDONLY);
    if (fd==-1) {
	Error(EL_ERROR,"AudioFile::PlaySound","Opening AudioFile:",sound_file,
	      Perror(),NULL);
	return 0;
    }
    // send start sound signal to the thread
    WorkCrew[member]->StartSound(fd,looped,volume,muffle);
    // create a sound id
    newId=soundId++;
    // put it in the dictionary
    SoundIdDictionary->insertKeyAndValue(newId,member);
    // return the id to the user
    return newId;
}

int AudioFile::StopSound(Audio::SoundID sid)
{
    unsigned target=-1; // try to get an error if something hoses up

    // we return 0 as the error sound id above... so check it here
    if (soundId==0) return 0;

    // look up the value in the dictionary;
    if (!SoundIdDictionary->findValue(sid,target)) {
	// this is a bogus id
	return 0;
    }
    Error(EL_DEBUG,"AudioFile::StopSound","Stopping Sound",NULL);
    // everything's cool
    WorkCrew[target]->StopSound();
    return 1;
}
