//
//	Note:  The use of this program shall be restricted to educational,
//	non-commercial use only.  All inquiries regarding licensing this
//	software for commercial use may be addressed to:
//	
//	Georgia Tech Research Corporation
//	Technology Licensing
//	Centennial Research Building - Rm. 275
//	400 Tenth Street, N.W.
//	Atlanta, Georgia  30332-0415
//	
//	*************************************************
//	
//	       Copyright 1996
//	Georgia Tech Research Corporation
//	    All Rights Reserved
//
//
//
// This is a baseclass for loadable Braille objects.  Note that we make this
// thing a subclass of FDInterest so that it can provide input to the system
// via its keys.
//
// We try to keep the philosophy of the other loadable classes here:  this
// class does not provide any interface to a specific device, but can still
// be instantiated.  When instantiated, anything passed to it for output will
// simply be displayed on the screen (marginally useful for debugging).
// When this class is instantiated (as opposed to one of its subclasses) it
// never registers an interest with FDInterest and hence will never produce
// keystrokes.
//
// The class provides two APIs for displaying Braille.  The DisplayBraille()
// method displays a single "chunk" of text at the position indicated by
// the position flag (which has a default value of zero).
//
// The SetText() method makes a copy of the text passed to it inside
// the class, and then displays it via DisplayBraille().  This method is
// primarily useful for Braille devices have have keys on them and
// wish to implement some simple scanning (left-right panning for
// example) internally to the class.  These devices can pan across the
// text and make repeated calls to DisplayBraille() to display the
// particular chunk they are interested in.
//
// Behavior of SetText on devices without internal panning capability
// is not defined.  It will probably clip the output.
//
// Future API candidates:
//
//	- Some way to query a device for the number of cells.
//	- Multi-line SetText (up/down scrolling handled internally).
//	- Some way to notify external code when scrolling has
//	  reached the boundary of text.
//	- Unified way to set display parameters.
//
//		keith edwards	26jul95
//

#ifndef _BRAILLE_H_
#define _BRAILLE_H_

#include "FDInterest.h"
#include "Util.h"
#include "Error.h"

#include <string.h>

#ifdef BEAM
#include <stdio.h>
#endif

#ifndef NULL
#define NULL 0
#endif /* NULL */

class Braille : public FDInterest {
public:
    enum { TRANS_TABLE_SIZE = 256 };
    typedef char  	*TransTable;
    
private:
    char		*brailleName;
    unsigned char	*translation;

public:
    Braille(const char *name = "Generic");
    virtual ~Braille();

    class Capabilities {
    public:
	int		displayCells;
	int		statusCells;
	int		highlightSupported;
	int		hasCursorKeys;
	int		hasProgKey;
	int		hasHomeKey;
	int		otherKeys;
	
	Capabilities();
	Capabilities(int, int, int, int, int, int, int);
	Capabilities(const Capabilities& other);
	Capabilities& operator=(const Capabilities& other);
    };

    virtual const Capabilities& GetCapabilities() const;

    virtual int		HandleActivity(int fd, short type);
    virtual void	Shutdown(int fd);
    
    virtual void	DisplayRaw(const char *text, int rawPosition = 0);
    virtual void	DisplayBraille(const char *text, int position = 0);
    virtual void	DisplayStatus(const char *text);

    virtual int         JumpScroll();
    virtual void        JumpScroll(int js);
    virtual void        ToggleJumpScroll();
    
    virtual void	SetText(const char *text[], int numRows);

    virtual void        ShutDown();
    
    void		Translation(const unsigned char *translation);
    const unsigned char	*Translation() const;
    
    const char		*GetBrailleName() const;

    //
    // This base class defines a few common translation sets.
    //
    static const unsigned char UCPT[TRANS_TABLE_SIZE];
    static const unsigned char LCPT[TRANS_TABLE_SIZE];
    static const unsigned char UCST[TRANS_TABLE_SIZE];
    static const unsigned char LCST[TRANS_TABLE_SIZE];
};

inline const char *
Braille::GetBrailleName() const
{
    return brailleName;
}

inline void
Braille::SetText(const char *t[], int numRows)
{
    if (t)
	DisplayBraille(t[0]);
}

inline void
Braille::Translation(const unsigned char *t)
{
    delete [] translation;
    translation = new unsigned char[TRANS_TABLE_SIZE];
    memcpy(translation, t, TRANS_TABLE_SIZE);
}

inline const unsigned char *
Braille::Translation() const
{
    return translation;
}

inline void
Braille::ShutDown()
{
    Error(EL_DEBUG, "Braille::ShutDown", "Shutting down braille subsystem",
	  NULL);
}

#endif /* _BRAILLE_H_ */
