// -*-Mode:C++-*-
#ifndef _CLIENT_H_
#define _CLIENT_H_
//
//
//
//	Note:  The use of this program shall be restricted to educational,
//	non-commercial use only.  All inquiries regarding licensing this
//	software for commercial use may be addressed to:
//	
//	Georgia Tech Research Corporation
//	Technology Licensing
//	Centennial Research Building - Rm. 275
//	400 Tenth Street, N.W.
//	Atlanta, Georgia  30332-0415
//	
//	*************************************************
//	
//	       Copyright 1996
//	Georgia Tech Research Corporation
//	    All Rights Reserved
//
//
//
//
//  This is the class Client.  It is a derived class of Storage.
//  The Client class encapsulates all the information about a
//  particular client connected to the display.  It holds onto the
//  toplevel windows and hierarchy.  
//  The data that is stored is specifically:
//
//    Data
//    ----
//    List of the XtObjects, stored by its Object.
//    Array of the top level windows.
//    List of TextReps, stored by their window Ids.
//    List of event grabs for all applications (i.e. global key grabs).
//    
//  A Client should only be created by the AppModelMgr instance (there should
//  only be one instance of the AppModelMgr class).
//  

//
// $Author: keith $
// $Source: /net/hm10/keith/proj/mercator/6.0/RCS/Client.h,v $
// $Revision: 6.8 $
// $Date: 1995/09/28 19:54:41 $
//


class Client;
typedef Client *ClientPtr;


#include <X11/Intrinsic.h>
#include "Storage.h"
#include "XtObject.h"
#include "mList.h"
#include "mDictValPtr.h"

// forward references
//class array(XtObjectPtr);
// class mList<XtObject>;
// class mDictValPtr<Object,XtObject>;
// class mDictValPtr<Window,TextRep>
class Action;
typedef Action *ActionPtr;
class RAP;
class FontCache;
class XtObject;
typedef XtObject *XtObjectPtr;


class Client : public Storage {
private:
    RAP      		*rap;

    char 		*clientName;
    char 		*clientClass;
    XID   		id;          // server client number.
    char 		*uniqueName; // a unique string id for this client.

    // Will be used if this client has any objects with ProtoTextReps.
    FontCache		*fontCache;

    // Map from object IDs to Mercator XtObjects.
    mDictValPtr<Object, XtObject> *xtobjectObjectIds;

    // The current XtObject for this client.
    XtObject		*currentLocation;
    
    // A list of the top-level windows of our application.
    mList<XtObject>	*topWindows;
    
    // Map from events to the actions that may fire on those events.
    mDictPtrPtr<EventStruct, Action> *eventActions; 
    
    //
   // If you look in XtHook and XlibHook::SendInitialMessages, you'll
    // see that there are a couple of messages that we send to clients
    // whenever they're first created and connected.  Once these respond
    // then the client is considered to be "ready."
    //
    // NOTE that the implementation of this is currently in XtHook ONLY.
    // That is, the receipt of the initial xlib messages is ignored (it's
    // more difficult to determine).  In the future the xlib protocol will
    // probably get restructured a bit anyway, so I'm just ignoring the
    // problem for now.			-k
    int			ready;

public:
    // NOTE that for now we can't get the XID.
#ifdef NO_XID
    Client(RAP *rap, XID id);
#else
    Client(RAP *rap);
#endif
    virtual ~Client();

    // set and get the name and class.  Get the XID.
    void		ClientName(const char *c);
    const char 		*ClientName() const;
    void         	ClientClass(const char *c);
    const char 		*ClientClass() const;
    XID          	GetID() const;
    const char 		*GetUniqueName() const;

    // Called by RAP when a HelloReply comes in (containing the
    // toplevel window id).
    void		SetWindowID(XID xid);
    
    int			Ready() const;
    void		Ready(int r);
    
    int			NumChildren() const;

    FontCache&		GetFontCache() const;

    XtObject 		*CreateXtObject(Object id, const char *name,
					const char *className,
					XtObject *parent, Window win);
    
    void       		DeleteXtObject(Object);
    XtObject 		*GetXtObject(Object);
    XtObject 		*GetXtObject(Window);
    int        		XtObjectExists(Object);

    void       		SetCurrentLocation(XtObject *l);
    XtObject 		*GetCurrentLocation() const;

    void                AddToplevelWindow(const XtObject *);
    void                DeleteToplevelWindow(const XtObject *);
    int			IsToplevelWindow(const XtObject *);
    const mList<XtObject> *GetTopWindows() const;
    
    void 		PrintHierarchy() const;
    
    TextRep 		*FindTextRep(Window win) const;
    
    RAP 		*GetRAP() const;
    void		SetRAP(RAP *r);
    
    // virtual functions required to be implemented by the base class Storage
    int  		DispatchEvent(EventPtr event);
    void 		CreateEventBind(EventPtr event, Action *function);
    int  		DeleteEventBind(EventPtr event);

    mDictPtrPtr<EventStruct,Action> *GetActions() const;
    // need these to allow us to put the client in a data structure
    // we do these based on address in core comparison
    int operator==(const Client &);
    int operator<(const Client &);

};

typedef Client * ClientPtr;

////////////////////////////////////////////////////////////////////////

inline const char *
Client::ClientName() const
{
    return clientName;
}

inline const char *
Client::ClientClass() const
{
    return clientClass;
}

inline XID
Client::GetID() const
{
    return id;
}

inline const char *
Client::GetUniqueName() const
{
    return uniqueName;
}
    
inline int
Client::Ready() const
{
    return ready;
}

inline XtObject *
Client::GetCurrentLocation() const
{
    return currentLocation;
}

inline const mList<XtObject> *
Client::GetTopWindows() const
{
    return topWindows;
}

inline RAP *
Client::GetRAP() const
{
    return rap;
}

inline void
Client::SetRAP(RAP *r)
{
    rap = r;
}
inline mDictPtrPtr<EventStruct,Action> *
Client::GetActions() const
{
    return eventActions;
}

#endif _CLIENT_H_
