//
//	Note:  The use of this program shall be restricted to educational,
//	non-commercial use only.  All inquiries regarding licensing this
//	software for commercial use may be addressed to:
//	
//	Georgia Tech Research Corporation
//	Technology Licensing
//	Centennial Research Building - Rm. 275
//	400 Tenth Street, N.W.
//	Atlanta, Georgia  30332-0415
//	
//	*************************************************
//	
//	       Copyright 1996
//	Georgia Tech Research Corporation
//	    All Rights Reserved
//
//
//
//	Support code for parsing and storing options from the Mercator
//	configuration file.
//
//		keith edwards	30jan95
//

#include <stdio.h>
#include <string.h>
#include "Config.h"
#include "Error.h"
#include "Attribute.h"
#include "Mercator.h"

/////
///// Stuff needed to link with the config parser.
/////
extern "C" int configparse(void);
extern FILE *configin;
int configLineno = 1;

/////
///// Miscellaneous crap.
/////
#ifndef TRUE
#define TRUE 1
#endif

#ifndef FALSE
#define FALSE 0
#endif

/////
///// Used during parsing.
/////
static mListVal<mString> *currentStringList = NULL;

/////
///// These are routines which are called directly from yacc-generated
///// C code.
/////

extern "C" {
    
void
ProcessBooleanDecl(int l, const char *id, unsigned char b)
{
    Config::InstallBooleanDecl(l, id, b);
}

void
ProcessIntegerDecl(int l, const char *id, long i)
{
    Config::InstallIntegerDecl(l, id, i);
}

void
ProcessStringDecl(int l, const char *id, const char *str)
{
    Config::InstallStringDecl(l, id, str);
}

void
ProcessIdentifierDecl(int l, const char *id, const char *str)
{
    Config::InstallIdentifierDecl(l, id, str);
}

void 
ProcessProcDecl(int l, const char *id, const char *str)
{
    Config::InstallProcDecl(l, id, str);
}

void
ProcessStringListDecl(int l, const char *id)
{
    if (currentStringList == NULL) {
	mListVal<mString> emptyStringList;
	Config::InstallStringListDecl(l, id, emptyStringList);
    } else {
	Config::InstallStringListDecl(l, id, *currentStringList);
	delete currentStringList;
	currentStringList=NULL;
    }
}

void
ConfigAddToStringList(int l, const char *str)
{
    if (currentStringList == NULL)
	currentStringList = new mListVal<mString>;

    //    copy=str; // I just feel better w/this
    // vs using all this implicit coversion crap
    currentStringList->append(str);
}

void
ConfigLexWarning(int l, const char *s)
{
    char tmp[16];
    sprintf(tmp, "%d", l);
    Error(EL_ERROR, "ConfigLexWarning", "syntax error on line ", tmp,
	  ": ", s, NULL);
}

} // extern "C"


static unsigned StringHasher(const mString& str) { return str.hash(); }

mDictValPtr<mString, FieldData> *Config::configDictionary =
  new mDictValPtr<mString, FieldData>(StringHasher);

int Config::validDict = 0;


/////////////////////////////////////////////////////////////////////////////

/////
///// Note constructor is private -- we should never create a Config.
/////
Config::Config()
{
    Error(EL_ABORT, "Config::Config", "Cannot call constructor for Config ",
	  "class\n\tThis is a nonrecoverable error as it indicates a ",
	  "programmer logic error.",
	  NULL);
}

static const char *
MakeTagName(FieldType tag)
{
    switch (tag) {
    case BOOLEAN:
	return "boolean";
    case INTEGER:
	return "integer";
    case STRING:
	return "string";
    case STRINGLIST:
	return "string list";
    default:
	return "unknown";
    }
}

int
Config::ValidateConfigFile(const char *filename)
{
    int success = 1;
    
    mDictValPtrIterator<mString, Attribute> classIter(*Attribute::classAttributeDictionary);

    while (classIter.next()) {
	// The !#@% mDictValPtrIterator doesn't have an API for accessing
	// the value!
	Attribute *a = Attribute::GetClassAttributeByName(classIter.current());

	if (a == NULL) {
	    Error(EL_ABORT, "Config::ValidateConfigFile",
		  "Internal dictionary consistency failure.", NULL);
	}
	
	FieldData *fd = Config::GetDefaultByName(a->RefName());

	if (fd == NULL) {
	    Error(EL_ERROR, "Config::ValidateConfigFile",
		  "Your configuration file\n",
		  "\t", filename, "\n",
		  "\tdoes not provide a default value for the attribute\n",
		  "\t", a->RefName(), ".\n",
		  "\tA default value for this attribute is required, either\n",
		  "\tbecause it is implicitly used by Sonic X, or because it\n",
		  "\tis specified in your attributes file.", NULL);
	    success = 0;
	    continue;
	}
	
	if (fd->Tag() != a->Tag()) {
	    Error(EL_ERROR, "Config::ValidateConfigFile",
		  "Your configuration file\n",
		  "\t", filename, "\n",
		  "\tspecifies a type for the attribute\n",
		  "\t", a->RefName(), " of ", MakeTagName(fd->Tag()), ".\n",
		  "\tThis attribute must have a type of ",
		  MakeTagName(a->Tag()), "\n",
		  "\teither because it is implicitly used by Sonic X, or\n",
		  "\tbecause it is specified with this type in your \n",
		  "\tattributes file.", NULL);
	    success = 0;
	    continue;
	}
    }

    mDictValPtrIterator<mString, Attribute> appIter(*Attribute::appAttributeDictionary);

    while (appIter.next()) {
	// The !#@% mDictValPtrIterator doesn't have an API for accessing
	// the value!
	Attribute *a = Attribute::GetAppAttributeByName(appIter.current());

	if (a == NULL) {
	    Error(EL_ABORT, "Config::ValidateConfigFile",
		  "Internal dictionary consistency failure.", NULL);
	}
	
	FieldData *fd = Config::GetDefaultByName(a->RefName());

	if (fd == NULL) {
	    Error(EL_ERROR, "Config::ValidateConfigFile",
		  "Your configuration file\n",
		  "\t", filename, "\n",
		  "\tdoes not provide ta default value for the attribute\n",
		  "\t", a->RefName(), ".\n",
		  "\tA default value for this attribute is required, either\n",
		  "\tbecause it is implicitly used by Sonic X, or because it\n",
		  "\tis specified in your attributes file.", NULL);
	    success = 0;
	    continue;
	}
	
	if (fd->Tag() != a->Tag()) {
	    Error(EL_FATAL, "Config::ValidateConfigFile",
		  "Your configuration file\n",
		  "\t", filename, "\n",
		  "\tspecifies a type for the attribute\n",
		  "\t", a->RefName(), " of ", MakeTagName(fd->Tag()), ".\n",
		  "\tThis attribute must have a type of ",
		  MakeTagName(a->Tag()), "\n",
		  "\teither because it is implicitly used by Sonic X, or\n",
		  "\tbecause it is specified with this type in your\n",
		  "\tattributes file.", NULL);
	    success = 0;
	    continue;
	}
    }

    return success;
}
	
int
Config::ParseConfigFile(const char *filename)
{
    FILE *configFile = NULL;

    if (Config::validDict)
	Config::EmptyDict();
#if 0
    Config::InstallImplicitAttributes();
#endif
    if (filename == NULL) {
	Error(EL_ERROR, "Config::ParseConfigFile",
	      "No filename specified.", NULL);
	return 0;
    }

    if ((configFile = fopen(filename, "r")) == NULL) {
	Error(EL_ERROR, "Config::ParseConfigFile", "couldn't open file ",
	      filename, ": ", Perror(), NULL);
	return 0;
    }

    // configin is the yyin for the config parser.
    configin = configFile;

    configLineno = 1;
    
    // configparse is the yyparse for the config parser.
    if (configparse() != 0) {
	Error(EL_ERROR, "Config::ParseConfigFile", "Parser failed for ",
	      filename, NULL);
	fclose(configin);
	return 0;
    }
    // ok, we need to close up the config file
    fclose(configin);

    // Make sure we got config entries for all of the attribute prototypes
    // specified in the attribs database.
    if (!Config::ValidateConfigFile(filename)) {
	Error(EL_FATAL, "Config::ParseConfigFile",
	      "Because of the errors reported above, Sonic X\n",
	      "\tcannot continue.  The system must have a consistent set of\n",
	      "\tconfiguration settings.  Probably you have added attributes\n",
	      "\tto your mercator.attributes file without adding the\n",
	      "\tdefaults to your mercator.config file, or you have removed\n",
	      "\tor changed an entry in the config file that Sonic X\n",
	      "\trequires.  Exiting...\n", NULL);
    }
    
    Config::validDict = 1;
    return 1;
}


FieldData *
Config::GetDefaultByName(const char *n)
{
    static FieldData fieldData(BOOLEAN);	// tag doesn't matter here.
    mString IndexName(n);
    FieldData *fd;

    // search through the ConfigDictionary first

    fd = configDictionary->lookup(IndexName);
    
    if (fd != NULL) {
	fieldData = *fd;
	return &fieldData;
    } else {
	return NULL;
    }

    return &fieldData;
}

/* performs a lookup to see if a varibable exists */
int
Config::ClearDictionaryEntry(const char *id)
{
    FieldData *fdPtr;

    fdPtr = configDictionary->lookup(id);
    if (fdPtr) {
	delete fdPtr;
	return 1;
    }
  
    return 0;
}

void
Config::InstallBooleanDecl(int l, const char *id, unsigned char b)
{
    FieldData *newFD = new FieldData(BOOLEAN);
    newFD->Boolean(b);
    Config::ClearDictionaryEntry(id);
    configDictionary->insert(id, newFD);
}

void
Config::InstallIntegerDecl(int l, const char *id, long i)
{
    FieldData *newFD = new FieldData(INTEGER);
    newFD->Integer(i);
    Config::ClearDictionaryEntry(id);
    configDictionary->insert(id, newFD);
}

void
Config::InstallStringDecl(int l, const char *id, const char *str)
{
    FieldData *newFD = new FieldData(STRING);
    newFD->String(str);
    Config::ClearDictionaryEntry(id);
    configDictionary->insert(id, newFD);
}

void
Config::InstallIdentifierDecl(int l, const char *id, const char *str)
{
    Config::InstallStringDecl(l, id, str);
}

void
Config::InstallStringListDecl(int l, const char *id,
			      const mListVal<mString>& sl)
{
    FieldData *newFD = new FieldData(STRINGLIST);
    newFD->StringList(sl);
    Config::ClearDictionaryEntry(id);
    configDictionary->insert(id, newFD);
}

void
Config::InstallProcDecl(int l, const char *id, const char *str)
{
    // Houston, we have a problem.  The FieldData type expects to know
    // what it's simple type so it can set it's value correctly after returning
    // from the procedure.  I initally tried to hack the FieldData code
    // so that it would always pass the type into FieldData::InvokeProc.
    // However, the Interp code manipulates FieldData instances and
    // has it's own way of calling procedures - expecting the instance to
    // know its own type!  Solution: Check the attribute dictionary for type.
    // If not found, then check implicitly loaded Config defaults!
 
    FieldData *searchFD;
    FieldType installtype = INTEGER;
    Attribute *A;
    mString ms;
    int success = 0;

    // iterator to search the attribute Class Dictionary
    mDictValPtrIterator<mString, Attribute> classDictIter(*Attribute::classAttributeDictionary);

    // iterator to search the attribute App Dictionary
    mDictValPtrIterator<mString, Attribute> appDictIter(*Attribute::appAttributeDictionary);


    while (classDictIter.next()) {
	ms = classDictIter.current();
	A = (Attribute::classAttributeDictionary)->lookup(ms);
	if (0 == strcmp(id, A->RefName()) ) {
	    installtype = A->Tag();
	    success = 1;
	    break;  // get out while
	}
    }
       
    if (!success)
	while (appDictIter.next()) {
	    ms = appDictIter.current();
	    A = (Attribute::appAttributeDictionary)->lookup(ms);
	    if (0 == strcmp(id, A->RefName()) ) {
		installtype = A->Tag();
		success = 1;
		break;  // get out while
	    }
	}	   
       
    // try seeing if the attribute is a standard-implicitly loaded one
    if (!success) {
	searchFD = configDictionary->lookup(id);
	if (searchFD != NULL) {
	    installtype = searchFD->Tag();
	    success = 1;
	}
    }

    if (!success) {
	Error(EL_WARNING, "Config::InstallProc ", "Attribute ", id,
              " does not appear in attribute prototype file.",
              " skipping!");
	return;
    }

    // I originally had newFD declared at the top, and then tried
    // to reset the tag.  However, FieldData::Tag() set "initialized"
    // to true and then SetProc() calls FreeOldData().  Core dump!
    // I need to investigate if FData::Tag() really has a need to
    // set the instance to true.  For now, we'll use this:

    FieldData *newFD = new FieldData(installtype);

    newFD->SetProc(str);
    Config::ClearDictionaryEntry(id);
    configDictionary->insert(id, newFD);
}


// GET VALUE METHODS -----------------------------------------
unsigned char
Config::GetBoolean(char *id, const Client& obj)
{
    FieldData *fdTmp = configDictionary->lookup(id);
    
    if (fdTmp == NULL)
	return -1;
    
    return(fdTmp->Boolean(obj));
}

unsigned char
Config::GetBoolean(char *id, const XtObject& obj)
{
    FieldData *fdTmp = configDictionary->lookup(id);
    
    if (fdTmp == NULL)
	return -1;
    
    return(fdTmp->Boolean(obj));
}

unsigned char
Config::GetBoolean(char *id)
{
    FieldData *fdTmp = configDictionary->lookup(id);
    
    if (fdTmp == NULL)
	return -1;
    
    return(fdTmp->Boolean());
}



long
Config::GetInteger(char *id, const Client& obj)
{
    FieldData *fdTmp = configDictionary->lookup(id);
    
    if (fdTmp == NULL)
	return -1;
    
    return(fdTmp->Integer(obj));
}

long
Config::GetInteger(char *id, const XtObject& obj)
{
    FieldData *fdTmp = configDictionary->lookup(id);
    
    if (fdTmp == NULL)
	return -1;
    
    return(fdTmp->Integer(obj));
}


long
Config::GetInteger(char *id)
{
    FieldData *fdTmp = configDictionary->lookup(id);
    
    if (fdTmp == NULL)
	return -1;
    
    return(fdTmp->Integer());
}

const char *
Config::GetString(char *id, const Client& obj)
{
    FieldData *fdTmp = configDictionary->lookup(id);
    
    if (fdTmp == NULL)
	return NULL;
    
    return(fdTmp->String(obj));
}

const char *Config::GetString(char *id, const XtObject& obj)
{
    FieldData *fdTmp = configDictionary->lookup(id);
    
    if (fdTmp == NULL)
	return NULL;
    
    return(fdTmp->String(obj));
}

const char *
Config::GetString(char *id)
{
    FieldData *fdTmp = configDictionary->lookup(id);
    
    if (fdTmp == NULL)
	return NULL;
    
    return(fdTmp->String());
}



const mListVal<mString>&
Config::GetStringList(char *id, const Client& obj)
{
    FieldData *fd = configDictionary->lookup(id);

    if (fd == NULL)
	return *new mListVal<mString>;	// empty list -- MEMORY LEAK!
    
    return(fd->StringList(obj));
}

const mListVal<mString>&
Config::GetStringList(char *id, const XtObject& obj)
{
    FieldData *fd = configDictionary->lookup(id);

    if (fd == NULL)
	return *new mListVal<mString>;	// empty list -- MEMORY LEAK!
    
    return(fd->StringList(obj));
}

const mListVal<mString>&
Config::GetStringList(char *id)
{
    FieldData *fd = configDictionary->lookup(id);

    if (fd == NULL)
	return *new mListVal<mString>; 	// empty list -- MEMORY LEAK
    
    return(fd->StringList());
}

void
Config::EmptyDict()
{
    Config::configDictionary->kill_all();
    delete Config::configDictionary;

    Config::configDictionary =
	new mDictValPtr<mString, FieldData>(StringHasher);
              
    Config::validDict = 0;
}

void
Config::FreeConfig()
{
    Config::EmptyDict();
    
    delete Config::configDictionary;

    Config::validDict = 0;
}
