// -*-C++-*-
//
//	Note:  The use of this program shall be restricted to educational,
//	non-commercial use only.  All inquiries regarding licensing this
//	software for commercial use may be addressed to:
//	
//	Georgia Tech Research Corporation
//	Technology Licensing
//	Centennial Research Building - Rm. 275
//	400 Tenth Street, N.W.
//	Atlanta, Georgia  30332-0415
//	
//	*************************************************
//	
//	       Copyright 1996
//	Georgia Tech Research Corporation
//	    All Rights Reserved
//
//
//
// FontCache provides a mechanism for associating font ids with XFontStructs.
// We probably want to keep one FontCache per client, since font IDs will
// vary between clients.
//
// This stuff is only needed by ProtoTextRep.  Perhaps the client should reserve
// a slot for a pointer to a FontCache, but not install it unless one of its
// XtObjects requires a ProtoTextRep.
//
// Here's how the thing works:
//	- FontCache is basically a wrapper around a dictionary which maps from
//	  font IDs to XFontStruct pointers.
//	- The primary use is to keep a client-side cache of XFontStructs for
//	  performance reasons (these are used by XTextWidth and
//	  XTextExtents).
//	- You can directly add and remove mappings to and from the FontCache
//	  via Add and Remove.  Get will retrieve an association.  If you pass
//	  an XFontStruct to Add, the data is stored in the dictionary and will
//	  be freed when the dictionary goes away (you should not free it
//	  after adding it!)  This matches the common usage of XQueryFont.
//	- The easier API is to simply call TextWidth and TextExtents on this
//	  object.  It will use the cached XFontStructs if they've been seen before;
//	  otherwise it will generate protocol to fill the cache.
//	- TextExtents and TextWidth return -1 on failure.
//	- Note that Get "rolls over" to XQueryFont.
//
//	keith edwards	10aug95
//

#ifndef _FONT_CACHE_H_
#define _FONT_CACHE_H_

#include <X11/Xlib.h>
#include "mDictValPtr.h"


class FontCache {
public:
    FontCache(Display *dpy);
    virtual ~FontCache();

    // Adding a font gives the memory to the dict.  It'll free it.
    void		Add(XFontStruct *);
    void		Remove(Font);
    XFontStruct		*Get(Font) const;

    // First is 8-bit version, second is 16-bit version.
    int			TextWidth(Font fid, const char *s) const;
    int			TextWidth(Font fid, const XChar2b *s) const;
    
    // First is 8-bit version, second is 16-bit version.
    int			TextExtents(Font fid, const char *s,
				    int& directionRet,
				    int& fontAscentRet,
				    int& fontDescentRet,
				    XCharStruct& overallRet) const;
    int			TextExtents(Font fid, const XChar2b *s,
				    int& directionRet,
				    int& fontAscentRet,
				    int& fontDescentRet,
				    XCharStruct& overallRet) const;
    static unsigned	FontHasher(const Font& fid);
    
private:
    FontCache();					// disallowed
    FontCache(const FontCache& other);			// disallowed
    FontCache& operator=(const FontCache& other);	// disallowed

    Display		*dpy;
    mDictValPtr<Font, XFontStruct>	fonts;
};

inline FontCache::FontCache(Display* dpy_) :
    dpy(dpy_), fonts(FontCache::FontHasher) { }
inline void FontCache::Add(XFontStruct *f) { if (f) fonts.insert(f->fid, f); }
inline unsigned FontCache::FontHasher(const Font& fid) { return (unsigned) fid;}

#endif /* _FONT_CACHE_H_ */
