//
//
//
//	Note:  The use of this program shall be restricted to educational,
//	non-commercial use only.  All inquiries regarding licensing this
//	software for commercial use may be addressed to:
//	
//	Georgia Tech Research Corporation
//	Technology Licensing
//	Centennial Research Building - Rm. 275
//	400 Tenth Street, N.W.
//	Atlanta, Georgia  30332-0415
//	
//	*************************************************
//	
//	       Copyright 1996
//	Georgia Tech Research Corporation
//	    All Rights Reserved
//
//
//
//
//
//	Loadable module for Genovations Keypad.
//
//	keith edwards 11jul95
//

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdlib.h>
#include <unistd.h>
#include <termio.h>

#include "Genovations.h"
#include "Loadable.h"
#include "Error.h"
#include "Util.h"
#include "Action.h"

LOADABLE_CLASS_DEFN(Genovations);

//
// This is fairly screwy.  The Genovations keypad generates key codes in
// the range 65 - 121, with gaps.  I keep a big table that's indexed by the
// keycode - 65 and produces a KeyMapEntry, which contains a character and
// the numlock state which correspond to that keycode.  If the character is
// zero, this indicates that the requested keycode was bogus.
//

Genovations::KeyMapEntry Genovations::keymap[] = {
    { 'n',	0 },			// index 0 (65) - numlock, OFF
    { 0,	0 },			// index 1 (66)
    { 0,	0 },			// index 2 (67)
    { 0,	0 },			// index 3 (68)
    { 0,	0 },			// index 4 (69)
    { 0,	0 },			// index 5 (70)
    { 0,	0 },			// index 6 (71)
    { 0,	0 },			// index 7 (72)
    { 0,	0 },			// index 8 (73)
    { 0,	0 },			// index 9 (74)
    { 0,	0 },			// index 10 (75)
    { 0,	0 },			// index 11 (76)
    { 0,	0 },			// index 12 (77)
    { 0,	0 },			// index 13 (78)
    { 0,	0 },			// index 14 (79)
    { '1',	1 },			// index 15 (80) - end/1, ON
    { '2',	1 },			// index 16 (81) - down/2, ON
    { '3',	1 },			// index 17 (82) - pagedown/3, ON
    { '4',	1 },			// index 18 (83) - left/4, ON
    { '5',	1 },			// index 19 (84) - 5, ON
    { '6',	1 },			// index 20 (85) - right/6, ON
    { '7',	1 },			// index 21 (86) - home/7, ON
    { '8',	1 },			// index 22 (87) - up/8, ON
    { '9',	1 },			// index 23 (88) - pageup/9, ON
    { 0,	0 },			// index 24 (89)
    { '/',	1 },			// index 25 (90) - divide, ON
    { '*',	1 },			// index 26 (91) - multiply, ON
    { '-',	1 },			// index 27 (92) - minus, ON
    { '+',	1 },			// index 28 (93) - plus, ON
    { '.',	1 },			// index 29 (94) - period, ON
    { '0',	1 },			// index 30 (95) - insert/0, ON
    { '1',	0 },			// index 31 (96) - end/1, OFF
    { '2',	0 },			// index 32 (97) - down/2, OFF
    { '3',	0 },			// index 33 (98) - pagedown/3, OFF
    { '4',	0 },			// index 34 (99) - left/4, OFF
    { '5',	0 },			// index 35 (100) - 5, OFF
    { '6',	0 },			// index 36 (101) - right/6, OFF
    { '7',	0 },			// index 37 (102) - home/7, OFF
    { '8',	0 },			// index 38 (103) - up/8, OFF
    { '9',	0 },			// index 39 (104) - pageup/9, OFF
    { 0,	0 },			// index 40 (105)
    { '/',	0 },			// index 41 (106) - divide, OFF
    { '*',	0 },			// index 42 (107) - multiply, OFF
    { '-',	0 },			// index 43 (108) - minus, OFF
    { '+',	0 },			// index 44 (109) - plus, OFF
    { '.',	0 },			// index 45 (110) - period, OFF
    { '0',	0 },			// index 46 (111) - insert/0, OFF
    { 0,	0 },			// index 47 (112)
    { 0,	0 },			// index 48 (113)
    { 0,	0 },			// index 49 (114)
    { 0,	0 },			// index 50 (115)
    { 'e',	0 },			// index 51 (116) - enter, ON/OFF both
    { 0,	0 },			// index 52 (117)
    { 0,	0 },			// index 53 (118)
    { 0,	0 },			// index 54 (119)
    { 0,	0 },			// index 55 (120)
    { 'n',	0 },			// index 56 (121) - numlock, OFF
};

Genovations::Genovations() :
    FDInterest("Genovations"),
    device(NULL),
    fd(-1)
{
    char *d;

    // We should try the config file first...
    
    if ((d = getenv("GENOVATIONSDEV")) != NULL)
	device = Strdup(d);
    else
	device = Strdup("/dev/ttya");
    
    if ((fd = Genovations::OpenDevice(device)) == -1) {
	Error(EL_ERROR, "Genovations::Genovations", "Cannot open device ",
	      device, NULL);
    } else {
	FDInterest::AddFDInterest(fd);
	SetIOMask(fd, ARead);
    }
}

Genovations::~Genovations()
{
    Shutdown(fd);
    delete [] device;
}

int
Genovations::OpenDevice(const char *devName)
{
    int fd;
    struct termio ttyconfig;

    // open the device.  Should we be setting O_NONBLOCK???
    if ((fd = open(devName, O_RDONLY|O_NONBLOCK)) == -1) {
	Error(EL_ERROR, "Genovations::OpenDevice", "Error opening ",
	      devName, ": ", Perror(), NULL);
	return -1;
    }

    if (!isatty(fd)) {
	Error(EL_ERROR, "Genovations::OpenDevice", devName,
	      " is not a tty.", NULL);
	return -1;
    }

    // get current configuration
    if (ioctl(fd, TCGETA, &ttyconfig) == -1) {
	Error(EL_ERROR, "Genovations::OpenDevice",
	      "Error getting terminal configuration on ", devName, ": ",
	      Perror(), NULL);
	return -1;
    }

    // set baud rate to 1200
    ttyconfig.c_cflag &= ~(CBAUD|CIBAUD);   // clear baud rate bits
    ttyconfig.c_cflag |= B1200;

    // turn off many port defaults 
    ttyconfig.c_lflag &= ~(ICANON|ECHO);
    ttyconfig.c_iflag &= ~(BRKINT|INLCR|ICRNL|IUCLC|INPCK|ISTRIP);

    // set handshaking
    ttyconfig.c_cflag |= CRTSCTS;
    ttyconfig.c_iflag &= ~IXON;

    ttyconfig.c_oflag &= ~OPOST;  // No output post processing

    // this is the hack, parity, but 8 bit values
    // if this is suspect of a problem, set Parity to even
    ttyconfig.c_cflag |= PARENB;
    ttyconfig.c_cflag &= ~(CSIZE);
    ttyconfig.c_cflag |= (CS8|CREAD);

    // disable input buffering
    ttyconfig.c_cc[VMIN] = 0; 
    ttyconfig.c_cc[VTIME] = 0;
    
    if (ioctl(fd, TCSETA, &ttyconfig) == -1) {
	Error(EL_ERROR, "Genovations::OpenDevice",
	      "Error setting terminal configuration on ", devName, ": ",
	      Perror(), NULL);
	return -1;
    }

    return fd;
}

void
Genovations::KeyPressed(char key, int numLockState) const
{
    char debugBuffer[256];

    sprintf(debugBuffer, "Key pressed = %c, numlock = %d", key, numLockState);
    Error(EL_DEBUG, "Genovations::KeyPressed", debugBuffer, NULL);
    
    switch(key) {
        case '0':
            Action::Invoke("StopSpeaking", 0, NULL);
            break;
        case '1':
            Action::Invoke("ReadThisChar", 0, NULL);
            break;
        case '2':
            Action::Invoke("DownPressed", 0, NULL);
            break;
        case '3':
            Action::Invoke("ReadThisWord", 0, NULL);
            break;
        case '4':
            Action::Invoke("LeftPressed", 0, NULL);
            break;
        case '5':
            Action::Invoke("FivePressed", 0, NULL);
            break;
        case '6':
            Action::Invoke("RightPressed", 0, NULL);
            break;
        case '7':
            Action::Invoke("ReadThisSentence", 0, NULL);
            break;
        case '8':
            Action::Invoke("UpPressed", 0, NULL);
            break;
        case '9':
            Action::Invoke("ReadThisSentence", 0, NULL);
            break;
        case '.':
            Action::Invoke("ChangeTextMode", 0, NULL);
            break;
        case 'e':
            Action::Invoke("SelCurrent", 0, NULL);
            break;
        default:
            Error(EL_DEBUG, "Genovation::KeyPressed",
		  "Default case, key not used", NULL);
    }
}

int
Genovations::HandleActivity(int _fd, short type)
{
    if (type & ARead) {
	unsigned char byte;
	int result;

	// Reading one byte is inefficient but what the heck... people are slow.
	if ((result = read(_fd, &byte, 1)) == -1) {
	    Error(EL_ERROR, "Genovations::HandleActivity",
		  "Error reading from ", device, ": ", Perror(), NULL);
	    return 0;
	} else if (result != 1) {
	    Error(EL_ERROR, "Genovations::HandleActivity",
		  "Read wrong number of bytes (expected 1)", NULL);
	    return 0;
	} else {
	    int offset = byte - KEYMAP_START;

	    if ((offset < 0) || (offset > (KEYMAP_MAX - KEYMAP_START))) {
		Error(EL_ERROR, "Genovations::HandleActivity",
		      "Keycode out of range.", NULL);
		return 0;
	    }

	    KeyMapEntry *entry = &(Genovations::keymap[offset]);

	    if (entry->key == 0) {
		Error(EL_ERROR, "Genovations::HandleActivity",
		      "Unexpected keycode", NULL);
		return 0;
	    }

	    KeyPressed(entry->key, entry->numLockState);
	}
    } else if (type & AExcept) {
	Error(EL_ERROR, "Genovations::HandleActivity", "exception on device ",
	      device, "...shutting down.", NULL);
	Shutdown(fd);
	return 0;
    } else if (type & AWrite) {
	Error(EL_ERROR, "Genovations::HandleActivity", "Device ", device,
	      " is writable?!  Shutting down...", NULL);
	Shutdown(fd);
	return 0;
    } else {
	Error(EL_ERROR, "Genovations::HandleActivity", "unexpected condition "
	      "on device ", device, "...shutting down.", NULL);
	Shutdown(fd);
	return 0;
    }

    return 1;
}

void
Genovations::Shutdown(int _fd)
{
    if (_fd != fd) {
	Error(EL_ERROR, "Genovations::Shutdown",
	      "Requested file descriptor is not ours!?", NULL);
	return;
    }
    
    RemoveFDInterest(_fd);

    if (!close(_fd))
	Error(EL_ERROR, "Genovations::Shutdown", "Error closing ",
	      device, ": ", Perror(), NULL);

    MarkForCleanup();
}

