//
//	Note:  The use of this program shall be restricted to educational,
//	non-commercial use only.  All inquiries regarding licensing this
//	software for commercial use may be addressed to:
//	
//	Georgia Tech Research Corporation
//	Technology Licensing
//	Centennial Research Building - Rm. 275
//	400 Tenth Street, N.W.
//	Atlanta, Georgia  30332-0415
//	
//	*************************************************
//	
//	       Copyright 1996
//	Georgia Tech Research Corporation
//	    All Rights Reserved
//
//
//
// ProcessManager.cc
//
//	keith edwards	14jul95
//

#include <sys/wait.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <stropts.h>

#include "ProcessManager.h"
#include "Error.h"
#include "Util.h"
#include "mDictValPtr.h"

struct sigaction ProcessManager::action;
struct sigaction ProcessManager::oldAction;
int ProcessManager::pipeFds[2];
int ProcessManager::initialized = FALSE;

// Hash function for our pid map.  Not very good...
static inline unsigned PidHasher(const pid_t& pid) { return pid; }

// Destructor for ProcessNode
ProcessManager::ProcessNode::~_ProcessNode()
{
    for (register int i=0 ; i<argc ; i++)
	delete [] argv[i];

    delete [] argv;
    delete [] basename;
}

ProcessManager::ProcessManager() :
    FDInterest("ProcessManager"),
    processes(new mDictValPtr<pid_t, ProcessNode>(PidHasher)),
    shuttingDown(FALSE)
{
    if (initialized)
	Error(EL_ABORT, "ProcessManager::ProcessManager",
	      "Only one instance of a ProcessManager can exist in a program.",
	      NULL);

    pipeFds[0] = -1;
    pipeFds[1] = -1;
    
    initialized = TRUE;
    
    memset(&ProcessManager::action, 0, sizeof(struct sigaction));
    memset(&ProcessManager::oldAction, 0, sizeof(struct sigaction));
    
    if (pipe(pipeFds) < 0) {
	Error(EL_ERROR, "ProcessManager::ProcessManager",
	      "Could not open pipes: ", Perror(), NULL);
	MarkForCleanup();
    } else {
	AddFDInterest(pipeFds[1]);
	SetIOMask(pipeFds[1], ARead);
    }

    // Install reaper as our SIGCHLD handler.
#ifdef BEAM
    action.sa_handler = (void (*)(int))ProcessManager::Reaper;
#else
    action.sa_handler = ProcessManager::Reaper;
#endif

    if (sigaction(SIGCHLD, &ProcessManager::action,
		  &ProcessManager::oldAction) == -1) {
	Error(EL_ERROR, "ProcessManager::ProcessManager",
	      "sigaction failed: ", Perror(), NULL);
	Shutdown(-1);
	MarkForCleanup();
    }
#ifdef SUNOS_5_5
    if ((ProcessManager::oldAction.sa_handler) != (void(*)(int)) SIG_DFL)
#else
    if ((ProcessManager::oldAction.sa_handler) != (void(*)()) SIG_DFL)
#endif /* SUNOS_5_5 */
	Error(EL_WARNING, "ProcessManager::ProcessManager",
	      "Overwrote SIGCHLD signal handler.", NULL);
}

//
// Here's a closure to delete all process nodes.
//
#ifdef BEAM
class DeleteProcNodeWalk : public mDictValPtrClosure<pid_t, ProcessManager::ProcessNode> {
#else
class DeleteProcNodeWalk : public mDictValPtrClosure<pid_t, ProcessNode> {
#endif
public:
#ifdef BEAM
    int action(pid_t pid, ProcessManager::ProcessNode *node) {
#else
    int action(pid_t pid, ProcessNode *node) {
#endif
	delete node;
	return 1;		// continue
    }
};

ProcessManager::~ProcessManager()
{
    Shutdown(-1);		// arg is don't care
    
    if ((pipeFds[0] != -1) && (close(pipeFds[0]) == -1))
	Error(EL_ERROR, "ProcessManager::~ProcessManager",
	      "close failed:  ", Perror(), NULL);
    
    if ((pipeFds[1] != -1) && (close(pipeFds[1]) == -1))
	Error(EL_ERROR, "ProcessManager::~ProcessManager",
	      "close failed:  ", Perror(), NULL);

    DeleteProcNodeWalk closure;

    processes->walk(closure, NULL);
    
#if 0
    // Free process nodes ...
    mDictValPtrIterator<pid_t, ProcessNode> iter(*processes);

    while (iter.next()) {
	ProcessNode *node = processes->lookup(iter.current());
	if (node)
	    processes->del_item(iter.current());
    }
#endif
    
    delete processes;

    if (sigaction(SIGCHLD, &ProcessManager::oldAction, NULL) == -1)
	Error(EL_ERROR, "ProcessManager::~ProcessManager",
	      "sigaction failed: ", Perror(), NULL);
}

void
#ifdef SUNOS_5_5
ProcessManager::Reaper(int dummy)
#else
ProcessManager::Reaper()
#endif /* SUNOS_5_5 */
{
    pid_t pid;
    int status;

    while (pid = waitpid((pid_t) -1, &status, WNOHANG)) {
	if (pid < 0) {
	    if (errno == EINTR)
		continue;

	    if (errno == ECHILD)  /* no child process left */
		break;
	    
	    Error(EL_ERROR, "ProcessManager::Reaper", "waidpid failed: ",
		  Perror(), NULL);
	} else {
	    // write the pid and status into the pipe
	    if (write(pipeFds[0], &pid, sizeof(pid)) != sizeof(pid_t)) {
		Error(EL_ERROR, "ProcessManager::Reaper",
		      "Error writing pid: ", Perror(), NULL);
		return;
	    }
	    if (write(pipeFds[0], &status, sizeof(status)) != sizeof(status)) {
		Error(EL_ERROR, "ProcessManager::Reaper",
		      "Error writing pid: ", Perror(), NULL);
		return;
	    }
	}
    }
}

pid_t
ProcessManager::StartProcess(int argc, char * const argv[], char * const envp[],
			     int killOnShutdown,
			     ProcessManager::UnexpectedDeathProc unexpected,
			     ProcessManager::BeforeDeathProc avant,
			     ProcessManager::AfterDeathProc apres)
{
    pid_t pid;

    // SVR4 fork1 behaves like vfork?
#ifdef BEAM
    if ((pid = vfork()) == 0) {
#else
    if ((pid = fork1()) == 0) {
#endif
	// child - exec new image

	if (envp)
	    execve(argv[0], argv, envp);
	else
	    execv(argv[0], argv);

	//
	// If exec fails, call _exit() rather than exit() to keep from
	// corrupting the parent's environment.
	//
	Error(EL_ERROR, "ProcessManager::StartProcess", "exec failed for ",
	      argv[0], ": ", Perror(), NULL);
	
	_exit(1);
	
    } else if (pid > 0){
	// parent - install in processes dict
	ProcessNode *newNode = new ProcessNode;
	newNode->argc = argc;

	newNode->argv = new char*[argc];
	for (register int i = 0 ; i<argc ; i++)
	    newNode->argv[i] = Strdup(argv[i]);

	newNode->basename = Strdup(argv[0]);		// for now
	
	newNode->killOnShutdown = killOnShutdown;

	newNode->flaggedForKill = 0;

	newNode->unexpected = unexpected;
	newNode->avant = avant;
	newNode->apres = apres;

	processes->insert(pid, newNode);
    } else {
	// error - complain
	Error(EL_ERROR, "ProcessManager::StartProcess", "vfork failed for ",
	      argv[0], ": ", Perror(), NULL);
    }

    return pid;
}

int
ProcessManager::StopProcess(pid_t process)
{
    ProcessNode *node = processes->lookup(process);

    if (node == NULL) {
	Error(EL_WARNING, "ProcessManager::StopProcess",
	      "Process is not under control of ProcessManager", NULL);
    } else {
	// Mark this one as flagged-for-kill
	node->flaggedForKill = TRUE;

	// Run the avant-death function, if there is one
	if (node->avant)
	    (*node->avant)(process);
    }
    
    return SignalProcess(process, SIGTERM);
}

int
ProcessManager::SignalProcess(pid_t process, int sig)
{
    ProcessNode *node = processes->lookup(process);

    if (node == NULL)
	Error(EL_WARNING, "ProcessManager::StopProcess",
	      "Process is not under control of ProcessManager", NULL);
    
    if (kill(process, sig) < 0) {
	if (node && node->basename) 
	    Error(EL_ERROR, "ProcessManager::SignalProcess",
		  "kill failed for ", node->basename, ": ", Perror(), NULL);
	else
	    Error(EL_ERROR, "ProcessManager::SignalProcess",
		  "kill failed: ", Perror(), NULL);
	
	return 0;
    }

    return 1;
}

int
ProcessManager::HandleActivity(int fd, short mask)
{
    if (fd != pipeFds[1]) {
	Error(EL_ERROR, "ProcessManager::HandleActivity",
	      "unexpected descriptor, shutting down...", NULL);
	Shutdown(-1);
	MarkForCleanup();
	return 0;
    }

    if (mask & ARead) {
	pid_t pid;
	int status;
	int bytes = 0;
	int result;

	// See how much data is available, we don't want to hang.  We'll
	// wait until a full pid_t + status is available.
	if (ioctl(pipeFds[1], I_NREAD, &bytes) == -1) {
	    Error(EL_ERROR, "ProcessManager::HandleActivity",
		  "ioctl failed: ", Perror(), NULL);
	    return 0;
	}

	if (bytes < (sizeof(pid) + sizeof(status))) {
	    char buf[1024];
	    Error(EL_ERROR, "ProcessManager::HandleActivity",
		  "Bogus read on process manager pipe.", NULL);
	    // consume and trash all waiting data...
	    while (bytes > 0) {
		read(pipeFds[1], buf, (bytes <= 1024) ? bytes : 1024);
		bytes -= 1024;
	    }
	    
	    return 0;
	}
	
	// Read pid and status, clean up data structures, report exit
	while ((result = read(pipeFds[1], &pid, sizeof(pid))) < 0) {
	    if (errno == EINTR)
		continue;
	    else {
		Error(EL_ERROR, "ProcessManager::HandleActivity",
		      "Read of pid failed: ", Perror(), NULL);
		return 0;
	    }
	}

	if (result != sizeof(pid)) {
	    Error(EL_ERROR, "ProcessManager::HandleActivity",
		  "Read unexpected number of bytes", NULL);
	    return 0;
	}
	
	while ((result = read(pipeFds[1], &status, sizeof(status))) < 0) {
	    if (errno == EINTR)
		continue;
	    else {
		Error(EL_ERROR, "ProcessManager::HandleActivity",
		      "Read of status failed: ", Perror(), NULL);
		return 0;
	    }
	}

	if (result != sizeof(pid)) {
	    Error(EL_ERROR, "ProcessManager::HandleActivity",
		  "Read unexpected number of bytes", NULL);
	    return 0;
	}

	// Look it up.
	ProcessNode *node = processes->lookup(pid);

	if (!node) {
	    // It's a child that was not managed by us.  This is ok, just print
	    // a debugging message.
	    Error(EL_DEBUG, "ProcessManager::HandleActivity",
		  "A child process NOT under the control of ProcessManager "
		  "has terminated.", NULL);
	    return 1;
	}

	//
	// See if the death was caused by us.  If it was, run the apres-death
	// function if it was.  If it was not, then run the unexpected
	// death proc.
	//
	if (node->flaggedForKill) {
	    if (node->apres)
		(*node->apres)(pid);
	} else {
	    if (node->unexpected)
		(*node->unexpected)(pid);
	}

	// Remove the data from the process dictionary
	processes->del_item(pid);
	
	return 1;
    } else {
	// Error - mark for cleanup
	Error(EL_ERROR, "ProcessManager::HandleActivity",
	      "Unexpected condition on pipe, shutting down...", NULL);
	Shutdown(-1);
	MarkForCleanup();
	return 0;
    }
}

void
ProcessManager::Shutdown(int fd)
{
    if (shuttingDown)
	return;
    
    // iterate through all children and kill the ones with killOnShutdown flagged.
    mDictValPtrIterator<pid_t, ProcessNode> iter(*processes);

    while (iter.next()) {
	pid_t pid = iter.current();
	ProcessNode *node = processes->lookup(pid);
	
	if (!node)
	    continue;

	if (node->killOnShutdown) {
	    Error(EL_DEBUG, "ProcessManager::Shutdown",
		  "Killing ", node->basename, NULL);
	    StopProcess(pid);
	}
    }

    shuttingDown = TRUE;
}

			      
