//
//	Note:  The use of this program shall be restricted to educational,
//	non-commercial use only.  All inquiries regarding licensing this
//	software for commercial use may be addressed to:
//	
//	Georgia Tech Research Corporation
//	Technology Licensing
//	Centennial Research Building - Rm. 275
//	400 Tenth Street, N.W.
//	Atlanta, Georgia  30332-0415
//	
//	*************************************************
//	
//	       Copyright 1996
//	Georgia Tech Research Corporation
//	    All Rights Reserved
//
//
//
//	keith edwards	10aug95
//

#include <exception.h>

#include "XtObject.h"
#include "Client.h"
#include "Error.h"
#include "FontCache.h"
#include "XServer.h"
#include "RAP.h"
#include "Audio.h"
#include "Mercator.h"
#include "ProtoTextRep.h"
#include "ScreenReader.h"
#include "Template.h"

#ifndef MIN
#define MIN(a,b) ((a) < (b) ? (a) : (b))
#endif /* MIN */

//
// The Line representation is inefficient for several reasons:
//	- We dup all incoming strings (probably not necessary).
//	- We don't use a sparse representation for attributes.
//
// This constructor creates a new line containing spaces for the width
// of the text rep it's contained in, except for the amount of text specified
// by str starting at basex and rendered in the specified gc.
//
// Note that we don't check for compatible sizes here; we assume that
// the ProtoTextRep will do it for us by calling VerifyGC before we're
// called.
//
ProtoTextRep::Line::Line(ProtoTextRep& rep, int basex, int basey, int uppery,
			 const char *str, int stringLength, GCValue& gcvalues) :
    textRep(&rep),
    data(),
    baselineX(basex),
    baselineY(basey),
    upperY(uppery)
{
    //
    // Note some weirdage here.  Text may not necessarily get drawn at
    // the starting coordinate.  So Whenever we see new text appearing,
    // we have to calculate how many "blanks" are at the start and end,
    // based on window width and margins.
    //
    // The starting column is based on the X offset, minus the left margin,
    // divided by the size of a character.  NOTE that this only supports
    // fixed-width fonts reliably!
    //
    // Alloc one more char than needed for a NULL...
    //
    int startingColumn = textRep->PixelToColumn(basex);
    char *chars = new char[textRep->Columns() + 1];

    // Set everything to space.
    memset(chars, ' ', textRep->Columns());

    // If the drawn text would overrun the line, clip it.
    int extent = stringLength;
    if (startingColumn + stringLength > textRep->Columns()) {
	extent = textRep->Columns() - startingColumn;
    }
	    
    // Overwrite the "real" data.
    memcpy((chars + startingColumn), str, extent);
    // Tag on our NULL
    chars[textRep->Columns()] = NULL;
    
    TextAttr *attrs = new TextAttr[textRep->Columns()];	// default gcv
    
    //    data.Length(MIN(stringLength, textRep->Columns()));
    data.Length(textRep->Columns());
    data.Chars(chars);

    for (register i=0 ; i<textRep->Columns() ; i++)
	attrs[i].SetGCValue(gcvalues);		// reset to real gcv

    data.Attrs(attrs);
}

ProtoTextRep::Line::Line(ProtoTextRep& r, const TextData& d,
			 int bx, int by, int uy) :
    textRep(&r),
    data(),
    baselineX(bx),
    baselineY(by),
    upperY(uy)
{
    //
    // Note some weirdage here.  Text may not necessarily get drawn at
    // the starting coordinate.  So Whenever we see new text appearing,
    // we have to calculate how many "blanks" are at the start and end,
    // based on window width and margins.
    //
    // The starting column is based on the X offset, minus the left margin,
    // divided by the size of a character.  NOTE that this only supports
    // fixed-width fonts reliably!
    //
    // Alloc one more char than needed for a NULL...
    //
    int startingColumn = textRep->PixelToColumn(baselineX);
    char *chars = new char[textRep->Columns() + 1];

    // Set everything to space and default GCV
    memset(chars, ' ', textRep->Columns());
    TextAttr *attrs = new TextAttr[textRep->Columns()];

    // Only do all this junk if length != 0.  We often insert blank text datas with
    // length == 0, so this saves us some time.
    if (d.Length() != 0) {
	// If the drawn text would overrun the line, clip it.
	int extent = d.Length();
	if (startingColumn + d.Length() > textRep->Columns()) {
	    extent = textRep->Columns() - startingColumn;
	}
	    
	// Overwrite the "real" data.
	//
	// BOGUS??
	memcpy((chars + startingColumn), d.Chars(), extent);

	// Overwrite the "real" attrs
	for (register int i=startingColumn ; i<extent ; i++) {
	    attrs[i] = d.Attrs()[i - startingColumn];
	}
    }
    
    // Tag on our NULL onto the characters
    chars[textRep->Columns()] = NULL;

    //
    // BOGUS??
    //    memcpy((attrs + (startingColumn * sizeof(TextAttr))), d.Attrs(),
    //	   (extent * sizeof(TextAttr)));
    
    //    data.Length(MIN(stringLength, textRep->Columns()));
    data.Length(textRep->Columns());
    data.Chars(chars);
    data.Attrs(attrs);
}

//
// Merge updates an existing line with some new text.  Only the text specified
// at the given baseline X is replaced.
//
int
ProtoTextRep::Line::MergeData(const TextData& newData, int x)
{
    if (!textRep) {
	Error(EL_ERROR, "ProtoTextRep::Line::MergeData",
	      "Destination line is unparented.", NULL);
	return 0;
    }

    // Figure out the starting and ending columns of the replacement text.
    FontCache &fc = textRep->xobj.GetClient()->GetFontCache();

    // Arrrgs... we determine text width via the first GCV.
    const GCValue& gcv = newData.Attrs()[0].GetGCValue();

    // Determine boundaries.  Note we don't allow the line to grow beyond
    // it's current length.
    int startingColumn = textRep->PixelToColumn(x);
    int textWidth = fc.TextWidth(gcv.GetFontID(), newData.Chars());
    int endingColumn = MIN(textRep->PixelToColumn(x + textWidth),
			   data.Length());
    
    // Overwrite the sequence between {startingColumn, endingColumn} in
    // our TextData.
    for (register i=startingColumn ; i<endingColumn ; i++) {
	data.Chars()[i] = newData.Chars()[i - startingColumn];
	data.Attrs()[i] = newData.Attrs()[i - startingColumn];
    }
    
    return 1;
}

int
ProtoTextRep::Line::Clear(short x, short width)
{
    // Figure out the starting and ending column that this area corresponds to.
    int startingColumn = textRep->PixelToColumn(x);
    int endingColumn = MIN(textRep->PixelToColumn(x + width), data.Length());

    // Overwrite this sequence with spaces.  Note we leave the gcv intact.
    char *chars = data;
    TextAttr *attrs = data;

    for (register i=startingColumn ; i<endingColumn ; i++)
	chars[i] = ' ';

    return 1;
}

///////////////////////////////////////////////////////////////////////////

ProtoTextRep::ProtoTextRep(XtObject& o, RAP& r) :
    TextRep(o),
    rap(r),
    maxChar(NULL),
    minInterlineSpace(-1),
    columns(0),
    xPixels(0), yPixels(0),
    rightMargin(0), leftMargin(0), topMargin(0), bottomMargin(0),
    ascent(-1),
    lines(NULL),
    numLines(0),
    allocated(0),
    appRow(-1),	// important!  These must be intialized to -1 for
    appCol(-1)	// ScreenReader::SyncEditToApp to work correctly!
{
}

ProtoTextRep::~ProtoTextRep()
{
    delete maxChar;

    for (register int i=0 ; i<numLines ; i++)
	delete lines[i];
    
    delete [] lines;
}

//
// VerifyGC is called each time one of the drawing routines executes.  It makes
// sure that the GC is in our model, and returns it.
//
GCValue *
ProtoTextRep::VerifyGC(GContext gid) const
{
    Display *dpy = mercator->GetXServer()->GetDisplay();
    GCValue *gcv = rap.GetGCDict()->lookup(gid);

    // Lookup the GC...
    if (gcv == NULL) {
	char tmp[32];
	sprintf(tmp, "%lx", gid);
	Error(EL_ERROR, "ProtoTextRep::VerifyGC",
	      "GID ", tmp, " is not in our model!  This indicates a\n"
	      "\tproblem with the GC handling code.  Likely this TextRep\n"
	      "\twill not work...", NULL);
	return NULL;
    }

    return gcv;
}

//
// VerifyFont ensures that the font specified in the GCV is in the model.
// It also compares character boundaries, and sets maxChar if it is not
// set.
//
// The specific tasks it performs are:
//	- If maxChar is NULL, we haven't been drawn to before.  Get the
//	  GCValue from our GCCache for this GID, update the font cache,
//	  and set MaxChar.
//	- If maxChar is non-NULL, we get the new GCValue from our cache,
//	  get the font, and compare MaxChar.  They must match for us to
//	  continue.
//
// NOTE that efficiency can be greatly improved here.  We're requerying
// the font cache everytime we enter this code.  We should keep the
// current font id in the ProtoTextRep class and, if it matches, simply
// bail out.
//
XFontStruct *
ProtoTextRep::VerifyFont(const FontCache& fc, const GCValue& gcv)
{
    Font fid = gcv.GetFontID();
    XFontStruct *font = NULL;
	
    if ((font = fc.Get(fid)) == NULL) {
	//
	// This is a MOST HEINOUS bit of code.  Refer to O'Reilly, Vol II
	// page 350.  XQueryFont takes a font id and returns an
	// XFontStruct ONLY if we're not using the default font (at
	// least this is my interpretation of it).  If this fails, you can
	// pass in the GContext and it'll look up the fid from it and exhibit
	// the correct behavior.
	//
	// NOTE that this only gets done here; not in the FontCache or
	// anyplace else where it could be done... problem?
	//
	if ((font = fc.Get(gcv.GetGID())) == NULL) {
	    char tmp[32];
	    sprintf(tmp, "%lx", fid);
	    Error(EL_ERROR, "ProtoTextRep::VerifyFont",
		  "Font ", tmp, " is not in our model!  This indicates a\n"
		  "\tproblem with the font/GC code.  Likely this TextRep\n"
		  "\twill not work...", NULL);
	    return NULL;
	}
    }

    if (!maxChar) {
	//
	// We've (newly) got a valid MaxChar.  Set it. 
	//
	maxChar = new XCharStruct(font->max_bounds);
	//
	// This is our first maxChar, so refresh our size.
	//
	RefreshSize();
    } else {
	//
	// We had an existing maxChar.  We need to compare against it and
	// ensure that the drawing is compatible.
	//
	if ((maxChar->width != font->max_bounds.width) ||
	    (maxChar->ascent != font->max_bounds.ascent) ||
	    (maxChar->descent != font->max_bounds.descent)) {
	    
	    Error(EL_ERROR, "ProtoTextRep::VerifyFont",
		  "Font size has changed during rendering, not supported.",
		  NULL);
	    
	    delete maxChar;
	    maxChar = NULL;
	    
	    return NULL;
	}
    }

    return font;
}

//
// VerifyBaseline is called whenever a new line of text is rendered to the screen.
// It checks the baseline of the text being drawn, and uses the GC it is being
// rendered with to determine the spacing/height/etc.  of the text.
//
// We do not allow the baseline to be between the baseline and height of any
// existing line (note that we DO allow the baseline to match exactly -- this
// indicates replacement of an existing line, as opposed to interspersed text.
//
// Further, the height of all lines must be the same.
//
int
ProtoTextRep::VerifyBaseline(int baselineY, XFontStruct& font) const
{
    int upperY = baselineY - font.max_bounds.ascent;

    // Verify that the height (ascent) matches for all fonts.  If this->ascent
    // is -1, this is the first text drawn (fake out constness).
    if (ascent == -1)
	((ProtoTextRep *) this)->ascent = font.max_bounds.ascent;
    else if (font.max_bounds.ascent != ascent) {
	Error(EL_ERROR, "ProtoTextRep::VerifyBaseline",
	      "Font ascent does not match model.", NULL);
	return FALSE;
    }

    // Now we know the upper and lower Y coordinates of the box containing
    // the text.  Walk through the (ordered) list and verify that we have no
    // overlaps.

    for (int position = 0; position < numLines ; position++) {
	Line *current = lines[position];
	
	if (current == NULL)
	    Error(EL_ABORT, "ProtoTextRep::VerifyBaseline",
		  "NULL line in model, cannot continue.", NULL);

	// See if the baseline matches this one's exactly.  If it does, then
	// we're seeing a redraw of an existing line.
	if (baselineY == current->BaselineY())
	    return TRUE;

	// See if the baseline falls between this one's baseline and upper Y.
	// If it does, then we're seeing interspersed text (which we don't allow).
	if ((baselineY > current->UpperY()) &&
	    (baselineY < current->BaselineY())){
	    
	    Error(EL_ERROR, "ProtoTextRep::VerifyBaseline",
		  "New text is interspersed with old; not allowed.", NULL);
	    return FALSE;
	}

	// Text is ordered by baselines.  So if we've gone past the baseline we're
	// looking for, we can break out.
	if (baselineY < current->UpperY())
	    break;
    }

    // If we've made it to here, we must be drawing a new line.  This is OK.
    return TRUE;
}

//
// InsertData takes a newly-formed line and inserts it into the model.  This
// code assumes that the appropriate verification has been done.
//
int
ProtoTextRep::InsertData(const TextData& data, int x, int y, int upperY)
{
    Line *previous = NULL;

    // Find where to insert.
    for (int position = 0 ; position < numLines ; position++) {
	Line *current = lines[position];

	if (current == NULL)
	    Error(EL_ABORT, "ProtoTextRep::InsertData",
		  "NULL line in model, cannot continue.", NULL);

	// If the baseline the data is rendered at matches an existing line,
	// we're overwriting it and must Merge the two lines together.
	if (y == current->BaselineY()) {
	    current->MergeData(data, x);
	    return TRUE;
	}

	// If the new line's baseline is less than the current line's upper Y,
	// and there is no previous line, we need to insert at the head of
	// the list.  If there's no previous and the above condition does not
	// hold, then we can skip out.
	if (!previous) {
	    if (y < current->UpperY()) {
		Line *line = new Line(*this, data, x, y, upperY);
		PrependLine(*line);

		// Reset minimum interline spacing if we have to.
		RecalculateLineSpacing(current->UpperY() - y);
		//RecalculateLineSpacing(current->UpperY() - y - maxChar->descent);
	    
		return TRUE;
	    } else {
		previous = current; // Bug fix --keith
		continue;
	    }
	}

	// If the new line's baseline falls between the previous line's baseline
	// and the current line's upper Y, then we need to insert the new line
	// between previous and current.
	if ((y > previous->BaselineY()) &&
	    (y < current->UpperY())) {
	    Line *line = new Line(*this, data, x, y, upperY);
	    InsertAfter(position-1, *line);

	    // Reset minimum interline spacing if we have to.
	    RecalculateLineSpacing(upperY - previous->BaselineY());
	    //RecalculateLineSpacing(upperY - previous->BaselineY() - maxChar->descent);
	
	    return TRUE;
	}

	// We haven't found where to place it this time through.  Reset
	// previous and keep going.
	previous = current;
    }

    // If we've gone through all the lines and still haven't found where to
    // insert the new guy, it must go at the end.  Append it.
    Line *line = new Line(*this, data, x, y, upperY);
    AppendLine(*line);
    
    // Reset minimum interline spacing if we have to.  Reset based on the
    // line before us.
    if (numLines > 1)
    	RecalculateLineSpacing(upperY - lines[numLines-2]->BaselineY());
	//RecalculateLineSpacing(upperY - lines[numLines-2]->BaselineY() - maxChar->descent);
	
    return TRUE;
}

//
// RecalculateLineSpacing() is called after a new line is inserted.  Its responsibility
// is to (1) update line numbering, (2) recompute the MINIMUM interline spacing
// currently present in the model, (3) insert blank lines if needed.
//
// If any blanks are inserted, causing the minimum interline spacing to change,
// the process is repeated until the model stabilizes.
//
int
ProtoTextRep::RecalculateLineSpacing(int difference)
{
    //    TextData blankData;
    
    // If the spacing has been initialized and difference is still larger than
    // the old minimum interline spacing, we can ignore it.
    if ((minInterlineSpace != -1) && (difference >= minInterlineSpace))
	return 1;

    minInterlineSpace = difference;

    // We will be called when the first text is drawn to a window (we get
    // maxChar and resize based on it).  If numLines is bogus, just bail...
    if (numLines <= 0)
	return 1;
    
    //
    // This is the amount of space required to store a line of text plus the
    // interline spacing.  We scan the lines in the text rep looking for
    // interline spacing greater or equal to this amount.  Whenever we see one
    // we call InsertData to put a blank line in.  NOTE that InsertData will
    // itself recalculate the minimum interline spacing and call us, so this
    // procedure may recurse if we keep whacking the line spacing.
    //
    int textPlusSpaceSize = maxChar->ascent + minInterlineSpace;
    //int textPlusSpaceSize = maxChar->ascent + maxChar->descent + minInterlineSpace;

    // Pad the start of the model with blanks, if needed.
    if (lines[0]->UpperY() >= textPlusSpaceSize) {
	// INSERT BLANKS AT THE FIRST
	int spaceToFill = lines[0]->UpperY() - topMargin;
	while (spaceToFill >= textPlusSpaceSize) {
	    TextData blankData;
	    int baselineY = spaceToFill - minInterlineSpace;
	    int upperY = baselineY - maxChar->ascent;
	    Line *blank = new Line(*this, blankData, leftMargin,
				   baselineY, upperY);

	    PrependLine(*blank);
	    spaceToFill -= textPlusSpaceSize;
	}
    }

    // Note that we start counting with 1 here.  We only need to do
    // between-line padding in this loop.  Start-padding and end-padding
    // are done separately.
    for (register int i=1 ; i<numLines ; i++) {
	Line *current = lines[i];

	// Check between two items
	if (current->UpperY() - lines[i-1]->BaselineY() >= textPlusSpaceSize) {
	    int spaceToFill = current->UpperY() - lines[i-1]->BaselineY();
	    while (spaceToFill >= textPlusSpaceSize) {
		TextData blankData;
		int baselineY = lines[i-1]->BaselineY() +
		    spaceToFill - minInterlineSpace;
		int upperY = baselineY - maxChar->ascent;
		Line *blank = new Line(*this, blankData, leftMargin,
				       baselineY, upperY);
	    
		InsertAfter(i-1, *blank);
		spaceToFill -= textPlusSpaceSize;
	    }
	}
    }

    // Now pad the end of the thing
    if (Height() - lines[numLines - 1]->BaselineY() >= textPlusSpaceSize) {
	int spaceToFill = Height() - lines[numLines - 1]->BaselineY()
	    - bottomMargin;
	while (spaceToFill >= textPlusSpaceSize + maxChar->descent) {
	    // note numLines will increment whenever we append a line.
	    // So this will place lines consecutively after one another.
	    TextData blankData;
	    int baselineY = lines[numLines - 1]->BaselineY()
		+ textPlusSpaceSize;
	    int upperY = baselineY - maxChar->ascent;
	    Line *blank = new Line(*this, blankData, leftMargin,
				   baselineY, upperY);

	    AppendLine(*blank);
	    spaceToFill -= textPlusSpaceSize;
	}
    }

    cerr << "RECALCULATE LINE SPACING:  height = " << xobj.Height() << ", spacePerLine = " << maxChar->ascent + minInterlineSpace << ", topMargin = " << topMargin << ", bottomMargin = " << bottomMargin << ", numLines = " << numLines << "\n" << flush;

    return TRUE;
}

void
ProtoTextRep::PrintContents(FILE *fp) const
{
    fprintf(fp, "ProtoTextRep: %s %s (%s) %d lines\n",
	    xobj.GetUniqueName(), xobj.GetLongName(), xobj.GetClass(),
	    numLines);
    
    for (register int i=0 ; i<numLines ; i++) {
	if (lines[i] == NULL) {
	    fprintf(fp, "%d: NULL LINE, VERY BOGUS...\n");
	} else {
	    fprintf(fp, "%d: Y=%d X=%d U=%d GC[0]=%ld |%s|\n", i,
		    lines[i]->BaselineY(), lines[i]->BaselineX(),
		    lines[i]->UpperY(),
		    lines[i]->Data().Attrs()[0].GetGCValue().GetGID(),
		    lines[i]->Data().Chars());
	}
    }

    fflush(fp);
}

TextData
ProtoTextRep::GetLine(int line) const throw(xmsg)
{
    if (line < 0 || line > (numLines-1))
	throw xmsg("Line index out of bounds.");
	    
    return lines[line]->Data();
}

TextData *
ProtoTextRep::GetData() const
{
    //
    // I wish there were some way to avoid the calls to the default ctor here...
    //
    TextData *ret = new TextData[numLines];

    for (register int i=0 ; i<numLines ; i++)
	ret[i] = lines[i]->Data();

    return ret;
}

//
// NOTE that GetSegment may return a segment of size < the requested
// length if there is not enough data.
//
TextData
ProtoTextRep::GetSegment(int line, int pos, int len) const
{
    int sizeSoFar = 0;
    TextData ret;

    for (register int i=0 ; i<numLines ; i++) {
	Line *line = lines[i];

	if (line == NULL)
	    Error(EL_ABORT, "ProtoTextRep::GetSegment",
		  "NULL line in model; cannot continue.", NULL);

	if ((sizeSoFar + line->Length()) < len) {
	    // If the new line won't surpass len, just glom the whole thing on.
	    ret += line->Data();
	    sizeSoFar += line->Length();
	} else {
	    // Else just add on a portion of it and bail.
	    ret += line->Data().Segment(0, (len-sizeSoFar));;
	    break;
	}
    }
    
    return ret;
}

int
ProtoTextRep::GetPixelCoords(int line, int pos, int& x, int& y) const
{
    if (!maxChar)
	return 0;

    if (line >= numLines)
	return 0;

    if (pos >= Columns())
	return 0;

    // if minInterlineSpace is -1, we've probably (hopefully?) only got one
    // line, so ignore it...
    if (minInterlineSpace == -1) {
	y = xobj.BorderWidth() + topMargin +
	    (line + 1) * maxChar->ascent;
    } else {
	y = xobj.BorderWidth() + topMargin +
	    ((line + 1) * (maxChar->ascent + minInterlineSpace));
    }

    cerr << "GETPIXELCOORDS:  line " << line << " is at " << y << "\n";
    x = xobj.BorderWidth() + leftMargin + (pos * maxChar->width);

    return 1;
}
    
void
ProtoTextRep::GetEditCursor(int& x, int& y) throw(xmsg)
{
    Error(EL_WARNING, "ProtoTextRep::GetEditCursor", "Not implemented", NULL);
    
    x = 0;
    y = 0;
}

void
ProtoTextRep::HandleClearArea(const xReq *request)
{
    const xClearAreaReq *req = (const xClearAreaReq *) request;
    short x = req->x, y = req->y, width = req->width, height = req->height;

    // The protocol specifies this behavior -- 0 is special case behavior.
    if (width == 0)
	width = xobj.Width() - x;

    if (height == 0)
	height = xobj.Height() - y;

    //
    // Search for baselines between y and y+height.
    // For each one, set the text between x and x+width to blank.
    //
    // What's the "correct" behavior if only part of the line is blanked out?
    // We take the conservative approach of saying that the *entire* line
    // must be cleared before WE clear it; we try to leave text in the model
    // as long as possible.
    //
    // Thus, we only run Clear() on lines that have a baseline and
    // upper Y completely within the cleared area.
    //
    for (register int i=0 ; i<numLines ; i++) {
	Line *current = lines[i];

	if (current == NULL)
	    Error(EL_ABORT, "ProtoTextRep::HandleClearArea",
		  "Blank line in model; cannot continue.", NULL);

	// Clear if needed
	if ((current->UpperY() >= y) && (current->BaselineY()) <= y+height)
	    current->Clear(x, width);

	// Bail as soon as we pass the cleared area completely.
	if (current->UpperY() > (y+height))
	    break;
    }

    if (GetTextRepDebug())
	GetTextRepDebug()->Update();
}

//
// CopyAreas are a big pain; we have to make some assumptions here that may
// not always be valid.
//
// HandleCopyArea() works by first identifying all of the lines present in the
// source block.  We make the assumption that if *any* part of a line is being
// copied, we will copy the entire line (we can't really have half-lines in our
// model anyway, so this is a simplifying assumption that will keep our model
// accurate in nearly all cases).
//
// Next, we identify the character positions (columns) within the source block
// that are affected.
//
// We then identify the destination lines that are affected.  Like the computation
// of source lines, if any portion of a line lies within the boundary area, we
// claim that the entire line is affected.
//
// If the baseline X coordinates of the copied lines, offset by the amount the
// line is moved, equal the baseline X coordinates of the destination lines, then
// the data is merged (this handles the case of horizontal scrolling, and
// "perfectly matched" vertical scrolling).
//
// This code could be made *much* faster.  I'm separating things out into
// different loops to keep the logic simpler, and to make things easier to
// test.
//
void
ProtoTextRep::HandleCopyArea(const xReq *request)
{
    const xCopyAreaReq *req = (const xCopyAreaReq *) request;
    short width = req->width, height = req->height;

    if (req->srcDrawable != req->dstDrawable) {
	Error(EL_WARNING, "ProtoTextRep::HandleCopyArea",
	      "Got a CopyArea with window IDs that are not equal!", NULL);
	return;
    }

    //
    // Note that xCopyAreaReq *combines* the source and destination
    // rectangles, based on the function operator specified in the GC.
    // We ignore the function and assume that everything is getting
    // overwritten (which is usually the case, with text anyway).
    //

    //
    // Construct the boundaries of the copy in character coordinate space.
    //
    int startSrcLine = PixelToRow(req->srcY);
    int startSrcColumn = PixelToColumn(req->srcX);
    int endSrcLine = PixelToRow(req->srcY + req->height);
    int endSrcColumn = PixelToColumn(req->srcX + req->width);

    int startDstLine = PixelToRow(req->dstY);
    int startDstColumn = PixelToColumn(req->dstX);
    int endDstLine = PixelToRow(req->dstY + req->height);
    int endDstColumn = PixelToColumn(req->dstX + req->width);

    // Sanity checking...
    if (startSrcLine >= startDstLine) {
	//
	// We're scrolling down, so text is copied up.  We must start at the
	// top of the source area to make sure we don't get overwrites while
	// copying.
	//
	int srcLine = startSrcLine;
	int dstLine = startDstLine;

	for ( ; srcLine <= endSrcLine ; srcLine++, dstLine++) {

	    if (srcLine >= numLines || srcLine < 0) continue;
	    if (dstLine >= numLines || dstLine < 0) continue;
	    
	    const char *srcData = lines[srcLine]->Data().Chars();
	    char *dstData = lines[dstLine]->Data().Chars();
	    const TextAttr *srcAttrs = lines[srcLine]->Data().Attrs();
	    TextAttr *dstAttrs = lines[dstLine]->Data().Attrs();
	    int srcCol = startSrcColumn;
	    int dstCol = startDstColumn;

	    for ( ; srcCol <= endSrcColumn ; srcCol++, dstCol++) {
		if (srcCol >= columns || srcCol < 0) continue;
		if (dstCol >= columns || dstCol < 0) continue;
		
		dstData[dstCol] = srcData[srcCol];
		dstAttrs[dstCol] = srcAttrs[srcCol];
	    }

	    // Update the baseline X coordinate for this line
	    lines[dstLine]->BaselineX(lines[srcLine]->BaselineX()
				      - req->height);
	}
    } else { 
	//
	// We're scrolling up, so text is copied down.  We must start at the
	// bottom of the source area to make sure we don't get overwrites while
	// copying.
	//
	int srcLine = endSrcLine;
	int dstLine = endDstLine;

	for ( ; srcLine >= startSrcLine ; srcLine--, dstLine--) {

	    if (srcLine >= numLines || srcLine < 0) continue;
	    if (dstLine >= numLines || dstLine < 0) continue;
	    
	    const char *srcData = lines[srcLine]->Data().Chars();
	    char *dstData = lines[dstLine]->Data().Chars();
	    const TextAttr *srcAttrs = lines[srcLine]->Data().Attrs();
	    TextAttr *dstAttrs = lines[dstLine]->Data().Attrs();
	    int srcCol = startSrcColumn;
	    int dstCol = startDstColumn;

	    for ( ; srcCol <= endSrcColumn ; srcCol++, dstCol++) {
		if (srcCol >= columns || srcCol < 0) continue;
		if (dstCol >= columns || dstCol < 0) continue;
		
		dstData[dstCol] = srcData[srcCol];
		dstAttrs[dstCol] = srcAttrs[srcCol];
	    }
	    
	    // Update the baseline X coordinate for this line
	    lines[dstLine]->BaselineX(lines[srcLine]->BaselineX()
				      + req->height);
	}
    }

    //
    // This is bogus, but I'm not sure what else to do in this case.
    // For some reason (probably a lack of understanding of the protocol
    // on my part), when we copy from source to dest lines, the source
    // lines seem to get cleared... at least we don't get any clearareas
    // or blanks being drawn or anything.
    //
    // So I'm going to make the whopper assumption that the behavior
    // that I'm seeing is an artifact of CopyArea and clear the lines in
    // the source region that don't overlap with the destination...
    //
    // I'm not very happy about this...
    //
    // NOTE - the code as it exists clears all cells in lines that are in the
    // source area but not in the destination area.  It does not handle the
    // cells in lines that are in the vertical overlap.  On these, the horizontal,
    // non-overlapping areas must be cleared.  Hence, this code does not
    // support non-vertical panning.
    //
    if (startSrcLine >= startDstLine) {
	// copy up
	if (startSrcColumn >= startDstColumn) {
	    // copy down/left
	    for (register int i=endSrcLine ; i>endDstLine ; i--) {
		if (i >= numLines || i < 0) continue;
		if (i >= numLines || i < 0) continue;
		for (register int j=0 ; j<lines[i]->Length() ; j++)
		    lines[i]->Data().Chars()[j] = ' ';
#if 0
		for (register int j=endSrcColumn ; j>endDstColumn ; j--) {
		    if (j >= columns || j < 0) continue;
		    if (j >= columns || j < 0) continue;
		    lines[i]->Data().Chars()[j] = ' ';
		}
#endif
	    }
	} else {
	    // copy down/right
	    for (register int i=endSrcLine ; i>endDstLine ; i--) {
		if (i >= numLines || i < 0) continue;
		if (i >= numLines || i < 0) continue;
		for (register int j=0 ; j<lines[i]->Length() ; j++)
		    lines[i]->Data().Chars()[j] = ' ';
#if 0
		for (register int j=startSrcColumn ; i<startDstColumn ; j++) {
		    if (j >= columns || j < 0) continue;
		    if (j >= columns || j < 0) continue;
		    lines[i]->Data().Chars()[j] = ' ';
		}
#endif
	    }
	}
    } else {
	// copy down
	if (startSrcColumn >= startDstColumn) {
	    // copy down/left
	    for (register int i=startSrcLine ; i<startDstLine ; i++) {
		if (i >= numLines || i < 0) continue;
		if (i >= numLines || i < 0) continue;
		for (register int j=0 ; j<lines[i]->Length() ; j++)
		    lines[i]->Data().Chars()[j] = ' ';
#if 0
		for (register int j=endSrcColumn ; j>endDstColumn ; j--) {
		    if (j >= columns || j < 0) continue;
		    if (j >= columns || j < 0) continue;
		    lines[i]->Data().Chars()[j] = ' ';
		}
#endif
	    }
	} else {
	    // copy down/right
	    for (register int i=startSrcLine ; i<startDstLine ; i++) {
		if (i >= numLines || i < 0) continue;
		if (i >= numLines || i < 0) continue;
		for (register int j=0 ; j<lines[i]->Length() ; j++)
		    lines[i]->Data().Chars()[j] = ' ';
#if 0
		for (register int j=startSrcColumn ; i<startDstColumn ; j++) {
		    if (j >= columns || j < 0) continue;
		    if (j >= columns || j < 0) continue;
		    lines[i]->Data().Chars()[j] = ' ';
		}
#endif
	    }
	}
    }
    // What should we pass in here?
    //    RecalculateLineSpacing();
    
    if (GetTextRepDebug())
	GetTextRepDebug()->Update();
}

void
ProtoTextRep::HandleImageText8(const xReq *request)
{
    const xImageText8Req *req = (const xImageText8Req *) request;
    
    GCValue *gcv = VerifyGC(req->gc);
    
    if (!gcv) {
	Error(EL_ERROR, "ProtoTextRep::HandleImageText8",
	      "GC IS NOT IN OUR MODEL!", NULL);
	return;
    }

    FontCache& fc = xobj.GetClient()->GetFontCache();
    XFontStruct *font = VerifyFont(fc, *gcv);
    
    if (!font) {
	Error(EL_ERROR, "ProtoTextRep::HandleImageText8",
	      "FONT IS NOT IN OUR MODEL!", NULL);
	return;
    } 
    
    if (!VerifyBaseline(req->y, *font)) {
	Error(EL_ERROR, "ProtoTextRep::HandleImageText8",
	      "Line dimensions will not fit our model.", NULL);
	return;
    }

    // Create a new chunk of text, represented by a TextData.
    char *chars = new char[req->nChars + 1];
    memcpy(chars, ((const char *) req) + sz_xImageText8Req, req->nChars);
    chars[req->nChars] = '\0';
    TextAttr *attrs = new TextAttr[req->nChars];
    for (register int i=0 ; i<req->nChars ; i++)
	attrs[i].SetGCValue(*gcv);
    
    TextData data(chars, attrs, req->nChars);

    // And place it in the model.  We pass x, y, and upper y.
    InsertData(data, req->x, req->y, req->y - font->max_bounds.ascent);
    
    if (GetTextRepDebug())
	GetTextRepDebug()->Update();
}

void
ProtoTextRep::HandleImageText16(const xReq *req)
{
    Error(EL_ERROR, "ProtoTextRep::HandleImageText16",
	  "16-bit text is unsupported!", NULL);

    if (GetTextRepDebug())
	GetTextRepDebug()->Update();
}

//
// Here's the request structure used by PolyText8 and PolyText16.
//
//	typedef struct {
//	       CARD8 reqType;
//	       CARD8 pad;
//	       CARD16 length B16;
//	       Drawable drawable B32;
//	       GContext gc B32;
//	       INT16 x B16, y B16
//	} xPolyTextReq
//
//	- items (xTextElt) start after struct
//
//	- nonaligned, big-endian font ID follows this struct
//
//	typedef struct {		// followed by string
//	       CARD8 len;		// number of characters in string,
//				// or FontChange (255) for font change, or
//				// 0 if just delta given
//	       INT8 delta;
//	} xTextElt;
//
void
ProtoTextRep::HandlePolyText8(const xReq *request)
{
    const xPolyText8Req *req = (const xPolyText8Req *) request;
    int length, x;
    char *str;
    
    GCValue *gcv = VerifyGC(req->gc);
    
    if (!gcv) {
	Error(EL_ERROR, "ProtoTextRep::HandlePolyText8",
	      "GC IS NOT IN OUR MODEL!", NULL);
	return;
    }

    FontCache& fc = xobj.GetClient()->GetFontCache();
    XFontStruct *font = VerifyFont(fc, *gcv);
    
    if (!font) {
	Error(EL_ERROR, "ProtoTextRep::HandlePolyText8",
	      "FONT IS NOT IN OUR MODEL!", NULL);
	return;
    } 
    
    if (!VerifyBaseline(req->y, *font)) {
	Error(EL_ERROR, "ProtoTextRep::HandlePolyText8",
	      "Line dimensions will not fit our model.", NULL);
	return;
    }

    length = (req->length * 4) - sz_xPolyText8Req;
    // Str points to the byte after the end of the fixed-length portion.
    str = ((char *) req) + sz_xPolyText8Req;

    x = req->x;

    while (length > 3) {
	// Get the number of chars in the string.
	int size = (unsigned char) *str++;

	length -= 1;		// eat the size;

	// Is it a font change?
	if (size == 255) {
	    Font oldFont = gcv->GetFontID();
	    union {
		Font newFont;
		char arr[4];
	    } f;
	    
	    // Explicit byte order -- this is off the wire.  Change for
	    // little endian CPUs
	    f.arr[0] = *str++;
	    f.arr[1] = *str++;
	    f.arr[2] = *str++;
	    f.arr[3] = *str++;

	    gcv->SetFontID(f.newFont);
	    
	    length -= 4;	// eat the font id

	    // Not quite sure what we should do here...
	    if ((font = VerifyFont(fc, *gcv)) == NULL) {
		Error(EL_ERROR, "ProtoTextRep::HandlePolyText8",
		      "FONT IS NOT IN OUR MODEL!", NULL);
		return;
	    } 

	    continue;
	}

	// Normal text drawing.  The next byte contains the delta.
	int delta = (unsigned char)  *str++;

	length -= 1;		// eat the delta

	x += delta;

	// This check seems to be necessary.  XmText seems to generate
	// lots of empty xPolyText8Reqs
	
	if (size != 0) {
	    // Allocate chars...
	    char *chars = new char[size + 1];
	    memcpy(chars, str, size);
	    chars[size] = '\0';

	    // Allocate attrs...
	    TextAttr *attrs = new TextAttr[size];
	    for (register int i=0 ; i<size ; i++)
		attrs[i].SetGCValue(*gcv);

	    // Create a text segment
	    TextData data(chars, attrs, size);

	    // And place it in the model.  We pass x, y, and upper y.
	    InsertData(data, x, req->y, req->y - font->max_bounds.ascent);
	    
	    // Update X after string is drawn?
	    x += fc.TextWidth(gcv->GetFontID(), chars);
	    
	    // Update length...
	    length -= size;
	}
    }
    
    if (GetTextRepDebug())
	GetTextRepDebug()->Update();
}

void
ProtoTextRep::HandlePolyText16(const xReq *req)
{
    Error(EL_ERROR, "ProtoTextRep::HandlePolyText16",
	  "16-bit text is unsupported!", NULL);

    if (GetTextRepDebug())
	GetTextRepDebug()->Update();
}


int
ProtoTextRep::PixelToColumn(short x) const throw(xmsg)
{
    if (!maxChar)
	throw xmsg("Uninitialized ProtoTextRep");

    // To produce the column we subtract the border width and left-margin,
    // then divide by the width of a character.
    //return (x - xobj.BorderWidth() - leftMargin) / maxChar->width;
    int tmp = (x - leftMargin) / maxChar->width;

    if (tmp >= 0)
	return tmp;
    else
	return 0;
}

int
ProtoTextRep::PixelToRow(short y) const throw(xmsg)
{
    if (!maxChar)
	throw xmsg("Uninitialized ProtoTextRep");
    
    // This is weird.  If minInterlineSpace is uninitialized, we only have one line
    // so they *must* be on line 0, regardless of y position.
    if (minInterlineSpace == -1)
	return 0;

    int spacePerLine = maxChar->ascent + minInterlineSpace;
    //int spacePerLine = maxChar->ascent + + maxChar->descent + minInterlineSpace;
    //return (y - xobj.BorderWidth() - topMargin) / spacePerLine;
    int tmp = (y - topMargin) / spacePerLine;

    if (tmp >= 0)
	return tmp;
    else
	return 0;
}

void
ProtoTextRep::RefreshSize()
{
    int newLeftMargin, newRightMargin, newTopMargin, newBottomMargin;
    
    // If no text has been drawn, we can't convert to rows/columns.  Bail.
    if (!maxChar)
	return;

    // Make sure our margin information is current
    try {
	newLeftMargin = Template::GetInteger(xobj, "leftMargin");
	newRightMargin = Template::GetInteger(xobj, "rightMargin");
	newTopMargin = Template::GetInteger(xobj, "topMargin");
	newBottomMargin = Template::GetInteger(xobj, "bottomMargin");
    }
    catch (xmsg& message) {
	Error(EL_ERROR, "ProtoTextRep::RefreshSize",
	      "Error determining margins: ", message.why(), NULL);
	newLeftMargin = leftMargin;
	newRightMargin = rightMargin;
	newTopMargin = topMargin;
	newBottomMargin = bottomMargin;
    }
    
    // Has width (number of columns) changed?
    if (xPixels != xobj.Width() ||
	newLeftMargin != leftMargin ||
	newRightMargin != rightMargin) {
	
	leftMargin = newLeftMargin;
	rightMargin = newRightMargin;
	xPixels = xobj.Width();
	
	// Don't count room required for right margin as a column...
	//int newNumColumns = PixelToColumn(objectWidth - rightMargin);
	int newNumColumns = PixelToColumn(xobj.Width() - rightMargin);
	
	HandleHorizontalResize(newNumColumns);
	// numColumns is reset as a side effect of HorizontalResize.
    }

    // Has height (number of rows) changed?
    if (yPixels != xobj.Height() ||
	newTopMargin != topMargin ||
	newBottomMargin != bottomMargin) {

	topMargin = newTopMargin;
	bottomMargin = newBottomMargin;
	yPixels = xobj.Height();
	
	// We can only compute lines once text is drawn.  Don't count room
	// required for bottom margin as a row.
	if (minInterlineSpace != -1)
	    HandleVerticalResize(PixelToRow(xobj.Height() - bottomMargin));

	// numLines is reset as a side effect of VerticalResize...
    }
    
    if (GetTextRepDebug())
	GetTextRepDebug()->Update();
}

//
// The old number of columns is represented by 'columns', the new is represented
// by the input parameter, 'newColumns.'  We reallocate all the lines in the model
// (expensive).  If columns < newColumns, we reassign, padding with blank space.
// If columns > newColumns, we clip by the new size.
//
// Number of columns is reset as a side effect.
//
void
ProtoTextRep::HandleHorizontalResize(int newColumns)
{
    int oldColumns = columns;
    Line **newLines;
    
    if (columns == newColumns)
	return;

    columns = newColumns;

    if (numLines == 0)
	return;

    newLines = new Line*[numLines];

    // Reallocate all lines to be the new size.
    for (register int i=0 ; i<numLines ; i++) {
	if (lines[i] == NULL)
	    Error(EL_ABORT, "ProtoTextRep::HandleHorizontalResize",
		  "NULL line in model; this should never happen.", NULL);

	newLines[i] = new Line(*this,
			       lines[i]->Data(),
			       lines[i]->BaselineX(),
			       lines[i]->BaselineY(),
			       lines[i]->UpperY());
	delete lines[i];
    }

    lines = newLines;
}

//
//  numLines will be changed as a side effect.
//
void
ProtoTextRep::HandleVerticalResize(int newNumLines)
{
    int oldLines = numLines;

    // No change?
    if (oldLines == newNumLines)
	return;

    // Uninitialized?
    if (maxChar == NULL)
	return;

    if (oldLines < newNumLines) {
	Line **newLines = new Line*[newNumLines];
	register int i;

	// copy the old lines...
	for (i=0 ; i<oldLines ; i++)
	    newLines[i] = lines[i];

	// and pad with blanks...
	int textPlusSpaceSize = maxChar->ascent + minInterlineSpace;
	int baselineY = lines[oldLines - 1]->BaselineY() + textPlusSpaceSize;
	
	for (i=oldLines ; i<newNumLines ; i++) {
	    TextData blankData;
	    int upperY = baselineY - maxChar->ascent;
	    
	    Line *blank = new Line(*this, blankData, leftMargin,
				   baselineY, upperY);
	    newLines[i] = blank;
	    baselineY += textPlusSpaceSize;
	}
	
	delete [] lines;
	lines = newLines;
    } else {
	// If we shrink, reset numLines and NULL-out any extra lines.
	for (register int i=newNumLines ; i<numLines ; i++) {
	    delete lines[i];
	    lines[i] = NULL;
	}
    }
    
    numLines = newNumLines;
}

void
ProtoTextRep::AllocLines(int newSize)
{
    if (newSize == numLines)
	return;

    Line **newLines = new Line*[newSize];
    int min = MIN(numLines, newSize);

    if (lines) {
	for (register int i=0 ; i<min ; i++)
	    newLines[i] = lines[i];

	delete [] lines;
    }

    numLines = newSize;
    lines = newLines;
}

void
ProtoTextRep::PrependLine(Line& line)
{
    int newSize = numLines + 1;
    Line **newLines = new Line*[newSize];

    if (lines) {
	for (register int i=0 ; i<numLines ; i++)
	    newLines[i+1] = lines[i];

	delete [] lines;
    }

    newLines[0] = &line;
    
    numLines = newSize;
    lines = newLines;
}

void
ProtoTextRep::AppendLine(Line& line)
{
    AllocLines(numLines + 1);
    lines[numLines - 1] = &line;
}

void
ProtoTextRep::InsertAfter(int l, Line& line)
{
    int newSize = numLines + 1;
    Line **newLines = new Line*[newSize];

    if (!lines || !numLines) {
	Error(EL_ERROR, "ProtoTextRep::InsertAfter",
	      "Uninitialized textrep.", NULL);
	return;
    }

    if (lines) {
	for (register int i=0 ; i<numLines ; i++) {
	    if (i < l) {
		newLines[i] = lines[i];
	    } else if (i == l) {
		newLines[i] = lines[i];
		newLines[i+1] = &line;
	    } else /* i > l */ {
		newLines[i+1] = lines[i];
	    }
	}

	delete [] lines;
    }
    
    numLines = newSize;
    lines = newLines;
}

int
ProtoTextRep::SetAppCursor(int row, int col)
{
    if ((appRow == row) && (appCol == col))
	return 1;
    
    // Optionally invoke an action whenever the app cursor moves...only
    // do this if the cursorFeedback attribute is true.
    ScreenReader *s;
    s = xobj.GetScreenReader();
    if (s->GetOperationMode() == ScreenReader::CURSOR_EDIT)
	s->SetCursorPosition(col, row);
	
    appRow = row;
    appCol = col;

    const FieldData *data = Template::FindByName(*rap.GetClient(),
						 "cursorFeedback");

    if (data && (data->Tag() == BOOLEAN) && data->Boolean(*rap.GetClient())) {
	Action::Invoke("CursorFeedback", 0, NULL);

	const FieldData *snd = Template::FindByName(*rap.GetClient(),
						    "cursorSound");

	//
	// We could just put the sound in the action, but I'm trying to make
	// this as fast as possible.
	//
	if (snd && (snd->Tag() == STRING)) {
	    const char *file = snd->String(*rap.GetClient());
	    float rate = 1.0;

	    if (file)
		mercator->GetAudio()->PlaySound(snd->String(*rap.GetClient()),
						&rate, 100, 0, 0);
	}
    }
    
    return 1;
}

void
ProtoTextRep::GetAppCursor(int& row, int& col)
{
    row = appRow;
    col = appCol;
}

int
ProtoTextRep::ValidateCursorSize(int width, int height)
{
    if (!maxChar)
	return 0;

    return ((maxChar->width == width) &&
	    ((maxChar->ascent + maxChar->descent) == height));
}
