// -*-C++-*-
//
//	Note:  The use of this program shall be restricted to educational,
//	non-commercial use only.  All inquiries regarding licensing this
//	software for commercial use may be addressed to:
//	
//	Georgia Tech Research Corporation
//	Technology Licensing
//	Centennial Research Building - Rm. 275
//	400 Tenth Street, N.W.
//	Atlanta, Georgia  30332-0415
//	
//	*************************************************
//	
//	       Copyright 1996
//	Georgia Tech Research Corporation
//	    All Rights Reserved
//
//
//
// ProtoTextRep is derived from TextRep and provides a model for on-screen
// text built from protocol monitoring.
//
// Caveats:
// - This class currently only supports 8-bit text.
// - This class only supports fixed-width text.
// - No  changes to character width within a line are allowed.
// - Interline spacing (detection of blank lines) is not robust.
//
//	keith edwards	10aug95
//

#ifndef _PROTO_TEXTREP_H_
#define _PROTO_TEXTREP_H_

#include "TextRep.h"

class XtObject;
class RAP;
class xmsg;

class ProtoTextRep : public TextRep {
private:
    ProtoTextRep();					// disallowed
    ProtoTextRep(const ProtoTextRep& other);		// disallowed
    ProtoTextRep& operator=(const ProtoTextRep& other);	// disallowed

    class Line {
    private:
	TextData	data;
	ProtoTextRep	*textRep;
	int		baselineX;		// only used for merging
	int 		baselineY;
	int		upperY;

    public:
	Line();
	Line(const Line& other);
	Line(ProtoTextRep& rep, const TextData& data,
	     int baseX, int baseY, int upperY);
	Line(ProtoTextRep& rep, int baseX, int baseY, int upperY,
	     const char *str, int strLen, GCValue& gcv);
	
	Line& operator=(const Line& other);

	TextData	Data() const;
	TextData&	Data();			// breaks ref-counting of data
	void		Data(const TextData data);

	int		BaselineY() const;
	void		BaselineY(int baselineY);

	int		BaselineX() const;
	void		BaselineX(int baselineXY);

	int		UpperY() const;
	void		UpperY(int upperY);

	// Draw new data over an existing line.
	int		MergeData(const TextData& line, int x);
	// "Blankify" region between pixel X and X + width
	int		Clear(short x, short width);

	int		Length() const;
	TextData	Segment(int start, int length);
    };

    friend class Line;
    
    RAP&		rap;

    //
    // The ProtoTextRep requires all text to be in a fixed width font.
    // Further, multiple fixed-width fonts cannot be mixed if they
    // have different sizes.
    //
    // This XCharStruct represents the max_char for the font we're
    // using.
    //
    XCharStruct		*maxChar;
    //
    // minInterlineSpace represents the smallest between-line spacing
    // that we've seen.  This is used for calculation of number of rows and
    // for blank line determination.
    //
    int			minInterlineSpace;

    int			columns;
    int			xPixels, yPixels;
    int			leftMargin, rightMargin, topMargin, bottomMargin;
    int			ascent;

    Line		**lines;
    int			numLines, allocated;

    // Our best guestimate of where the application cursor is.
    int			appRow, appCol;
    
    //
    // Helper methods
    //
    GCValue		*VerifyGC(GContext gid) const;
    XFontStruct		*VerifyFont(const FontCache&, const GCValue&);
    int			VerifyBaseline(int baselineY, XFontStruct& font) const;
    
    int			InsertData(const TextData& data, int x, int y, int upY);
    int			RecalculateLineSpacing(int newSpacing);
    void		AllocLines(int newSize);
    void		PrependLine(Line& line);
    void		AppendLine(Line& line);
    void		InsertAfter(int num, Line& line);

    void		HandleHorizontalResize(int newNumColumns);
    void		HandleVerticalResize(int newNumLines);

public:
    ProtoTextRep(XtObject& obj, RAP& rap);
    virtual ~ProtoTextRep();

    virtual void	PrintContents(FILE *fp) const;

    virtual int		Height() const;
    virtual int		Rows() const;
    virtual int		Width() const;
    virtual int		Columns() const;
    
    virtual void	RefreshSize();

    virtual TextData	GetLine(int line) const throw(xmsg);
    virtual TextData   *GetData() const;
    virtual TextData	GetSegment(int line, int pos, int len) const;

    virtual int		GetPixelCoords(int line, int pos,
				       int& x, int& y) const;
    
    virtual void	GetEditCursor(int& x, int& y) throw(xmsg);

    //
    // These are our input methods, used by RAP to set text data.
    //
    void		HandleClearArea(const xReq *);
    void		HandleCopyArea(const xReq *);
    void		HandleImageText8(const xReq *);
    void		HandleImageText16(const xReq *);
    void		HandlePolyText8(const xReq *);
    void		HandlePolyText16(const xReq*);

    // Convert between pixel units and character units.
    int			PixelToRow(short y) const throw(xmsg);
    int			PixelToColumn(short x) const throw(xmsg);

    // Called by RAP whenever it detects a change in the app cursor
    // position.
    int			SetAppCursor(int row, int col);
    void		GetAppCursor(int& row, int& col);
    // See if the size of our rectangle is character-sized (bogus...)
    int			ValidateCursorSize(int width, int height);
};

inline ProtoTextRep::Line::Line() :
    textRep(NULL), data(), baselineX(-1), baselineY(-1), upperY(-1) { }
inline ProtoTextRep::Line::Line(const Line& other) :
    textRep(other.textRep), data(other.data), baselineY(other.baselineY),
    baselineX(other.baselineX), upperY(other.upperY) { }
inline int ProtoTextRep::Line::Length() const { return data.Length(); }

inline TextData ProtoTextRep::Line::Data() const { return data; }
inline TextData& ProtoTextRep::Line::Data() { return data; }
inline void ProtoTextRep::Line::Data(const TextData d) { data = d; }
inline int ProtoTextRep::Line::BaselineY() const { return baselineY; }
inline void ProtoTextRep::Line::BaselineY(int n) { baselineY = n; }
inline int ProtoTextRep::Line::BaselineX() const { return baselineX; }
inline void ProtoTextRep::Line::BaselineX(int n) { baselineX = n; }
inline int ProtoTextRep::Line::UpperY() const { return upperY; }
inline void ProtoTextRep::Line::UpperY(int n) { upperY = n; }

inline int ProtoTextRep::Height() const { return yPixels; }
inline int ProtoTextRep::Rows() const { return numLines; }
inline int ProtoTextRep::Width() const { return xPixels; }
inline int ProtoTextRep::Columns() const { return columns; }

#endif /* _PROTO_TEXT_REP_H_ */

