//
//
//
//	Note:  The use of this program shall be restricted to educational,
//	non-commercial use only.  All inquiries regarding licensing this
//	software for commercial use may be addressed to:
//	
//	Georgia Tech Research Corporation
//	Technology Licensing
//	Centennial Research Building - Rm. 275
//	400 Tenth Street, N.W.
//	Atlanta, Georgia  30332-0415
//	
//	*************************************************
//	
//	       Copyright 1996
//	Georgia Tech Research Corporation
//	    All Rights Reserved
//
//
//
//
//
// RAP implementation - one of these is created per client-connection.  The
// RAP object is actually what causes clients to be created (it calls out to
// AppModelMgr to perform this task when a connection to a client has been
// fully established.
//
//
//		     ! ! !   B E W A R E   ! ! !
//
// Don't even *THINK* of changina so much as a *line* of this code unless
// you're absolutely, positively sure that you know what you're doing.  If
// you do change anything, you should test the change rigorously.
//
// This code represents the bowels of Mercator and has been hardened
// over the years to have reasonable behavior in the face of all sorts of
// ill-behaved applications.  There are lots of special cases, order dependencies,
// boundary-conditions, and the like which are embodied in this code.  If
// you change things on a whim it's likely that you'll break the world.
//
// I've tried to put in comments to explain the assumptions that I've made
// to try to make the behavior of the code as explicit as possible.  I'm sure
// there's still lots of weirdage here though.
//
// Good luck!
//				keith edwards
//				keith@cc.gatech.edu
//

#include <exception.h>		// for xmsg
#include <sys/time.h>		// for gettimeofday()
#include <poll.h>

#include <X11/StringDefs.h>

#include "RapProto.h"		// Protocol definition

#include "RAP.h"
#include "AppModelMgr.h"
#include "Client.h"
#include "Template.h"
#include "Util.h"
#include "XtObject.h"
#include "TextRep.h"
#include "ScreenReader.h"	// for Process_Incoming_Text
#include "ProtoTextRep.h"
#include "Mercator.h"
#include "Action.h"
#include "Error.h"
#include "mDictValPtr.h"

// hashing function for Window : just use value, its 32 bits
static unsigned windowHash(const Window &w) {
    return (unsigned) w;
}
// hashing function for a IceConn ... we assume same FD 
// is same object 
static unsigned IceConnHash(const IceConn &ice) {
    union {
	unsigned uns;
	unsigned char c[sizeof(unsigned)];
    } result;
    result.uns=0;
    int i=0;
    while (ice->connection_string[i]) {
	result.c[i%(sizeof(unsigned))]^=(unsigned char)ice->connection_string[i];
	++i;
    }
    return result.uns;
}
// hashing function for GContext (which is a 32 bit number)
static unsigned hashGContext(const GContext &x) { return x;}

#ifdef BEAM
mDictValPtr<IceConn, RAP>::mDictValPtr(unsigned (*hashKey)(const IceConn&)) :
    impl(new RWTPtrHashDictionary<IceConn, RAP>(hashKey))
{
}
mDictValPtr<Window, Client>::mDictValPtr(unsigned (*hashKey)(const Window&)) :
    impl(new RWTPtrHashDictionary<Window, Client>(hashKey))
{
}
#endif

RAPInstanceDict		RAP::instances(IceConnHash);
WinClientDict		RAP::clients(windowHash);


////////////////////////////////////////////////////////////////////////
//
// Class special functions.
//
////////////////////////////////////////////////////////////////////////

//
// Default constructor should never be called.  It's private, and calls abort!
//
RAP::RAP() :
    FDInterest("RAP"),
    client(NULL),
    gcDict(NULL),
    clientWindow(None)
{
    Error(EL_ABORT, "RAP:RAP", "Default constructor should never be called.\n"
	  "\tThis is unrecoverable as it indicates a programmer logic error.",
	  NULL);
}

//
// The constructor merely instantiates itself as an FDInterest and waits for
// stuff to happen.
//
RAP::RAP(IceConn _iceConn, int _majorOpcode) :
    FDInterest("RAP"),
    iceConn(_iceConn),
    majorOpcode(_majorOpcode),
    client(NULL),
    clientWindow(None),
    gcDict(new GCValueDict(hashGContext)),
    connectStatus(IceConnectStatus(iceConn))
{
    AddFDInterest(IceConnectionNumber(iceConn));
    SetIOMask(IceConnectionNumber(iceConn), ARead);

    RAP::instances.insert(iceConn, this);
}

RAP::~RAP()
{
    IceCloseStatus status;
    
    RemoveFDInterest(IceConnectionNumber(iceConn));

    RAP::instances.del(iceConn);

    // Other ICE shutdown junk?

    IceSetShutdownNegotiation(iceConn, False);
    
    if (!IceProtocolShutdown(iceConn, majorOpcode)) {
	Error(EL_ERROR, "RAP::~RAP",
	      "Could not shutdown RAP protocol on this ICE connection.", NULL);
    }

    status = IceCloseConnection(iceConn);

    switch (status) {
    case IceClosedNow:
    case IceClosedASAP:
	break;			// both these are fine...
    case IceConnectionInUse:
	Error(EL_ERROR, "RAP::~RAP", "Can't close ICE connection, in use.",
	      NULL);
	break;
    case IceStartedShutdownNegotiation:
	Error(EL_WARNING, "RAP::~RAP", "ICE shutdown negotiation started.",
	      NULL);
	break;
    default:
	Error(EL_ERROR, "RAP::~RAP", "Unknown IceCloseStatus on shutdown.",
	      NULL);
    }
}

////////////////////////////////////////////////////////////////////////
//
// Class static functions.
//
////////////////////////////////////////////////////////////////////////

void
RAP::ProcessProtoMsgProc(IceConn iceConn, IcePointer clientData, int opcode,
			 unsigned long length, Bool swap)
{
    RapMessageHeader *header = NULL;
    RAP *instance = NULL;
    char tmp[16];
    int blocked = 0;

    // Determine the instance to dispatch the message to.
    if ((instance = RAP::instances.lookup(iceConn)) == 0) {
 	//
 	// If we bail without reading data, we're gonna be screwed... we should
 	// probably just call this a client-terminating error and call it quits.
 	// But to call Shutdown we have to have an instance.  Arrgh.  Maybe
	// we should actually call abort if this happens...
	//
	Error(EL_ABORT, "RAP::ProcessProtoMsgProc",
	      "Cannot determine which instance is associated with this ICE "
	      "connection.\n\tThis is nonrecoverable as it indicates an "
	      "internal\n\tconsistency failure.", NULL);
	return;
    }
    
    // I've tried to order these by (guestimated) probability for speed.

    switch (opcode) {
    case RapRequestNotify: {
	RequestNotif *data = HandleRapRequestNotify(iceConn);
	instance->ProcessRequestNotify(data);
	FreeRequestNotify(data);
	break;
    }
    case RapEventNotify: {
	EventNotif *data = HandleRapEventNotify(iceConn);
	instance->ProcessEventNotify(data);
	FreeEventNotify(data);
	break;
    }
    case RapChangeNotify: {
	ChangeNotif *data = HandleRapChangeNotify(iceConn);
	instance->ProcessChangeNotify(data);
	FreeChangeNotify(data);
	break;
    }
    case RapConfigureNotify: {
	ConfigureNotif *data = HandleRapConfigureNotify(iceConn);
	instance->ProcessConfigureNotify(data);
	FreeConfigureNotify(data);
	break;
    }
    case RapCreateNotify: {
	CreateNotif *data = HandleRapCreateNotify(iceConn);
	instance->ProcessCreateNotify(data);
	FreeCreateNotify(data);
	break;
    }
    case RapDestroyNotify: {
	DestroyNotif *data = HandleRapDestroyNotify(iceConn);
	instance->ProcessDestroyNotify(data);
	FreeDestroyNotify(data);
	break;
    }
    case RapGeometryNotify: {
	GeometryNotif *data = HandleRapGeometryNotify(iceConn);
	instance->ProcessGeometryNotify(data);
	FreeGeometryNotify(data);
	break;
    }
    case RapGetValuesReply: {
	GetValuesReply *data = HandleRapGetValuesReply(iceConn);
	instance->ProcessGetValuesReply(data);
	FreeGetValuesReply(data);
	break;
    }
    case RapGetGCValuesReply: {
	GetGCValuesReply *data = HandleRapGetGCValuesReply(iceConn);
	instance->ProcessGetGCValuesReply(data);
	FreeGetGCValuesReply(data);
	break;
    }
    case RapLocateObjectReply: {
	LocateObjectReply *data = HandleRapLocateObjectReply(iceConn);
	instance->ProcessLocateObjectReply(data);
	FreeLocateObjectReply(data);
	break;
    }
    case RapGetResourcesReply: {
	GetResourcesReply *data = HandleRapGetResourcesReply(iceConn);
	instance->ProcessGetResourcesReply(data);
	FreeGetResourcesReply(data);
	break;
    }
    case RapAcknowledge: {
	(void) HandleRapAcknowledge(iceConn);
	break;
    }
    case RapError: {
	(void) HandleRapError(iceConn);
	break;
    }
    case RapFullQueryTreeReply: {
	FullQueryTreeReply *data = HandleRapFullQueryTreeReply(iceConn);
	instance->ProcessFullQueryTreeReply(data);
	FreeFullQueryTreeReply(data);
	break;
    }
    case RapQueryTreeReply: {
	QueryTreeReply *data = HandleRapQueryTreeReply(iceConn);
	instance->ProcessQueryTreeReply(data);
	FreeQueryTreeReply(data);
	break;
    }
    case RapHelloReply: {
	HelloReply *data = HandleRapHelloReply(iceConn);
	instance->ProcessHelloReply(data);
	FreeHelloReply(data);
	break;
    }
    default:
	sprintf(tmp, "%d", opcode);
	Error(EL_ERROR, "RAP::ProcessProtoMsgProc", "Unknown opcode: ",
	      tmp, NULL);
    }
}

////////////////////////////////////////////////////////////////////////
//
// Member functions.
//
////////////////////////////////////////////////////////////////////////

//
// HandleActivity will have to be rewritten to support both pseudo-blocking and
// non-blocking behavior (currently it only supports non-blocking behavior).
//
// Pseudo-blocking occurs something like this:  When a request goes out which
// requires that a reply arrives before any further processing is done, information
// about the requst is put into a list of "pending" requests.  Each node in the
// pending request list contains the major opcode, the minor opcode, and the
// sequence number.
//
// The routine which sends the request then calls into a dispatcher-loop itself
// (it might even be able to call MainLoop).   Thus, rather than popping the stack
// and returning to mainloop (and thereby losing the context in which the
// request was made, a new stack frame is pushed.
//
// This new mainloop continues processing input from all sources.  The only
// difference is that when there is activity on a RAP socket, we check to see
// if the incoming data is a reply to a request in the pending list.
//
// Ice provides some functionality to make this easier:  you can call
// IceProcessMessages with a non-NULL second argument.  This indicates that
// "pseudo-blocking" behavior is desired, and the "IceReplyWaitInfo" struct
// is returned in the second arg which indicates the major and minor opcodes,
// and sequence number of the incoming message.
//
// If this matches one of the requests in the pending list, we know we've satisfied
// a block.  (We should probably also keep timestamp information in the pending
// list so that we can determine if a request has timed out.)
//
// But how do we return?  If we've got several pending requests, our call stack
// may look like this:
//
//
//		RAP::HandleActivity	- the *first* reply comes in here!
//		MainLoop			- called by GetValuesRequest for blocking
//		RapGetValuesRequest	- we block to get the resource value
//		ProcessReadThisWord	- on a diff fd, the user wants a label read
//		MainLoop				- called to implement blocking behavior
//		RapGetGCValuesRequest	- blocks to get the value
//		EnterTextMode			- needs to know a GC value
//		MainLoop
//
// This is a problem!  We can't just pop once because that won't get us back
// to the GetGCValuesRequest which is waiting on the reply.  But if we pop
// multiple times, then we've lost the context of GetValuesRequest, which is
// also waiting on a reply.
//
// There seem to be a few possible solutions:
//	- Use some setjmp/longjmp shenanigans... very dangerous, probably
//	  will result in destructors for automatic objects not being called,
//	  memory leaks, and who knows what else...
//	- We record code memory addresses in the pending list and try to
//	  clean up the stack ourselves... very unportable, probably impossible
//	  to get working correctly.
//	- Prohibit multiple pending requests... easy to implement but restricts
//	  flexibility.
//	- Require that there be no outstanding pending requests whenever we
//	  pop the stack.
//
// Of these I think only the third and the fourth are close to workable.  Of
// course, the real solution is to multithread mercator.  :-)
//
// In solution #3, we keep a single "pending" variable for the pending request.
// If we try to issue a pseudo-blocking request and pending is set, we refuse
// it (perhaps issuing an exception).  This doesn't have nice failure properties,
// but keeps the model the same as it would be in a multithreaded implementation.
//
// In solution #4, if several pending requests are issued (and our call stack grows
// as shown above), we require *all* pending requests to be satisfied before we
// pop the stack.  This mechanism has a few interesting properties:  (1) it's
// possible for unbounded wait.  If multiple pending requests are issued, the
// poor first request will never get serviced, even if its reply has come in a year
// ago!  (2) The stack can grow quite large.  (3) It doesn't preserve ordering.
// That is, the replies must be handed out to the blocked caller procedures in
// stack order.  The most recently-issued requests get their replies first, even
// though the replies may have not come in in that order.
//
// A modification to #3 which gives it some of the same properties of #4 is possible:
// we call a procedure which issues pseudo-blocking request.  This procedure
// then enters its own mainloop code for processing.  Any messages received which
// are *not* the reply are queued.  As soon as the reply shows up, it is dispatched
// and the procedure returns.  MainLoop() then notices that several queued,
// unprocessed messages are available and dispatches them.
//
// This modification really does implement blocking semantics.  Its drawbacks are
// that (1) we lose ordering, and (2) we have a negative impact on performance.
// (the app essentiaily freezes until the reply comes in).  This freezing behavior
// may not be too bad though...
//
// After some more thinking I'm not going to try to implement *any* of these
// for the short-term.  All have some bogus side effects and implications for
// the rest of the system...  I'm thinking that the only real solution is to
// multithread Mercator on a per-client basis.
//
int
RAP::HandleActivity(int _fd, short mask)
{
    if (mask & ARead) {
	Status status;

	// IceProcessMessages will dispatch our protocol for us.
	status = IceProcessMessages(iceConn, NULL, NULL);

	if (status != IceProcessMessagesSuccess) {
	    switch (status) {
	    case IceProcessMessagesIOError:
		Error(EL_ERROR, "RAP::HandleActivity",
		      "IO Error during process messages", NULL);
		Shutdown(_fd);
		break;
	    case IceProcessMessagesConnectionClosed:
		Error(EL_ERROR, "RAP::HandleActivity",
		      "Connection broken", NULL);
		Shutdown(_fd);
		break;
	    default:
		Error(EL_WARNING, "RAP::HandleActivity",
		      "Unknown status from IceProcessMessages", NULL);
		break;
	    }

	    // If we've encounted an error, bail out.
	    return 0;
	}

	//
	// Update the connection status each time through.   If we don't have
	// a client yet and our connectStatus changes to ConnectAccepted,
	// then ask the AppModelMgr to create us a client instance.  If any
	// kind of error or rejection occurred, shut ourselves down.
	//
	// NOTE that there are some weird ordering effects here.  Before
	// we update the connect status, we look to see if the OLD connect
	// status was IceConnectAccepted and if client is still NULL.  If
	// it is, then we create a new client.   We cannot simply set the
	// connection status and then create a client because we need to
	// return *once* after the connection is accepted for the protocol
	// reply to get sent.  This structure ensures that the client instance
	// will get created after the protocol reply is sent.
	//
	if ((connectStatus == IceConnectAccepted) && !client) {
	    // FIX ME once we can get XIDs
	    Error(EL_DEBUG, "RAP::HandleActivity",
		  "Creating new client via AMM", NULL);
	    client = mercator->GetAppModelMgr()->CreateClient(this);

	    if (client == NULL) {
		Error(EL_ERROR, "RAP::HandleActivity",
		      "COULD NOT CREATE CLIENT!  Shutting down.\n");
		Shutdown(_fd);
	    }
	}
	    
	connectStatus = IceConnectionStatus(iceConn);
	
	if (connectStatus == IceConnectRejected) {
	    Error(EL_WARNING, "RAP::HandleActivity",
		  "Connection attempt was rejected.", NULL);
	    Shutdown(_fd);
	} else if (connectStatus == IceConnectIOError) {
	    Error(EL_ERROR, "RAP::HandleActivity",
		  "IO Error on client connection.", NULL);
	    Shutdown(_fd);
	} else if (connectStatus == IceConnectPending) {
	    Error(EL_DEBUG, "RAP::HandleActivity",
		  "Connection still pending...", NULL);
	}

	return 1;
    } 

    if (mask & AWrite) {
	// IGNORE?  Let ICE worry about output buffering.

	return 1;
    } 
    
    if (mask & AExcept) {
	//
	// An exceptional condition has occurred on a listener descriptor.
	// In theory these should never happen; if we get an exception,
	// shut down Mercator.
	//
	if (mask & POLLERR) 
	    Error(EL_ERROR, "RAP::HandleActivity",
		  "Exceptional condition detected: POLLERR", NULL);
	if (mask & POLLHUP)
	    Error(EL_ERROR, "RAP::HandleActivity",
		  "Exceptional condition detected: POLLHUP", NULL);
	if (mask & POLLNVAL)
	    Error(EL_ERROR, "RAP::HandleActivity",
		  "Exceptional condition detected: POLLNVAL", NULL);

	Shutdown(_fd);

	return 0;
    }
}

void
RAP::Shutdown(int fd)
{
    char *argv[10];
    int argc = 0;
    
    //
    // Shutdown() is idempotent - if we call Shutdown() again, just return.
    // This is indicated by NeedsCleanupP() returning TRUE.
    //
    if (NeedsCleanupP())
	return;
    
    MarkForCleanup();

    if (client) {
	argv[argc++] = Strdup(client->GetUniqueName());
	client->SetRAP(NULL);
	mercator->GetAppModelMgr()->DeleteClient(client->GetID());
    } else
	argv[argc++] = Strdup("NULL");

    Action::Invoke("ClientShutdown", argc, argv);

    while (argc--) delete [] argv[argc];

    client = NULL;
}


////////////////////////////////////////////////////////////////////////

//
// At the end of querytreereply, if we weren't ready before, we are now.
// (Actually we should probably double-check to make sure that we got some
// widgets back and that they're mapped, etc.)
//
void
RAP::ProcessQueryTreeReply(QueryTreeReply *data)
{
    TreeList trees;
    register int i;
    
    if (data == NULL)
	return;
    
    trees = data->trees;

    for (i=0 ; i<data->numShells ; i++) {
	register int j;
	long numWidgets = trees[i].numWidgets;

	for (j=0 ; j<numWidgets ; j++) {
	    XtObject *xtobj = NULL, *parent = NULL;
	    Object obj = (Object) trees[i].widgets[j].widget;
	    Object p = (Object) trees[i].widgets[j].parent;
	    char *n = trees[i].widgets[j].name;
	    char *c = trees[i].widgets[j].className;
	    Window w = (Window) trees[i].widgets[j].window;
	    int managed = trees[i].widgets[j].managed;
	    XServer *xserver = mercator->GetXServer();
	    XWindowAttributes wattr;

	    //
	    // The protocol requires that our parent be defined before we
	    // are.  Look up a corresponding XtObject.  Only look up the
	    // parent if there is one though!
	    //
	    if (p && (parent = client->GetXtObject(p)) == NULL) {
		Error(EL_ERROR, "RAP::ProcessQueryTreeReply",
		      "Unknown parent, can't create object", NULL);
		continue;
	    }

	    // NOTE parent defaults to NULL if p == 0
	    xtobj = client->CreateXtObject(obj, n, c, parent, w);

	    if (xtobj == NULL) {
		Error(EL_ERROR, "RAP::ProcessQueryTreeReply",
		      "Couldn't create XtObject?!", NULL);
		continue;
	    }

	    //
	    // Set the management state.
	    //
	    xtobj->SetManaged(managed);
	    
	    //
	    // POSSIBLE HACK ALERT -
	    // See the comment in ProcessCreateNotify.  Mercator's use of the
	    // "managed" flag doesn't really equal that of Xt's.  We use managed
	    // to determine if navigation, et al., is possible.  We set it to TRUE on
	    // objects with no parent (even though this makes no sense in Xt)
	    // so that navigation works properly without special-case checks
	    // everyplace.
	    //
	    if (parent == NULL)
		xtobj->SetManaged(TRUE);

	    //
	    // Determine if it's mapped or not.  I looked at putting this
	    // into RAP but I don't think we can determine it reliably there
	    // without generating protocol... I'd rather generate more
	    // protocol here than do it in the client.		-keith
	    //

	    // First off, if it's not realized, then we KNOW its not mapped.
	    if (w == RapUnrealized || w == RapGadget) {
		xtobj->SetMapped(0);
	    } else if (xserver->GetWindowAttributes(w, &wattr) == 0) {
		Error(EL_ERROR, "RAP::ProcessQueryTreeReply",
		      "GetWindowAttributes failed, can't determine map state ",
		      "of window for widget ", xtobj->GetName(), NULL);
		xtobj->SetMapped(0);
	    } else {
		// IsViewable means mapped and all our ancestors are mapped.
		int mapped = (wattr.map_state == IsViewable);
		xtobj->SetMapped(mapped);
	    }

	    // We may have unrealized widgets; don't grab on them.
	    if (w != RapUnrealized && w != RapGadget && !xtobj->GrabbedYet())
		mercator->GetXServer()->GrabInThisWindow(xtobj);

	    //
	    // Here we have a repeat of the code that goes on in
	    // HandleCreateNotify.  We need to figure out if the thing is a
	    // toplevel and, if so, stick it in the toplevel list.
	    //
	    // Also, if it's a toplevel being mapped, we have to call
	    // HandleToplevelMapped, which will do all our special toplevel
	    // handling (stuff like setting initial client and location if needed).
	    // The TRUE parameter means that the window we are getting
	    // reported as mapped was already on screen, so just add it to
	    // the model.  Don't do "popup" behavior as if it were NEWLY
	    // mapped.
	    //
	    if (xtobj->CheckIsShell()) {
		client->AddToplevelWindow(xtobj);
		if (xtobj->GetMapped()) {
		    ProcessToplevelMapped(xtobj, TRUE);
		}
	    }
	}
    }

    if (!client->Ready())
	client->Ready(TRUE);
}

//
// At the end of querytreereply, if we weren't ready before, we are now.
// (Actually we should probably double-check to make sure that we got some
// widgets back and that they're mapped, etc.)
//
void
RAP::ProcessFullQueryTreeReply(FullQueryTreeReply *data)
{
    FullTreeList trees;
    register int i;
    
    if (data == NULL)
	return;
    
    trees = data->trees;

    for (i=0 ; i<data->numShells ; i++) {
	register int j;
	long numWidgets = trees[i].numWidgets;

	for (j=0 ; j<numWidgets ; j++) {
	    XtObject *xtobj = NULL, *parent = NULL;
	    Object obj = (Object) trees[i].widgets[j].widget;
	    Object p = (Object) trees[i].widgets[j].parent;
	    char *n = trees[i].widgets[j].name;
	    char *c = trees[i].widgets[j].className;
	    Window w = (Window) trees[i].widgets[j].window;
	    int managed = trees[i].widgets[j].managed;
	    XServer *xserver = mercator->GetXServer();
	    XWindowAttributes wattr;
	    register int k;
	    
	    //
	    // The protocol requires that our parent be defined before we
	    // are.  Look up a corresponding XtObject.  Only look up the
	    // parent if there is one though!
	    //
	    if (p && (parent = client->GetXtObject(p)) == NULL) {
		Error(EL_ERROR, "RAP::ProcessQueryTreeReply",
		      "Unknown parent, can't create object", NULL);
		continue;
	    }

	    // NOTE parent defaults to NULL if p == 0
	    xtobj = client->CreateXtObject(obj, n, c, parent, w);

	    if (xtobj == NULL) {
		Error(EL_ERROR, "RAP::ProcessQueryTreeReply",
		      "Couldn't create XtObject?!", NULL);
		continue;
	    }

	    // Set the managed state.
	    xtobj->SetManaged(managed);
	    
	    //
	    // POSSIBLE HACK ALERT -
	    // See the comment in ProcessCreateNotify.  Mercator's use of the
	    // "managed" flag doesn't really equal that of Xt's.  We use managed
	    // to determine if navigation, et al., is possible.  We set it to TRUE on
	    // objects with no parent (even though this makes no sense in Xt)
	    // so that navigation works properly without special-case checks
	    // everyplace.
	    //
	    if (parent == NULL)
		xtobj->SetManaged(TRUE);

	    //
	    // Determine if it's mapped or not.  I looked at putting this
	    // into RAP but I don't think we can determine it reliably there
	    // without generating protocol... I'd rather generate more
	    // protocol here than do it in the client.		-keith
	    //

	    // First off, if it's not realized, then we KNOW its not mapped.
	    if ((w == RapUnrealized) || (w == RapGadget)) {
		xtobj->SetMapped(0);
	    } else if (xserver->GetWindowAttributes(w, &wattr) == 0) {
		Error(EL_ERROR, "RAP::ProcessQueryTreeReply",
		      "GetWindowAttributes failed, can't determine map state "
		      "of window for widget ", xtobj->GetName(), NULL);
		xtobj->SetMapped(0);
	    } else {
		// IsViewable means mapped and all our ancestors are mapped.
		int mapped = (wattr.map_state == IsViewable);
		xtobj->SetMapped(mapped);
		// Set width and height while we're at it...
		char tmpw[16], tmph[16];
		sprintf(tmpw, "%d", wattr.width);
		sprintf(tmph, "%d", wattr.height);
		
		xtobj->Width(wattr.width);
		xtobj->Height(wattr.height);
		xtobj->X(wattr.x);
		xtobj->Y(wattr.y);
		xtobj->BorderWidth(wattr.border_width);
	    }

	    // We may have unrealized widgets; don't grab on them.
	    if (w != RapUnrealized && w != RapGadget && !xtobj->GrabbedYet())
		mercator->GetXServer()->GrabInThisWindow(xtobj);

	    //
	    // Here we have a repeat of the code that goes on in
	    // HandleCreateNotify.  We need to figure out if the thing is a
	    // toplevel and, if so, stick it in the toplevel list.
	    //
	    // Also, if it's a toplevel being mapped, we have to call
	    // HandleToplevelMapped, which will do all our special toplevel
	    // handling (stuff like setting initial client and location if needed).
	    // The TRUE parameter means that the window we are getting
	    // reported as mapped was already on screen, so just add it to
	    // the model.  Don't do "popup" behavior as if it were NEWLY
	    // mapped.
	    //
	    if (xtobj->CheckIsShell()) {
		client->AddToplevelWindow(xtobj);
		if (xtobj->GetMapped()) {
		    ProcessToplevelMapped(xtobj, TRUE);
		}
	    }

	    //
	    // Now do all of its resources.
	    //
	    for (k=0 ; k<trees[i].widgets[j].numResources ; k++) {
		char *rNm =trees[i].widgets[j].resourceData[k].name;
		char *rCl =trees[i].widgets[j].resourceData[k].className;
		char *rTy =trees[i].widgets[j].resourceData[k].returnType;
		char *rVl =trees[i].widgets[j].resourceData[k].value;
		//
		// Note that resources come off the wire in a pointer-to-data
		// format, regardless of how long they are.  We leave it up to
		// the resource type management infrastructure to figure out
		// how to extract the actual resource from the data.
		//
		xtobj->CreateResource(rNm, rCl, rTy, rVl);
	    }
	}
    }

    if (!client->Ready())
	client->Ready(TRUE);
}

void
RAP::ProcessHelloReply(HelloReply *data)
{
    if (data == NULL)
	return;

    client->SetWindowID(data->window);
    clients.insert(data->window, client);
}

void
RAP::ProcessGetValuesReply(GetValuesReply *data)
{
    register int i;
    XtObject *obj = NULL;
    
    if (data == NULL)
	return;

    obj = client->GetXtObject((Object) data->widget);

    if (obj == NULL) {
	char tmp[64];
	sprintf(tmp, "%lx (%ld)", data->widget, data->widget);
	Error(EL_ERROR, "RAP::ProcessGetValuesReply",
	      "Received values for object not in our model: ", tmp, NULL);
	return;
    }

    for (i=0 ; i<data->numValues ; i++) {
	// Was there an error?
	if (data->values[i].errorCode != RapNoError) {
	    Error(EL_ERROR, "RAP::ProcessGetValuesReply",
		  "Error retrieving value of ", data->values[i].name,
		  ": ", data->values[i].data.errorString, NULL);
	    continue;
	}

	try {
	    obj->SetResource(data->values[i].name,
			     data->values[i].data.v.returnType,
			     data->values[i].data.v.value);
	}
	catch (const xmsg& msg) {
	    Error(EL_ERROR, "RAP::ProcessGetValuesReply",
		  "Problem setting value of ", data->values[i].name,
		  ": ", msg.why(), NULL);
	}
    }
}

void
RAP::ProcessGetGCValuesReply(GetGCValuesReply *data)
{
    register int i;
    
    if (data == NULL)
	return;

    //
    // Loop through all the incoming GCValues and process them (basically we
    // do a create, overwriting whatever previous contents were associated
    // with the GC ID in our cache).
    //
    for (i=0 ; i<data->numGCs ; i++) {
	if (data->GCValues[i].errorCode != RapNoError) {
	    char tmp[16];
	    sprintf(tmp, "%lx", data->GCValues[i].gc);
	    Error(EL_ERROR, "RAP::ProcessGetGCValuesReply",
		  "Error in retrieving GC ", tmp, ": ",
		  data->GCValues[i].data.errorString, NULL);
	    continue;
	}

	// have constructor just send over entire gcvalues struct
        //GCValue *gcv = new GCValue( &(data->GCValues[i]) );
        GCValue *gcv = new GCValue(data->GCValues[i]);
	gcDict->insert(data->GCValues[i].gc, gcv);

	char tmp[16];
	sprintf(tmp, "%lx", data->GCValues[i].gc);
    }
}

void
RAP::ProcessLocateObjectReply(LocateObjectReply *data)
{
    register int i;
    
    if (data == NULL)
	return;

    for (i=0 ; i<data->numLocations ; i++) {
	char tmp[16];
	
	sprintf(tmp, "%lx (%ld)", data->locations[i].widget,
		data->locations[i].widget);
    
	if (data->locations[i].errorCode != RapNoError) {
	    Error(EL_ERROR, "RAP::ProcessLocateObjectReply",
		  "Error in retrieving location for ", tmp, ": ",
		  data->locations[i].data.errorString, NULL);
	    continue;
	}

	XtObject *obj = client->GetXtObject((Object) data->locations[i].widget);

	if (obj == NULL) {
	    Error(EL_ERROR, "RAP::ProcessLocateObjectReply",
		  "Got a locate reply for an object not in our model: ",
		  tmp, NULL);
	    continue;
	}

	if (obj->GetWindowId() != data->locations[i].data.location.window) {
	    char tmp1[16], tmp2[16];
	    Window newWindow = data->locations[i].data.location.window;
	    
	    sprintf(tmp1, "%lx", obj->GetWindowId());
	    sprintf(tmp2, "%lx", newWindow);
	    
	    Error(EL_WARNING, "RAP::ProcessLocateObjectReply",
		  "Window ID of object ", tmp, " seems to have changed from ",
		  tmp1, " to ", tmp2, NULL);

	    mercator->GetAppModelMgr()->ChangeWindowId(obj, newWindow);
	}
	
	obj->SetManaged(data->locations[i].data.location.managed);
	obj->X(data->locations[i].data.location.x);
	obj->Y(data->locations[i].data.location.y);
	obj->Width(data->locations[i].data.location.width);
	obj->Height(data->locations[i].data.location.height);
	obj->BorderWidth(data->locations[i].data.location.borderWidth);
    }
}

void
RAP::ProcessGetResourcesReply(GetResourcesReply *data)
{
    Error(EL_WARNING, "RAP::ProcessGetResourcesReply",
	  "Not implemented yet.", NULL);
}

//
// Here is how I'm trying to handle deletions.  
//  - If the object being deleted is not the current object, then
//    we've got the easy case.  We have to:
//	- Tell its parent to delete it from its list of children.
//	- If it was a toplevel, delete it from the list of toplevels.
//  - If the object being deleted is the current object, then life
//    is a little tougher.  In addition to doing all of the stuff
//    above, we also must compute a new current object.  There are
//    two subcases here.
//	- If the object is a toplevel window, then we need to look
//	  at the popReturn member to see where to go to.  Case closed.
//	- If the object is not a toplevel, then we look at its parent.
//	  If it is navigable (including being mapped) then we go to
//	  it.  If it is not navigable, then we loop up the object
//	  tree, looking for the first parent that is navigable.
//
// There's another case which I *think* won't ever happen.  That is
// where the object being deleted is the *parent* of the current
// object.  If this happens we can be fairly certain that we'll
// get an individual deletion message for the child, the current
// object, at some point.  In fact, it should come before the parent
// is deleted.
//
// There's probably another case in which the object being deleted
// is the *last* object in a given application.  At this point we
// need to pick a new client to go to.  THIS IS NOT YET IMPLEMENTED.
//
// This is a fairly complicated algorithm, and may best be implemented
// in TCL.
//			-k
//
// OPEN ISSUE - What about the case where a PARENT of the current object
// is being deleted?  I *think* that we'll get individual deletions,
// from the bottom of the tree up, as things get deleted.  That means
// that a string of deletions will start below the current, and move upward
// eventually hitting us.  As it hits us, it carries us along with it
// (sort of like riding a wave) until the toplevel gets deleted and we're
// popped back out to another shell.
//
// ...We should make sure that things really work this way though.
//
void
RAP::ProcessDestroyNotify(DestroyNotif *data)
{
    XtObject *xtobj = NULL, *gotoObj = NULL, *parent = NULL;;
    Object objID;
    
    if (data == NULL)
	return;

    objID = (Object) data->widget;

    xtobj = client->GetXtObject(objID);

    if (xtobj == NULL) {
	char tmp[64];
	sprintf(tmp, "%lx (%ld)", data->widget, data->widget);
	Error(EL_ERROR, "RAP::ProcessDestroyNotify",
	      "Got a DestroyNotify for an object not in our model: ",
	      tmp, NULL);
	return;
    }

    parent = xtobj->GetParent();
    
    //
    // First, if this is the current object, we need to figure out where
    // to go to before we start deleting things.
    //
    if (client->GetCurrentLocation() == xtobj) {
	Error(EL_DEBUG, "RAP::HandleDestroyNotify",
	      "CURRENT LOCATION IS BEING DELETED: ",
	      xtobj->GetName(), NULL);
	//
	// If it's a toplevel, then we will want to return to its
	// popReturn.
	//
	if (client->IsToplevelWindow(xtobj)) {
	    if (xtobj->GetPopReturn() == NULL) {
		Error(EL_ERROR, "RAP::HandleDestroyNotify",
		      "OBJECT BEING DELETED IS A TOPLEVEL BUT IT"
		      " HAS NO POPUP RETURN.\n", NULL);
		gotoObj = NULL;
	    } else {
		Error(EL_DEBUG, "RAP::HandleDestroyNotify",
		      "OBJECT BEING DELETED IS A TOPLEVEL.  "
		      "ITS RETURN IS ", xtobj->GetPopReturn()->GetName(),
		      NULL);
		Action::Invoke("PopupReturn", 0, NULL);
		gotoObj = xtobj->GetPopReturn();
	    }
	} else {
	    int done = 0;
	    //
	    // The object we're deleting was *not* a toplevel, so
	    // loop up the hierarchy, searching for a parent which is
	    // navigable.
	    //
	    gotoObj = xtobj->GetParent();
	    while (!done) {
		if (gotoObj == NULL) {
		    // We're fucked!  We should probably change toplevels or
		    // even change clients.
		    Error(EL_ERROR, "RAP::HandleDestroyNotify",
			  "CAN'T FIND A NAVIGABLE PARENT FOR OBJECT BEING"
			  " DELETED.  object = ", xtobj->GetName(), NULL);
		    done = 1;
		} else {
		    if (Template::GetBoolean(*gotoObj, "navigable"))
			done = 1;
		    else
			gotoObj = gotoObj->GetParent();
		}
	    }
	}
	    
	// END OF THE CASE where the object being deleted is the current
	// object
    }
	
    //
    // At this point, we've got the object we need to go to stuffed into
    // gotoObj (which may be NULL if we encountered an error above,
    // or NULL if the object being deleted didn't cause us to have to
    // move anyplace.  
    //
    // Now we can go through the mechanics of actually deleting the
    // thing from the model.
    //
	
    // WE SHOULD PROBABLY invoke a Tcl action here for deletion.  Do
    // it before the object is actually deleted, so the Tcl code will
    // be able to query stuff about it.
	
    // Tell the parent that the child is being deleted.
    if (parent != NULL)
	parent->DeleteChild(xtobj);
	
    // If the object was a toplevel window, delete it.
    if (client->IsToplevelWindow(xtobj)) {
	client->DeleteToplevelWindow(xtobj);
    }

    char *argv[10];
    int argc = 0;
    argv[argc++] = Strdup(xtobj->GetUniqueName());
    Action::Invoke("XtObjectDestroyed", argc, argv);
    delete [] argv[--argc];

    // Tell the appmodelmgr to delete the object.  AMM will call the
    // client to delete the object from its dict.  At this point xtobj points
    // to invalid data.
    mercator->GetAppModelMgr()->DeleteXtObject(xtobj);
    xtobj = NULL;
    
    // If this was the last object in the client, then delete the client.
    if (client->NumChildren() == 0) {
	Error(EL_DEBUG, "RAP::HandleDestroyNotify",
	      "Last object in client destroyed, destroying client.",
	      NULL);
	mercator->GetAppModelMgr()->DeleteClient(client->GetID());
    }

    //
    // FINALLY, if gotoObj is non-NULL, it means that we need to
    // change the current location to the new object.
    //
    if (gotoObj != NULL) {
	Error(EL_DEBUG, "RAP::HandleDestroyNotify",
	      "THE OBJECT WE WERE ON WAS DELETED,  GOING TO ",
	      gotoObj->GetName(), NULL);
	//
	// I'm not sure if this is the way things *should* work, but
	// here it is.  The TCL command setFocus is what actually does
	// the movement stuff.  It calls SetCurrentLocation for us, so
	// we don't do it here.  FURTHER, SetCurrentLocation doesn't
	// actually do any warping for us (I think it probably *should*).
	// We call out to the TCL routine GoTo to move.  It warps,
	// raises, calls setFocus, the works.
	//			-k
	char *argv[1];
	int result = 0;
	    
	argv[0] = Strdup(gotoObj->GetUniqueName());
	result = Action::Invoke("GoTo", 1, argv);
	delete [] argv[0];
	    
	if (result) {
	    Error(EL_DEBUG, "RAP::HandleDestroyNotify",
		  "SUCCESSFULLY called TCL Action GoTo to move.", NULL);
	} else {
	    Error(EL_ERROR, "RAP::HandleDestroyNotify",
		  "Call of TCL Action GoTo FAILED!", NULL);
	}
    }
}

void
RAP::ProcessConfigureNotify(ConfigureNotif *data)
{
    Object objID;
    XtObject *xtobj = NULL;
    XtGeometryMask mask;
    
    if (data == NULL)
	return;
    
    objID = (Object) data->widget;
    mask = data->geometryMask;

    xtobj = client->GetXtObject(objID);

    if (xtobj == NULL) {
	char tmp[64];
	sprintf(tmp, "%lx (%ld)", data->widget, data->widget);
	Error(EL_ERROR, "RAP::ProcessConfigureNotify",
	      "Got a ConfigureNotify for an object not in our model: ",
	      tmp, NULL);
	return;
    }

    if (mask & CWX)
	xtobj->X(data->x);

    if (mask & CWY)
	xtobj->Y(data->y);

    //
    // Note that on width and height changes, we also have to resize our textrep
    // (if we have one).
    //
    if (mask & CWWidth)
	xtobj->Width(data->width);

    if (mask & CWHeight)
	xtobj->Height(data->height);

    ///// MESSAGE TO TEXT REP?
    if (mask & CWBorderWidth)
	xtobj->BorderWidth(data->borderWidth);

    //
    // NOTE that we currently ignore:
    //		CWSibling
    //		CWStackMode
    //
}

void
RAP::ProcessCreateNotify(CreateNotif *data)
{
    XtObject *xtobj = NULL, *parent = NULL;
    register int i;
    
    if (data == NULL)
	return;

    if (data->parent)
	parent = client->GetXtObject((Object) data->parent);

    // Create the object
    xtobj = client->CreateXtObject((Object) data->widget,
				   data->name,
				   data->className,
				   parent,
				   (Window) data->window);

    char tmp[16];
    sprintf(tmp, "%lx", data->widget);
    if (parent)
	Error(EL_DEBUG, "RAP::ProcessCreateNotify: Created ", data->name,
	      " (", data->className, ") id = ", tmp, " parent = ",
	      parent->GetName(), NULL);
    else
	Error(EL_DEBUG, "RAP::ProcessCreateNotify: Created ", data->name,
	      " (", data->className, ") id = ", tmp, " parent = NONE", NULL);
    
    //
    // POSSIBLE HACK ALERT -
    // If we have no parent then I set the "managed" flag for this object to
    // TRUE.  Note that this really has no meaning in Xt-speak:  management
    // is the property of an object being made aware to its parent.  But much
    // of the rest of the Mercator code uses the management flag to determine
    // if its possible to move to a widget and so forth.  So to make things work
    // out I throw the managed switch.
    //
    // Just keep in mind that the value of the flag won't exactly match up with
    // the managed state in the application!
    //
    if (parent == NULL)
	xtobj->SetManaged(TRUE);
    
    // And add its resources
    for (i=0 ; i<data->numResources ; i++) {
	xtobj->CreateResource(data->resources[i].name,
			      data->resources[i].className,
			      data->resources[i].returnType,
			      data->resources[i].value);
    }

    if (xtobj->CheckIsShell())
	client->AddToplevelWindow(xtobj);
}

void
RAP::ProcessChangeNotify(ChangeNotif *data)
{
    XtObject *xtobj = NULL;
    char *type;
    
    if (data == NULL)
	return;

    xtobj = client->GetXtObject((Object) data->widget);

    if (xtobj == NULL) {
	char tmp[64];
	sprintf(tmp, "%lx (%ld)", data->widget, data->widget);
	Error(EL_ERROR, "RAP::ProcessChangeNotify",
	      "Got a ChangeNotify for an object not in our model: ",
	      tmp, NULL);
	return;
    }

    type = data->type;
    
    if (strcmp(type, XtHsetValues) == 0) {
	register int i;
	for (i=0  ;  i<data->c.setValues.numResources; i++) {
	    char *rNm = data->c.setValues.resources[i].name;
	    char *rTy = data->c.setValues.resources[i].type;
	    char *rVl = data->c.setValues.resources[i].value;
	    xtobj->SetResource(rNm,rTy,rVl);
	}
    } else if (strcmp(type, XtHmanageChildren) == 0) {
	ProcessWidgetChildrenManaged(xtobj, data->c.manageChildren.widgets,
				     data->c.manageChildren.numWidgets);
    } else if (strcmp(type, XtHunmanageChildren) == 0) {
	ProcessWidgetChildrenUnmanaged(xtobj, data->c.unmanageChildren.widgets,
				       data->c.unmanageChildren.numWidgets);
    } else if (strcmp(type, XtHmanageSet) == 0) {
    } else if (strcmp(type, XtHunmanageSet) == 0) {
    } else if (strcmp(type, XtHrealizeWidget) == 0) {
	ProcessWidgetRealized(xtobj, data->c.realizeWidget.window);
    } else if (strcmp(type, XtHunrealizeWidget) == 0) {
	ProcessWidgetUnrealized(xtobj);
    } else if (strcmp(type, XtHaddCallback) == 0) {
    } else if (strcmp(type, XtHaddCallbacks) == 0) {
    } else if (strcmp(type, XtHremoveCallback) == 0) {
    } else if (strcmp(type, XtHremoveCallbacks) == 0) {
    } else if (strcmp(type, XtHsetKeyboardFocus) == 0) {
    } else if (strcmp(type, XtHaugmentTranslations) == 0) {
    } else if (strcmp(type, XtHoverrideTranslations) == 0) {
    } else if (strcmp(type, XtHuninstallTranslations) == 0) {
    } else if (strcmp(type, XtHsetKeyboardFocus) == 0) {
    } else if (strcmp(type, XtHsetWMColormapWindows) == 0) {
    } else if (strcmp(type, XtHsetMappedWhenManaged) == 0) {
    } else if (strcmp(type, XtHmapWidget) == 0) {
	ProcessWidgetMapped(xtobj);
    } else if (strcmp(type, XtHunmapWidget) == 0) {
	ProcessWidgetUnmapped(xtobj);
    } else if (strcmp(type, XtHpopup) == 0) {
    } else if (strcmp(type, XtHpopupSpringLoaded) == 0) {
    } else if (strcmp(type, XtHpopdown) == 0) {
    } else {
	Error(EL_WARNING, "RAP::ProcessChangeNotify",
	      "Unknown change type: ", type, NULL);
    }
}

void
RAP::ProcessGeometryNotify(GeometryNotif *data)
{
    XtObject *xtobj;
    
    if (data == NULL)
	return;

    xtobj = client->GetXtObject((Object) data->widget);
    
    if (xtobj == NULL) {
	char tmp[64];
	sprintf(tmp, "%lx (%ld)", data->widget, data->widget);
	Error(EL_ERROR, "RAP::ProcessGeometryNotify",
	      "Got a GeometryNotify for an object not in our model: ",
	      tmp, NULL);
	return;
    }
    
    if (data->final.request_mode & CWX)
	xtobj->X(data->final.x);

    if (data->final.request_mode & CWY)
	xtobj->Y(data->final.y);

    //
    // Note that on width and height changes, we also have to resize our textrep
    // (if we have one).
    //
    if (data->final.request_mode & CWWidth)
	xtobj->Width(data->final.width);

    if (data->final.request_mode & CWHeight)
	xtobj->Height(data->final.height);

    ///// MESSAGE TO TEXT REP?
    if (data->final.request_mode & CWBorderWidth)
	xtobj->BorderWidth(data->final.border_width);

    //
    // NOTE that we currently ignore:
    //		CWSibling
    //		CWStackMode
    //
}

void
RAP::ProcessRequestNotify(RequestNotif *data)
{
    char tmp[16];
    
    if (data == NULL) return;
    
    //
    // Currently we're only interested in GC and text-related requests.
    //
    switch (data->type) {
    case X_CreateGC:
	ProcessCreateGCRequest(data->request);
	break;
    case X_ChangeGC:
	ProcessChangeGCRequest(data->request);
	break;
    case X_CopyGC:
	ProcessCopyGCRequest(data->request);
	break;
    case X_FreeGC:
	ProcessFreeGCRequest(data->request);
	break;
    case X_ImageText16:
    case X_PolyText16:
    case X_ImageText8:
    case X_PolyText8:
    case X_ClearArea:
    case X_CopyArea:
    case X_PolyFillRectangle:
	ProcessTextRequests(data->request);
	break;
    case X_ReparentWindow:
	ProcessReparentWindowRequest(data->request);
	break;
    default:
	sprintf(tmp, "%d\n", data->type);
	Error(EL_WARNING, "RAP::ProcessRequestNotify",
	      "Got an unexpected request type: ", tmp, NULL);
	break;
    }
}

//
// The old XlibHook implementation had *two* versions of this call - one was
// invoked whenever we got an X protocol GC creation, and one was invoked
// from the GCFlushProc client-side extension.  I don't understand (still)
// why the special GC client-side extensions are there and why they catch
// stuff that the protocol catcher doesn't.			-keith
//
void
RAP::ProcessCreateGCRequest(xReq *request)
{
    GCValue *gcv;
    xCreateGCReq *req = (xCreateGCReq *) request;
    XID *pval = (XID *) &req[1];	// get ptr to data after request

    if ((gcv = gcDict->lookup(req->gc)) == NULL) {
 	// The GC we've just seen isn't in our model, so create a new
	// one.  (CURRENTLY THIS LEAKS - WE HAVE NO FREEING FOR
 	// THE GCVALUES DICT.)
	gcv = new GCValue(req->gc);
	gcDict->insert(req->gc, gcv);
    }
    // We've already encountered it, so do a change only.

    gcv->ChangeGC(req->mask, pval);
}

// 
// typedef struct {
//     CARD8 reqType;
//     BYTE pad;
//     CARD16 length B16;
//     GContext gc B32;
//     CARD32 mask B32;
// } xChangeGCReq;    
// 
void
RAP::ProcessChangeGCRequest(xReq *request)
{
    GCValue * gcv;
    xChangeGCReq * req = (xChangeGCReq *) request;

    gcv = FindGC(req->gc);
    if (gcv == NULL) {
	char tmp[16];
    
	sprintf(tmp, "%lx", req->gc);
	Error(EL_WARNING, "RAP::ChangeGCRequest",
	      "Couldn't find GC, sending GetGCValues... ", tmp, "\n", NULL);
	GetGCValuesRequest(&req->gc, 1);
	return;
    }

    gcv->ChangeGC(req->mask, (XID *) &req[1]);
}

//
// typedef struct {
//     CARD8 reqType;
//     BYTE pad;
//     CARD16 length B16;
//     GContext srcGC B32, dstGC B32;
//     CARD32 mask B32;
// } xCopyGCReq;    
// 
void
RAP::ProcessCopyGCRequest(xReq *request)
{
    GCValue * src, * dest;
    xCopyGCReq * req = (xCopyGCReq *) request;

    src = FindGC(req->srcGC);
    if (src == NULL) {
	char tmp[16];
	sprintf(tmp, "%lx", req->srcGC);
	Error(EL_WARNING, "RAP::ProcessCopyGCRequest",
	      "couldn't find source GC ", tmp,
	      ", sending GetGCValuesRequest.", "\n", NULL);
	GetGCValuesRequest(&req->srcGC, 1);
	return;
    }
    dest = FindGC(req->dstGC);
    if (dest == NULL) {
	char tmp[16];
	sprintf(tmp, "%lx", req->dstGC);
	Error(EL_WARNING, "\nRAP::ProcessCopyGCRequest",
	      "couldn't find dest GC ", tmp,
	      ", sending GetGCValuesRequest.", "\n", NULL);
	GetGCValuesRequest(&req->dstGC, 1);
	return;
    }
    
    *dest = *src;
}

//
// Wrapper around dictionary lookup.  We probably don't really need this...
//
GCValue *
RAP::FindGC(GContext gid)
{
    GCValue *gcv;

    // See if it already exists
    gcv = gcDict->lookup(gid);

    // If it doesn't, then return NULL
    if (gcv == NULL)
	return NULL;

    // Otherwise return it.
    return gcv;
}

//
// The X protocol request ResourceReq is used for any request which has
// a resource ID (or Atom or Time) as its one and only argument.
//
// typedef struct {
//     CARD8 reqType;
//     BYTE pad;
//     CARD16 length B16;
//     CARD32 id B32;  /* a Window, Drawable, Font, GContext, Pixmap, etc. */
//     } xResourceReq;
// 
void
RAP::ProcessFreeGCRequest(xReq *request)
{
    xResourceReq * req = (xResourceReq *) request;
    GCValue * gcv;

    gcv = FindGC((GContext) req->id);
    if (gcv == NULL) {
	char tmp[16];
  
	sprintf(tmp, "%lx", req->id);
	Error(EL_ERROR, "RAP::ProcessFreeGCRequest", "can't find GC: ", 
	      tmp, NULL);
	return;
    }
    
    // remove it from our internal dictionary.
    gcDict->del(req->id);
    delete gcv;
}

//
// I believe that the only time we'll really get these is when we're using
// #*!%@ CDE tear-off menus...
//
void
RAP::ProcessReparentWindowRequest(xReq *request)
{
    xReparentWindowReq *req = (xReparentWindowReq *) request;
    Window window = req->window;
    Window parent = req->parent;
    XtObject *childObj, *parentObj;

    if ((childObj = mercator->GetAppModelMgr()->GetXtObject(window)) == NULL) {
	Error(EL_WARNING, "RAP::ProcessReparentWindowRequest",
	      "Window being reparented is not in our model", NULL);
	return;
    }

    if ((parentObj = mercator->GetAppModelMgr()->GetXtObject(parent)) == NULL) {
	Error(EL_WARNING, "RAP::ProcessReparentWindowRequest",
	      "Window getting a new child is not in our model", NULL);
	return;
    }

    // If this guy had a parent before, remove us from it's children list.
    if (childObj->GetParent())
	childObj->GetParent()->DeleteChild(childObj);
    
    childObj->SetParent(parentObj);
    parentObj->AddChild(childObj);
}

//
// ProcessTextRequests() basically just grabs the text request, resolves
// the XtObject on which the request occurred, and then forwards the
// request to the object's TextRep for processing.
//
void
RAP::ProcessTextRequests(xReq *request)
{
    switch (request->reqType) {
    case X_ImageText8:
    case X_ImageText16: {
	xImageTextReq *req = (xImageTextReq *) request;
	XtObject *obj = mercator->GetAppModelMgr()->GetXtObject(req->drawable);
	
	if (obj == NULL) {
	    char tmp[16];
	    sprintf(tmp, "%lx (%ld)", req->drawable, req->drawable);
	    Error(EL_ERROR, "RAP::ProcessTextRequests",
		  "ImageText drawn to window not in our model: ", tmp, NULL);
	    return;
	}
	
	ProtoTextRep *text = (ProtoTextRep*) client->FindTextRep(req->drawable);
	
	if (text == NULL)
	    return;
	
	if (FindGC(req->gc) == NULL) {
	    char tmp[16];
	    sprintf(tmp, "%lx", req->gc);
	    Error(EL_ERROR, "RAP::ProcessTextRequests",
		  "Can't find GC used in ImageText: ", tmp, NULL);
	    GetGCValuesRequest(&req->gc, 1);
	}

	if (request->reqType == X_ImageText8)
	    text->HandleImageText8(request);
	else
	    text->HandleImageText16(request);
	
	break;
    }
    
    case X_PolyText8:
    case X_PolyText16: { 
	xPolyTextReq *req = (xPolyTextReq *) request;
	XtObject *obj = mercator->GetAppModelMgr()->GetXtObject(req->drawable);
	
	if (obj == NULL) {
	    char tmp[16];
	    sprintf(tmp, "%lx (%ld)", req->drawable, req->drawable);
	    Error(EL_ERROR, "RAP::ProcessTextRequests",
		  "PolyText drawn to window not in our model: ", tmp, NULL);
	    return;
	}
	
	ProtoTextRep *text = (ProtoTextRep*) client->FindTextRep(req->drawable);
	
	if (text == NULL)
	    return;
	
	if (FindGC(req->gc) == NULL) {
	    char tmp[16];
	    sprintf(tmp, "%lx", req->gc);
	    Error(EL_ERROR, "RAP::ProcessTextRequests",
		  "Can't find GC used in PolyText: ", tmp, NULL);
	    GetGCValuesRequest(&(((xPolyTextReq *) req)->gc), 1);
	}

	if (request->reqType == X_PolyText8)
	    text->HandlePolyText8(request);
	else
	    text->HandlePolyText16(request);
	
	break;
    }
    
    case X_CopyArea: {
	xCopyAreaReq *req = (xCopyAreaReq *) request;

	//
	// There are five cases to consider when we see a CopyArea.
	//
	// (1) If the source and destination are the same, and are known to
	//      us (i.e., they have text reps associated with them), then we
	//      pass the request down to ProtoTextRep::HandleCopyArea, which
	//      is responsible for scrolling text areas.
	// (2) If the source and destination are the same, but are unknown
	//      to us, we ignore the request.
	// (3) If the source and destination are different, but both are known
	//      to us, we SHOULD copy between text areas, but this is
	//      unimplemented.  I don't think many widgets will generate this
	//      kind of protocol though...  FIX ME LATER.
	// (4) The source is unknown, but the destination represents
	//      something we have a text rep for.  In this case, we treat the
	//      request like a ClearArea on the destination (we assume that
	//      they're copying background or something...)
	// (5) The destination is unknown, but the source is something in
	//      our model.  In this case we just throw away the request.
	//
	
	if (req->srcDrawable == req->dstDrawable) {
	    XtObject *obj =
		mercator->GetAppModelMgr()->GetXtObject(req->srcDrawable);

	    if (obj == NULL)
		return;
	
	    ProtoTextRep *text = (ProtoTextRep *)
		client->FindTextRep(req->srcDrawable);
	
	    if (text == NULL)
		return;

	    text->HandleCopyArea(request);
	} else {
	    XtObject *srcObj, *dstObj;
	    ProtoTextRep *srcText, *dstText;

	    srcObj = mercator->GetAppModelMgr()->GetXtObject(req->srcDrawable);
	    dstObj = mercator->GetAppModelMgr()->GetXtObject(req->dstDrawable);

	    if (dstObj == NULL)
		return;

	    srcText = (ProtoTextRep *) client->FindTextRep(req->srcDrawable);
	    dstText = (ProtoTextRep *) client->FindTextRep(req->dstDrawable);

	    if (dstText == NULL)
		return;

	    if (srcText == NULL) {
		// Treat this as a clear area.  In fact, we'll just build a
		// bogus xClearAreaReq and pass it along to the destination
		// text rep...
		xClearAreaReq clearAreaReq;

		clearAreaReq.reqType = X_ClearArea;
		clearAreaReq.window = req->dstDrawable;
		clearAreaReq.exposures = 1;	// not sure...
		clearAreaReq.length = sz_xClearAreaReq;
		clearAreaReq.x = req->dstX;
		clearAreaReq.y = req->dstY;
		clearAreaReq.width = req->width;
		clearAreaReq.height = req->height;

		dstText->HandleClearArea((xReq *) &clearAreaReq);
	    } else {
		Error(EL_WARNING, "RAP::ProcessTextRequests",
		      "Text is being copied between two different text ",
		      "areas.\n\tThis is not currently supported.", NULL);
	    }
	}
	break;
    }
    
    case X_ClearArea: {
	xClearAreaReq *req = (xClearAreaReq *) request;
	ProtoTextRep *text = (ProtoTextRep *) client->FindTextRep(req->window);

	if (text == NULL)
	    return;
	
	text->HandleClearArea(request);

	break;
    }

    case X_PolyFillRectangle: {
	xPolyFillRectangleReq *req = (xPolyFillRectangleReq *) request;
	ProtoTextRep *text = (ProtoTextRep*) client->FindTextRep(req->drawable);
	// Note cheesy way to get to end of request.
	XRectangle *rects = (XRectangle *) &req[1];
	register int i = 0, size = req->length * 4;

	if (text == NULL)
	    return;

	size -= sz_xPolyFillRectangleReq;
	
	while (size >= sz_xRectangle) {
	    //
	    // We don't always want to treat a filled rectangle as a cleared
	    // area.  Rects are also used for cursors and highlighting.
	    // So we look at the GC the rectangle is being drawn in, and
	    // only treat it as a clear if the caller is doing GXcopy.
	    //
	    GCValue *gcv = GetGCDict()->lookup(req->gc);

	    if (gcv != NULL) {
		XGCValues values = *gcv;
		
		if ((values.function == GXcopy) ||
		    (values.function == GXclear) ||
		    (values.function == GXset)) {
		    // Treat this as a clear area.  In fact, we'll just build a
		    // bogus xClearAreaReq and pass it along to the destination
		    // text rep...
		    xClearAreaReq clearAreaReq;

		    clearAreaReq.reqType = X_ClearArea;
		    clearAreaReq.window = req->drawable;
		    clearAreaReq.exposures = 1;	// not sure...
		    clearAreaReq.length = sz_xClearAreaReq;
		    clearAreaReq.x = rects->x;
		    clearAreaReq.y = rects->y;
		    clearAreaReq.width = rects->width;
		    clearAreaReq.height = rects->height;

		    text->HandleClearArea((xReq *) &clearAreaReq);
		}
	    }
	    
	    //
	    // This is mighty questionable.  Applications like dtterm use a
	    // box cursor drawn via X_PolyFillRectangle to denote the
	    // application cursor.  We treat the rectangle as a cursor
	    // position for the text rep.
	    //
	    // NOTE that we ONLY do this if (1) the blockCursor attribute
	    // for the application is true, and (2) the rectangle being drawn
	    // is character-sized.
	    //
	    if (text->ValidateCursorSize(rects->width, rects->height)) {
		const FieldData *data = Template::FindByName(*client,
							     "blockCursor");
	    
		if (data && (data->Tag() == BOOLEAN) && data->Boolean(*client)){
		    try {
			int row = text->PixelToRow(rects->y);
			int col = text->PixelToColumn(rects->x);
			text->SetAppCursor(row, col);
		    }
		    catch (xmsg& msg) {
			// dummy...
#if 0
			Error(EL_WARNING, "RAP::ProcessTextRequests",
			      "FillRectangle is not a cursor, ",
			      (const char *) msg.why(), NULL);
#endif
		    }
		}
	    }

	    i++;
	    rects++;
	    size -= sz_xRectangle;
	}

	break;
	}
    
    default:
	// Do nothing
	Error(EL_ERROR, "RAP::ProcessTextRequests",
	      "Called for a non-text related request?", NULL);
	break;
    }
}

//
// ProcessEventNotify is used to dispatch any X Events that the CLIENT
// receives.  These are forwarded to Mercator.  Note that this is completely
// separate from any events that Mercator itself may receive from the
// X Server.
//
// Currently the events that we process are:
//    KeyPress
//    EnterNotify
//    MapNotify
//    UnmapNotify
//    ButtonPress
//
// Pretty much we throw all of these away, only firing actions, with two
// exceptions:
//    KeyPress 		- forward the keypress to the object's screenreader for
//			  	  keyboard feedback.
//    MapNotify		- ensure that our grabs remain in place on the mapped
//				  window.
//
void
RAP::ProcessEventNotify(EventNotif *data)
{
    char tmp[16];
    
    if (data == NULL) return;
    
    //
    // Currently we're only interested in a handful of events.
    //
    switch (data->type) {
    case KeyPress:
	ProcessKeyPressEvent(data->event);
	break;
    case EnterNotify:
	ProcessEnterNotifyEvent(data->event);
	break;
    case UnmapNotify:
	ProcessUnmapNotifyEvent(data->event);
	break;
    case MapNotify:
	ProcessMapNotifyEvent(data->event);
	break;
    case ButtonPress:
	ProcessButtonPressEvent(data->event);
	break;
    default:
	sprintf(tmp, "%ld", data->type);
	Error(EL_WARNING, "RAP::ProcessEventNotify",
	      "Got an unexpected event, type = ", tmp, NULL);
	break;
    };
}

//
// Keypresses are detected to perform keyboard feedback.
//
void
RAP::ProcessKeyPressEvent(xEvent *event)
{
    XtObject *obj;
    ScreenReader *sreader;
    char *s;
    XServer  *xServer = mercator->GetXServer();

    // Find the appropriate screenreader
    obj = client->GetXtObject((Window) event->u.keyButtonPointer.event);

    if (obj == NULL) {
	char tmp[16];
	sprintf(tmp, "%lx (%ld)", event->u.keyButtonPointer.event,
		event->u.keyButtonPointer.event);
	Error(EL_ERROR, "RAP::ProcessKeyPressEvent",
	      "Key press occurred in a window not in our model: ", tmp, NULL);
	return;
    }

    //
    // Not having a screenreader isn't necessarily an error - keypresses
    // might occur in objects that have nothing to do with text.
    //
    if ((sreader = obj->GetScreenReader()) == NULL) {
	Error(EL_DEBUG, "RAP::ProcessKeyPressEvent",
	      "Key press occurred in an object without a screenreader.", NULL);
	return;
    }

    //
    // Send the keypress to the object's screenreader.
    //
    s = xServer->KeysymToString(xServer->KeycodeToKeysym(event->u.u.detail));

    // Ignore metacharacters.  This shouldn't be hard-wired.
    if ((strcmp(s, "Shift_R") == 0) || (strcmp(s, "Shift_L") == 0) ||
	(strcmp(s, "Control_R") == 0) || (strcmp(s, "Control_L") == 0))
	return;
    else
	sreader->Process_Incoming_Text(s);
}

//
// Currently enter notifies don't do much of anything but cause an action
// to be invoked.  We could get smart and have Mercator's concept of
// current location track the mouse pointer if it is moved.  This is where
// that code would go (probably with some supporting code in other places
// as well).
//
void
RAP::ProcessEnterNotifyEvent(xEvent *event)
{
    XtObject	*obj = NULL;
    char	*argv[10];
    int		argc = 0;

    if (event->u.u.type != EnterNotify) {
	Error(EL_ERROR, "RAP::ProcessEnterNotifyEvent",
	      "Why did we get called!?", NULL);
	return;
    }
    
    obj = client->GetXtObject((Window) event->u.enterLeave.event);
    
    if (obj == NULL) {
	char tmp[16];
	sprintf(tmp, "%lx", event->u.keyButtonPointer.event);
	Error(EL_ERROR, "RAP::ProcessEnterNotifyEvent",
	      "No XtObject found for window ", tmp, NULL);
	return;
    } else {
        argv[argc++] = Strdup(obj->GetUniqueName());
    }

    argv[argc++] = Strdup(client->GetUniqueName());

    Action::Invoke("EnterNotify", argc, argv);

    while (argc--) delete [] argv[argc];
}

//
// Currently, our processing for map and unmap notifies doesn't really do
// much except fire actions.  For simplicity, we may even want to remove
// these calls (this would also save on some bandwidth).
//
// ProcessMapNotify does ensure that out grabs remain on our windows though.
//
void
RAP::ProcessUnmapNotifyEvent(xEvent *event)
{
    XtObject *obj = NULL;

    if (event->u.u.type != UnmapNotify) {
	Error(EL_ERROR, "RAP::ProcessUnmapNotifyEvent",
	      "Why did we get called!?", NULL);
	return;
    }
    
    obj = client->GetXtObject((Window)event->u.unmapNotify.window);
    
    if (obj == NULL) {
	char tmp[64];
	sprintf(tmp, "%lx", event->u.keyButtonPointer.event);
	Error(EL_ERROR, "RAP::ProcessUnmapNotifyEvent",
	      "UnmapNotify in window not in our model: ", tmp, NULL);
	return;
    }

    ProcessWidgetUnmapped(obj);
}

void
RAP::ProcessMapNotifyEvent(xEvent *event)
{
    XtObject *obj = NULL;

    if (event->u.u.type != MapNotify) {
	Error(EL_ERROR, "RAP::ProcessMapNotifyEvent", 
	      "why did we get called!?", NULL);
	return;
    }
    
    obj = client->GetXtObject((Window)event->u.mapNotify.window);
    
    if (obj == NULL) {
	char tmp[64];
	sprintf(tmp, "%lx", event->u.keyButtonPointer.event);
	Error(EL_ERROR, "RAP::ProcessMapNotifyEvent",
	      "MapNotify in window not in our model: ", tmp, NULL);
	return;
    }

    ProcessWidgetMapped(obj);
}

//
// ButtonPresses don't do anything now except cause an action to be fired.
// If eventually we want to keep the auditory state in sync with the visual
// state, we might want to do some more stuff here (at least play the
// select sound when a seletion occurs).
//
void
RAP::ProcessButtonPressEvent(xEvent *event)
{
    XtObject	*obj = NULL;
    char	*argv[10];
    int		argc = 0;
    char	tmp[16];

    if (event->u.u.type != ButtonPress) {
	Error(EL_ERROR, "RAP::ProcessButtonPressEvent",
	      "why did we get called!?", NULL);
	return;
    }
   
    obj = client->GetXtObject((Window) event->u.keyButtonPointer.event);
    
    if (obj == NULL) {
	sprintf(tmp, "%lx (%ld)", event->u.keyButtonPointer.event,
		event->u.keyButtonPointer.event);
	Error(EL_ERROR, "RAP::ProcessButtonPressEvent",
	      "ButtonPress in window not in our model: ", tmp, NULL);
	return;
    } else {
        argv[argc++] = Strdup(obj->GetUniqueName());
    }


    sprintf(tmp, "%d", event->u.keyButtonPointer.state);
    argv[argc++] = Strdup(tmp);
    argv[argc++] = Strdup(client->GetUniqueName());

    Action::Invoke("ButtonPress", argc, argv);

    while (argc--) delete [] argv[argc];
}

//
// This routine encapsulates all of the junk that we have to do when we
// determine that a toplevel has been mapped.
//
// There are two cases under which this routine may be called:  (1) If Mercator
// has already "stabilized" its connection to an application and a new
// window is mapped (for example, a popup appears).  In this case we probably
// already have a current location (and maybe a current client), but we
// might have to warp to the popup.  (2) We've just connected to an application
// and are retrieving its widget tree.  We probably need to set the initial
// location within the app (and, if it is the first app, set the initial
// value of currentClient), but we only want to warp to the first popup
// created, not successive ones.
//
// The 'initial' parameter indicates whether we're in case (2).
//
// This routine does the following:
//   - Invokes the ShellMapped action if we're not in the initial state.
//   - Sets the popReturn member on the object if we're not in the initial
//     state.
//   - If the client has no current location, and the mapped object is
//     navigable, set the client's current location to it.
//   - If there is no current client, set the current client to the client
//     owning the viable (mapped, navigable) object and WARP to it.
//   - If we're not in the initial case, treat the object as a popup and
//     WARP to it.
//
void
RAP::ProcessToplevelMapped(XtObject *xtobj, int initial)
{
    char *argv[10];
    int argc = 0;
    int navigable = 0;
    int warped = 0;

    if (xtobj == NULL) {
	Error(EL_ERROR, "RAP::ProcessToplevelMapped",
	      "NULL object argument", NULL);
	return;
    }

    //
    // Fire the ShellMapped action, only if it appears after we've stabilized.
    //
    if (!initial) {
        argv[argc++] = Strdup(xtobj->GetUniqueName());
        argv[argc++] = Strdup(xtobj->GetClient()->GetUniqueName());
   
        Error(EL_DEBUG, "RAP::ProcessToplevelMapped",
              "CALLING ShellMapped for ", xtobj->GetName(), NULL);
 
        Action::Invoke("ShellMapped", argc, argv);
 
        while (argc--) delete [] argv[argc];
    }
    //
    // Set the popReturn member on the newly-mapped object.  If we're being
    // called as the widget tree is being downloaded, we don't know the
    // causality of any of the popups, so we can't logically set a popreturn
    // for them.  Only set a "meaningful" popReturn if initial == FALSE.  If
    // initial == TRUE, popReturn will retain its initialized value of NULL.
    //
    if (!initial) {
        //
        // If this is the first shell in the application then we'll have no
        // return point from it.  Otherwise the current location is the return
        // point.
        //
        if (client->GetCurrentLocation() == NULL) {
            Error(EL_DEBUG, "RAP::ProcessToplevelMapped",
                  "FIRST SHELL IS MAPPED, NO RETURN: ", xtobj->GetName(), NULL);
            xtobj->SetPopReturn(NULL);
        } else {
            Error(EL_DEBUG, "RAP::ProcessToplevelMapped",
                  "SHELL, NOT FIRST, IS MAPPED: ", xtobj->GetName(),
                  "RETURN IS ", client->GetCurrentLocation()->GetName(), NULL);
            xtobj->SetPopReturn(client->GetCurrentLocation());
        }
    }
    //
    // We want to set this client's current location to this newly-mapped
    // shell IF:
    //    - There is no current location for this client.  AND
    //    - The shell is navigable.
    //
    // IF we do indeed have a viable location within a client, then we should
    // check to see if current client is set at all; it's possible that this
    // is the first client.  If this is in fact the first client, then we
    // should actually move to it.
    //
 
    navigable = Template::GetBoolean(*xtobj, "navigable");
 
    if (navigable && client->GetCurrentLocation() == NULL) {
        Error(EL_DEBUG, "RAP::ProcessToplevelMapped",
              "SETTING INITIAL CLIENT LOCATION: ", xtobj->GetName(), NULL);
 
        client->SetCurrentLocation(xtobj);
 
        if (mercator->GetAppModelMgr()->GetCurrentClient() == NULL) {
            mercator->GetAppModelMgr()->SetCurrentClient(client);
 
            Error(EL_DEBUG, "RAP::ProcessToplevelMapped",
                  "SETTING INITIAL CLIENT AND WARPING TO ",
                  xtobj->GetName(), NULL);
 
            argv[0] = Strdup(xtobj->GetUniqueName());
            Action::Invoke("GoTo", 1, argv);
            delete [] argv[0];
 
            warped = 1;
        }
    }
    //
    // If we're not in the initial case AND the object is navigable, treat
    // the object as a popup and warp to it.
    //
    // If we've already warped, above, then we don't need to do it again.
    //
    if (!initial && navigable && !warped) {
        Error(EL_DEBUG, "RAP::ProcessToplevelMapped",
              "WARPING to POPUP ", xtobj->GetName(), NULL);
 
        argv[0] = Strdup(xtobj->GetUniqueName());
        Action::Invoke("GoTo", 1, argv);
        delete [] argv[0];
    }
}

//
// ProcessWidgetMapped may have one potential problem still in it:  it
// should probably check to see that all of the object's ancestors are
// mapped before setting map state to TRUE.  An object cannot be mapped
// if one of its ancestors is not mapped.
//
// NOTE THE SPECIAL CASE FOR TOPLEVELS.  If the widget is a toplevel, then
// the window manager will be responsible for mapping it, and hence we
// can't trust this RAP-generated map notification.  If the object is a
// toplevel, we basically bail without
//	- Setting map state to TRUE
//	- Grabbing keys
//	- Calling ProcessToplevelMapped
//	- Dinking with the object's children.
// We wait for XServer::HandleMapNotify to do these things when the
// object is "really" mapped.
//
void
RAP::ProcessWidgetMapped(XtObject *xtobj)
{
    char   *argv[10];
    int    argc = 0;

    if (xtobj == NULL) {
	Error(EL_ERROR, "RAP::ProcessWidgetMapped",
	      "NULL object passed.", NULL);
	return;
    } else {
	Error(EL_DEBUG, "RAP::ProcessWidgetMapped",
	      "Widget ", xtobj->GetName(), " is MAPPED.", NULL);
    }

    if (xtobj->GetMapped())
	return;

    ///// ADDED BY BETH+KEITH
    //if (client->IsToplevelWindow(xtobj))
    //	return;
    /////

    if (client->IsToplevelWindow(xtobj)) {
	if (mercator->GetXServer()->IsMappedToplevel(xtobj->GetWindowId()))
	    ProcessXServerMapNotify(xtobj->GetWindowId());
        return;
    }

    xtobj->SetMapped(True);
    if (!xtobj->GrabbedYet())
	mercator->GetXServer()->GrabInThisWindow(xtobj,1);

    argv[argc++] = Strdup(xtobj->GetUniqueName());
    argv[argc++] = Strdup(client->GetUniqueName());
    Action::Invoke("MapNotify", argc, argv);
    while (argc--) delete [] argv[argc];
    
    //
    // Now we have to toggle the map state of all of the children of the
    // object which was just mapped.
    //
    MapAndGrabChildren(xtobj, TRUE);
}

void
RAP::ProcessWidgetUnmapped(XtObject *xtobj)
{
    XtObject *gotoObj = NULL;
    int done = 0;
    char *argv[10];
    int argc = 0;

    if (xtobj == NULL) {
	Error(EL_ERROR, "RAP::ProcessWidgetUnmapped",
	      "NULL object passed.", NULL);
	return;
    } else {
	Error(EL_DEBUG, "RAP::ProcessWidgetUnmapped",
	      "Widget ", xtobj->GetName(), " is UNMAPPED.", NULL);
    }
    if (!xtobj->GetMapped())
	return;

    xtobj->SetMapped(False);
	
    // We need to do the same stuff as in Delete -- if the current
    // object was what was unmapped, we need a new location.
	
    // ONE SLIGHT DIFFERENCE between here and delete.  If we're in
    // an interior widget in a popup and XtPopdown is called, we only
    // get an unmap for the parent (this implicitly unmaps all children,
    // but we don't get an actual message for them).  So in determining
    // if the current object is the one being mapped, we also have to
    // consider whether its parent is being unmapped.

    // First, if this is the current object, we need to figure out where
    // to go to before we start unmapping things.
	
    XtObject *tester = client->GetCurrentLocation();
	
    while (!done) {
	if (tester == NULL) {
	    // We got to the top and it looks like nothing between the
	    // current location and the root was being unmapped.  No
	    // problem.
	    done = 1;
	} else if (tester == xtobj) {
	    // We found the thing being unmapped.
	    done = 1;
	} else {
	    tester = tester->GetParent();
	}
    }
	
    // At this point, if tester is set, then we have to worry about
    // changing locations.
	
    if (tester != NULL) {
	Error(EL_DEBUG, "RAP::ProcessWidgetUnmapped",
	      "CURRENT LOCATION (OR ITS PARENT) IS BEING UNMAPPED: ",
	      xtobj->GetName(), NULL);
	//
	// If it's a toplevel, then we will want to return to its
	// popReturn.  We may just want to go to the popReturn anytime
	// it's non-NULL (and scrap the toplevel check stuff).
	//
	if (client->IsToplevelWindow(tester)) {
	    if (tester->GetPopReturn() == NULL) {
		Error(EL_ERROR, "RAP::ProcessWidgetUnmapped",
		      "OBJECT BEING UNMAPPED IS A TOPLEVEL BUT IT"
		      " HAS NO POPUP RETURN.\n", NULL);
		gotoObj = NULL;
	    } else {
		Error(EL_DEBUG, "RAP::ProcessWidgetUnmapped",
		      "OBJECT BEING UNMAPPED IS A TOPLEVEL.  "
		      "ITS RETURN IS ", tester->GetPopReturn()->GetName(),
		      NULL);
		Action::Invoke("PopupReturn", 0, NULL);
		gotoObj = tester->GetPopReturn();
	    }
	} else {
	    int done = 0;
	    //
	    // The object we're unmapping was *not* a toplevel, so
	    // loop up the hierarchy, searching for a parent which is
	    // navigable.
	    //
	    // WE KNOW we at least have to start at tester's parent,
	    // since it's the thing being unmapped.
	    //
	    gotoObj = tester->GetParent();
	    while (!done) {
		if (gotoObj == NULL) {
		    // We're fucked!  We should probably change shells, or
		    // even change clients.
		    Error(EL_ERROR, "RAP::ProcessWidgetUnmapped",
			  "CAN'T FIND A NAVIGABLE PARENT FOR OBJECT BEING"
			  " UNMAPPED.  object = ", xtobj->GetName(), NULL);
		    done = 1;
		} else {
		    if (Template::GetBoolean(*gotoObj, "navigable"))
			done = 1;
		    else
			gotoObj = gotoObj->GetParent();
		}
	    }
	}
	    
	// END OF THE CASE where the object being unmapped is the current
	// object.  If we have a gotoObj set, then make it the current
	// object.
	
	if (gotoObj != NULL) {
	    Error(EL_DEBUG, "RAP::ProcessStateChangeRequest",
		  "THE OBJECT WE WERE ON WAS UNMAPPED (", xtobj->GetName(),
		  ").  GOING TO ", gotoObj->GetName(), NULL);
	    
	    // See the comment in the Delete clause for why we call out
	    // to TCL rather than just calling SetCurrentLocation.
	     
	    char *argv[1];
	    int result = 0;
	    
	    argv[0] = Strdup(gotoObj->GetUniqueName());
	    result = Action::Invoke("GoTo", 1, argv);
	    delete [] argv[0];
	    
	    if (result) {
		Error(EL_DEBUG, "RAP::ProcessWidgetUnmapped",
		      "SUCCESSFULLY called TCL Action GoTo to move.",
		      NULL);
	    } else {
		Error(EL_ERROR, "RAP::ProcessStateChangeNotify",
		      "Call of TCL Action GoTo FAILED!", NULL);
	    }
	}
    }
	
    argv[argc++] = Strdup(xtobj->GetUniqueName());
    argv[argc++] = Strdup(client->GetUniqueName());
    Action::Invoke("XtObjectUnmapped", argc, argv);
    while (argc--) delete [] argv[argc];
    
    // Now recursively mark all our children as unmapped.
    MapAndGrabChildren(xtobj, FALSE);
}

//
// SEE THE NOTE in ProcessWidgetMapped about why we bail out of this
// code when we see a toplevel being realized.
//
void
RAP::ProcessWidgetRealized(XtObject *xtobj, Window w)
{
    if (xtobj == NULL) {
	Error(EL_ERROR, "RAP::ProcessWidgetRealized",
	      "NULL object passed.", NULL);
    } else {
	Error(EL_DEBUG, "RAP::ProcessWidgetRealized",
	      "Widget ", xtobj->GetName(), " is REALIZED.", NULL);
    }

    char *argv[10];
    int argc = 0;
    argv[argc++] = Strdup(xtobj->GetUniqueName());
    Action::Invoke("XtObjectRealized", argc, argv);
    delete [] argv[--argc];
    
    mercator->GetAppModelMgr()->ChangeWindowId(xtobj, w);
    
    //
    // XtRealizeWidget does the following (see man page):
    //    - Makes a post-order traversal of the widget tree rooted
    //      at the specified object and called the change_managed
    //      procedure of each composite widget that has one or more
    //      managed children (DO WE HAVE TO WORRY ABOUT THIS?  WHAT
    //      DOES THIS MEAN?)
    //    - If the object is a subclass of compositeWidgetClass, it
    //      realizes each managed child.  It also maps all managed
    //      children that have mapped_when_managed set to TRUE.
    //
    // I'm not sure what the first case entails, so I'm not doing anything
    // about it yet.
    //
    // For the second case there are two subparts.
    // 
    // For the first (realizing each managed child), I'm not doing anything
    // YET, pending on figuring out whether we'll get a WidgetRealized
    // message for the managed children or not.
    //
    // For the second, I'll call into MapAndGrabChildren and let it figure
    // out whether or not to map.
    //
    // The manual goes on to say:
    // {various realization-oriented routines} maintain the following
    // invariants:
    //    - If a widget is realized, then all of its managed children are
    //	    REALIZED.
    //    - If a widget is realized, then all of its managed children which
    //      are mapped_when_managed are MAPPED.
    //
    
    // WE STILL NEED TO REALIZE OUR CHILDREN, IF IT DOESN'T HAPPEN
    // AUTOMATICALLY.
  
    // Can we assume that if we're being realized that we MUST have been
    // managed?  Let's try that assumption and see what happens.
    xtobj->SetManaged(TRUE);

    ///// ADDED BY BETH+KEITH
    //if (client->IsToplevelWindow(xtobj))
    //	return;
    /////

    if (client->IsToplevelWindow(xtobj)) {
	if (mercator->GetXServer()->IsMappedToplevel(w))
	    ProcessXServerMapNotify(w);
	return;
    }

    // Should we check mapped_when_managed here too?  Hmmmm...
    xtobj->SetMapped(True);
	
    if (!xtobj->GrabbedYet())
	mercator->GetXServer()->GrabInThisWindow(xtobj,1);
    
    MapAndGrabChildren(xtobj, True);
}

void
RAP::ProcessWidgetChildrenManaged(XtObject *xtobj, CARD32 *widgets,
				  CARD32 numWidgets)
{
    register int i;

    if (xtobj == NULL) {
	Error(EL_ERROR, "RAP::ProcessWidgetChildrenManaged",
	      "NULL object passed.", NULL);
    }

    for (i=0 ; i<numWidgets ; i++) {
	XtObject *child = client->GetXtObject((Object) widgets[i]);
	if (child)
	    ProcessWidgetManaged(child);
	else {
	    char tmp[16];
	    sprintf(tmp, "%lx", widgets[i]);
	    Error(EL_WARNING, "RAP::ProcessWidgetChildrenManaged",
		  "Unknown widget reported as managed: ", tmp, NULL);
	}
    }
}

void
RAP::ProcessWidgetChildrenUnmanaged(XtObject *xtobj, CARD32 *widgets,
				    CARD32 numWidgets)
{
    register int i;

    if (xtobj == NULL) {
	Error(EL_ERROR, "RAP::ProcessWidgetChildrenUnmanaged",
	      "NULL object passed.", NULL);
    }

    for (i=0 ; i<numWidgets ; i++) {
	XtObject *child = client->GetXtObject(widgets[i]);
	if (child)
	    ProcessWidgetUnmanaged(child);
	else {
	    char tmp[16];
	    sprintf(tmp, "%lx", widgets[i]);
	    Error(EL_WARNING, "RAP::ProcessWidgetChildrenUnanaged",
		  "Unknown widget reported as unmanaged: ", tmp, NULL);
	}
    }
}

void
RAP::ProcessWidgetUnrealized(XtObject *xtobj)
{
    if (xtobj == NULL) {
	Error(EL_ERROR, "RAP::ProcessWidgetUnrealized",
	      "NULL object passed.", NULL);
    } else {
	Error(EL_DEBUG, "RAP::ProcessWidgetUnrealized",
	      "WIDGET UNREALIZED: ", xtobj->GetName(),
	      NULL);
    }
    
    char *argv[10];
    int argc = 0;
    argv[argc++] = Strdup(xtobj->GetUniqueName());
    Action::Invoke("XtObjectUnrealized", argc, argv);
    delete [] argv[--argc];
    
    //
    // According to the man page, XtUnrealizeWidget does the following:
    //    - Destroys the window of the current widget.
    //    - Destroys the window of the children of the current widget
    //      recursively.
    //    - Unmanage any widgets whose windows it destroys.
    //
    mercator->GetAppModelMgr()->ChangeWindowId(xtobj, RapUnrealized);
    
    xtobj->SetManaged(FALSE);
    xtobj->SetMapped(FALSE);
    
    MapAndGrabChildren(xtobj, FALSE);
    ManageChildren(xtobj, FALSE);
}

void
RAP::ProcessWidgetManaged(XtObject *xtobj)
{
    if (xtobj == NULL) {
	Error(EL_ERROR, "RAP::ProcessWidgetManaged",
	      "NULL object passed.", NULL);
    } else {
	Error(EL_DEBUG, "RAP::ProcessWidgetManaged",
	      "WIDGET MANAGED:  ", xtobj->GetName(), NULL);
    }

    xtobj->SetManaged(TRUE);
    
    char *argv[10];
    int argc = 0;
    argv[argc++] = Strdup(xtobj->GetUniqueName());
    Action::Invoke("XtObjectManaged", argc, argv);
    delete [] argv[--argc];
    
    //
    // The man page states that XtManageChildren does the following:
    //    - Calls the change_managed routine of the widget's parent.
    //      (DO WE HAVE TO WORRY ABOUT THIS?)
    //    - If our parent is realized then XtRealizeWidget is called for
    //      its children (and thus, we don't have to worry about twiddling
    //      the realize state since we'll get an extra message for that.
    //    - Maps each previously unmanaged child that has map_when_managed
    //      set to TRUE.
    //
    // I only deal with the third case, since it looks like the other
    // two are noops.
    //
    //
    ManageChildren(xtobj, TRUE);
    MapAndGrabChildren(xtobj, TRUE);
}

void
RAP::ProcessWidgetUnmanaged(XtObject *xtobj)
{
    if (xtobj == NULL) {
	Error(EL_ERROR, "RAP::ProcessWidgetUnmanaged",
	      "NULL object passed.", NULL);
    } else {
	Error(EL_ERROR, "RAP::ProcessWidgetUnmanaged",
	      "WIDGET UNMANAGED:  ", xtobj->GetName(),
	      NULL);
    }

    // If we're already unmanaged, do nothing.
    if (xtobj->GetManaged() == FALSE)
	return;
    
    xtobj->SetManaged(FALSE);
    
    char *argv[10];
    int argc = 0;
    argv[argc++] = Strdup(xtobj->GetUniqueName());
    Action::Invoke("XtObjectUnmanaged", argc, argv);
    delete [] argv[--argc];
    
    //
    // The man page states that XtUnmanageChildren does the following:
    //    - Ignores the widget if it is already unmanaged.
    //    - If the widget is realized, it is made nonvisible via
    //      unmapping.
    //
    // I call MapAndGrabChildren to unrealize the child widgets.
    //
    //
    ManageChildren(xtobj, FALSE);
    MapAndGrabChildren(xtobj, FALSE);
}


////////////////////////////////////////////////////////////////////////

//
// MapAndGrabChildren is called recursively from ProcessWidgetMapped to
// map an object's children and establish the required key grabs on it.
// There are two code paths depending on the value of the mapState argument.
//
// If mapState is TRUE, this routine checks to see that the object is 
// (1) managed, (2) the resource mapped_when_managed is set to TRUE, and 
// (3) is realized.  If all of these are true then the widget's map state
// is set to TRUE and the grab is placed on it.
//
// If mapState is FALSE, then the map state of all children is set to
// FALSE.  Do we need to ungrab on an unmapped window?  Probably not.
//

void
RAP::MapAndGrabChildren(XtObject *obj, int mapState)
{
    mDictValPtr<Object,XtObject> *children = obj->GetChildren();
    mDictValPtrIterator<Object,XtObject> iter(*children);

    // No children dictionary or empty children dictionary.
    if (children == NULL || children->size() == 0) {
	Error(EL_DEBUG, "RAP::MapAndGrabChildren", 
	      "obj ", obj->GetName(), " has no children.", NULL);
	return;
    }
    
    if (mapState) {
	// We're mapping.  Work through the dictionary and decide whether
	// or not to set the map state on the objects and their children.
	while (iter.next()) {
	    XtObject *current=children->lookup(iter.current());
	    // this is the old LEDA code
	    //	forall_dic_items(item, *children) {
	    //	    XtObject *current = children->info(item);
	    
	    //
	    // This *may* be somewhat questionable.  I'm doing this since
	    // popup performance went to hell in this code.  If the child
	    // is marked as already mapped, then we'll *assume* that 
	    // we've already done all the requisite crap to it and its
	    // subchildren.  This *seems* to be the case for Athena, where
	    // MapNotifies and RealizeNotifies work up from the bottom of
	    // the tree.  May not be the case for other toolkits, or
	    // "weird" Athena apps though.		-k
	    //
	    if (current->GetMapped())
		continue;
	    
	    if (current->GetManaged() && current->GetWindowId() != 0) {
		// Get the value of mapped_when_managed
		if (Template::GetBoolean(*current, "mappedWhenManaged")) {
		    // This widget is legitimately mapped!  Set its map state
		    // to TRUE, add a grab on it, and call its children
		    // recursively.
		    current->SetMapped(TRUE);
		    if (!current->GrabbedYet())
			mercator->GetXServer()->GrabInThisWindow(current, 1);
		    
		    Error(EL_DEBUG, "RAP::MapAndGrabChildren",
			  "mapped and grabbed object ",
			  current->GetName(), NULL);
		
		    MapAndGrabChildren(current, mapState);
		}
	    }
	}
    } else {
	// We're unmapping.  Loop through and set map state to FALSE for
	// all children.
	// keith's old LEDA code
	//	forall_dic_items(item, *children) {
	//	    XtObject *current = children->info(item);
	while (iter.next()) {
	    XtObject *current=children->lookup(iter.current());
	    //
	    // Same as above.  In an attempt to get a little more
	    // performance, if the child is already unmapped, then we'll believe
	    // that it's children have already been dealt with, and save
	    // the recusion.
	    //
	    if (!current->GetMapped())
		continue;
	    
	    current->SetMapped(FALSE);
	    
	    Error(EL_DEBUG, "RAP::MapAndGrabChildren",
		  "unmapped object ", obj->GetName(),  NULL);
	    
	    MapAndGrabChildren(current, mapState);
	}
    }
}
    
//
// ManageChildren is used to recursively set/unset the manage state
// for all of an object's children.
//
void
RAP::ManageChildren(XtObject *xtobj, int manageState)
{
    mDictValPtr<Object,XtObject> *children = xtobj->GetChildren();
    mDictValPtrIterator<Object,XtObject> iter(*children);

    
    // No children dictionary or empty children dictionary.
    if (children == NULL || children->size() == 0) {
	Error(EL_DEBUG, "RAP::ManageChildren", 
	      "obj ", xtobj->GetName(), " has no children.", NULL);
	return;
    }
    
    while (iter.next()) {
	XtObject *current = children->lookup(iter.current());
	
	current->SetManaged(manageState);
	
	if (manageState) {
	    Error(EL_DEBUG, "RAP::ManageChildren",
		  "object managed: ", xtobj->GetName(), NULL);
	} else {
	    Error(EL_DEBUG, "RAP::ManageChildren",
		  "object unmanaged: ", xtobj->GetName(), NULL);
	}
	
	ManageChildren(current, manageState);
    }
}
    
////////////////////////////////////////////////////////////////////////

//
// NOTE THAT since we're building on the C language API, these things
// take a RapClient argument.  This is obsolete and should go away once
// we move to a "true" C++ protocol API.
//
int
RAP::HelloRequest(int block)
{
    RapClient c = new RapClientStruct;
    int ret;
    
    c->window = clientWindow;
    c->iceConn = iceConn;
    c->majorOpcode = majorOpcode;
    c->connectStatus = connectStatus;
    c->fd = IceConnectionNumber(iceConn);

    if (IceConnectionStatus(iceConn) != IceConnectAccepted) {
	Error(EL_ERROR, "RAP::HelloRequest", "Connection NOT accepted", NULL);
	// If error, we should clean up.
	return 0;
    } else {
	ret = RAPHelloRequest(c);
    }

    return ret;
}

int
RAP::QueryTreeRequest(int block)
{
    RapClient c = new RapClientStruct;
    c->window = clientWindow;
    c->iceConn = iceConn;
    c->majorOpcode = majorOpcode;
    c->connectStatus = connectStatus;
    c->fd = IceConnectionNumber(iceConn);
    
    if (IceConnectionStatus(iceConn) != IceConnectAccepted) {
	Error(EL_ERROR, "RAP::QueryTreeRequest",
	      "Connection NOT accepted", NULL);
	// If error, we should clean up.
	return 0;
    } else {
	return RAPQueryTreeRequest(c);
    }
}


int
RAP::FullQueryTreeRequest(int block)
{
    RapClient c = new RapClientStruct;
    c->window = clientWindow;
    c->iceConn = iceConn;
    c->majorOpcode = majorOpcode;
    c->connectStatus = connectStatus;
    c->fd = IceConnectionNumber(iceConn);
    
    if (IceConnectionStatus(iceConn) != IceConnectAccepted) {
	Error(EL_ERROR, "RAP::FullQueryTreeRequest",
	      "Connection NOT accepted", NULL);
	// If error, we should clean up.
	return 0;
    } else {
	return RAPFullQueryTreeRequest(c);
    }
}

int
RAP::GetResourcesRequest(WidgetList widgets, long numWidgets, int block)
{
    RapClient c = new RapClientStruct;
    c->window = clientWindow;
    c->iceConn = iceConn;
    c->majorOpcode = majorOpcode;
    c->connectStatus = connectStatus;
    c->fd = IceConnectionNumber(iceConn);
    
    return RAPGetResourcesRequest(c, widgets, numWidgets);
}

int
RAP::GetValuesRequest(Object widget, const char **resourceNames,
		      long numResources, int block)
{
    RapClient c = new RapClientStruct;
    c->window = clientWindow;
    c->iceConn = iceConn;
    c->majorOpcode = majorOpcode;
    c->connectStatus = connectStatus;
    c->fd = IceConnectionNumber(iceConn);
    
    return RAPGetValuesRequest(c, (Widget) widget, resourceNames, numResources);
}

int
RAP::SetValuesRequest(Object object, const ResourceKeyValue *values,
		      long numKeyValues, int block)
{
    Error(EL_ERROR, "RAP::SetValuesRequest", "NOT IMPLEMENTED YET", NULL);
    return 0;
}

int
RAP::GetGCValuesRequest(unsigned long *gcList, long numGCs, int block)
{
    Error(EL_DEBUG, "RAP::GetGCValuesRequest", "SENDING...", NULL);

    RapClient c = new RapClientStruct;
    c->window = clientWindow;
    c->iceConn = iceConn;
    c->majorOpcode = majorOpcode;
    c->connectStatus = connectStatus;
    c->fd = IceConnectionNumber(iceConn);
    
    return RAPGetGCValuesRequest(c, (GContext *) gcList, numGCs);
}

#ifdef NOT_IMPLEMENTED

int
RAP::SetValuesRequest(const char *name, const char *type, const char *value,
		      WidgetList widgets, short numWidgets, int block)
{
    return RAPSetValuesRequest(c, name, type, value,
			       widgets, numWidgets);
}

int
RAP::WindowToObjectRequest(int block)
{
    return 0;
}

int
RAP::ObjectToWindowRequest(int block)
{
    return 0;
}

#endif /* NOT_IMPLEMENTED */


int
RAP::LocateObjectRequest(Object w, int block)
{
    RapClient c = new RapClientStruct;
    
    c->window = clientWindow;
    c->iceConn = iceConn;
    c->majorOpcode = majorOpcode;
    c->connectStatus = connectStatus;
    c->fd = IceConnectionNumber(iceConn);

    return RAPLocateObjectRequest(c, (WidgetList) &w, 1);
}

int
RAP::SelectEventRequest(short *events, short numEvents, int block)
{
    RapClient c = new RapClientStruct;
    c->window = clientWindow;
    c->iceConn = iceConn;
    c->majorOpcode = majorOpcode;
    c->connectStatus = connectStatus;
    c->fd = IceConnectionNumber(iceConn);
    
    return RAPSelectEventRequest(c, events, numEvents);
}

int
RAP::SelectRequestRequest(short *requests, short numRequests, int block)
{
    RapClient c = new RapClientStruct;
    c->window = clientWindow;
    c->iceConn = iceConn;
    c->majorOpcode = majorOpcode;
    c->connectStatus = connectStatus;
    c->fd = IceConnectionNumber(iceConn);
    
    return RAPSelectRequestRequest(c, requests, numRequests);
}

int
RAP::AddNotifyRequest(short *notifies, short numNotifies, int block)
{
    RapClient c = new RapClientStruct;
    c->window = clientWindow;
    c->iceConn = iceConn;
    c->majorOpcode = majorOpcode;
    c->connectStatus = connectStatus;
    c->fd = IceConnectionNumber(iceConn);
    
    return RAPAddNotifyRequest(c, notifies, numNotifies);
}

int
RAP::RemoveNotifyRequest(short *notifies, short numNotifies, int block)
{
    RapClient c = new RapClientStruct;
    c->window = clientWindow;
    c->iceConn = iceConn;
    c->majorOpcode = majorOpcode;
    c->connectStatus = connectStatus;
    c->fd = IceConnectionNumber(iceConn);
    
    return RAPRemoveNotifyRequest(c, notifies, numNotifies);
}

int
RAP::CloseConnectionRequest(int block)
{
    return 0;
}

//
// This thing gets called from 
//    - XServer when a new child of the root window is mapped.
//    - ProcessWidgetMapped if the widget is a toplevel and we've
//      seen it already be mapped.
//    - ProcessWidgetRealized if the widget is a toplevel and we've
//      seen it already be mapped.
//
// The basic idea is that we may get mapnotifies from the X protocol stream
// before OR after we get the RAP realize/map notifies.  So the XServer keeps
// a list of currently mapped windows.  As windows are mapped, we call this
// code if they're in our model.
//
// The RAP routines call this if the window being mapped or realized has
// already "really" been mapped, as told by XServer.
//
// We try to keep the knowledge in both places and have them cross-check
// to make sure that things will work regardless of the order that they
// come in.
//
void
RAP::ProcessXServerMapNotify(Window w)
{
    XtObject *obj = client->GetXtObject(w);

    if (!obj) {
	Error(EL_ERROR, "RAP::ProcessXServerMapNotify",
	      "A toplevel window is mapped, but is unknown in our model.",
	      NULL);
	return;
    }

    if (obj->GetMapped())
	return;

    obj->SetMapped(TRUE);

    if (!obj->GrabbedYet())
	mercator->GetXServer()->GrabInThisWindow(obj, 1);

    ProcessToplevelMapped(obj, FALSE);

    MapAndGrabChildren(obj, TRUE);
}

Client *
RAP::FindClientByWindow(Window w)
{
    Client *client = NULL;

    if ((client = RAP::clients.lookup(w)) == nil)
	return NULL;
    else
	return client;
}
