//
//
//
//	Note:  The use of this program shall be restricted to educational,
//	non-commercial use only.  All inquiries regarding licensing this
//	software for commercial use may be addressed to:
//	
//	Georgia Tech Research Corporation
//	Technology Licensing
//	Centennial Research Building - Rm. 275
//	400 Tenth Street, N.W.
//	Atlanta, Georgia  30332-0415
//	
//	*************************************************
//	
//	       Copyright 1996
//	Georgia Tech Research Corporation
//	    All Rights Reserved
//
//
//
//

#include <poll.h>

#include "RAPListener.h"
#include "RapProto.h"		// Protocol definition

extern "C" {
#include "agent.h"		// C language agent API
}
#include <X11/Xatom.h>		// For XA_STRING
#include "RAP.h"
#include "XServer.h"
#include "Error.h"
#include "Client.h"
#include "Mercator.h"
#include "AppModelMgr.h"

//
// There is one almighty global instance of "Mercator"
//
extern Mercator *mercator;

////////////////////////////////////////////////////////////////////////
//
// Xtea-related stuff
//
////////////////////////////////////////////////////////////////////////

const char *EXTAGENT_NAME	= "ExternalAgent";
const char *ICE_NETWORK_IDS	= "IceNetworkIds";

const long EXTAGENT_FORMAT	= 32;

// FOR NOW
static IcePaVersionRec protoVersions[1];

////////////////////////////////////////////////////////////////////////
//
// Class special functions.
//
////////////////////////////////////////////////////////////////////////

RAPListener::RAPListener(XServer& xserver) :
    FDInterest("RAPListener"),
    protocolVersions(new IcePaVersionRec),
    listeners(NULL),
    listenerCount(0),
    majorOpcode(-1),
    networkIds(NULL),
    agentWindow(NULL),
    protoNameAtom(NULL)
{
    IceSetIOErrorHandler(RAPListener::IOErrorHandler);
    
    protoVersions[0].major_version = RapMajorVersion;
    protoVersions[0].minor_version = RapMinorVersion;

    // Note protocol msg handler lives in RAP, not RAPListener.
    protoVersions[0].process_msg_proc = RAP::ProcessProtoMsgProc; 
}

RAPListener::~RAPListener()
{
    register int i;
    
    delete protocolVersions;

    for (i=0 ; i<listenerCount ; i++) {
	RemoveFDInterest(IceGetListenConnectionNumber(listeners[i]));
    }
    
    // Call ICE to free listeners, networkIds
    IceFreeListenObjs(listenerCount, listeners);
}


////////////////////////////////////////////////////////////////////////
//
// Static member functions - mostly ICE-required boilerplate.
//
////////////////////////////////////////////////////////////////////////

Bool
RAPListener::AuthorizeProc(char *hostname)
{
    Error(EL_DEBUG, "RAPListener::AuthorizeProc", "auth for host ",
	  hostname, NULL);
    return TRUE;
}

void
RAPListener::IOErrorHandler(IceConn iceConn)
{
    RAP *instance = NULL;
    
    if ((instance = RAP::instances.lookup(iceConn)) == 0) {
	Error(EL_ERROR, "RAPListener::IOErrorHandler",
	      "No client is associated with this ice connection.  ",
	      "What do we do?", NULL);
	return;
    }

    if (instance->GetClient())
        Error(EL_DEBUG, "RAPListener::IOErrorHandler", "Client ",
	      instance->GetClient()->ClientName(),
	      " seems to have exited.", NULL);
    else 
        Error(EL_DEBUG, "RAPListener::IOErrorHandler",
	      "A client seems to have exited.", NULL);

    instance->Shutdown(IceConnectionNumber(iceConn));
}

void
RAPListener::IOErrorProc(IceConn iceConn)
{
    RAP *instance = NULL;
    
    // Shutdown stuff?  This is what Will's sample code does.  I need to
    // understand how these two error procedures will get called better.

    if ((instance = RAP::instances.lookup(iceConn)) == 0) {
	Error(EL_ERROR, "RAPListener::IOErrorProc",
	      "No client is associated with this ice connection.  ",
	      "What do we do?", NULL);
	return;
    }

    Error(EL_DEBUG, "RAPListener::IOErrorProc", "Client ",
	  instance->GetClient()->ClientName(),
	  " seems to have exited.", NULL);

    instance->Shutdown(IceConnectionNumber(iceConn));
}

void
RAPListener::ProtoActivateProc(IceConn iceConn, IcePointer clientData)
{
    Error(EL_DEBUG, "RAPListener::ProtoActivateProc", "called.", NULL);
}

Status
RAPListener::ProtoSetupProc(IceConn iceConn, int majorVersion, int minorVersion,
			    char *vendorString, char *releaseString,
			    IcePointer *clientDataRet, char **failureReasonRet)
{
    char tmp[16];
    
    sprintf(tmp, "%d.%d", majorVersion, minorVersion);

    Error(EL_DEBUG, "RAPListener::ProtoSetupProc", "\n",
	  "\tVersion: ", tmp, "\n",
	  "\tVendor: ", vendorString, "\n",
	  "\tRelease: ", releaseString, NULL);

    return 1;
}

void
RAPListener::SetupAgent(XServer& xserver)
{
    XTextProperty	textProp;

    textProp.value = (unsigned char *) networkIds;
    textProp.encoding = XA_STRING;
    textProp.format = 8;
    textProp.nitems = strlen(networkIds);

    xserver.SetTextProperty(agentWindow, &textProp,
			    xserver.InternAtom(ICE_NETWORK_IDS, FALSE));
}

int
RAPListener::Register(XServer& xserver)
{
    Status status;
    char errorStr[256];
    register int i;
    
    majorOpcode = IceRegisterForProtocolReply(RapName,
					      "X Consortium",
					      "1.0",
					      1,
					      protoVersions,
					      0,
					      NULL,
					      NULL,
					      RAPListener::AuthorizeProc,
					      RAPListener::ProtoSetupProc,
					      RAPListener::ProtoActivateProc,
					      RAPListener::IOErrorProc);

    if (majorOpcode == -1) {
	Error(EL_FATAL, "RAPListener::Register",
	      "IceRegisterForProtocolReply failed!  Cannot continue.", NULL);
	return 0;
    }

    status = IceListenForConnections(&listenerCount,
				     &listeners,
				     256,
				     errorStr);

    if (!status) {
	Error(EL_FATAL, "RAPListener::Register",
	      "IceListenForConnections failed: ", errorStr, NULL);
	return 0;
    }

    RAPListener::networkIds = IceComposeNetworkIdList(listenerCount, listeners);

    RAPListener::agentWindow = xserver.CreateWindow(xserver.GetRootWindow(),
					    0, 0, 1, 1, 1,
					    xserver.GetBlackPixel(),
					    xserver.GetBlackPixel());

    if (agentWindow == None) {
	Error(EL_FATAL, "RAPListener::Register",
	      "Could not create agent window.", NULL);
	return 0;
    }

    SetupAgent(xserver);
    protoNameAtom = xserver.InternAtom(RapName, FALSE);

    for (i=0 ; i<listenerCount ; i++) {
	IceSetHostBasedAuthProc(listeners[i], RAPListener::AuthorizeProc);
	
	AddFDInterest(IceGetListenConnectionNumber(listeners[i]));
	SetIOMask(IceGetListenConnectionNumber(listeners[i]), ARead);
    }
    
    return 1;
}

int
RAPListener::InitiateConnection(XServer& xserver, Window clientWindow)
{
    XClientMessageEvent cme;
    char tmp[16], tmpHex[16];
    
    sprintf(tmp, "%ld", clientWindow);
    sprintf(tmpHex, "%lx", clientWindow);

    //
    // We used to send an event to ANY window which is not the application
    // shell window for an already-seen client.  Instead, we now send
    // an event to any window which is mapped on the root window and
    // is not in our model.
    //
    if (mercator->GetAppModelMgr()->GetXtObject(clientWindow)) 
	return FALSE;
#if 0
    //
    // Return if this window ID has already been seen before.
    //
    if (RAP::clients.lookup(clientWindow) != nil) {
	Error(EL_DEBUG, "RAPListener::InitiateConnection",
	      "Already seen window ", tmp, "(", tmpHex,
	      "), ignoring...", NULL);
	return FALSE;
    }
#endif
    Error(EL_DEBUG, "RAPListener::InitiateConnection",
	  "Initiating connection to client window ", tmp, "(",
	  tmpHex, ")...", NULL);
    
    cme.window = clientWindow;
    cme.type = ClientMessage;
    cme.message_type = xserver.InternAtom(EXTAGENT_NAME, FALSE);
    cme.format = EXTAGENT_FORMAT;
    cme.data.l[0] = protoNameAtom;
    cme.data.l[1] = xserver.InternAtom(ICE_NETWORK_IDS, FALSE);
    cme.data.l[2] = agentWindow;
    cme.data.l[3] = 0L;
    cme.data.l[4] = 0L;

    if (!xserver.SendEvent(clientWindow, TRUE, 0, (XEvent *) &cme))
	Error(EL_ERROR, "RAPListener::InitiateConnection",
	      "SendEvent failed!", NULL);

    xserver.Flush();

    return TRUE;
}



////////////////////////////////////////////////////////////////////////
//
// Member functions.
//
////////////////////////////////////////////////////////////////////////

int
RAPListener::HandleActivity(int _fd, short mask)
{
    // we probably don't have to do this... let ICE do it for us...
    if (mask & ARead) {
	IceConn newIceConn;
	IceListenObj l;
	IceAcceptStatus statusRet;
	register int i;
	RAP *rap;
	
	//
	// Find the listener object which corresponds to this descriptor.
	// This is cheezy, but (1) we probably won't have too many
	// listeners, and (2) connection setup doesn't happen too often.
	//
	for (i=0 ; i<listenerCount ; i++) {
	    if (_fd == IceGetListenConnectionNumber(listeners[i])) {
		l = listeners[i];
	    }
	}

	Error(EL_DEBUG, "RAPListener::HandleActivity",
	      "connection request on addr ",
	      IceGetListenConnectionString(l), NULL);

	newIceConn = IceAcceptConnection(l, &statusRet);

	if (statusRet != IceAcceptSuccess) {
	    Error(EL_ERROR, "RAPListener::HandleActivity",
		  "Could not accept ICE connection.", NULL);
	    return 0;
	}

	//
	// At this point we've been contacted by a RAP-speaking client and
	// we're engaged in the startup negotiation process.  We create a
	// new RAP instance to process the connection; it'll finish the
	// connection process and continue on to hanlding actual RAP
	// messages.  We need to padd it the IceConn and major opcode for
	// the protocol.
	//
	rap = new RAP(newIceConn, majorOpcode);

	return 1;
    }
    
    if (mask & AWrite) {
	// IGNORE?

	return 1;
    }
    if (mask & AExcept) {
	//
	// An exceptional condition has occurred on a listener descriptor.
	// In theory these should never happen; if we get an exception,
	// shut down Mercator.
	//
	if (mask & POLLERR) 
	    Error(EL_ERROR, "RAPListener::HandleActivity",
		  "Exceptional condition detected: POLLERR\n",
		  "\tCannot continue.", NULL);
	if (mask & POLLHUP)
	    Error(EL_ERROR, "RAPListener::HandleActivity",
		  "Exceptional condition detected: POLLHUP\n",
		  "\tCannot continue.", NULL);
	if (mask & POLLNVAL)
	    Error(EL_ERROR, "RAPListener::HandleActivity",
		  "Exceptional condition detected: POLLNVAL\n",
		  "\tCannot continue.", NULL);

	this->MarkForCleanup();
	FDInterest::Quit();

	return 0;
    }
}

