//
//
//
//	Note:  The use of this program shall be restricted to educational,
//	non-commercial use only.  All inquiries regarding licensing this
//	software for commercial use may be addressed to:
//	
//	Georgia Tech Research Corporation
//	Technology Licensing
//	Centennial Research Building - Rm. 275
//	400 Tenth Street, N.W.
//	Atlanta, Georgia  30332-0415
//	
//	*************************************************
//	
//	       Copyright 1996
//	Georgia Tech Research Corporation
//	    All Rights Reserved
//
//
//
//
//
// NOTE use of class hiding here to hide details, reduce recompilations, and
// speed compile times.
//
// [ I deleted Keith's not of 17mar95 which was about Rogue Wave.]
// We no longer are using Rogue Wave directly, but rather are doing
// our own data structures which are currently implemented with 
// Rogue wave.
//

#include <strstream.h>
#include <X11/StringDefs.h>

#include "Resource.h"
#include "Error.h"

//
// String hasher for dictionaries.
//
static unsigned StringHasher(const mString& str) { return str.hash(); }

//
// The resource type manager provides type-based services for the rest
// of the resource management infrastructure.  The RTM allows us to register
// procedures which describe how different resource types will be handled;
// in essence, it lets us associate semantics with resource type names.
//
// Users must bind three functions to each type registered in the type manager.
//	- Copy		Return a copy of the data, cast a a char *.
//	- Free		Delete the data if it would have been allocated off the heap.
//	- Print		Print the data to an ostream.
//	- Size		Return the size of the data.
//
// It is the ResourceTypeManager (or more accurately, the routines provided
// to it) which must decide whether to return data as an immediate or as an
// indirect value, depending on the size of the word length of the machine (eg,
// on 32-bit machines, 64-bit values should be returned indirect.  But on 64-bit
// machines they should be returned immediate).  The RTM must also decide
// whether storage for a type would have been dynamically allocated (that is,
// the size is greater than the natural word length, so it would be stored as
// an indirect) and hence whether or not it must free it.
//
// Each type also has a "StorageClass" associated with it which indicates how
// the type is actually represented.  The current storage types are 8-bits,
// 16-bits, 32-bits, 64-bits, NULL-terminated character arrays, and sized
// byte arrays (for storing values larger than the native word size, or for
// storing variable length, but non-NULL-terminated data).
//
// NOTE that the code only currently supports 8, 16, 32, and string storage
// classes.
//
// StorageClass is *different* from the size indicator, because string (and
// possibly other types) have variable lengths.  This is an important distinction
// because in the _ResourceValue code for converting to and from system-
// provided types, we must do a check to see if the conversion is legitimate.
// A simple size check is insufficient in the case of variable-length data.  Hence
// the StorageClass.
//
//  A "helper" function, InitializeResourceTypeManager(), is where a few common
// types which are already understood by Mercator reside.
//
// Note that the ResourceTypeManager class provides static functions only (it's
// a class to prevent namespace pollution).  It should never be instantiated.
//
// The ResourceTypeManager gets and passes all values as pointers.  The
// primary caller of this class, _ResourceValue, does its own imposition of
// casted types on top of pointers.  This code will probably fail to work on
// machines where sizeof(ulong) > sizeof(char *), since we're casting ulongs
// to char *'s to get them into the ResourceTypeManager.
//

//
// Stupid !@#% LEDA chokes if we use a scoped type name in one of its
// class declaration macros.  So I've had to move the declaration of
// TypeInformation outside the ResourceTypeManager.  This should be
// fixed if (WHEN!) we get rid of LEDA.
//
// The Copy, ... , functions should also be nested, but there's a cyclic
// reference problem if they are (all caused by TypeInformation having
// to be outside ResourceTypeManager).  If this gets moved, also move
// the typedefs for the Copy, ... , functions.
//
typedef char *(*CopyFunc)(const char *v);
typedef void (*FreeFunc)(char **v);
typedef ostream& (*PrintFunc)(const char *v, ostream& str);
typedef ulong (*SizeFunc)(const char *v);

typedef enum { TI_8, TI_16, TI_32, TI_64, TI_STRING, TI_SIZED } StorageClass_t;

typedef struct _TypeInformation {
    StorageClass_t	storageClass;
    CopyFunc		copyFunc;
    FreeFunc		freeFunc;
    PrintFunc		printFunc;
    SizeFunc		sizeFunc;
} TypeInformation, *TypeInformationPtr;

//
// This could profitably be made an mDictValPtr, and replace const char *
// arguments with const mString&'s.
//
typedef mDictPtrPtr<mString, TypeInformation> TypeInformationDict;

#ifdef BEAM
mDictPtrPtr<mString, _TypeInformation>::mDictPtrPtr(unsigned (*hashKey)(const mString&)) :
    impl(new RWTPtrHashDictionary<mString, _TypeInformation>(hashKey))
{
}
#endif

////////////////////////////////////////////////////////////////////////


class ResourceTypeManager {
private:
    ResourceTypeManager();
    static TypeInformationDict	types;
    
public:
    static int			Register(const char *type, StorageClass_t,
					 CopyFunc, FreeFunc, PrintFunc,
					 SizeFunc);
    static int			Registered(const char *type);
    static TypeInformation	*Find(const char *repType);
};



////////////////////////////////////////////////////////////////////////

TypeInformationDict ResourceTypeManager::types(StringHasher);

////////////////////////////////////////////////////////////////////////



class Resource {

    friend class ResourceCache;
    friend ostream& operator<<(ostream&, const Resource&);
    
private:
    //
    // Resource value is where all of the size-dependent junk associated
    // with storing resource values goes on.  Basically all values are represented
    // by a char *.  If the size of the thing being stored is <= sizeof(char*),
    // then it is stored as an immediate in this variable (via casting, ick).
    // If the size is greater than can be accomodated in a char *, the value
    // variable is interpreted as a pointer.
    //
    // _ResourceValue communicates with the ResourceTypeManager to
    // determine how a particular value will be stored (depending on its
    // size).  The RTM also handles copying and freeing of values, since the
    // knowledge of whether a value is going to be allocated off the heap is
    // type-dependent.
    //
    // This code supports conversion to and from a handful of system-provided
    // types which I felt were (1) common, and (2) easily mapped onto likely
    // types for actual resource values.  Right now I'm using unchar, ushort,
    // ulong, and const char * as my supported types.  This is non-portable, as
    // the size of a ulong may not be 32-bits (as assumed by the StorageClass
    // stuff in the ResourceTypeManager).  For this reason, the type names should
    // be replaced with CARD8, CARD32 and the like, so that we get data with
    // sizes that we expect.
    //
    // Access to the value itself is gained via several methods.  For setting
    // the value, versions of the Value() method are provided which take
    // unchars (1-byte), ushorts (2-bytes), ulongs (4-bytes), and strings
    // (NULL-terminated char arrays).  We currently don't support storage
    // of length>4, , or variable-length but non-NULL-terminated data.  All of
    // these take an optional repType argument which specifies the type of the
    // representation used to store the value.  If omitted, the current repType
    // is assumed (and if no repType has been set, an exception will be raised).
    //
    // Constructors and type converters are also supplied to-and-from the
    // supported formats.
    //
    // Any size mismatches will cause an exception to be raised.
    //
    // The goal with this class is to completely hide any size-related
    // shenanigans from callers.  The class is fairly large, but we prevent
    // duplication of effort from spreading through our callers.  We also
    // gain an ease of notational use in callers.
    //
    class _ResourceValue {

	friend ostream& operator<<(ostream&, const _ResourceValue&);
	
    private:
	char				*repType;
	char				*value;
	// This points into data maintained by the RTM.  Never delete it!
	TypeInformation			*typeInfo;
	
    public:
	// Empty, uninitialized value.
	_ResourceValue();
	// Copy constructor.
	_ResourceValue(const _ResourceValue& rv);
	// Create from unchar.
	_ResourceValue(const char *repType, unchar value) throw(xmsg);
 	// Create from ushort.
	_ResourceValue(const char *repType, ushort value) throw(xmsg);
 	// Create from ulong.
	_ResourceValue(const char *repType, ulong value) throw(xmsg);
 	// Create from string.
	_ResourceValue(const char *repType, const char *value) throw(xmsg);
	
	virtual ~_ResourceValue();
	
	_ResourceValue& 	operator=(const _ResourceValue& rv);
	//
 	// For the type-based assignments, repType *MUST* have been set
	// previously.
	//
	_ResourceValue& 	operator=(unchar v) throw(xmsg);
	_ResourceValue& 	operator=(ushort v) throw(xmsg);
	_ResourceValue& 	operator=(ulong v) throw(xmsg);
	_ResourceValue& 	operator=(const char *v) throw(xmsg);

	operator unchar() const throw(xmsg);
	operator ushort() const throw(xmsg);
	operator ulong() const throw(xmsg);
	operator char *() const throw(xmsg);
	
	const char		*RepType() const;
	ulong			Size() const throw(xmsg);
	int			Initialized() const;
	
 	//
	// These methods should *only* be used to generate pointers for
 	// passing into and out of the RTM.
	//
	const char * const	Value() const throw(xmsg);
	//
	// Set the value based on "raw" data, not necessarily a string.  This
	// also allows us to change the repType.
 	//
	void			Value(const char *v,
				      const char *r = NULL) throw(xmsg);
    };

    const char			*name;
    const char			*className;
    const char			*type;
    _ResourceValue		value;

    Resource();			// should never be called.
    
    // Simple constructor - no value storage.
    Resource(const char *name, const char *className, const char *type);
    // Provide a value as a string.
    Resource(const char *n, const char *cl, const char *type,
	     const char *rt, const char *val);
    // Provide a value as a unchar.
    Resource(const char *n, const char *cl, const char *type,
	     const char *rt, unchar val);
    // Provide a value as a ushort.
    Resource(const char *n, const char *cl, const char *type,
	     const char *rt, ushort val);
    // Provide a value as a ulong.
    Resource(const char *n, const char *cl, const char *type,
	     const char *rt, ulong val);
    
    // Return resource info.
    const char			*Name() const;
    const char			*Class() const;
    const char			*Type() const;
    ulong			Size() const;
    const char			*RepType() const;
    int				Initialized() const;

    Resource&			operator=(unchar v);
    Resource&			operator=(ushort v);
    Resource&			operator=(ulong v);
    Resource&			operator=(const char *v);

    operator unchar() const;
    operator ushort() const;
    operator ulong() const;
    operator char *() const;

    // Print just the value (primarily for use by Interp).
    ostream&			PrintValue(ostream& s) const;

    //
    // This method should *only* be used to generate a pointer for passing
    // to the RTM.
    const char * const		Value() const;
    //
    // This method passes in a "raw" value (char *, as received from the
    // RAP protocol).  The v arg is *not* to be interpreted as a string.
    //
    void			Value(const char *v, const char *r = NULL);

public:
    virtual ~Resource();
};


////////////////////////////////////////////////////////////////////////

inline
Resource::_ResourceValue::_ResourceValue() :
    repType(NULL),
    value(NULL),
    typeInfo(NULL)
{
}

inline
Resource::_ResourceValue::_ResourceValue(const _ResourceValue& rv) :
    repType(Strdup(rv.repType)),
    value(NULL),
    typeInfo(rv.typeInfo)
{
    if (typeInfo)
	value = typeInfo->copyFunc(rv.value);
}

inline
Resource::_ResourceValue::_ResourceValue(const char *_repType,
					 unchar _value) throw(xmsg) :
    repType(Strdup(_repType)),
    value(NULL),
    typeInfo(ResourceTypeManager::Find(repType))
{
    if (typeInfo)
	value = typeInfo->copyFunc((const char *) _value);
    else
	throw xmsg("NoTypeForResourceValue");
}

inline
Resource::_ResourceValue::_ResourceValue(const char *_repType,
					 ushort _value) throw(xmsg) :
    repType(Strdup(_repType)),
    value(NULL),
    typeInfo(ResourceTypeManager::Find(repType))
{
    if (typeInfo)
	value = typeInfo->copyFunc((const char *) _value);
    else
	throw xmsg("NoTypeForResourceValue");
}

inline
Resource::_ResourceValue::_ResourceValue(const char *_repType,
					 ulong _value) throw(xmsg) :
    repType(Strdup(_repType)),
    value(NULL),
    typeInfo(ResourceTypeManager::Find(repType))
{
    if (typeInfo)
	typeInfo->copyFunc((const char *) _value);
    else
	throw xmsg("NoTypeForResourceValue");
}

inline
Resource::_ResourceValue::_ResourceValue(const char *_repType,
					 const char *_value) throw(xmsg) :
    repType(Strdup(_repType)),
    value(NULL),
    typeInfo(ResourceTypeManager::Find(repType))
{
    if (typeInfo)
	typeInfo->copyFunc((const char *) _value);
    else
	throw xmsg("NoTypeForResourceValue");
}

inline
Resource::_ResourceValue::~_ResourceValue()
{
    typeInfo->freeFunc(&value);
    delete repType;
}
	
inline Resource::_ResourceValue&
Resource::_ResourceValue::operator=(const Resource::_ResourceValue& rv)
{
    if (&rv == this)
	return *this;

    typeInfo->freeFunc(&value);
    
    delete [] repType;

    repType = Strdup(rv.repType);
    typeInfo = rv.typeInfo;

    if (typeInfo)
	value = typeInfo->copyFunc(rv.value);
}

inline Resource::_ResourceValue&
Resource::_ResourceValue::operator=(unchar v) throw(xmsg)
{
    if (!Initialized())
	throw xmsg("NoTypeForResourceValue");

    // Note assumption that sizeof(unchar) == TI_8
    if (typeInfo->storageClass != TI_8)
	throw xmsg("TypeStorageClassMismatch");
    
    typeInfo->freeFunc(&value);
    value = typeInfo->copyFunc((const char *) v);

    return *this;
}

inline Resource::_ResourceValue&
Resource::_ResourceValue::operator=(ushort v) throw(xmsg)
{
    if (!Initialized())
	throw xmsg("NoTypeForResourceValue");
    
    // Note assumption that sizeof(ushort) == TI_16
    if (typeInfo->storageClass != TI_16)
	throw xmsg("TypeStorageClassMismatch");
    
    typeInfo->freeFunc(&value);
    value = typeInfo->copyFunc((const char *) v);

    return *this;
}

inline Resource::_ResourceValue&
Resource::_ResourceValue::operator=(ulong v) throw(xmsg)
{
    if (!Initialized())
	throw xmsg("NoTypeForResourceValue");

    // Note assumption that sizeof(ulong) == TI_32
    if (typeInfo->storageClass != TI_32)
	throw xmsg("TypeStorageClassMismatch");
    
    typeInfo->freeFunc(&value);
    value = typeInfo->copyFunc((const char *) v);

    return *this;
}

inline Resource::_ResourceValue&
Resource::_ResourceValue::operator=(const char *v) throw(xmsg)
{
    if (!Initialized())
	throw xmsg("NoTypeForResourceValue");

    if (typeInfo->storageClass != TI_STRING)
	throw xmsg("TypeStorageClassMismatch");

    typeInfo->freeFunc(&value);
    value = typeInfo->copyFunc((const char *) v);

    return *this;
}

inline
Resource::_ResourceValue::operator unchar() const throw(xmsg)
{
    unchar ret;
    
    if (!Initialized())
	throw xmsg("NoTypeForResourceValue");

    // Note assumtion that sizeof(unchar) == TI_8
    if (typeInfo->storageClass != TI_8)
	throw xmsg("TypeStorageClassMismatch");

    ret = (unchar) ((int) typeInfo->copyFunc(value) & 0xFF);

    return ret;
}

inline
Resource::_ResourceValue::operator ushort() const throw(xmsg)
{
    ushort ret;
    
    if (!Initialized())
	throw xmsg("NoTypeForResourceValue");

    // Note assumtion that sizeof(ushort) == TI_16
    if (typeInfo->storageClass != TI_16)
	throw xmsg("TypeStorageClassMismatch");

    ret = (ushort) ((int) typeInfo->copyFunc(value) & 0xFFFF);

    return ret;
}

inline
Resource::_ResourceValue::operator ulong() const throw(xmsg)
{
    ulong ret;
    
    if (!Initialized())
	throw xmsg("NoTypeForResourceValue");

    // Note assumtion that sizeof(ulong) == TI_32
    if (typeInfo->storageClass != TI_32)
	throw xmsg("TypeStorageClassMismatch");

    ret = (long) typeInfo->copyFunc(value);

    return ret;
}

inline
Resource::_ResourceValue::operator char *() const throw(xmsg)
{
    if (!Initialized())
	throw xmsg("NoTypeForResourceValue");

    if (typeInfo->storageClass != TI_STRING)
	throw xmsg("TypeStorageClassMismatch");
    
    return typeInfo->copyFunc(value);
}

	
inline const char *
Resource::_ResourceValue::RepType() const
{
    return repType;
}

inline ulong
Resource::_ResourceValue::Size() const throw(xmsg)
{
    if (!Initialized())
	throw xmsg("NoTypeForResourceValue");
   
    return typeInfo->sizeFunc(value);
}

inline int
Resource::_ResourceValue::Initialized() const
{
    return (typeInfo != NULL);
}

inline const char * const
Resource::_ResourceValue::Value() const throw(xmsg)
{
    if (!Initialized())
	throw xmsg("ResourceNotInitialized");
    
    return (const char *) value;
}

inline void
Resource::_ResourceValue::Value(const char *v,
				const char *newRepType) throw(xmsg)
{
    if (Initialized()) {
	typeInfo->freeFunc(&value);

	if (newRepType != NULL) {
	    delete [] repType;
	    repType = Strdup(newRepType);
	}
    } else {
	repType = Strdup(newRepType);
    }

    //
    // We've reset the repType, so retrieve a new TypeInformation struct
    // from the RTM.
    //
    if ((typeInfo = ResourceTypeManager::Find(newRepType)) == NULL) {
	Error(EL_WARNING, "Resource::_ResourceValue::Value",
	      "unregistered resource type: ", newRepType, NULL);
	return;
    }

    //
    // This routine may be somewhat non-intuitive.  The deal here is that
    // this method is the mechanism for *initially* installing resource values
    // in resources when they come hot off the wire.  Wire values are always
    // sent as pointer-to-data, regardless of the size of the value.  But we
    // store data more efficiently:  as an immediate if the size is <= the size
    // of a pointer, and only indirect otherwise.
    //
    // Thus we check the storage class for the type we have (basically to see
    // how many bytes we need to store it).  If the storage class indicates that
    // it will take more bytes than we have available in a char * to store the
    // thing, then we use indirect storage.  Otherwise we use direct.
    //
    // This code will probably need to be tuned (read: fixed) on 64-bit
    // architectures.
    //
    // NOTE THAT THIS CODE assumes cooperation by the routines installed
    // in the RTM (that is, that they Do The Right Thing with respect to immediate
    // and indirect storage.  This assumption is implicit in the fact that we
    // dereference the wire value pointer if we're storing immediate, and don't
    // otherwise.
    //
    if (v) {
	//	switch (ResourceTypeManager::StorageClass(repType)) {
	switch (typeInfo->storageClass) {
	case TI_8:
	    value = typeInfo->copyFunc((const char *) *((const char*) v));
	    break;
	case TI_16:
	    value = typeInfo->copyFunc((const char *) *((short *) v));
	    break;
	case TI_32:
	    // Store immediate.  Note cast to get to proper arg format.
	    // value = ResourceTypeManager::Copy(repType, (const char *) *v);
	    value = typeInfo->copyFunc((const char *) *((long *) v));
	    break;
	    //	    value = typeInfo->copyFunc((const char *) *v);
	    //	    break;
	case TI_STRING:
	case TI_SIZED:
	    // Store indirect.  No cast or dereference needed.
	    // value = ResourceTypeManager::Copy(repType, v);
	    value = typeInfo->copyFunc(v);
	    break;
	default:
	    throw xmsg("UnknownStorageClass");
	}
    }
}

ostream&
operator<<(ostream& s, const Resource::_ResourceValue& v)
{
    if (!v.typeInfo) {
    	s << "<unregistered type>";
	return s;
    } else {
	return v.typeInfo->printFunc(v.value, s);
    }
}

////////////////////////////////////////////////////////////////////////


Resource::Resource()
{
    Error(EL_ABORT, "Resource::Resource", "Default constructor called.\n"
	  "\tThis is a nonrecoverable error as it indicates a ",
	  "programmer logic error.",
	  NULL);
}

inline
Resource::Resource(const char *_name, const char *_class, const char *_type) :
    name(Strdup(_name)),
    className(Strdup(_class)),
    type(Strdup(_type)),
    value()
{
}

inline
Resource::Resource(const char *n, const char *c, const char *t,
		   const char *rt, const char *v) :
    name(Strdup(n)),
    className(Strdup(c)),
    type(Strdup(t)),
    value(rt, v)
{
}

inline
Resource::Resource(const char *n, const char *c, const char *t,
		   const char *rt, unchar v) :
    name(Strdup(n)),
    className(Strdup(c)),
    type(Strdup(t)),
    value(rt, v)
{
}

inline
Resource::Resource(const char *n, const char *c, const char *t,
		   const char *rt, ushort v) :
    name(Strdup(n)),
    className(Strdup(c)),
    type(Strdup(t)),
    value(rt, v)
{
}

inline
Resource::Resource(const char *n, const char *c, const char *t,
		   const char *rt, ulong v) :
    name(Strdup(n)),
    className(Strdup(c)),
    type(Strdup(t)),
    value(rt, v)
{
}

inline
Resource::~Resource()
{
    delete [] (char *) name;
    delete [] (char *) className;
    delete [] (char *) type;
}

inline const char * Resource::Name() const 	{ return name; }
inline const char * Resource::Class() const 	{ return className; }
inline const char * Resource::Type() const 	{ return type; }
inline ulong Resource::Size() const 		{ return value.Size(); }
inline const char * Resource::RepType() const 	{ return value.RepType(); }
inline int Resource::Initialized() const       	{ return value.Initialized();}

inline Resource& Resource::operator=(unchar v) 	{ value = v; return *this; }
inline Resource& Resource::operator=(ushort v)	{ value = v; return *this; }
inline Resource& Resource::operator=(ulong v) 	{ value = v; return *this; }
inline Resource& Resource::operator=(const char*v){ value = v; return *this; }

#ifdef BEAM
inline Resource::operator unchar() const	{ return (ushort)value; }
inline Resource::operator ushort() const	{ return (ushort)value; }
#else
inline Resource::operator unchar() const	{ return value; }
inline Resource::operator ushort() const	{ return value; }
#endif
inline Resource::operator ulong() const		{ return value; }
inline Resource::operator char *() const	{ return value; }

inline const char * const Resource::Value() const { return value.Value(); }
inline void Resource::Value(const char *v, const char *r) {value.Value(v, r);}

ostream&
Resource::PrintValue(ostream& s) const
{
    s << value;
    return s;
}

ostream&
operator<<(ostream& s, const Resource& r)
{
    s << r.Name() << " [" << r.Class() << "] (" << r.Type() << ") = "
      << r.value;
    return s;
}


////////////////////////////////////////////////////////////////////////


ResourceTypeManager::ResourceTypeManager()
{
    Error(EL_ABORT, "ResourceTypeManager::ResourceTypeManager",
	  "Should never be instantiated.\n"
	  "\tThis is a nonrecoverable error as it indicates a "
	  "programmer logic error.",
	  NULL);
}

int
ResourceTypeManager::Register(const char *t, StorageClass_t sc,
			      CopyFunc c, FreeFunc f,
			      PrintFunc p, SizeFunc s)
{
    // Allocate one to insert
    mString *typeString = new mString(t);

    if (types.lookup(typeString)) {
	Error(EL_ERROR, "ResourceTypeManager::Register",
	      "Type ", t, ", already registered.", NULL);
	delete typeString;
	return 0;
    }

    if (c == NULL || f == NULL || p == NULL || s == NULL) {
	Error(EL_ERROR, "ResourceTypeManager::Register",
	      "Bogus function(s) passed for type ", t, ".", NULL);
	delete typeString;
	return 0;
    }
    
    TypeInformation *ti = new TypeInformation;

    ti->storageClass = sc;
    ti->copyFunc = c;
    ti->freeFunc = f;
    ti->printFunc = p;
    ti->sizeFunc = s;

    types.insert(typeString, ti);

    return 1;
}

inline int
ResourceTypeManager::Registered(const char *t)
{
    if (t == NULL) return 0;

    mString typeString(t);

    return (types.lookup(&typeString) != NULL);
}

inline TypeInformation *
ResourceTypeManager::Find(const char *repType)
{
    mString repString(repType);
    return types.lookup(&repString);
}


////////////////////////////////////////////////////////////////////////

class ResourceCacheDestroyer : public mDictPtrPtrClosure<mString, Resource> {
public:
    int action(mString *k, Resource *v)
		    {
			delete k; k = NULL;
			delete v; v = NULL;
			return 1;
		    }
};

ResourceCache::ResourceCache()
{
    resources = new ResourceDict(StringHasher);
}

ResourceCache::~ResourceCache()
{
    ResourceCacheDestroyer paranoia;

    //
    // This is a very cheezy (and expensive) way to free the resources.
    // The closure class above doesn't do the trick, since we stomp on the
    // key as we walk the dictionary...this makes subsequent operations
    // on the dict read freed memory.
    //
    // This instantiates mutliple iterators, resetting them each time, to
    // destroy the resources in the cache....
    //
    while (resources->size() > 0) {
	mDictPtrPtrIterator<mString, Resource> iter(*resources);
	mString *key;
	if (key = iter.next())
	    resources->del_item(key);
    }

    //    resources->walk(paranoia, NULL);
    delete resources;
}

int
ResourceCache::Exists(const char *name) const
{
    mString nameString(name);

    return (resources->lookup(&nameString)==NULL);
}

int
ResourceCache::Initialized(const char *name) const throw(xmsg)
{
    mString nameString(name);
    Resource *resource;

    if ((resource = resources->lookup(&nameString)) == NULL)
	throw xmsg("NoSuchResource");
    
    return resource->Initialized();
}

ostream&
ResourceCache::Print(const char *name, ostream& s) const
{
    mString nameString(name);
    Resource *resource;
    
    if ((resource = resources->lookup(&nameString)) == NULL)
	return s << "<not found>";
    else
	return s << *resource;
}

ostream&
ResourceCache::PrintValue(const char *name, ostream& s) const
{
    mString nameString(name);
    Resource *resource;

    if ((resource = resources->lookup(&nameString)) == NULL)
	return s << "<not found>";
    else
	return resource->PrintValue(s);
}

int
ResourceCache::Add(const char *name, const char *className, const char *type,
		   const char *value) throw(xmsg)
{
    mString *nameString = new mString(name);
    Resource *resource;

    if (resources->lookup(nameString)!=NULL) {
	delete nameString;
	throw xmsg("ResourceAlreadyExists");
    }

    //
    // For now, we refuse to add a resource whose type is not registered
    // with the ResourceTypeManager.  In the future we may wish to raise
    // an exception.
    //
    if (!ResourceTypeManager::Registered(type)) {
	delete nameString;
	return 0;
    }

    resource = new Resource(name, className, type);
    
    //
    // Resource values *always* come in off the wire in a pointer-to-data
    // format, regardless of length.  But we store values in a more efficient way,
    // as immediates if their size is less than or equal to the size of a pointer;
    // we only use indirect storage when we have to.  Thus, we call the Value()
    // method on the resource which takes data in the wire (raw) format and
    // figures out what to do with it.  This is probably the ONLY time we should
    // call this method.
    //
    if (value)
	resource->Value(value, type);

    resources->insert(nameString, resource);

    return 1;
}

int
ResourceCache::Set(const char *name, const char *type,
		   const char *value) throw(xmsg)
{
    mString *nameString = new mString(name);
    Resource *resource;

    if (!ResourceTypeManager::Registered(type)) {
	delete nameString;
	return 0;
    }

    if ((resource = resources->lookup(nameString)) == NULL) {
	delete nameString;
	throw xmsg("NoSuchResource");
    }
    
    if (value)
	resource->Value(value, type);
    else 
	Error(EL_DEBUG, "ResourceCache::Set", "Value for resource ", name,
	    "is NULL\n", NULL);
    
    return 1;
}



int
ResourceCache::Remove(const char *name)
{
    mString nameString(name);
    Resource *resource=resources->lookup(&nameString);

    // if its not there, then just return 0
    if (!resource) return 0;

    // remove from dict... I think this is going to leak the Key
    // since we are using == on a different key 
    resources->del(&nameString);
    delete resource;

    return 1;
}

int
ResourceCache::SetValueRaw(const char *n, const char *val,
			   const char *rep) throw(xmsg)
{
    mString nameString(n);
    Resource *resource;

    if ((resource = resources->lookup(&nameString)) == NULL)
	throw xmsg("NoSuchResource");

    resource->Value(val, rep);
    return 1;
}

int
ResourceCache::SetValue(const char *name, const char *value) throw(xmsg)
{
    mString nameString(name);
    Resource *resource;

    if ((resource = resources->lookup(&nameString)) == NULL)
	throw xmsg("NoSuchResource");
    
    *(resource) = value;
    return 1;
}

int
ResourceCache::SetValue(const char *name, unchar value) throw(xmsg)
{
    mString nameString(name);
    Resource *resource;

    if ((resource = resources->lookup(&nameString)) == NULL)
	throw xmsg("NoSuchResource");

    *(resource) = value;
    return 1;
}

int
ResourceCache::SetValue(const char *name, ushort value) throw(xmsg)
{
    mString nameString(name);
    Resource *resource;

    if ((resource = resources->lookup(&nameString)) == NULL)
	throw xmsg("NoSuchResource");
    
    *(resource) = value;
    return 1;
}

int
ResourceCache::SetValue(const char *name, ulong value) throw(xmsg)
{
    mString nameString(name);
    Resource *resource;

    if ((resource = resources->lookup(&nameString)) == NULL)
	throw xmsg("NoSuchResource");
    
    *(resource) = value;
    return 1;
}

int
ResourceCache::GetValue(const char *name, char **value,
			const char **repType) const throw(xmsg)
{
    mString nameString(name);
    Resource *resource;
    ulong length;

    if ((resource = resources->lookup(&nameString)) == NULL)
	throw xmsg("NoSuchResource");

    if (repType)
	*repType = resource->RepType();
    
    *value = *resource;
    return 1;
}

int
ResourceCache::GetValue(const char *name, unchar *value,
			const char **repType) const throw(xmsg)
{
    mString nameString(name);
    Resource *resource;
    ulong length;

    if ((resource = resources->lookup(&nameString)) == NULL)
	throw xmsg("NoSuchResource");

    if (resource->Size() != sizeof(unchar))
	throw xmsg("SizeMismatch");

    if (repType)
	*repType = resource->RepType();
    
    *value = *resource;
    return 1;
}

int
ResourceCache::GetValue(const char *name, ushort *value,
			const char **repType) const throw(xmsg)
{
    mString nameString(name);
    Resource *resource;
    ulong length;

    if ((resource = resources->lookup(&nameString)) == NULL)
	throw xmsg("NoSuchResource");

    if (resource->Size() != sizeof(ushort))
	throw xmsg("SizeMismatch");

    if (repType)
	*repType = resource->RepType();
    
    *value = *resource;
    return 1;
}

int
ResourceCache::GetValue(const char *name, ulong *value,
			const char **repType) const throw(xmsg)
{
    mString nameString(name);
    Resource *resource;
    ulong length;

    if ((resource = resources->lookup(&nameString)) == NULL)
	throw xmsg("NoSuchResource");

    if (resource->Size() != sizeof(ulong)) 
	throw xmsg("SizeMismatch");

    if (repType)
	*repType = resource->RepType();
    
    *value = *resource;
    return 1;
}

ostream&
operator<<(ostream& s, const ResourceCache& c)
{
    Resource *r;
    mString *m;
    mDictPtrPtrIterator<mString, Resource> iter(*(c.resources));

    while (m=iter.next()) {
	r=c.resources->lookup(m);
	s << *r << "\n";
    }

    return s;
}

////////////////////////////////////////////////////////////////////////

//
// Below is InitializeResourceTypeManager.  It installs type registrations
// for several common Xt resource types.  Its "helper" functions are
// declared static here.
//
// NOTE that this implementation is not portable.  Different Xt implementations
// may use different implementations of this type.  Eventually we will
// probably want a per-client ResourceTypeManager.  The installation of
// types is done during RAP negotiation or as resources come in (so that
// Mercator sees that a particular client is using long longs to represent
// Booleans or something...)
//

// Immediate funcs work with any immediate, of any size.
inline static char *ImmediateCopy(const char *v){ return (char *) v; }
inline static void ImmediateFree(char **v)	{ *v = NULL; return; }

// Versions of immediates where size is relevant.
inline static ostream& PrintByte(const char *v, ostream& s)
			{ s << (int) v; return s; }
inline static ostream& PrintShort(const char *v, ostream& s)
			{ s << (int) v; return s; }
inline static ostream& PrintUShort(const char *v, ostream& s)
			{ s << (unsigned int) v; return s; }
inline static ostream& PrintLong(const char *v, ostream& s)
			{ s << (long) v; return s; }
inline static ostream& PrintLongHex(const char *v, ostream& s)
			{ s << hex << (long) v; return s; }
inline static ostream& PrintULong(const char *v, ostream& s)
			{ s << (ulong) v; return s; }
inline static ostream& PrintULongHex(const char *v, ostream& s)
			{ s << hex << (ulong) v; return s; }

inline static ostream& PrintScrollMode(const char *v, ostream& s)
			{ s << (int) v; return s; }

inline static ulong SizeByte(const char *v)	{ return sizeof(unchar); }
inline static ulong SizeShort(const char *v)	{ return sizeof(ushort); }
inline static ulong SizeLong(const char *v)	{ return sizeof(ulong);  }

inline static ostream& PrintBoolean(const char *v, ostream&s )
			{ ((int) v) ? s << "TRUE" : s << "FALSE"; return s; }

// String funcs assume NULL-terminated char arrays.
inline static char *StringCopy(const char *v) 	{ return Strdup(v); }
inline static void StringFree(char **v) 	{ delete [] *v; return; }
inline static ostream& StringPrint(const char *v, ostream& s)
						{ s << v; return s; }
inline static ulong StringSize(const char *v)	{ return strlen(v); }

//
// We currently support:
//	- XtRPosition			short
//	- XtRBoolean			char
//	- XtRBool			char
//	- XtRDimension			unsigned short
//	- XtRString			char *
//	- XtRCardinal			unsigned int
//	- XtRInt			int
//	- XtRWidget			unsigned int
//	- XtRWindow			unsigned int
//
//	- XmRXmString			char *
//
//	- XtRScrollMode (Athena)	short
//

/////
///// UBER HACK
///// I want to put in a converter from XmStrings without having to pull
///// in motif itself.  Hence this #define.
/////
#define XmRXmString "XmString"
#define XmRHorizontalPosition "HorizontalPosition"
#define XmRVerticalPosition "VerticalPosition"
#define XmRHorizontalDimension "HorizontalDimension"
#define XmRVerticalDimension "VerticalDimension"

/////
///// From Athena
/////
#define XtRScrollMode "ScrollMode"

void
InitializeResourceTypeManager()
{
    //
    // Xt internal type names.
    //
    ResourceTypeManager::Register(XtRPosition, TI_16,
				  ImmediateCopy,
				  ImmediateFree,
				  PrintShort,
				  SizeShort);
    ResourceTypeManager::Register(XtRBoolean, TI_8,
				  ImmediateCopy,
				  ImmediateFree,
				  PrintBoolean,
				  SizeByte);
    ResourceTypeManager::Register(XtRBool, TI_8,
				  ImmediateCopy,
				  ImmediateFree,
				  PrintBoolean,
				  SizeByte);
    ResourceTypeManager::Register(XtRDimension, TI_16,
				  ImmediateCopy,
				  ImmediateFree,
				  PrintUShort,
				  SizeShort);
    ResourceTypeManager::Register(XtRString, TI_STRING,
				  StringCopy,
				  StringFree,
				  StringPrint,
				  StringSize);
    ResourceTypeManager::Register(XtRCardinal, TI_32,
				  ImmediateCopy,
				  ImmediateFree,
				  PrintULong,
				  SizeLong);
    ResourceTypeManager::Register(XtRInt, TI_32,
				  ImmediateCopy,
				  ImmediateFree,
				  PrintLong,
				  SizeLong);
    ResourceTypeManager::Register(XtRWidget, TI_32,
				  ImmediateCopy,
				  ImmediateFree,
				  PrintULongHex,
				  SizeLong);
    ResourceTypeManager::Register(XtRWindow, TI_32,
				  ImmediateCopy,
				  ImmediateFree,
				  PrintULongHex,
				  SizeLong);

    //
    // Xm Type Names.
    //
    ResourceTypeManager::Register(XmRXmString, TI_STRING,
				  StringCopy,
				  StringFree,
				  StringPrint,
				  StringSize);
    ResourceTypeManager::Register(XmRHorizontalPosition, TI_16,
				  ImmediateCopy,
				  ImmediateFree,
				  PrintShort,
				  SizeShort);
    ResourceTypeManager::Register(XmRVerticalPosition, TI_16,
				  ImmediateCopy,
				  ImmediateFree,
				  PrintShort,
				  SizeShort);
    ResourceTypeManager::Register(XmRHorizontalDimension, TI_16,
				  ImmediateCopy,
				  ImmediateFree,
				  PrintUShort,
				  SizeShort);
    ResourceTypeManager::Register(XmRVerticalDimension, TI_16,
				  ImmediateCopy,
				  ImmediateFree,
				  PrintUShort,
				  SizeShort);

    //
    // Xaw Type Names.
    //
    ResourceTypeManager::Register(XtRScrollMode, TI_32,
				  ImmediateCopy,
				  ImmediateFree,
				  PrintScrollMode,
				  SizeLong);
}

