//
//
//
//	Note:  The use of this program shall be restricted to educational,
//	non-commercial use only.  All inquiries regarding licensing this
//	software for commercial use may be addressed to:
//	
//	Georgia Tech Research Corporation
//	Technology Licensing
//	Centennial Research Building - Rm. 275
//	400 Tenth Street, N.W.
//	Atlanta, Georgia  30332-0415
//	
//	*************************************************
//	
//	       Copyright 1996
//	Georgia Tech Research Corporation
//	    All Rights Reserved
//
//
//
//
#include <assert.h>
#include <ctype.h>
#include <exception.h>
#include <fstream.h>
#include <memory.h>
#include <stdio.h>
#include <string.h>
#include "Config.h"
#include "Error.h"
#include "Mercator.h"
#include "ScreenReader.h"
#include "Template.h"
#include "TextRep.h"
#include "Util.h"
#include "XServer.h"
#include "mRegexp.h"
#include "mString.h"

#ifndef MIN
#define MIN(a,b) (((a) <= (b)) ? (a) : (b))
#endif

#ifndef MAX
#define MAX(a,b) (((a) >= (b)) ? (a) : (b))
#endif

extern int filterparse(void);

// This was taken from the original ScreenReader code.  I think it's Ian's.

static unsigned StringHash(const mString &s) {
    union {
        unsigned uns;
        unsigned char c[sizeof(unsigned)];
    } result;
    result.uns=0;
    for (int i=0; i<s.length(); ++i) {
        result.c[i%(sizeof(unsigned))]^=
	  (unsigned char)((const char *)s)[i];
    }
    return result.uns;
}

CursorRepData::~CursorRepData()
{
    //    delete tdata;
}

#if 0
void
CursorRepData::SetData(const char *input) {
  data = Strdup(input);
}

char *
CursorRepData::GetData(void) {
  if (data)
    return data;
  else
    return Strdup("");
}
#endif

// The CursorRep object serves as the interface between the TextRep base
//    class and the ScreenReader class.  Data is stored in line form in
//    the TextRep class -- this is accessed in row, column format by
//    the CursorRep, which maintains a cache of a "current line" from
//    the TextRep which it locally queries until the current row is changed.

// Currently, there is no provision to update the current line cache from
//    the TextRep short of an explicit refresh call.  Hopefully, this won't
//    be necessary (we assume that the TextRep contents can't change while
//    we're in the CursorRep code).

//****************************************************************************
//** Methods for CursorRep objects
//****************************************************************************

CursorRep::CursorRep(TextRep *t) {
  if (!(textrep = t)) {
    Error(EL_ERROR, "CursorRep::CursorRep", "received a NULL TextRep pointer.",
	  NULL);
  }
  x = y = bogusline = 0;
  //  boundsx = textrep->Columns(); assert(boundsx >= 0);
  //  boundsy = textrep->Rows(); assert(boundsy >= 0);
  try {
    current_line = textrep->GetLine(0);
  }
  catch (xmsg) {
    bogusline = 1;
    Error(EL_ERROR, "CursorRep::CursorRep", "attempted to initialize ",
	  "using an empty TextRep.", NULL);
  }
}

// Cursor set routines.  The cursor is not updated if the new value would
//   result in an out-of-bounds condition.

void
CursorRep::SetCursorX(int newx) throw (CursorOutOfBounds) {
  //  assert(textrep != NULL);
  if ((newx < 0) || (newx > textrep->Columns() - 1))
    throw CursorOutOfBounds(newx, y, textrep->Columns(), textrep->Rows());
  else
    x = newx;
}

void
CursorRep::SetCursorY(int newy) throw (CursorOutOfBounds) {
  //  assert(textrep != NULL);
  if ((newy < 0) || (newy > textrep->Rows() - 1))
    throw CursorOutOfBounds(x, newy, textrep->Columns(), textrep->Rows());
  else if (y != newy) {
    bogusline = 0;
    try {
      current_line = textrep->GetLine(newy);
    }
    catch (xmsg) {
      bogusline = 1;
      char buffer[16];
      sprintf(buffer, "%d", newy);
      Error(EL_ERROR, "CursorRep::SetCursorY", "could not read line ",
	    buffer, " from current TextRep.", NULL);
    }
    y = newy;
  }
}

void
CursorRep::SetCursor(int newx, int newy) throw (CursorOutOfBounds) {
    //  assert(textrep != NULL);
    int textCol = textrep->Columns(), textRow = textrep->Rows();
    if ((newx < 0) || ((textCol > 0) && (newx > textCol - 1)))
	throw CursorOutOfBounds(newx, y, textCol, textRow);
    if ((newy < 0) || ((textRow > 0) && (newy > textRow - 1)))
	throw CursorOutOfBounds(x, newy, textCol, textRow);
    if (y != newy) {
	bogusline = 0;
	try {
	    current_line = textrep->GetLine(newy);
	}
	catch (xmsg) {
	    bogusline = 1;
	    char buffer[16];
	    sprintf(buffer, "%d", newy);
	    Error(EL_ERROR, "CursorRep::SetCursorY", "could not read line ",
		  buffer, " from current TextRep.", NULL);
	}
    }
    x = newx; y = newy;
}

CursorRep &
CursorRep::operator++(void) throw (CursorOutOfBounds) {
  //  assert(textrep != NULL);
  if (++x >= textrep->Columns()) {
    x = 0;
    if (++y >= textrep->Rows()) {
	// x = textrep->Columns() - 1; 
	// y = textrep->Rows() - 1;  // give cursor well-defined position
	// Can x and y be less than zero? -wl
	x = MAX(textrep->Columns() - 1, 0); 
	y = MAX(textrep->Rows() - 1, 0);  // give cursor well-defined position
      throw CursorOutOfBounds(0, textrep->Rows(), textrep->Columns(), 
			      textrep->Rows());
    }
    bogusline = 0;
    try {
      current_line = textrep->GetLine(y);
    }
    catch (xmsg) {
      bogusline = 1;
      char buffer[16];
      sprintf(buffer, "%d", y);
      Error(EL_ERROR, "CursorRep::SetCursorY", "could not read line ",
	    buffer, " from current TextRep.", NULL);
    }
  }
  return (*this);
}

CursorRep &
CursorRep::operator++(int dummy) {
  return ++(*this);
}


CursorRep &
CursorRep::operator--(void) throw (CursorOutOfBounds) {
  //  assert(textrep != NULL);
  if (--x < 0) {
    x = MAX(textrep->Columns() - 1, 0);
    if (--y < 0) {
      x = y = 0;   // give cursor well-defined position
      throw CursorOutOfBounds(textrep->Columns() - 1, -1, textrep->Columns(), 
			      textrep->Rows());
    }
    bogusline = 0;
    try {
      current_line = textrep->GetLine(y);
    }
    catch (xmsg) {
      bogusline = 1;
      char buffer[16];
      sprintf(buffer, "%d", y);
      Error(EL_ERROR, "CursorRep::SetCursorY", "could not read line ",
	    buffer, " from current TextRep.", NULL);
    }
  }
  return (*this);
}

CursorRep &
CursorRep::operator--(int dummy) {
  return --(*this);
}

CursorRep &
CursorRep::operator+=(const int count) throw (CursorOutOfBounds) {
  //  assert(textrep != NULL);
  int new_pos = count + x + (textrep->Columns() * y);
  if ((new_pos < 0) || (new_pos >= (textrep->Columns() * textrep->Rows())))
    throw CursorOutOfBounds(new_pos % textrep->Columns(), 
			    new_pos / textrep->Rows(),
			    textrep->Columns(), textrep->Rows());
  int newy = new_pos / textrep->Columns();
  if (y != newy) {
    bogusline = 0;
    try {
      current_line = textrep->GetLine(newy);
    }
    catch (xmsg) {
      bogusline = 1;
      char buffer[16];
      sprintf(buffer, "%d", newy);
      Error(EL_ERROR, "CursorRep::SetCursorY", "could not read line ",
	    buffer, " from current TextRep.", NULL);
    }
    y = newy;
  }
  x = new_pos % textrep->Columns();
  return *this;
}

CursorRep &
CursorRep::operator-=(const int count) {
  return (*this+= -count);
}

char
CursorRep::GetChar(void) {
  //  assert(textrep != NULL);
  if (bogusline) {
    bogusline = 0;
    try {
      current_line = textrep->GetLine(y);
    }
    catch (xmsg) {
      bogusline = 1;
      char buffer[16];
      sprintf(buffer, "%d", y);
      Error(EL_ERROR, "CursorRep::GetChar", "could not read line ",
	    buffer, " from current TextRep.", NULL);
    }
  }
  return ((const char *)current_line)[x];
}

TextAttr &
CursorRep::GetAttr(void) {
  //  assert(textrep != NULL);
  if (bogusline) {
    bogusline = 0;
    try {
      current_line = textrep->GetLine(y);
    }
    catch (xmsg) {
      bogusline = 1;
      char buffer[16];
      sprintf(buffer, "%d", y);
      Error(EL_ERROR, "CursorRep::GetAttr", "could not read line ",
	    buffer, " from current TextRep.", NULL);
    }
  }
  //return ((const TextAttr *)current_line)[x];
  return ((TextAttr *) current_line)[x];
}

//----------------------------------------------------------------------------
// Retrieve data from the current cursor position.  If the length of the
//   data requested would cause a read past the end of the TextRep area,
//   throw an exception.  The overloaded version of this function takes
//   row/column parameters, and basically sets the cursor position,
//   grabs its data, and then resets the cursor.
//----------------------------------------------------------------------------

// Storage for the TextData field is allocated in this routine, and must be
//   freed by the caller.  Storage for the subfields of the TextData structure
//   is also allocated by this routine, but TextData should manage the
//   disposal of those fields.

CursorRepData *
CursorRep::GetData(const int len) throw (CursorOutOfBounds) {
  CursorRepData *ptr;
  int length, index, end_offset, tmpx = x, tmpy = y, copylen , i;
  TextAttr *setattrs, *retattrs;
  TextData newdata;
  char *text;
  //  assert(textrep != NULL);
  // Verify that we are not attempting to read past the end of the 
  //   text area
  end_offset = tmpx + (tmpy * textrep->Columns()) + len - 1;
  if (end_offset >= (textrep->Columns() * textrep->Rows()))
    throw CursorOutOfBounds(end_offset % textrep->Columns(), 
			    end_offset / textrep->Columns(),
			    textrep->Columns(), textrep->Rows());
  ptr = new CursorRepData();
  ptr->SetStart(tmpx, tmpy);
  // Read the current line, if necessary
  // Hack -- force us to re-read line data whenever we try to get data
  //   from the textrep.  If that doesn't keep up up to date, nothing will.  B)
  //  if (bogusline) {
  if (1) {
    bogusline = 0;
    try {
      current_line = textrep->GetLine(tmpy);
    }
    catch (xmsg) {
      bogusline = 1;
      char buffer[16];
      sprintf(buffer, "%d", tmpy);
      Error(EL_ERROR, "CursorRep::GetData", "could not read line ",
	    buffer, " from current TextRep.", NULL);
    }
  }
  // Do not modify the current line for this cursor
  TextData line = current_line;
  text = new char[len+1];
  setattrs = new TextAttr[len];
  // Read data
  for (length = len, index = 0, copylen = MIN(textrep->Columns()-tmpx, length);
       length > 0;
       copylen = MIN(textrep->Columns()-tmpx, length)) {
    memcpy(text+index, ((const char *)line)+tmpx, copylen);
    //    for (retattrs = (const TextAttr *)line, i = 0; i < copylen; i++)
    for (retattrs = (TextAttr *)line, i = 0; i < copylen; i++)
      setattrs[index+i] = retattrs[tmpx+i];
    // increment append pointer, decrement size remaining
    index += copylen; length -= copylen;
    // read next line, if necessary
    if (length > 0) {
      tmpx = 0; tmpy++; 
      try {
	line = textrep->GetLine(tmpy);
      }
      // This is actually a pretty huge problem.  If we can't read data,
      //   we should probably panic, or something.
      catch (xmsg) {
	bogusline = 1;
	char buffer[16];
	sprintf(buffer, "%d", tmpy);
	Error(EL_ERROR, "CursorRep::GetData", "could not read line ",
	      buffer, " from current TextRep.", NULL);
      }
    }
  }
  text[len] = '\0';
  //  newdata = new TextData;
  newdata.Chars(text);
  newdata.Attrs(setattrs);
  newdata.Length(len);  // only include length of the data, not NULL
  ptr->SetTData(newdata);
  ptr->SetEnd(end_offset % textrep->Columns(), 
	      end_offset / textrep->Columns());
  return ptr;
}

CursorRepData *
CursorRep::GetData(const int cx, const int cy, const int len) {
  int oldx, oldy;
  CursorRepData *ptr;
  GetCursor(oldx, oldy);
  SetCursor(cx, cy);
  ptr = GetData(len);
  SetCursor(oldx, oldy);
  return ptr;
}

// Note that (endx, endy) >= (startx, starty).  May fix this later.
CursorRepData *
CursorRep::GetData(const int startx, const int starty, const int endx,
		   const int endy) {
  int len = endx + (textrep->Columns() * endy) - 
    (startx + (textrep->Columns() * starty)) + 1;
  CursorRepData *ptr = GetData(startx, starty, len);
  return ptr;
}

//----------------------------------------------------------------------------
// Requery the current textrep associated with this cursor.  I'm not 
//   actually sure when we need to do this.
//----------------------------------------------------------------------------
void 
CursorRep::Refresh(void) {
  //  boundsx = textrep->Columns();
  //  boundsy = textrep->Rows();
  try {
      current_line = textrep->GetLine(y);
  }
  catch (xmsg) {
    bogusline = 1;
    char buffer[16];
    sprintf(buffer, "%d", y);
    Error(EL_ERROR, "CursorRep::GetData", "could not read line ",
	  buffer, " from current TextRep.", NULL);
  }
}

//****************************************************************************
//** Methods for TextFilter objects
//****************************************************************************
extern FILE *filterin; // from filterlexer

mDictValPtr<mString, TextFilter> *FilterDict::Filters =
  new mDictValPtr<mString, TextFilter>(StringHash);
TextFilter *FilterDict::current_filter = NULL;
int FilterDict::filterLineCount = 0;

TextFilter::TextFilter(void) {
  active = num_commands = 0;
  max_commands = MAX_FILTER_COMMANDS;
  name = NULL;
  commands = new FilterEntryPtr[max_commands];
}

TextFilter::~TextFilter() {
  int i;
  delete [] name;
  for (i = 0; i < num_commands; i++)
    delete commands[i];
}

void
TextFilter::SetName(char *n) {
  if (TextFilter::name)
    delete [] TextFilter::name;
  TextFilter::name = Strdup(n);
}

void
TextFilter::Filter(CursorRepData &input) {
  int i, j, k, len;
  char *str;
  mString subst((const char *)input.GetTData()), command_subst, *ptr;
  for (i = 0, j = 0; i < num_commands; i++, j = 0) {
    // Perform substitution on the actual expression to be substituted into
    //   the text presented to the user.  This must be done every time a
    //   regular expression match is found.  Currently the only $substitution 
    //   allowed by the filter is $old.
    //   This may be extended in the future.  Maybe.
    if (commands[i]->action == FILTER_ACTION_REGEXP) {
      while ((k = subst.index(commands[i]->r, j)) != -1) {
	// Check for presence of $old in substitution string
	if ((command_subst = 
	     mString(commands[i]->arg)).index("$old", 0) != -1) {
	  // replace $old in substitution string with value which matched
	  //   the regular expression
	  ptr = subst.substString(commands[i]->r, j);
	  command_subst.subst(mRegexp("\\$old"), 0, *ptr);
	  delete ptr;
	}
	// Perform substitution operation
	j = subst.subst(commands[i]->r, j, command_subst);
	// Make sure we get a copy of this data
	str = (char *)subst;
	len = strlen(str);
	input.GetTData().Chars(str);
	input.GetTData().ResizeAttrs(len, (k == -1) ? 0 : k);
	input.GetTData().Length(len);
      }
    }
    else if (commands[i]->action == FILTER_ACTION_FONT) {
      TextAttr *attrs;
      unsigned long storemask = GCValue::GetCompFlags();
      GCValue::SetCompFlags(GCValue::GC_COMP_FONT);			    
      for (k = 1; k < input.GetTData().Length(); k++) {
	attrs = input.GetTData().Attrs();
	if (attrs[k-1] != attrs[k]) {
	  subst.insert(k, commands[i]->arg);
	  len = strlen((const char *)subst);
	  input.GetTData().Chars((char *)subst);
	  input.GetTData().ResizeAttrs(len, k);
	  input.GetTData().Length(len);
	  k += strlen(commands[i]->arg);
	}
      }
    }
    // Let's assume that the text is highlighted if any of the drawing
    //   GC values have been changed
    else if (commands[i]->action == FILTER_ACTION_HIGHLIGHT) {
      TextAttr *attrs;
      unsigned long storemask = GCValue::GetCompFlags();
      GCValue::SetCompFlags(GCValue::GC_COMP_FUNCTION |
			    GCValue::GC_COMP_FOREGROUND |
			    GCValue::GC_COMP_BACKGROUND);
      for (k = 1; k < input.GetTData().Length(); k++) {
	attrs = input.GetTData().Attrs();
	if (attrs[k-1] != attrs[k]) {
	  subst.insert(k, commands[i]->arg);
	  len = strlen((const char *)subst);
	  input.GetTData().Chars((char *)subst);
	  input.GetTData().ResizeAttrs(len, k);
	  input.GetTData().Length(len);
	  k += strlen(commands[i]->arg);
	}
      }
      GCValue::SetCompFlags(storemask);
    }
    if ((commands[i]->action >= FILTER_ACTION_REGEXP) &&
	(commands[i]->action <= FILTER_ACTION_HIGHLIGHT)) {
    }
    else {
      Error(EL_ERROR, "TextFilter::Filter", "illegal action code executed.",
	    NULL);
    }
  }
}

//----------------------------------------------------------------------------
// Insert a new filter command into the current TextFilter.
//----------------------------------------------------------------------------
void
TextFilter::Insert(char *regexp, int action, char *arg) {
  FilterEntryPtr *new_commands;
  mString remove_quotes(regexp);
  char buffer[16], *subst_regexp;
  int i;
  if (num_commands == max_commands) {
    new_commands = new FilterEntryPtr[max_commands += MAX_FILTER_COMMANDS];
    for (i = 0; i < num_commands; i++)
      new_commands[i] = commands[i];
    delete [] commands;
    commands = new_commands;
  }
  // Convert escaped quotes to normal quotes.  Escaped quotes are necessary
  //   to recognize the regular expressions within the filter lexer.
  for (i = 0; (i = remove_quotes.subst(mRegexp("\\\""), i, "\"")) != -1;);
  subst_regexp = (char *)remove_quotes;
  commands[num_commands] = new FilterEntry;
  // copy new pattern into this regular expression
  commands[num_commands]->r = subst_regexp;
  // we no longer need this character string, so we free it to avoid leaking
  delete [] subst_regexp;
  commands[num_commands]->action = action;
  // the argument passed in gets deleted by the caller; save a copy
  commands[num_commands++]->arg = Strdup(arg);
}

FilterDict::FilterDict(void) {
  num_filters = 0;

  max_filters = MAX_FILTER;
  filter_names = new char*[max_filters];
}

FilterDict::~FilterDict() {
  int i;
  for (i = 0; i < num_filters; i++)
    delete [] filter_names[i];
  delete filter_names;
}

//----------------------------------------------------------------------------
// Create a new TextFilter structure to be inserted into the dictionary
//----------------------------------------------------------------------------
int
FilterDict::NewFilterRegister(char *name) {
  current_filter = new TextFilter();
  current_filter->SetName(name);
  current_filter->SetActive(1);
  return 1;
}

//----------------------------------------------------------------------------
// Abort the current TextFilter structure due to errors.
//----------------------------------------------------------------------------
void
FilterDict::NewFilterAbort(void) {
  if (current_filter) {
    delete current_filter;
    current_filter = NULL;
  }
}

//----------------------------------------------------------------------------
// Add a command to the current registered filter.
//----------------------------------------------------------------------------
int
FilterDict::NewFilterAdd(char *regexp, int action, char *arg) {
  if (current_filter) {
    current_filter->Insert(regexp, action, arg);
    return 1;
  }
  else {
    Error(EL_ERROR, "FilterDict::AddFilter", 
	  "attempted to insert an action without a registered filter.", 
	  NULL);
    return 0;
  }
}

//----------------------------------------------------------------------------
// Commit the current registered filter to the dictionary by inserting it.
//----------------------------------------------------------------------------
int 
FilterDict::NewFilterStore(void) {
  if (current_filter) 
    if (Filters->insert(current_filter->GetName(), current_filter))
      Error(EL_ERROR, "FilterDict::StoreFilter", 
	    "replacing previous definition for",
	    current_filter->GetName(), ".", NULL);
    else
      return 1;
  else
    Error(EL_ERROR, "FilterDict::StoreFilter", 
	  "attempted to store without a registered filter.", NULL);
  return 0;
}

//----------------------------------------------------------------------------
// Add a filter to the current ScreenReader filter dictionary.
//----------------------------------------------------------------------------
int 
FilterDict::AddFilter(char *filter_name) {
  char **new_names;
  TextFilter *t;
  int i;
  mString filter(filter_name);
  if (num_filters == max_filters) {
    new_names = new char*[max_filters += MAX_FILTER];
    for (i = 0; i < num_filters; i++)
      new_names[i] = filter_names[i];
    delete [] filter_names;
    filter_names = new_names;
  }
  if (t = Filters->lookup(filter)) {
    filter_names[num_filters++] = Strdup(filter_name);
    return 1;
  }
  else
    return 0;
}

//----------------------------------------------------------------------------
// Return the active status of the filter in the specified position.
//----------------------------------------------------------------------------
int
FilterDict::GetFilterActive(const int position) {
  TextFilter *t;
  char buffer[10];
  if ((position >= num_filters) || (position < 0)) {
    sprintf(buffer, "%d", position);
    Error(EL_ERROR, "FilterDict::GetFilterActive", "no filter at position ",
	  buffer, ".", NULL);
    return 0;
  }
  else {
    t = Filters->lookup(filter_names[position]);
    return t->GetActive();
  }
}

//----------------------------------------------------------------------------
// Return the name of the filter in the specified position.
//----------------------------------------------------------------------------
char *
FilterDict::GetFilterName(const int position) {
  char buffer[10];
  if ((position >= num_filters) || (position < 0)) {
    sprintf(buffer, "%d", position);
    Error(EL_ERROR, "FilterDict::GetFilterActive", "no filter at position ",
	  buffer, ".", NULL);
    return NULL;
  }
  else
    return filter_names[position];
}

//----------------------------------------------------------------------------
// Set the active status of the filter in the specified position.
//----------------------------------------------------------------------------
void 
FilterDict::SetFilterActive(const int position, const int status) {
  TextFilter *t;
  char buffer[10];
  if (position >= num_filters) {
    sprintf(buffer, "%d", position);
    Error(EL_ERROR, "FilterDict::SetFilterActive", "no filter at position ",
	  buffer, ".", NULL);
  }
  else {
    t = Filters->lookup(filter_names[position]);
    t->SetActive(status);
  }
}

//char *
void
FilterDict::Filter(CursorRepData &input) {
  char *result;
  TextFilter *f;
  int i;
  for (i = 0; i < num_filters; i++) {
    f = Filters->lookup(filter_names[i]);
    if (!f)
      Error(EL_ERROR, "FilterDict::Filter","could not find filter ",
	    filter_names[i], ".", NULL);
    else if (f->GetActive()) {
      Error(EL_DEBUG, "FilterDict::Filter", "applying filter ",
	    filter_names[i], ".", NULL);
      // result = f->Filter(result);
      f->Filter(input);
    }
    else
      Error(EL_DEBUG, "FilterDict::Filter", "disabling filter ",
	    filter_names[i], ".", NULL);
  }
}

int
FilterDict::LoadFilters(char *filename) {
  FILE *filterFile = NULL;
  if (filename == NULL) {
    Error(EL_ERROR, "FilterDict::LoadFilters", "No filename specified.",
	  NULL);
    return 0;
  }
  if ((filterFile = fopen(filename, "r")) == NULL) {
    Error(EL_ERROR, "FilterDict::LoadFilters", "Error opening file",
	  filename, ".", NULL);
    return 0;
  }
  filterin = filterFile;
  if (filterparse()) {
    Error(EL_ERROR, "FilterDict::LoadFilters", "Error parsing file",
	  filename, ".", NULL);
    return 0;
  }
  fclose(filterin);
  return 1;
}

void
FilterDict::FreeFilters()
{
    FilterDict::Filters->kill_all();
    delete FilterDict::Filters;
}

//****************************************************************************
// The various Move* and Get* routines return the following things:
//
// Character: one character (duh).
// Word: a string of characters whose beginning and end are marked
//   by one of a number of delimiting characters.  Typically, these
//   will be spaces.
// Sentence: a series of words whose end is marked by a terminating
//   character, such as a period, question mark, etc.  The terminating
//   punctuation is assumed to be a part of the final word of the
//   sentence.
// Paragraph: a series of sentences whose start and end are marked by
//   one or more intervening blank lines. 
//
// I'm going to make a ton of assumptions about the way things are
//   punctuated.  In particular, I need to assume that things are 
//   spaced properly, since in typical usage, punctuation will be
//   associated with a nearby word.  For example, in the sentence:
//
// "This sentence is nonsensical, used purely for testing purposes."
//
// using purely whitespace as delimiters, the comma character will be
//   associated with the word "nonsensical", and the period character
//   will be associated with the word "purposes".  These characters
//   must be left with the words to provide context for the sentence
//   retrieval routine.  People who surround their punctuation with
//   spaces, or attempt to use punctuation marks are word delimiters
//   will get some strange results.
//
//****************************************************************************

//****************************************************************************
//** Methods for ScreenReader objects
//****************************************************************************

ScreenReader::ScreenReader(TextRep *text) {
  mListVal<mString> list;
  mListValIterator<mString> iter(list);
  int i, listsize;
  const char *delName = Config::GetString(TEMPLATE_DEL_STR);
  const char *termName = Config::GetString(TEMPLATE_TERM_STR);

  textrep = text;
  review = new CursorRep(textrep);
  edit = new CursorRep(textrep);

  delimitString = NULL;
  terminalString = NULL;
  //  if (name = Config::GetString(TEMPLATE_DEL_STR))
  if (delName)
    SetDelimiterString(delName);
  else
    SetDelimiterString(DEL_STR);
  //  if (name = Config::GetString(TEMPLATE_TERM_STR))
  if (termName)
    SetTerminalString(termName);
  else
    SetTerminalString(TERM_STR);
  operation_mode = CURSOR_EDIT;
  follow_mode = FOLLOW_ON;

  incoming_mode = INCOMING_LETTER;
  incoming_count = 0;

  debug_flag = 1;

  list = Template::GetStringList(textrep->obj(), TEMPLATE_FILTER_LIST);
  listsize = list.entries();
  if (!listsize) {
    Error(EL_DEBUG, "ScreenReader::ScreenReader", "could not locate filter ",
	  "list to be used in template file.", NULL);
    list = Config::GetStringList(CONFIG_FILTER_LIST);
    iter = mListValIterator<mString>(list);
    listsize = list.entries();
    if (listsize > 0)
      for (i = 0, iter.next(); i < listsize; i++, iter.next()) {
	  //	name = (char *)iter.current();
	char * name = (char *)iter.current();
	Error(EL_DEBUG, "ScreenReader::ScreenReader", "installing filter ",
	      name, ".", NULL);
	filter.AddFilter(name);
      }
    else {
      Error(EL_ERROR, "ScreenReader::ScreenReader", 
	    "no filters specified -- using default filters.",
	    NULL);
      filter.AddFilter("SpeakCaps");
      filter.AddFilter("SpeakUNIX");
    }
  }
  else {
    if (listsize > 0) {
      iter = mListValIterator<mString>(list);
      for (i = 0, iter.next(); i < listsize; i++, iter.next()) {
	  //	name = (char *)iter.current();
	char *name = (char *)iter.current();
	Error(EL_DEBUG, "ScreenReader::ScreenReader", "installing filter ",
	      name, ".", NULL);
	filter.AddFilter(name);
      }
    }
  }
}

ScreenReader::~ScreenReader(void) {
  // Alter this if we move to an array of cursors
  delete review;
  delete edit;
  // We may need to delete the delimiter/terminal strings, too.
}

// These routines set the delimiter and terminal strings for the
//   ScreenReader.  No one calls them yet.  B)
void
ScreenReader::SetDelimiterString(const char *new_delimit) {
  if (delimitString)
    delete [] delimitString;
  delimitString = new char[strlen(new_delimit)+1];
  strcpy(delimitString, new_delimit);
}

void
ScreenReader::SetTerminalString(const char *new_terminal) {
  if (terminalString)
    delete [] terminalString;
  terminalString = new char[strlen(new_terminal)+1];
  strcpy(terminalString, new_terminal);
}

int
ScreenReader::SetReviewCursor(int x, int y) {
  try {
    review->SetCursor(x, y);
  }
  catch (CursorOutOfBounds err) {
    return 0;
  }
  return 1;
};

int
ScreenReader::SetEditCursor(int x, int y) {
  try {
    edit->SetCursor(x, y);
  }
  catch (CursorOutOfBounds err) {
    return 0;
  }
  return 1;
};

void
ScreenReader::SetCursorPosition(int x, int y) {
  // update TextRep information before all ops
  review->Refresh();
  edit->Refresh();
  switch (operation_mode) {
  case CURSOR_REVIEW: SetReviewCursor(x, y); break;
  case CURSOR_EDIT: SetEditCursor(x, y); break;
  default: return;		    
  }
}

void
ScreenReader::SetOperationMode(enum cursor_mode_t new_mode) {
  if ((new_mode == CURSOR_REVIEW) || (new_mode == CURSOR_EDIT))
    operation_mode = new_mode;
  else if (debug()) {
    char tmp_buffer[16];
    sprintf(tmp_buffer, "%d", (int)new_mode);
    Error(EL_ERROR, "ScreenReader::SetOperationMode",
	  "Incorrect operation mode: ", new_mode, ".", NULL);
  }
}

void
ScreenReader::SetFollowMode(enum follow_mode_t new_mode) {
  if ((new_mode == FOLLOW_ON) || (new_mode == FOLLOW_OFF))
    follow_mode = new_mode;
  else if (debug()) {
    char tmp_buffer[16];
    sprintf(tmp_buffer, "%d", (int)new_mode);
    Error(EL_ERROR, "ScreenReader::SetFollowMode",
	  "Incorrect follow mode: ", new_mode, ".", NULL);
  }
}

void
ScreenReader::SetLiteralMode(enum literal_mode_t new_mode) {
  if ((new_mode == LITERAL_ON) || (new_mode == LITERAL_OFF))
    literal_mode = new_mode;
  else if (debug()) {
    char tmp_buffer[16];
    sprintf(tmp_buffer, "%d", (int)new_mode);
    Error(EL_ERROR, "ScreenReader::SetLiteralMode",
	  "Incorrect literal mode: ", new_mode, ".", NULL);
  }
}

void 
ScreenReader::SyncReviewToEdit(void) {
  review->SetCursor(edit->GetCursorX(), edit->GetCursorY());
}

void 
ScreenReader::SyncEditToReview(void) {
  edit->SetCursor(review->GetCursorX(), review->GetCursorY());
}

//
// This is an odd bit of code.  Ideally, we know where the app cursor is, and
// we can always force the edit cursor to wind up where the app cursor is
// to make sure they don't lose sync with each other.
//
// In practice this only works with "block cursor" apps, or apps with other
// mechanisms for us to authoritatively detect the position of the app cursor.
//
// As a result this routine has two code paths.
//
// If the app cursor has ever been initialized (that is, it is not  at (-1, -1), then
// we assume that we have authoritative app cursor tracking via the
// block cursor (or some other) trick.  In this case we reset the position of
// the edit cursor to be wherever the app cursor is.
//
// If the app cursor is uninitialized, we assume that we don't have app cursor
// tracking and leave the edit cursor where it is (since we have no app cursor
// to sync it to).
//
// Note that a better solution might be to retrieve some attribute for the app
// that says whether app cursor tracking is provided, and key off that, perhaps
// updating the cached app cursor inside the textrep.  That solution is quite
// slow, however...
//
void
ScreenReader::SyncEditToApp(void)
{
    static int x, y;
    textrep->GetAppCursor(y, x);

    if ((x != -1) && (y != -1)) {
	try {
	    edit->SetCursor(x, y);
	}
	catch (CursorOutOfBounds err) {
	    char tmp[100];
	    sprintf(tmp, "x: %d, y: %d", x, y);
	    Error(EL_ERROR, "CursorRep::SetCursor", "caught CursorOutOfBounds ",
		  "error ", tmp, NULL);
	}
    }
}

//----------------------------------------------------------------------------
// EndWord()
//   Determines whether a particular word is a terminal word; that is, 
//   if it ends a properly-punctuated sentence.  This will be true if
//   the final character in the string is itself a terminal punctuator,
//   or if the final character is a quotation mark (single or double) and
//   the immediately preceding character is a terminal.
//
// I might need to add some more cases for parentheses, braces, etc.
//----------------------------------------------------------------------------
int
ScreenReader::EndWord(const char *w) {
  int len;
  len = strlen(w);
  if ((w[--len] == '"') || (w[len] == '\'') || (w[len] == '`'))
    len--;
  return TerminalChar(w[len]);
}

//----------------------------------------------------------------------------
// ResizeDataPtrArray()
//   Used to reallocate additional space for the arrays used to store 
//   intermediate values for the MoveSentence and MoveParagraph routines.
//----------------------------------------------------------------------------
TextData *
ScreenReader::ResizeDataPtrArray(TextData *&array, int &current_size,
				 const int block_size) {
  TextData *resize_array;
  int i, new_size;
  new_size = current_size + block_size;   // add increment to current size
  resize_array = new TextData[new_size];
  for (i = 0; i < current_size; i++)      // copy old array contents
    resize_array[i] = array[i];
  current_size = new_size;                // return new size to caller
#ifdef BEAM
  delete array;                        // delete old array
#else
  delete [] array;                        // delete old array
#endif
  return resize_array;
}

//----------------------------------------------------------------------------
// MoveChar()
//   Given a cursor (which will be one of the internal cursors stored by the
//   ScreenReader object) and an integer count, locate the xth char relative
//   to the given cursor position and return it in a two-character string.
//   The first character of the string is the character at the specified 
//   offset; the second character is a terminating null.  New storage is
//   allocated by this routine to return the string; this must be deleted
//   by the user to avoid memory leaks.  The cursor is updated to point to 
//   the character returned, if it exists.
//----------------------------------------------------------------------------
CursorRepData *
ScreenReader::MoveChar(CursorRep &c, int x, int update) 
throw (ScreenReaderNotFound) {
  CursorRepData *ptr;
  int x1, y1, x2, y2;
  char *text;
  CursorRep start = c;
  start.Refresh();
  start.GetCursor(x1, y1);
  try {
    start += x;
  }
  catch (CursorOutOfBounds err) {
    throw ScreenReaderNotFound(c.GetCursorX(), c.GetCursorY(), sr_char, x);
  }
  c = start;
  c.GetCursor(x2, y2);
  if (update && operation_mode == CURSOR_EDIT) {
    MoveAppCursor(x1, y1, x2, y2);
    SyncEditToApp();
  }
  return start.GetData(1);
}

//----------------------------------------------------------------------------
// MoveWord()
//   Given a cursor (which will be one of the internal cursors stored
//   by the ScreenReader object) and an integer count, locate the xth 
//   word relative to the given cursor position and return its representation
//   in a character string.  All characters flanked by delimiting chars
//   will be returned in the string, plus a null terminating character.
//   New storage is allocated by this routine to return the string; this must
//   be deleted by the user to avoid memory leaks.  The cursor is updated to
//   point to a position within the word returned, if it exists.
//----------------------------------------------------------------------------
CursorRepData * 
ScreenReader::MoveWord(CursorRep &c, int x, int update) 
throw (ScreenReaderNotFound) {
    int dir = (x >= 0) ? 1 : -1, count, index = (dir > 0) ? x : -x;
    int x1, y1, x2, y2;
    CursorRep start = c, end;
    start.Refresh();
    start.GetCursor(x1, y1);

    // Scan forward/backward to the next non-delimiter (between words)
    try {
	while (DelimitChar(start.GetChar())) (dir > 0) ? start++ : start--;
    } 
    catch (CursorOutOfBounds err) { 
	// No non-delimiter characters found
	throw ScreenReaderNotFound(c.GetCursorX(), c.GetCursorY(), 
				   sr_word, index);
    }
    // Scan to the start/end of the current word (forward/backward).  Set
    //   cursor to point to the "first" non-whitespace character.
    try {
	if (dir > 0) {
	    while (!DelimitChar(start.GetChar())) start--;
	    start++;
	}
	else {
	    while (!DelimitChar(start.GetChar())) start++;
	    start--;
	}
    }
    catch (CursorOutOfBounds err) {
	// Out of bounds locating start/end of word. Cursor is reset to the
	//   character before the out-of-bounds occurred.
    }
    end = start;
    // Loop over requested number of words
    do {
	count = 0;
	if (dir > 0)   // Update start/end cursor depending on direction
	    start = end;
	else 
	    end = start;
	// Scan a word.  Save its length in count.
	try {
	    while (!DelimitChar((dir > 0) ? end.GetChar() : start.GetChar())) {
		(dir > 0) ? end++ : start--;
		count++;
	    }
	}
	catch (CursorOutOfBounds err) {
	    if (!index) 
		break; // Out of bounds, but we've got a word.
	    else
		throw ScreenReaderNotFound(c.GetCursorX(), c.GetCursorY(), 
					   sr_word, index);
	}
	// If we need the "next" word, skip whitespace to get there.
	try {
	    if (index)
		while (DelimitChar((dir > 0) ? end.GetChar() : start.GetChar()))
		    (dir > 0) ? end++ : start--;
	}
	catch (CursorOutOfBounds err) {
	    // Couldn't find next word
	    throw ScreenReaderNotFound(c.GetCursorX(), c.GetCursorY(), 
				       sr_word, index);
	}
    } while (index--);
    // Place cursor at end/start of the last word we examined.
    if (dir > 0)
	c = --end;
    else
	c = ++start;
    c.GetCursor(x2, y2);
    if (update && operation_mode == CURSOR_EDIT) {
	MoveAppCursor(x1, y1, x2, y2);
	SyncEditToApp();
    }
    return start.GetData(count);
}

// Notes on MoveSentence and MoveParagraph

// In order to avoid re-scanning the text as much as possible, items
//   that are read in (words in MoveSentence, sentences in MoveParagraph)
//   are stored in an array.  The index value i defines the index of the
//   first element in the array current *not* assigned, and is used as
//   the insertion point for new elements.  Since the cursor may be located
//   in the center of the unit which it is retrieving, we scan backwards
//   from the current cursor position, and then forwards from the same
//   position, working from the "inside out".  Thus, elements up to the
//   start of the unit are stored backwards, and units up to the end of
//   the sentence are stored forwards; the start_index and end_index values
//   store the indices in the array which point to the first and last
//   stored units.

// Okay.  Looks like we'll have to do some extra work in MoveSentence and 
//   MoveParagraph to grab whitespace and return all of it.  We can do this
//   by computing the difference between the last word and calling 
//   GetData.
 
//----------------------------------------------------------------------------
// MoveSentence()
//   Given a cursor (which will be one of the internal cursors stored
//   by the ScreenReader object) and an integer count, locate the xth
//   sentence relative to the given cursor position and return its
//   representation in a character string.  All words between the first
//   word prior to the current cursor position which is a terminal 
//   word (q.v. EndWord(), above) and the first terminal word at or after
//   the current cursor position, inclusive.  The returned sentence is 
//   null-terminated.  New storage is allocated by this routine for the 
//   returned sentence and must be deleted by the user to avoid memory
//   leaks.  The cursor is updated to point to a location within the 
//   returned sentence, if it exists.
//----------------------------------------------------------------------------

CursorRepData *
ScreenReader::MoveSentence(CursorRep &c, int x, int update) 
throw (ScreenReaderNotFound) {
    CursorRepData *ptr = NULL;
    char *sentence;
    int i, start_index, end_index, dir = (x >= 0) ? 1 : -1;
    int x1, y1, x2, y2;
    int len, tmp, index, max_words = MAX_SENTENCE_LEN;
    int lastx, lasty, startx=0, starty=0, endx, endy;
    CursorRep start, end;

#ifndef BEAM
    TextData *twords, tdata;
    index = (dir > 0) ? x : -x;
    start = end = c;

    start.Refresh();
    start.GetCursor(x1, y1);
    twords = new TextData[max_words];

    // Loop over number of requested sentences
    do {
	// Erase words from previous sentence
	//	if (i >= 0)
	    //	    for (start_index = 0; start_index < i; start_index++)
	    //		delete twords[start_index];
	i = tmp = 0;
	// Scan backward to beginning of current sentence
	try {
	    do {
		lastx = startx; lasty = starty;
		ptr = MoveWord(start, -1, 0);
		twords[i] = ptr->GetTData();
		ptr->GetStart(startx, starty);
		if (!i)
		    ptr->GetEnd(endx, endy);
		delete ptr;
		if (i == (max_words - 1))
		    twords = ResizeDataPtrArray(twords, max_words,
						MAX_SENTENCE_LEN);
	    } while (!EndWord((const char *)twords[i++]));
	}
	catch (ScreenReaderNotFound err) {
	    if ((index) && (dir < 0)) {
		//	for (start_index = 0; start_index < i; start_index++)
		//	    delete twords[start_index];
		delete [] twords;
		delete ptr;
		throw ScreenReaderNotFound(c.GetCursorX(), c.GetCursorY(),
					   sr_sent, index);
	    }
	    else
		tmp = 1;
	}
	
	// Throw away the last word read, since it's the end of the previous
	//   sentence.  Don't do this if we've hit the start of the text
	//   area, though -- then it's the first word in this sentence.
	//	if (!tmp) {
	//	    delete twords[--i];
	//	}

	// note: i-1 indexes the last word processed, which is actually the
	// last word of the previous sentence. i-2 actually indexes the first
	// word of the current sentence. -wl
	start_index = i-2;
	
	// Read forward to end of current sentence
	try {
	    int run = 0;
	    do {
		if (i == max_words)
		    twords = ResizeDataPtrArray(twords, max_words,
						MAX_SENTENCE_LEN);
		ptr = MoveWord(end, (run++) ? 1 : 0, 0);
		twords[i] = ptr->GetTData();
		ptr->GetEnd(endx, endy);
		delete ptr;
	    } while (!EndWord((const char *)twords[i++]));
	}
	catch (ScreenReaderNotFound err) {
	    if ((index) && (dir > 0)) {
		// clean up before exiting
		//	for (start_index = 0; start_index < i; start_index++)
		//	    delete twords[start_index];
		delete [] twords;
		delete ptr;
		throw ScreenReaderNotFound(c.GetCursorX(), c.GetCursorY(), 
					   sr_sent, index);
	    }
	}

	end_index = --i;

	// Move to the next word in the next sentence.
	if (index) {
	    try {
		if (dir > 0) {
		    // move back to start of this sentence
		    delete MoveWord(end, 1, 0);
		    start = end;
		}
		else {
		    end = start;
		}
	    }
	    catch (ScreenReaderNotFound err) {
		//	for (start_index = 0; start_index < i; start_index++)
		//	    delete twords[start_index];
		delete [] twords;
		delete ptr;
		throw ScreenReaderNotFound(c.GetCursorX(), c.GetCursorY(), 
					   sr_sent, index);
	    }
	}
    } while (index--);
    
    // Update cursor to point to the new sentence.
    if (dir > 0) 
	c = end;
    else {
	if (!tmp)
	    delete MoveWord(start, 1, 0);
	c = start;
    }
    
    // Get all the words to this sentence and concatenate them together.
    // Right now this is pretty inefficient.  Maybe I'll eventually rewrite
    //   this to use the mString stuff.
    //### BIG-ASS MEMORY LEAK HERE -- DELETE TWORDS ###
    c.GetCursor(x2, y2);
    if (update && operation_mode == CURSOR_EDIT) {
	MoveAppCursor(x1, y1, x2, y2);
	SyncEditToApp();
    }

    TextAttr *setattrs, *retattrs;
    int j;
    for (len = 0, i = 0; i <= end_index; i++)
	len += twords[i].Length() + 1;
    sentence = new char[len]; sentence[0] = '\0';
    setattrs = new TextAttr[len];
    for (tmp=index=0, i = start_index; i >= 0; index+=twords[i--].Length()) {
	strcat(sentence, (const char *)twords[i]);

	// for (retattrs = (const TextAttr *)*twords[i], j = 0; 
	for (retattrs = (TextAttr *)twords[i], j = 0; 
	     j < twords[i].Length(); j++)
	    setattrs[index+j] = retattrs[j];

	// Set attribute for delimiter to attribute of last char
	setattrs[index+j] = setattrs[index+j-1];
	tmp = strlen(sentence);

	// ### POSSIBLE TROUBLE HERE IF NOTHING FORWARD FROM THIS POSITION ###
	sentence[tmp] = delimitString[0];
	sentence[tmp+1] = '\0';
    }
    // note: (start_index + 1) points the last word of the prev. sent. -wl
    for (i = start_index+2; i <= end_index; index+=twords[i++].Length()) {
	strcat(sentence, (const char *)twords[i]);

	//    for (retattrs = (const TextAttr *)*twords[i], j = 0; 
	for (retattrs = (TextAttr *)twords[i], j = 0; 
	     j < twords[i].Length(); j++)
	    setattrs[index+j] = retattrs[j];
	tmp = strlen(sentence);

	if (i != end_index) {
	    sentence[tmp++] = delimitString[0];
	    setattrs[index+j] = setattrs[index+j-1];
	}
	sentence[tmp] = '\0';
    }

    //    for (start_index = 0; start_index < i; start_index++)
    //	delete twords[start_index];
    delete [] twords;

    ptr = new CursorRepData;
    //    tdata = new TextData;
    tdata.Chars(sentence);
    tdata.Attrs(setattrs);
    tdata.Length(len-1);     // Don't include the terminating NULL in length
    //  ptr->SetData(sentence);
    ptr->SetTData(tdata);
    ptr->SetStart(startx, starty);
    ptr->SetEnd(endx, endy);
#endif
    return ptr;
}

//----------------------------------------------------------------------------
// MoveParagraph()
//   Given a cursor (which will be one of the internal cursors stored
//   by the ScreenReader object) and an integer count, locate the xth
//   paragraph relative to the given cursor position and return its
//   representation in a character string.  All sentences between text
//   separated by blanklines is returned, inclusive.  The returned paragraph
//   is null-terminated.  New storage is allocated by this routine for the 
//   returned paragraph and must be deleted by the user to avoid memory
//   leaks.  The cursor is updated to point to a location within the 
//   returned paragraph, if it exists.
//----------------------------------------------------------------------------

// I'm expecting a full line of whitespace (delimiter) characters as a 
//   blankline.  If they're not there, this returns a *lot* of text.
CursorRepData *
ScreenReader::MoveParagraph(CursorRep &c, int x, int update) 
throw (ScreenReaderNotFound) {
  CursorRepData *ptr;
  char *paragraph, *test_word_str = NULL;
  TextData *tsentences, tdata;
  int i = -1, start_index, end_index, dir = (x >= 0) ? 1 : -1;
  int x1, y1, x2, y2;
  int len, tmp, index, max_sentences = MAX_PARAGRAPH_LEN;
  int start_line, end_line;
  int startx=0, starty=0, endx, endy, lastx, lasty;
  CursorRep start, end, test_word;

#ifndef BEAM
  index = (dir > 0) ? x : -x;
  start = c;
  tsentences = new TextData[max_sentences];
  start.Refresh();
  start.GetCursor(x1, y1);
  // Back up to the beginning of the current sentence.  Search for a
  //   terminal word, then move forward one word.
  try {
    do {
      if (test_word_str)
	delete [] test_word_str;
      lastx = startx; lasty = starty;
      ptr = MoveWord(start, -1, 0);
      ptr->GetStart(startx, starty);
      //      test_word_str = (const char *)*ptr->GetTData();
      test_word_str = (char *) ptr->GetTData();
      delete ptr;
    } while (!EndWord(test_word_str));
    // advance back to first word in this sentence
    delete MoveWord(start, 1, 0);
    delete [] test_word_str; test_word_str = NULL;
  }
  catch (ScreenReaderNotFound err) {
    // If we hit the start of the text area, we can start from
    //   here.  Set start to this area.
    start.GetCursor(startx, starty);
  }
  end = start;
  do {
    // Erase previous sentences from the last paragraph parsed
    if (i >= 0) {
	//      for (start_index = 0; start_index < i; start_index++) {
	//	delete tsentences[start_index];
	//      }
    }
    i = tmp = 0;
    // Scan previous sentences in this paragraph
    // This loop may go forever if MoveSentence doesn't throw anything!
    try {
      do {
	test_word = start;
	// Get start of this word
	ptr = MoveWord(test_word, 0, 0); 
	start_line = ptr->GetStartY();
	delete ptr;
	// Get end of previous word
	ptr = MoveSentence(start, -1, 0);
	tsentences[i++] = ptr->GetTData();
	end_line = ptr->GetEndY();
	delete ptr;
	if (i == max_sentences) {
	  tsentences = ResizeDataPtrArray(tsentences, max_sentences,
					  MAX_PARAGRAPH_LEN);
	}
      } while ((start_line - end_line) <= 1);
    }
    catch (ScreenReaderNotFound err) {
      // If we're scanning forward, reaching the start of the text area isn't
      //   a problem.
      if ((index) && (dir < 0)) {
	  //	for (start_index = 0; start_index < i; start_index++)
	  //	  delete tsentences[start_index];
	delete tsentences;
	throw ScreenReaderNotFound(c.GetCursorX(), c.GetCursorY(),
				   sr_para, index);
      }
      else
	tmp = 1;
    }
    //    if (!tmp) {
    //      delete tsentences[--i];
    //    }
    start_index = i-1;
    // Read to end of paragraph
    try {
      do {
	ptr = MoveSentence(end, 0, 0);
	tsentences[i++] = ptr->GetTData();
	start_line = ptr->GetEndY();
	delete ptr;
	if (i == max_sentences) {
	  tsentences = ResizeDataPtrArray(tsentences, max_sentences,
					  MAX_PARAGRAPH_LEN);
	}
	// need to move cursor to start of next word
	ptr = MoveWord(end, 1, 0);
	end_line = ptr->GetStartY();
	delete ptr;
      } while ((end_line - start_line) <= 1);
    }
    catch (ScreenReaderNotFound err) {
      if ((index) && (dir > 0)) {
	  //	for (start_index = 0; start_index < i; start_index++)
	  //	  delete tsentences[start_index];
	delete tsentences;
	throw ScreenReaderNotFound(c.GetCursorX(), c.GetCursorY(), 
				   sr_para, index);
      }
    }
    end_index = --i;
    // sync the pointers for the next paragraph
    if (index) {
      try {
	if (dir > 0) 
	  start = end;
	else 
	  end = start;
      }
      catch (ScreenReaderNotFound err) {
      }
    }
  } while (index--);
  // Move back to the returned sentence
  if (dir > 0) {
    delete MoveWord(end, -1, 0);
    c = end;
  }
  else {
    //    delete MoveWord(test_word, 1);
    c = test_word;
  }
  c.GetCursor(x2, y2);
  if (update && operation_mode == CURSOR_EDIT){
    MoveAppCursor(x1, y1, x2, y2);
    SyncEditToApp();
  }
  int j;
  TextAttr *setattrs, *retattrs;
  for (len = 0, i = 0; i <= end_index; i++) 
    len += tsentences[i].Length() + 1;
  paragraph = new char[len]; paragraph[0] = '\0';
  setattrs = new TextAttr[len];
  for (index = 0, i = start_index; i >= 0; index+=tsentences[i--].Length()) {
    strcat(paragraph, (const char *)tsentences[i]);
    //n    for (retattrs = (const TextAttr *)*tsentences[i], j = 0; 
    for (retattrs = tsentences[i], j = 0; 
	 j < tsentences[i].Length(); j++)
      setattrs[index+j] = retattrs[j];
    setattrs[index+j] = setattrs[index+j-1];
    tmp = strlen(paragraph);
    paragraph[tmp] = delimitString[0];
    paragraph[tmp+1] = '\0';
  }
  for (i = start_index+1; i <= end_index; index+=tsentences[i++].Length()) {
    strcat(paragraph, (const char*) tsentences[i]);
    //    for (retattrs = (const TextAttr *)*tsentences[i], j = 0; 
    for (retattrs = tsentences[i], j = 0; 
	 j < tsentences[i].Length(); j++)
      setattrs[index+j] = retattrs[j];
    tmp = strlen(paragraph);
    if (i != end_index) {
      paragraph[tmp++] = delimitString[0];
      setattrs[index+j] = setattrs[index+j-1];
    }
    paragraph[tmp] = '\0';
  }
  //  for (start_index = 0; start_index < i; start_index++)
  //    delete tsentences[start_index];
  delete tsentences;
  //  tdata = new TextData;
  tdata.Chars(paragraph);
  tdata.Attrs(setattrs);
  tdata.Length(len);
  ptr = new CursorRepData();
  ptr->SetTData(tdata);
#endif
  return ptr;
}

// #### THIS IS PROBABLY BROKEN ####
// (not that anyone calls it B) )
CursorRepData *
ScreenReader::MovePage(CursorRep &c, int x, int update) {
  CursorRepData *ptr;
  CursorRep start = c;
  start.Refresh();
  try {
    ptr = start.GetData(0, 0, start.GetBoundsX() * start.GetBoundsY());
  }
  catch (CursorOutOfBounds err) {
    // This really shouldn't happen.
  }
  start.SetCursor(start.GetBoundsX()-1, start.GetBoundsY()-1);
  c = start;
  return ptr;
}

CursorRepData *
ScreenReader::MoveLine(CursorRep &c, int x, int update) 
throw (ScreenReaderNotFound) {
  int x1, y1, x2, y2;
  CursorRepData *ptr;
  CursorRep test = c;
  test.Refresh();
  test.GetCursor(x1, y1);
  try {
    test.SetCursor(test.GetCursorX(), test.GetCursorY() + x);
    ptr = test.GetData(0, test.GetCursorY(), test.GetBoundsX());
  }
  catch (CursorOutOfBounds err) {
    throw ScreenReaderNotFound(c.GetCursorX(), c.GetCursorY(), 
				   sr_line, x);
  }
  c = test;
  c.GetCursor(x2, y2);
  if (update && operation_mode == CURSOR_EDIT) {
    MoveAppCursor(x1, y1, x2, y2);
    SyncEditToApp();
  }
  return ptr;
}

//-----------------------------------------------------------------------------
// The following routines call the corresponding Move routine and reset the
//   cursor position when they're done.
//-----------------------------------------------------------------------------
CursorRepData *
ScreenReader::GetChar(CursorRep &c, int x) {
  CursorRepData *tmp;
  int tmpx, tmpy;
  c.GetCursor(tmpx, tmpy);
  tmp = MoveChar(c, x, 0);
  c.SetCursor(tmpx, tmpy);
  return tmp;
}

CursorRepData *
ScreenReader::GetWord(CursorRep &c, int x) {
  CursorRepData *tmp;
  int tmpx, tmpy;
  c.GetCursor(tmpx, tmpy);
  tmp = MoveWord(c, x, 0);
  c.SetCursor(tmpx, tmpy);
  return tmp;
}

CursorRepData *
ScreenReader::GetSentence(CursorRep &c, int x) {
  CursorRepData *tmp;
  int tmpx, tmpy;
  c.GetCursor(tmpx, tmpy);
  tmp = MoveSentence(c, x, 0);
  c.SetCursor(tmpx, tmpy);
  return tmp;
}

CursorRepData *
ScreenReader::GetParagraph(CursorRep &c, int x) {
  CursorRepData *tmp;
  int tmpx, tmpy;
  c.GetCursor(tmpx, tmpy);
  tmp = MoveParagraph(c, x, 0);
  c.SetCursor(tmpx, tmpy);
  return tmp;
}

CursorRepData *
ScreenReader::GetLine(CursorRep &c, int x) {
  CursorRepData *tmp;
  int tmpx, tmpy;
  c.GetCursor(tmpx, tmpy);
  tmp = MoveLine(c, x, 0);
  c.SetCursor(tmpx, tmpy);
  return tmp;
}

CursorRepData *
ScreenReader::GetPage(CursorRep &c, int x) {
  CursorRepData *tmp;
  int tmpx, tmpy;
  c.GetCursor(tmpx, tmpy);
  tmp = MovePage(c, x, 0);
  c.SetCursor(tmpx, tmpy);
  return tmp;
}

// The following routines are modifications of the original ScreenReader code.

// Kevin Chen  7/17/95

char *
ScreenReader::read(const char *location, const char *token, int raw) {
  int offset;
  static char error[] = "";  // return something that doesn't alloc memory
  CursorRepData *result;

  if (strcmp(location, "this") == 0)
    offset = 0;
  else if (strcmp(location, "next") == 0)
    offset = 1;
  else if (strcmp(location, "prev") == 0)
    offset = -1;
  else
    return Strdup(error);

  try {
    if (strcmp(token, "char") == 0)
      result = GetChar(((operation_mode == CURSOR_REVIEW) ? 
		      *review : *edit), offset);
    else if (strcmp(token, "word") == 0)
      result = GetWord(((operation_mode == CURSOR_REVIEW) ? 
		      *review : *edit), offset);
    else if (strcmp(token, "sentence") == 0)
      result = GetSentence(((operation_mode == CURSOR_REVIEW) ? 
			  *review : *edit), offset);
    else if (strcmp(token, "paragraph") == 0)
      result = GetParagraph(((operation_mode == CURSOR_REVIEW) ? 
			   *review : *edit), offset);
    else if (strcmp(token, "line") == 0)
      result = GetLine(((operation_mode == CURSOR_REVIEW) ? 
		      *review : *edit), offset);
    else if (strcmp(token, "screen") == 0)
      result = GetPage(((operation_mode == CURSOR_REVIEW) ? 
		      *review : *edit), offset);
    else
      return Strdup(error);
  }
  catch (ScreenReaderNotFound err) {
    return Strdup(error);
  }

  // Only apply the filter if we're not raw.
  if (!raw)
      filter.Filter(*result);
  
  char *retVal = Strdup(result->GetTData());
  delete result;
  return retVal;
  
  //  return (char *) *result->GetTData();
}

char *
ScreenReader::move(const char *location, const char *token) {
  int offset;
  static char error[] = "";  // return something that doesn't alloc memory
  CursorRepData *result;
  
  if (strcmp(location, "this") == 0)
    offset = 0;
  else if (strcmp(location, "next") == 0)
    offset = 1;
  else if (strcmp(location, "prev") == 0)
    offset = -1;
  else
      return Strdup(error);
      //    return error;

  try {
    if (strcmp(token, "char") == 0)
      result = MoveChar(((operation_mode == CURSOR_REVIEW) ? 
			 *review : *edit), offset);
    else if (strcmp(token, "word") == 0)
      result = MoveWord(((operation_mode == CURSOR_REVIEW) ? 
			 *review : *edit), offset);
    else if (strcmp(token, "sentence") == 0)
      result = MoveSentence(((operation_mode == CURSOR_REVIEW) ? 
			     *review : *edit), offset);
    else if (strcmp(token, "paragraph") == 0)
      result = MoveParagraph(((operation_mode == CURSOR_REVIEW) ? 
			      *review : *edit), offset);
    else if (strcmp(token, "line") == 0)
      result = MoveLine(((operation_mode == CURSOR_REVIEW) ? 
			 *review : *edit), offset);
    else if (strcmp(token, "screen") == 0)
      result = MovePage(((operation_mode == CURSOR_REVIEW) ? 
			 *review : *edit), offset);
    else
	return Strdup(error);
	//      return error;
  }
  catch (ScreenReaderNotFound err) {
    return Strdup(error);
  }
  filter.Filter(*result);
  char *retVal = Strdup(result->GetTData());
  delete result;
  return retVal;
  //  return (char *)*result->GetTData();
}

void
ScreenReader::MoveAppCursor(int x1, int y1, int x2, int y2)
{
    static KeySym left = mercator->GetXServer()->StringToKeysym("Left");
    static KeySym right = mercator->GetXServer()->StringToKeysym("Right");
    static KeySym up = mercator->GetXServer()->StringToKeysym("Up");
    static KeySym down = mercator->GetXServer()->StringToKeysym("Down");

    KeySym key;
    register int i;
    int x, y;
    
    if (y1 == y2) {
	key = (x1 > x2) ? left : right;
	// Move to new position on same line.
	for (i = x1; i != x2; (x1 > x2) ? i-- : i++)
	    mercator->GetXServer()->DoKeyEvent(&(textrep->obj()), key, 0, 2, 
					       0, True);
    } else {
#if 0
	// This one's more complicated...
	// Move to column zero on the current line...
	for (i = x1; i > 0; i--)
	    mercator->GetXServer()->DoKeyEvent(&(textrep->obj()), 
					       left, 0, 2, 0, True);
#endif

	// Move up or down to the new line...
	key = (y1 > y2) ? up : down;
	for (i = y1; i != y2; (y1 > y2) ? i-- : i++)
	    mercator->GetXServer()->DoKeyEvent(&(textrep->obj()), key, 0, 2, 
					       0, True);
#if 0
	// Move right out to the desired column...
	for (i = x2; i > 0; i--)
	    mercator->GetXServer()->DoKeyEvent(&(textrep->obj()), 
					       right, 0, 2, 0, True);
#endif
    }
    XSync(mercator->GetXServer()->GetDisplay(), False);
#if 0
    textrep->GetAppCursor(y, x);
    SetEditCursor(x, y);
#endif
}

void
ScreenReader::Speak(char *speak, int interrupt) {
  if (interrupt)
    mercator->GetSpeech()->StopSpeaking();
  mercator->GetSpeech()->Speak(speak);
}

void
ScreenReader::SetIncomingMode(int mode) {
  if ((mode < 0) || (mode > 5))
    return;
  else
    incoming_mode = mode;
}

void
ScreenReader::flush_buffer(void) {
  Speak(incoming_buffer);
  incoming_buffer[incoming_count = 0] = '\0';
}

void
ScreenReader::Process_Incoming_Text(char * a_char) {
  if (incoming_mode == INCOMING_LETTER) 
    if (a_char[0] == '\n')
      Speak("return");
    else if (a_char[0] == ' ') 
      Speak("space");
    else if (a_char[0] == '\t') 
      Speak("tab");
    else if (strcmp(a_char, "Left") && strcmp(a_char, "Right") &&
	     strcmp(a_char, "Up") && strcmp(a_char, "Down"))
	Speak(a_char); // don't say these words in letter mode -wl
  
  else if (incoming_mode == INCOMING_WORD) {
    if ((!strcmp(a_char,"space"))  || (a_char[0] == ' ') || 
	(!strcmp(a_char,"tab")) || (a_char[0] == '\t') || 
	(!strcmp(a_char,"Return")) || (a_char[0] == '\n')) { 
      if (incoming_count > 0)
	flush_buffer();
    }
    else if ((!strcmp(a_char, "Delete")) || 
	     (!strcmp(a_char, "BackSpace"))) {
      incoming_buffer[--incoming_count] = '\0';
    }
    else {
      if (incoming_count >= INCOMING_BUFFER_LEN-2) 
	flush_buffer();
      incoming_buffer[incoming_count++] = *a_char;
      incoming_buffer[incoming_count] = '\0';
    }
  }
  else if (incoming_mode == INCOMING_LINE) {
    if ((!strcmp(a_char, "Return")) || (a_char[0] != '\n'))
      flush_buffer();
    else if ((!strcmp(a_char, "Delete")) || 
	     (!strcmp(a_char, "BackSpace"))) {
      incoming_buffer[--incoming_count] = '\0';
    }
    else {
      if (incoming_count >= INCOMING_BUFFER_LEN-2) 
	flush_buffer();
      incoming_buffer[incoming_count++] = *a_char;
      incoming_buffer[incoming_count] = '\0';
    }
  }
  
  else if (incoming_mode == INCOMING_SENTENCE) { //HERE
  }
  else if (incoming_mode == INCOMING_CLICK) {
    float ratef = 1.0;
    incoming_count++;
  }
}
