//
//
//
//	Note:  The use of this program shall be restricted to educational,
//	non-commercial use only.  All inquiries regarding licensing this
//	software for commercial use may be addressed to:
//	
//	Georgia Tech Research Corporation
//	Technology Licensing
//	Centennial Research Building - Rm. 275
//	400 Tenth Street, N.W.
//	Atlanta, Georgia  30332-0415
//	
//	*************************************************
//	
//	       Copyright 1996
//	Georgia Tech Research Corporation
//	    All Rights Reserved
//
//
//
//
#ifndef _SCREENREADER_H
#define _SCREENREADER_H

#include <assert.h>
#include <string.h>

#include "Speech.h"
#include "TextRep.h"
#include "mList.h"
#include "mDictValPtr.h"
#include "mDictValVal.h"
#include "mRegexp.h"
#include "mString.h"

// Revised screenreader class -- 6/27/95
// Kevin Chen (kchen@cc.gatech.edu)

#define TERM_STR ".?!"
#define DEL_STR " "

#define TEXTBUF_LEN 512
#define INCOMING_BUFFER_LEN 1024

class CursorOutOfBounds  {
public:
  int x, y, boundsx, boundsy;
  CursorOutOfBounds(int x_, int y_, int boundsx_, int boundsy_) :
    x(x_), y(y_), boundsx(boundsx_), boundsy(boundsy_) { }
};

// Note that this class does not own the data it uses; it merely provides
//   storage for the data pointers.
class CursorRepData {
private:
  friend class CursorRep;
  int startx, starty, endx, endy;
  int boundsx, boundsy;
  TextData tdata;
  //  char *data;
public:
    //  CursorRepData(void) : tdata(NULL) {};
    CursorRepData() { };
  ~CursorRepData();
  //  void SetData(const char *input);
  void SetTData(TextData newdata) { tdata = newdata; }
  void Chars(char *chars) { tdata.Chars(chars); }
  void Attrs(TextAttr *attrs) { tdata.Attrs(attrs); }
  void Length(int l) { tdata.Length(l); }
  TextData GetTData(void) { return tdata; }
  //  char * GetData(void);
  int GetBoundsX(void) { return boundsx; }
  int GetBoundsY(void) { return boundsy; }
  void GetBounds(int &x, int &y) { x = boundsx; y = boundsy; }
  int GetStartX(void) { return startx; }
  int GetStartY(void) { return starty; }
  void GetStart(int &x, int &y) { x = startx; y = starty; }
  int GetEndX(void) { return endx; }
  int GetEndY(void) { return endy; }
  void GetEnd(int &x, int &y) { x = endx; y = endy; }
  void SetBounds(int x, int y) { boundsx = x; boundsy = y; }
  void SetStart(int x, int y) { startx = x; starty = y; }
  void SetEnd(int x, int y) { endx = x; endy = y; }
};

// Notes for original CursorRep implementation (this will go away soon)

// The CursorRep object acts as the interface between the TextRep and
//   ScreenReader classes.  It is the job of the CursorRep to provide
//   a cursor abstraction usable by the ScreenReader which provides
//   the functions of pointing and boundschecking, as well as retrieval
//   of the data abstracted by the TextRep.

// In order to speed up the character operations, we store an offset
//   representation of the x and y coordinates, and allow this offset
//   to be out of sync when we do character scans.

class CursorRep {
private:
  int x, y;                   // coordinates of cursor
  //int boundsx, boundsy;       // row, column limits
  TextRep *textrep;           // TextRep for this cursor
  TextData current_line;      // data representation for the current line
  int bogusline;
public:
  CursorRep(void) : x(0), y(0), textrep(NULL), 
      bogusline(1) {;}
  ~CursorRep() {}
  CursorRep(TextRep *t);
  int GetCursorX(void) { return x; }
  int GetCursorY(void) { return y; }
  void GetCursor(int &retx, int &rety) { retx = x; rety = y; }
  void SetCursorX(int newx) throw (CursorOutOfBounds);
  void SetCursorY(int newy) throw (CursorOutOfBounds);
  void SetCursor(int newx, int newy) throw (CursorOutOfBounds);

  int GetBoundsX(void) { return textrep->Columns(); }
  int GetBoundsY(void) { return textrep->Rows(); }
  void GetBounds(int &retx, int &rety) { assert(textrep != NULL); 
  retx = textrep->Columns(); 
  rety = textrep->Rows();}

  const int operator==(const CursorRep &c) { return (x == c.x) && (y == c.y); }
  const int operator!=(const CursorRep &c) { return (x != c.x) || (y != c.y); }
  const int operator>(const CursorRep &c) { return (y > c.y) || 
				  ((y == c.y) && (x > c.x)); }
  const int operator>=(const CursorRep &c) { return (y > c.y) ||
				   (( y == c.y) && (x >= c.x)); }
  const int operator<(const CursorRep &c) { return (y < c.y) ||
				  (( y == c.y) && (x < c.x)); }
  const int operator<=(const CursorRep &c) { return (y < c.y) ||
				   (( y == c.y) && (x <= c.x)); }

  CursorRep& operator++(void) throw (CursorOutOfBounds);
  CursorRep& operator++(int dummy);
  CursorRep& operator--(void) throw (CursorOutOfBounds);
  CursorRep& operator--(int dummy);
  CursorRep& operator+=(const int count) throw (CursorOutOfBounds);
  CursorRep& operator-=(const int count);

  char GetChar();
  TextAttr& GetAttr();
  CursorRepData * GetData(const int len = 1) throw (CursorOutOfBounds);
  CursorRepData * GetData(const int x, const int y, const int len = 1);
  CursorRepData * GetData(const int startx, const int starty,
			  const int endx, const int endy);
  void Refresh(void);
};

#define MAX_FILTER 8
#define MAX_FILTER_COMMANDS 20

#define FILTER_ACTION_REGEXP 0
#define FILTER_ACTION_FONT 1
#define FILTER_ACTION_HIGHLIGHT 2

// Each one of these objects encapsulates a single text filter.  It is
//   identified by a name identifier, and contains num_commands actual
//   command statements, up to max_commands.
class TextFilter {
public:
  TextFilter(void);
  ~TextFilter();
  void SetName(char *name);
  char *GetName(void) { return name; }
  //  char *Filter(char *input);
  void Filter(CursorRepData &input);
  void Insert(char *regexp, int action, char *arg);
  void SetActive(int active_) {active = active_; }
  int GetActive(void) {return active;}
private:
  typedef class FilterEntry {
  public:
    FilterEntry(void) : r("") {
      action = action_arg = -1;
      arg = NULL;
    }
    ~FilterEntry() { if (arg) delete [] arg; }
    mRegexp r;
    int action, action_arg;
    char *arg;
  } *FilterEntryPtr;
  
  char *name;
  int active;
  FilterEntryPtr *commands;
  int num_commands, max_commands;
};

// This is a static structure which holds Filter objects, keyed by their
//   name.
class FilterDict {
public:
  FilterDict(void);
  ~FilterDict();
  static int LoadFilters(char *filename);
  static void FreeFilters();
  // global storage for the current filter
  static int NewFilterRegister(char *name);
  static int NewFilterAdd(char *regexp, int action, char *arg);
  static int NewFilterStore(void);
  static void NewFilterAbort(void);
  static int filterLineCount;
  static TextFilter *current_filter;
  int GetFilterCount(void) {return num_filters;}
  int GetFilterActive(const int position);
  char *GetFilterName(const int position);
  void SetFilterActive(const int position, const int status);
  int AddFilter(char *filter_name);
    //  char *Filter(char *input);
    void Filter(CursorRepData &input);
private:
  static mDictValPtr<mString, TextFilter> *Filters;
  char **filter_names;
  int num_filters, max_filters;
};

enum unit_enum_t {sr_char, sr_word, sr_sent, sr_para, sr_page, sr_line};

class ScreenReaderNotFound {
public:
  enum unit_enum_t unit;
  int x, y, index;
  ScreenReaderNotFound(int x_, int y_, enum unit_enum_t unit_, int index_) :
      x(x_), y(y_), unit(unit_), index(index_) { }
};

#define CONFIG_FILE_LIST             "srFilterDefinitionFiles"
#define CONFIG_FILTER_LIST           "srFilters"

#define TEMPLATE_DEL_STR             "srDelimitChars"
#define TEMPLATE_TERM_STR            "srTerminalChars"
#define TEMPLATE_FILTER_LIST         "srFilters"

#define INCOMING_LETTER       0
#define INCOMING_WORD         1
#define INCOMING_LINE         2
#define INCOMING_SENTENCE     3
#define INCOMING_CLICK        4
#define INCOMING_NONE         5

#define MAX_SENTENCE_LEN 50        // Max. length of sentence in words
#define MAX_PARAGRAPH_LEN 20       // Max. length of paragraph in sentences

class ScreenReader {
public:
  enum cursor_mode_t {CURSOR_REVIEW, CURSOR_EDIT};
  enum follow_mode_t {FOLLOW_ON, FOLLOW_OFF};
  enum literal_mode_t {LITERAL_ON, LITERAL_OFF};

  ScreenReader(TextRep *text);
  ~ScreenReader(void);
  // These methods retrieve the "unit" relative to the current review
  //   cursor location.  That is, an argument value of -1 will read the
  //   previous character/word/sentence/etc.  Note that the "current"
  //   unit is defined as the first unit after the current cursor 
  //   position.

  void SetDelimiterString(const char *new_delimit);
  void SetTerminalString(const char *new_terminal);

  int GetIncomingMode(void) {return incoming_mode;}
  void SetIncomingMode(int mode);
  void Process_Incoming_Text(char *a_char);

  void SetOperationMode(enum cursor_mode_t new_mode);
  enum cursor_mode_t GetOperationMode(void) {return operation_mode;}
  void SetFollowMode(enum follow_mode_t new_mode);
  enum follow_mode_t GetFollowMode(void) {return follow_mode;}
  void SetLiteralMode(enum literal_mode_t new_mode);
  enum literal_mode_t GetLiteralMode(void) {return literal_mode;}

  void SetCursorPosition(int x, int y);
  CursorRep * GetCursorRep(void) 
    {return ((operation_mode == CURSOR_REVIEW) ? review : edit);}

  void SyncReviewToEdit(void);
  void SyncEditToReview(void);
  void SyncEditToApp(void);

  char * move(const char *location, const char *token);
  // Raw indicates whether filtering will be done.
  char * read(const char *location, const char *token, int raw = 0);

  void Speak(char *speak, int interrupt = 1);

  int GetFilterCount(void) {return filter.GetFilterCount();}
  int GetFilterActive(int pos) {return filter.GetFilterActive(pos);}
  char *GetFilterName(int pos) {return filter.GetFilterName(pos);}
  void SetFilterActive(int pos, int active) {
    filter.SetFilterActive(pos, active);
  }

  int debug(void) {return debug_flag;}
  void debug(int d) {debug_flag = d;}

private:
  FilterDict filter;                  // filter structure
  TextRep *textrep;                   // textrep for this object
  char *delimitString;                // string of delimiting characters
  char *terminalString;               // string of terminating punctuation

  // Locate the character we are searching for in a string of possible
  //   values.  Performance should be improved if more commonly-used values
  //   are stored toward the front of the string.
  int DelimitChar(const char c) {return (strchr(delimitString, c) != NULL);}
  int TerminalChar(const char c) {return (strchr(terminalString, c) != NULL);}

  int EndWord(const char *w);

  // Used to reallocate space for temporary strings used in sentence
  //   and paragraph parsing.
  TextData * ResizeDataPtrArray(TextData *&array, int &current_size,
				 const int block_size);
  void CopyTextAttrArray(TextAttr *dest, TextAttr *src, int start, int len);

  CursorRepData * MoveChar(CursorRep &c, int x = 0, int update = 1) 
    throw (ScreenReaderNotFound);
  CursorRepData * MoveWord(CursorRep &c, int x = 0, int update = 1)
    throw (ScreenReaderNotFound);
  CursorRepData * MoveSentence(CursorRep &c, int x = 0, int update = 1) 
    throw (ScreenReaderNotFound);
  CursorRepData * MoveParagraph(CursorRep &c, int x = 0, int update = 1)
    throw (ScreenReaderNotFound);
  CursorRepData * MovePage(CursorRep &c, int x = 1, int update = 1);
  CursorRepData * MoveLine(CursorRep &c, int x = 0, int update = 1)
    throw (ScreenReaderNotFound);
  
  CursorRepData * GetChar(CursorRep &c, int x = 0);
  CursorRepData * GetWord(CursorRep &c, int x = 0);
  CursorRepData * GetSentence(CursorRep &c, int x = 0);
  CursorRepData * GetParagraph(CursorRep &c, int x = 0);
  CursorRepData * GetPage(CursorRep &c, int x = 0);
  CursorRepData * GetLine(CursorRep &c, int x = 0);
  CursorRep *review, *edit;           // edit, review cursors

  void MoveAppCursor(int x1, int y1, int x2, int y2);

  int SetReviewCursor(int x, int y);
  int SetEditCursor(int x, int y);

  void flush_buffer(void);

  enum cursor_mode_t operation_mode;
  enum follow_mode_t follow_mode;
  enum literal_mode_t literal_mode;   // remove whitespace in non-literal mode

  char incoming_buffer[INCOMING_BUFFER_LEN];
  int incoming_count;
  int incoming_mode;
  char *user_abbre;
  
  int debug_flag;
};

#endif
