// -*- c++ -*-
//
//	Note:  The use of this program shall be restricted to educational,
//	non-commercial use only.  All inquiries regarding licensing this
//	software for commercial use may be addressed to:
//	
//	Georgia Tech Research Corporation
//	Technology Licensing
//	Centennial Research Building - Rm. 275
//	400 Tenth Street, N.W.
//	Atlanta, Georgia  30332-0415
//	
//	*************************************************
//	
//	       Copyright 1996
//	Georgia Tech Research Corporation
//	    All Rights Reserved
//
//
//
//	Code to support parsing and retrieval of rule templates.
//	In addition to being a C++ class definition, this file is
//	also included by the template parser to get prototypes for
//	the routines it needs to call out to.  Thus note the
//	#ifdef cplusplus to keep the C++ gunk from getting seen
//	by the C-code parser.
//
//		keith edwards	25jan95
//
// This code interacts a lot with the Config stuff... actually I'm not too
// happy with it.  I'd originally designed it for speed, and robustness in
// mind.  To this end, I put in explicit function calls to retrieve each
// template attribute.
//
// I probably should have done something more string based, where you
// pass in the name of the attribute you want.  Would have slowed it down
// and made it possible to make spelling errors, but it would also have been
// a lot easier to add new attributes.
//
// As it stands now, you have to jump through hoops to add new attributes...
// Hopefully these will stabilize soon though...
//
//		keith edwards 2feb95
//


#ifndef _TEMPLATE_H_
#define _TEMPLATE_H_

extern "C" {
#ifdef TCL7
#include "mercator-tcl-sol.h"
#else
#include "mercator-tcl.h"
#endif
}
#include "mString.h"
#include "mListVal.h"
#include "mDictValPtr.h"
#include <stdio.h>

class XtObject;
class Client;
class Mercator;
class Template;

#if 0
//
// RTTI codes... sorta hacky.
//
typedef enum { CLASS_TEMPLATE_TYPE = 17,
	       OBJECT_TEMPLATE_TYPE = 42,
	       APP_TEMPLATE_TYPE = 93
} ClassObjectType;
#endif

typedef enum {CLASS_TEMPLATE, OBJECT_TEMPLATE, APP_TEMPLATE} TemplateType;

/////
/////		These are called from the parser.
/////
extern "C" {
void NewObjTemplate(int l, const char *name);
void NewClassTemplate(int l, const char *name);
void NewAppTemplate(int l, const char *name);

void EndTemplate(int l);

void ProcessBooleanField(int l, const char *id, unsigned char b);
void ProcessIntegerField(int l, const char *id, long i);
void ProcessStringField(int l, const char *id, const char *s);
void ProcessStringListField(int l, const char *id);
void ProcessProcField(int l, const char *id, const char *proc);

void AddToStringList(int l, const char *s);
void TmplLexWarning(int l, const char *s);
}

////////////////////////////////////////////////////////////////////////////

#ifdef __cplusplus

#include "Mercator.h"

typedef enum { BOOLEAN, INTEGER, STRING, STRINGLIST } FieldType;
typedef enum { SIMPLE, PROC} FieldRep;

class FieldData {
  friend class Template;
private:
    // Tag is the type of the data that we're initialized to.  It should
    // probably never change.
    FieldType		tag;	       
    // Rep is the type of our representation; either SIMPLE or PROC.
    FieldRep 		rep;

    // Has our value been set.
    unsigned char	initialized;
    union {
	unsigned char	boolean;
	long		integer;
	char		*str;
	mListVal<mString> *stringList;
    };

    char		*proc;
    
    static Tcl_Interp	*interp;

    // Used to pass date out of InvokeProc.
    typedef union {
	unsigned char 	boolean;
	long		integer;
	char		*str;
	mListVal<mString> *stringList;
    } OutputStruct;
    
    void		FreeOldData();
    int	    		InvokeProc(const XtObject& obj, OutputStruct *) const;
    int	    		InvokeProc(const Client& c, OutputStruct *) const;
    
public:
    FieldData(FieldType);
    FieldData(const FieldData& fd);
    virtual ~FieldData();
    
    FieldData&		operator=(const FieldData& fd);
    
    FieldType		Tag() const;
    void		Tag(FieldType t);

    unsigned char	Initialized() const;
    
    // Keep the same tag, but set rep to PROC
    void		SetProc(const char *p);
    
    // Interpreter is maintained across instances.
    static void		SetInterpreter(Tcl_Interp *interp);
    
    unsigned char	Boolean() const;
    unsigned char	Boolean(const XtObject&) const;
    unsigned char	Boolean(const Client&) const;
    void		Boolean(unsigned char b);
    long		Integer() const;
    long		Integer(const XtObject&) const;
    long		Integer(const Client&) const;
    void		Integer(long i);
    const char		*String() const;
    const char		*String(const XtObject&) const;
    const char		*String(const Client&) const;
    void		String(const char *s);
    const mListVal<mString>& StringList() const;
    const mListVal<mString>& StringList(const XtObject&) const;
    const mListVal<mString>& StringList(const Client&) const;
    void		StringList(const mListVal<mString>& sl);

};

////////////////////////////////////////////////////////////////////////////

class Template;
typedef Template *TemplatePtr;

#if 0
class ClassObjectTemplate;
typedef ClassObjectTemplate *ClassObjectTemplatePtr;

class AppTemplate;
typedef AppTemplate *AppTemplatePtr;
#endif

class Template {
    
    friend int Mercator::LoadTemplates();
    friend int WriteOut(ClientData, Tcl_Interp *, int, char *[]);
    
private:
    // Have we found an error while constructing this template?
    int			valid;
    // What is our name?
    char		*name;
    // What kind of template are we?
    TemplateType	type;
    
    virtual int		Set(const char *n, FieldData& data);

protected:
    void		Name(const char *n);
    void		Valid(int v);
    
#if 0
    // Yes, Template knows about its subclasses.  Makes our interface to
    // templates simpler...
#endif

    static mDictValPtr<mString,Template> *classTemplates;
    static mDictValPtr<mString,Template> *objectTemplates;
    static mDictValPtr<mString,Template> *appTemplates;

    static void		EmptyDicts();
    static int 		validDicts;

    // attributes are added dynamicly at run time to this list
    mDictValPtr<mString, FieldData> attributeDict;
    
public:
    Template(const char *n, TemplateType type);
    virtual ~Template();
    
    int			Valid() const;
    const char		*Name() const;
    
    int			InstallBooleanField(const char *id, unsigned char b);
    int			InstallIntegerField(const char *id, long i);
    int			InstallStringField(const char *id, const char *s);
    int			InstallStringListField(const char *id, 
					       const mListVal<mString>& sl);
    int			InstallProcField(const char *id, const char *proc);
    
    //
    // A few statics that govern parsing, loading, and freeing.
    //
    static int		LoadTemplates(const mListVal<mString>& dirs);
    static int		ParseTemplateFile(const char *filename);
    static void		FreeTemplates();
    
    // Used while parsing.
    static Template	*currentTemplate;
    static mListVal<mString> *currentStringList;

    // for writing back out to file
    int 		WriteToFile(FILE *f) const;

    // return an RTTI code.
    virtual TemplateType Type() const;

    // Return the name of our class.
    virtual const char	*ClassName() const;

    // Install the instance in its subclass' dictionary of valid templates.
    virtual void	Install();

    // Used by the interpreter to go from strings to FieldData.
    static const FieldData	*FindByName(const XtObject&, const char *s);
    static const FieldData	*FindByName(const Client&, const char *s);

    static unsigned char 	GetBoolean(const XtObject& o, const char *s);
    static unsigned char 	GetBoolean(const Client& o, const char *s);
    static long 		GetInteger(const XtObject& o, const char *s);
    static long 		GetInteger(const Client& o, const char *s);
    static const char 	       *GetString(const XtObject& o, const char *s);
    static const char 	       *GetString(const Client& o, const char *s);
    static const mListVal<mString>& GetStringList(const XtObject& o,
						  const char *s);
    static const mListVal<mString>& GetStringList(const Client& o,
						  const char *s);
};


inline unsigned char Template::GetBoolean(const XtObject& obj, const char *s)
	      { return (FindByName(obj, s))->Boolean(obj); }
inline unsigned char Template::GetBoolean(const Client& obj, const char *s)
	      { return (FindByName(obj, s))->Boolean(obj); }
inline long Template::GetInteger(const XtObject& obj, const char *s)
	      { return (FindByName(obj, s))->Integer(obj); }
inline long Template::GetInteger(const Client& obj, const char *s)
	      { return (FindByName(obj, s))->Integer(obj); }
inline const char * Template::GetString(const XtObject& obj, const char *s)
	      { return (FindByName(obj, s))->String(obj); }
inline const char * Template::GetString(const Client& obj, const char *s)
	      { return (FindByName(obj, s))->String(obj); }
inline const mListVal<mString>& Template::GetStringList(const XtObject& obj,
							 const char *s)
	      { return (FindByName(obj, s))->StringList(obj); }
inline const mListVal<mString>& Template::GetStringList(const Client& obj,
							 const char *s)
	      { return (FindByName(obj, s))->StringList(obj); }

    
#endif /* __cplusplus */
    
#endif /* _TEMPLATE_H_ */








