//
//	Note:  The use of this program shall be restricted to educational,
//	non-commercial use only.  All inquiries regarding licensing this
//	software for commercial use may be addressed to:
//	
//	Georgia Tech Research Corporation
//	Technology Licensing
//	Centennial Research Building - Rm. 275
//	400 Tenth Street, N.W.
//	Atlanta, Georgia  30332-0415
//	
//	*************************************************
//	
//	       Copyright 1996
//	Georgia Tech Research Corporation
//	    All Rights Reserved
//
//
//
//	keith edwards 10aug95
//

#include <exception.h>

#include "TextRep.h"
#include "Mercator.h"
#include "XServer.h"

#ifndef MIN
#define MIN(a,b) ((a) < (b) ? (a) : (b))
#endif /* MIN */

		  
/////////////////////////////////////////////////////////////////////////
//
// TextAttr stuff
//
GCValue TextAttr::defaultValue;

/////////////////////////////////////////////////////////////////////////
//
// TextRepDebug methods
//
TextRepDebug::TextRepDebug(XServer& x, const TextRep& t) :
    xserver(x),
    textrep(t),
    rows(textrep.Rows()),
    columns(textrep.Columns())
{
    XSetWindowAttributes attr;
    XFontStruct *font;

    font = xserver.LoadQueryFont("9x15");

    fontWidth = font->max_bounds.width;
    fontHeight = font->ascent + font->descent;

    int width = columns * fontWidth;
    int height = rows * fontHeight;
    
    window = xserver.CreateWindow(xserver.GetRootWindow(), 0, 0,
				  width, height, 1,
				  xserver.GetBlackPixel(),
				  xserver.GetWhitePixel());
    
    attr.backing_store = WhenMapped;
    attr.backing_planes = AllPlanes;
    xserver.ChangeWindowAttributes(window, CWBackingStore|CWBackingPlanes,
				   &attr);

    xserver.MapWindow(window);

    textGC = xserver.CreateGC(window, 0, NULL);
    xserver.SetBackground(textGC, xserver.GetWhitePixel());
    xserver.SetForeground(textGC, xserver.GetBlackPixel());
    xserver.SetFont(textGC, font->fid);

    cursorGC = xserver.CreateGC(window, 0, NULL);
    xserver.SetBackground(cursorGC, xserver.GetBlackPixel());
    xserver.SetForeground(cursorGC, xserver.GetBlackPixel());
    xserver.SetFunction(cursorGC, GXxor);

    Update();
}

TextRepDebug::~TextRepDebug()
{
    if (window != None){
	xserver.DestroyWindow(window);
	xserver.Flush();
    }
}

//
// Note that this is currently very expensive.  We clear and redraw everything
// each time.
//
void
TextRepDebug::Update()
{
    int width = columns * fontWidth;
    int height = rows * fontHeight;
    
    // Resize the window if we need to
    if (textrep.Rows() != rows || textrep.Columns() != columns) {

	xserver.ResizeWindow(window, width, height);

	columns = textrep.Columns();
	rows = textrep.Rows();
    }

    xserver.ClearArea(window, 0, 0, width, height);

    for (register int i=0 ; i<textrep.Rows() ; i++) {
	char *line = textrep.GetLine(i);
	xserver.ImageText(window, textGC, 0,
			  (i * fontHeight + fontHeight), line);
    }
    
    xserver.Flush();
}

//
// For now, just to make things work, our expose handler redraws the world.
// (I don't think this is ever called anywhere though...)
//
void
TextRepDebug::Refresh()
{
    Update();
}

/////////////////////////////////////////////////////////////////////////
//
// TextDataRep methods
//
TextData::TextDataRep::TextDataRep(const TextDataRep& other) :
    chars(NULL),
    attrs(NULL),
    length(other.length),
    refCount(1)
{
    chars = new char[length + 1];
    attrs = new TextAttr[length];

    memcpy(chars, other.chars, length);
    // TextAttrs contain structures and may be padded...so we copy individually
    //    memcpy(attrs, other.attrs, (length * sizeof(TextAttr)));
    for (register int i=0 ; i<length ; i++) {
	attrs[i] = other.attrs[i];
    }

    chars[length] = '\0';
}

/////////////////////////////////////////////////////////////////////////
//
// TextData methods
//
TextData::TextData(char *chars, TextAttr *attrs, int length) :
    rep(new TextDataRep)
{
    rep->chars = chars;
    rep->attrs = attrs;
    rep->length = length;
}

TextData&
TextData::operator+=(const TextData& other)
{
    TextDataRep *newrep = new TextDataRep();

    newrep->length = rep->length + other.Length();
    newrep->chars = new char[rep->length + other.Length() + 1];
    newrep->attrs = new TextAttr[rep->length + other.Length()];

    // copy in old data...
    memcpy(newrep->chars, rep->chars, rep->length);
    memcpy(newrep->attrs, rep->attrs, (rep->length * sizeof(TextAttr)));

    // and append new data...
    memcpy(newrep->chars + rep->length, other.Chars(), other.Length());
    memcpy(newrep->attrs + (rep->length * sizeof(TextAttr)), other.Attrs(),
	   (other.Length() * sizeof(TextAttr)));

    // append NULL onto char array
    newrep->chars[rep->length + other.Length()] = '\0';
    
    if (--rep->refCount <= 0)
	delete rep;

    rep = newrep;
    
    return *this;
}

//
// Methods which return pointers to internal data causes us to make a copy
// of our TextDataRep since the caller may muck with the data.  We only do
// the copy if our ref count is greater than one though (which indicates that
// someone else has a handle to our rep).
//
TextData::operator char *()
{
    if (rep->refCount > 1) {
	TextDataRep *newrep = new TextDataRep(*rep);

	if (--rep->refCount <= 0)
	    delete rep;

	rep = newrep;
    }
    return rep->chars;
}

TextData::operator TextAttr*()
{
    if (rep->refCount > 1) {
	TextDataRep *newrep = new TextDataRep(*rep);

	if (--rep->refCount <= 0)
	    delete rep;

	rep = newrep;
    }
    return rep->attrs;
}

char *
TextData::Chars()
{
    if (rep->refCount > 1) {
	TextDataRep *newrep = new TextDataRep(*rep);

	if (--rep->refCount <= 0)
	    delete rep;

	rep = newrep;
    }
    return rep->chars;
}

TextAttr *
TextData::Attrs()
{
    if (rep->refCount > 1) {
	TextDataRep *newrep = new TextDataRep(*rep);

	if (--rep->refCount <= 0)
	    delete rep;

	rep = newrep;
    }
    return rep->attrs;
}

void
TextData::Chars(char *newchars)
{
    if (rep->refCount > 1) {
	TextDataRep *newrep = new TextDataRep(*rep);
	
	if (--rep->refCount <= 0)
	    delete rep;

	rep = newrep;
    }
    rep->chars = newchars;
}

void
TextData::Attrs(TextAttr *newattrs)
{
    if (rep->refCount > 1) {
	TextDataRep *newrep = new TextDataRep(*rep);

	if (--rep->refCount <= 0)
	    delete rep;

	rep = newrep;
    }
    rep->attrs = newattrs;
}

void
TextData::Length(int l)
{
    if (rep->refCount > 1) {
	TextDataRep *newrep = new TextDataRep(*rep);

	if (--rep->refCount <= 0)
	    delete rep;

	rep = newrep;
    }
    rep->length = l;
}

char *
TextData::Chars(int index) {
  if ((index < 0) || (index >= rep->length))
    return NULL;
  else
    return &rep->chars[index];
}

TextAttr *
TextData::Attrs(int index) {
  if ((index < 0) || (index >= rep->length))
    return NULL;
  else
    return &rep->attrs[index];
}

// We're going to make a dangerous-ass assumption and hope that the
//   length field is properly set for both the ResizeChars and 
//   ResizeAttrs functions.
void
TextData::ResizeChars(int newlen, int pos) {
  char *ptr;
  int diff = newlen - rep->length;
  if ((newlen < 0) || (pos >= rep->length))
    return;
  ptr = new char[newlen+1];
  // If increasing size, copy characters up to the break position.  Otherwise,
  //    copy all characters from shorter span.  Fill the empty span with
  //    spaces.
  strncpy(ptr, (const char *)*this, (diff > 0) ? pos : newlen);
  if (diff > 0) {
    memset((void *) ptr, (int)' ', diff);
    strncpy(ptr + pos + diff, (const char *)*(this + pos), rep->length - pos);
  }
  Chars(ptr);
}

void
TextData::ResizeAttrs(int newlen, int pos) {
    //  TextAttr *ptr, *attrs = (const TextAttr *)*this;
  TextAttr *ptr;
  const TextAttr *attrs = Attrs();
  int diff = newlen - rep->length, i;
  if ((newlen < 0) || (pos >= rep->length))
    return;
  ptr = new TextAttr[newlen];
  // If increasing size, copy attributes up to break position.  Otherwise,
  //    copy all characters from the shorter span.  Fill the empty span with
  //    the characters from the original break position.
  for (i = 0; i < ((diff > 0) ? pos : newlen); i++)
    ptr[i] = attrs[i];
  if (diff > 0) {
    for (i = 0; i < diff; i++)
      ptr[pos + i] = attrs[pos];
    for (i = 0; i < rep->length - pos; i++)
      ptr[i + pos + diff] = attrs[i + pos];
  }
  Attrs(ptr);
}

TextData
TextData::Segment(int start, int len) const
{
    if (start >= Length())
	return *this;
    
    int realLength = MIN(start + len, Length());
    char *newChars = new char[realLength + 1];
    TextAttr *newAttrs = new TextAttr[realLength];

    memcpy(newChars, rep->chars+start, realLength);
    memcpy(newAttrs, rep->attrs+(start*sizeof(TextAttr)),
	   realLength*sizeof(TextAttr));
    newChars[realLength] = '\0';

    return TextData(newChars, newAttrs, realLength);
}

//
// Implementation subject to change...
//
TextAttr&
TextData::Attr(int pos) throw(xmsg)
{
    if (pos >= Length())
	throw xmsg("Position out-of-bounds");

    return rep->attrs[pos];
}

const TextAttr&
TextData::Attr(int pos) const throw(xmsg)
{
    if (pos >= Length())
	throw xmsg("Position out-of-bounds");

    return rep->attrs[pos];
}

/////////////////////////////////////////////////////////////////////////
//
// TextRep methods
//
void
TextRep::Debug(int d)
{
    if (d && !debug) {
	debug = new TextRepDebug(*mercator->GetXServer(), *this);
    } else {
	delete debug;
	debug = NULL;
    }
}

/////////////////////////////////////////////////////////////////////////
