// -*-C++-*-
//
//	Note:  The use of this program shall be restricted to educational,
//	non-commercial use only.  All inquiries regarding licensing this
//	software for commercial use may be addressed to:
//	
//	Georgia Tech Research Corporation
//	Technology Licensing
//	Centennial Research Building - Rm. 275
//	400 Tenth Street, N.W.
//	Atlanta, Georgia  30332-0415
//	
//	*************************************************
//	
//	       Copyright 1996
//	Georgia Tech Research Corporation
//	    All Rights Reserved
//
//
//
// Revised TextRep stuff.  This file provides the following classes:
//
//	- TextRep - base class for ProtoTextRep and XmTextTextRep.
//	- TextData - a representation of a chunk of textual data that
//	  also contains attribute information.
//	- TextAttr - represent text attribute information.
//	- TextRepDebug - a visual output class for debugging textrep
//	  contents.
//
// TextRep enforces an output API for retrieving text (used by ScreenReader),
// but subclasses provide their own input APIs (for keeping their model
// up-to-date).
//
//	keith edwards	10aug95
//

#ifndef _TEXTREP_H_
#define _TEXTREP_H_

#include <stdio.h>		// for FILE
#include <X11/Xlib.h>
#include "GCValue.h"

class XServer;
class XtObject;
class CursorRep;
class TextAttr;
class TextData;
class TextRepDebug;
class TextRep;
class xmsg;

/////////////////////////////////////////////////////////////////////////

class TextAttr {
private:
  //GCValue&		gcv;
    GCValue		gcv;

    static GCValue	defaultValue;

public:
    TextAttr();
    TextAttr(const TextAttr& other);
    TextAttr(GCValue& gcv);

    TextAttr& operator=(const TextAttr& other);
    int operator==(const TextAttr& other);
    int operator!=(const TextAttr& other);

    //    GCValue&		GetGCValue() const;
    const GCValue&		GetGCValue() const;
    void		SetGCValue(GCValue& gcv);
};

inline TextAttr::TextAttr() : gcv(TextAttr::defaultValue) { }
inline TextAttr::TextAttr(const TextAttr& other) : gcv(other.gcv) { }
inline TextAttr::TextAttr(GCValue& v) : gcv(v) { }
//inline GCValue& TextAttr::GetGCValue() const { return gcv; }
inline const GCValue& TextAttr::GetGCValue() const { return gcv; }
inline void TextAttr::SetGCValue(GCValue& v) { gcv = v; }
inline TextAttr& TextAttr::operator=(const TextAttr& a)
	{ gcv = a.gcv; return *this; }
inline int TextAttr::operator==(const TextAttr& other) 
        { return (gcv == other.gcv); }
inline int TextAttr::operator!=(const TextAttr& other) 
        { return (gcv != other.gcv); }

/////////////////////////////////////////////////////////////////////////
//
// The TextData class stores a segment of textual data, along with its
// attributes, in a time- and space-efficient manner.  The TextData class
// is a wrapper around a pointer to a TextDataRep instance which is
// reference counted.  Copying and assignment between TextDatas merely
// copies the pointer and increments the reference count.  The ref-counted
// representation class is only freed when its reference count drops to zero.
//
// TextData implements copy-on-write semantics.  The const versions of
// the access functions below return pointers to internal data.  The non-const
// versions cause a copy of the representation (to ensure that no one else
// that points to the representation gets written to) and then the data is
// returned.
//
// TextDatas are a single word in size, so they can be passed on the stack
// without references or pointers.
//
// A few weird semantics and a caveat:
//	- If you set member data via the Chars() or Attrs() methods,
//	  the class owns this memory and will delete[] it when done.
//	- Setting member data via Chars() and Attrs() and retrieving
//	  data via the non-const access methods only forces a copy
//	  of the representation when the reference count is not one (this
//	  prevents an extra, bogus copy if the data is placed in a TextData
//	  piecemeal or if no one else refers to our data).
//	- The class could be a little smarter.  If we return a non-const
//	  pointer to member data, the representation should be flagged
//	  as non-sharable to ensure that if the TextData is copied noone
//	  will get a rep that may change out from under it because of the
//	  previous pass-out of data.
//
// NOTE that this class only supports 8-bit data.
//
class TextData {
private:
    class TextDataRep {
    public:
	char		*chars;
	TextAttr	*attrs;
	int		length;
	int		refCount;

	TextDataRep();				// empty, refcount = 1
	TextDataRep(const TextDataRep& other);	// copies data
	~TextDataRep();				// free data
    };

    TextDataRep		*rep;

public:
    TextData();
    TextData(char *, TextAttr *, int len);
    TextData(const TextData& other);
    virtual ~TextData();
    
    TextData& operator=(const TextData& other);		// cheap
    TextData& operator+=(const TextData& other);	// forces copy

    TextAttr&		Attr(int pos) throw(xmsg);
    const TextAttr&	Attr(int pos) const throw(xmsg);

    operator char *();					// forces copy
    operator const char *() const;			// cheap
    operator TextAttr*();				// forces copy
    operator const TextAttr*() const;			// cheap

    char		*Chars();			// forces copy
    const char		*Chars() const;			// cheap
    TextAttr		*Attrs();			// forces copy
    const TextAttr	*Attrs() const;			// cheap
    int			Length() const;			// cheap
    
    // These own the memory you pass into them; they'll free it themselves.
    void		Chars(char *chars);		// may force copy
    void		Attrs(TextAttr *attrs);		// may force copy
    void		Length(int l);			// may force copy

    char 		*Chars(int index);
    TextAttr 		*Attrs(int index);

    void 		ResizeChars(int newlen, int pos);
    void 		ResizeAttrs(int newlen, int pos);
    
    TextData		Segment(int start, int length) const;
};

inline TextData::TextDataRep::TextDataRep() :
    chars(NULL), attrs(NULL), length(0), refCount(1) { }
inline TextData::TextDataRep::~TextDataRep()
	{ delete [] chars; delete [] attrs; }

inline TextData::TextData() : rep(new TextDataRep) { }
inline TextData::TextData(const TextData& other) : rep(other.rep)
	{ (rep->refCount)++; }
inline TextData::~TextData()
	{ if (--(rep->refCount) <= 0) delete rep; }

inline TextData& TextData::operator=(const TextData& other)
	{
	    if (--(rep->refCount) <= 0)
		delete rep;
	    rep = other.rep ;
	    (rep->refCount)++;
	    return *this;
	}

inline TextData::operator const char*() const { return rep->chars; }
inline TextData::operator const TextAttr*() const { return rep->attrs; }
inline const char *TextData::Chars() const { return rep->chars;}
inline const TextAttr *TextData::Attrs() const { return rep->attrs; }
inline int TextData::Length() const { return rep->length; }

#if 0
inline TextData::operator char *() { return rep->chars; } 
inline TextData::operator TextAttr*() { return rep->attrs; } 
inline char * TextData::Chars() { return rep->chars; } 
inline TextAttr * TextData::Attrs() { return rep->attrs; } 
inline void TextData::Chars(char *newchars) { rep->chars = newchars; } 
inline void TextData::Attrs(TextAttr *newattrs) { rep->attrs = newattrs; } 
inline void TextData::Length(int l) { rep->length = l; }
#endif

/////////////////////////////////////////////////////////////////////////

class TextRepDebug {
private:
    TextRepDebug();					// disallowed
    TextRepDebug(const TextRepDebug&); 			// disallowed
    TextRepDebug& operator=(const TextRepDebug&);	// disallowed

    const TextRep&	textrep;
    XServer&		xserver;
    GC			textGC, cursorGC;
    int			rows, columns;			// cache these
    Window		window;
    int			fontWidth, fontHeight;
    
public:
    TextRepDebug(XServer& xserver, const TextRep& textrep);
    virtual ~TextRepDebug();

    virtual void	Update();	// query textrep and update view
    virtual void	Refresh();	// process exposures
};

/////////////////////////////////////////////////////////////////////////

class TextRep {
private:
    TextRepDebug	*debug;
    
protected:
    TextRepDebug	*GetTextRepDebug() const;
    XtObject&		xobj;
    
public:
    TextRep(XtObject& o);
    virtual ~TextRep();
    //
    // Get and set debug state.
    //
    int			Debug() const;
    void		Debug(int d);
    //
    // Dump the contents of the textrep to the specified descriptor.
    //
    virtual void	PrintContents(FILE *fp) const = 0;
    //
    // Get the size of the representation, by pixels or characters.
    //
    virtual int		Height() const = 0;
    virtual int		Rows() const = 0;
    virtual int		Width() const = 0;
    virtual int		Columns() const = 0;
    //
    // Called whenever the underlying object's size has been updated.
    virtual void	RefreshSize() = 0;
    //
    // The is the output API - return TextData.
    //
    virtual TextData    GetLine(int line) const throw(xmsg) = 0;
    virtual TextData   *GetData() const = 0;
    virtual TextData	GetSegment(int line, int pos, int len) const = 0;
    //
    // Return the X,Y pixel coordinates that correspond to the specified
    // line and character position within that line.  Returns 0 if out of bounds.
    //
    virtual int		GetPixelCoords(int line, int pos,
				       int& x, int& y) const = 0;
    //
    // Get the edit cursor position.
    //
    virtual void	GetEditCursor(int& x, int& y) throw(xmsg) = 0;
    const XtObject& obj(void) { return xobj; }
    virtual int		SetAppCursor(int row, int col) = 0;
    virtual void	GetAppCursor(int& row, int& col) = 0;
};

inline TextRep::TextRep(XtObject& o) : xobj(o), debug(NULL) { }
inline TextRep::~TextRep() { delete debug; }
inline TextRepDebug *TextRep::GetTextRepDebug() const { return debug; }
inline int TextRep::Debug() const { return (debug != NULL); }


#endif /* _TEXTREP_H_ */
