//  -*-Mode: C++-*-
//
//
//
//	Note:  The use of this program shall be restricted to educational,
//	non-commercial use only.  All inquiries regarding licensing this
//	software for commercial use may be addressed to:
//	
//	Georgia Tech Research Corporation
//	Technology Licensing
//	Centennial Research Building - Rm. 275
//	400 Tenth Street, N.W.
//	Atlanta, Georgia  30332-0415
//	
//	*************************************************
//	
//	       Copyright 1996
//	Georgia Tech Research Corporation
//	    All Rights Reserved
//
//
//
//

#include "Truetalk.h"

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/file.h>
#include <sys/wait.h>
 
#include <netinet/in.h>
 
#include <errno.h>
#include <netdb.h>
#include <signal.h>
#include <sgtty.h>
#include <fcntl.h>

#include <ctype.h>
#include <sysent.h>
#include <unistd.h>
#include <stdio.h>
#include <strings.h>
#include <stdlib.h>
#include <stream.h>

#include "Error.h"
#include "Util.h"
#include "Mercator.h"
#include "Config.h"
#include "Template.h"

LOADABLE_CLASS_DEFN(Truetalk);

#define is_member(c, s)  strchr((s), (c))

#define SPECIAL_CHARS   "!@#$%^&*()/[]{}"

// These are some Truetalk voice settings. I am not too happy with these
// settings but I don't know enough about speech synthesizers to make these
// voices more apparent.
#define DEF_M "\\!uSm\\!uF65\\!uB65\\!gq0.55\\!ga0.60\\!gt0.35\\!gd0.05"
#define DEF_F "\\!uSf\\!uF65\\!uB65\\!gq0.55\\!ga0.60\\!gt0.35\\!gd0.05"
#define DEEP_M "\\!uSm\\!uF45\\!uB50\\!gq0.55\\!ga0.60\\!gt0.35\\!gd0.05"
#define DEEP_F "\\!uSf\\!uF55\\!uB55\\!gq0.55\\!ga0.60\\!gt0.35\\!gd0.05"
#define OLD_M "\\!uSm\\!uF65\\!uB65\\!gq0.55\\!ga0.60\\!gt0.35\\!gd0.05"
#define CHILD "\\!uSf\\!uF220\\!uB220\\!gq0.55\\!ga0.60\\!gt0.35\\!gd0.35"
#define LIGHT_F "\\!uSf\\!uF80\\!uB80\\!gq0.9\\!ga0.9\\!gt0.95\\!gd0.05"
#define WHIS_F "\\!uSf\\!uF80\\!uB80\\!gq0.9\\!ga0.9\\!gt0.95\\!gd0.05"
#define WHIS_M "\\!uSm\\!uF80\\!uB80\\!gq0.9\\!ga0.9\\!gt0.95\\!gd0.05"

// use /dev/sound/1 for now
#define SOUND1

extern Mercator *mercator;

Speech::Capabilities 	Truetalk::capabilities;
int			Truetalk::capabilitiesInitialized = 0;


//////////////////////////////////////////////////////////////////////
//
// Truetalk now assumes that there is a secondary audio device in
// /dev/sound/1. In the future truetalk should use an audio server
// to play its speech output.
//
//////////////////////////////////////////////////////////////////////
Truetalk::Truetalk(const char *hostname) : Speech("Truetalk"){

    // initialization
    _server = NULL;
    _client = NULL;
    _port[0] = 0;
    _port[1] = 0;
    _encoding = ENCODING_MULAW;
    _freq = 8000;
#ifdef SOUND1
    _who = OUTPUT_CLIENT;
    _where = OUTPUT_FILE;    // a hack to get truetalk to work with netaudio
#else    
    _who = OUTPUT_CLIENT;
    _where = OUTPUT_AUDIO;
#endif

    Truetalk::StartServer();
    Truetalk::ConnectServer(hostname);

    if (!Truetalk::capabilitiesInitialized++) {
	Truetalk::capabilities.languages.append("English");

	Truetalk::capabilities.voices.append("Perfect Paul");
	Truetalk::capabilities.voices.append("Beautiful Betty");
	Truetalk::capabilities.voices.append("Huge Harry");
	Truetalk::capabilities.voices.append("Frail Frank");
	Truetalk::capabilities.voices.append("Kit the Kid");
	Truetalk::capabilities.voices.append("Rough Rita");
	Truetalk::capabilities.voices.append("Uppity Ursula");
	Truetalk::capabilities.voices.append("Doctor Dennis");
	Truetalk::capabilities.voices.append("Whispering Wendy");
	Truetalk::capabilities.voices.append("Variable Val");

	Truetalk::capabilities.minRate = 85;
	Truetalk::capabilities.maxRate = 355;

	Truetalk::capabilities.gainSupported = 0;
    }
}

Truetalk::~Truetalk(){
    
    Error(EL_DEBUG, "Truetalk::~Truetalk", "TRUETALK BEING DESTROYED!\n", NULL);

    TT_Close(_client);
#ifdef SOUND1
    fclose(_outArg);
#endif    
}

int
Truetalk::ConnectServer(const char* truetalkhost)
{
    FieldData *maxTriesData = Config::GetDefaultByName("truetalkTimeout");
    int maxTries = 5, res;
    
    if (maxTriesData && (maxTriesData->Tag() == INTEGER))
	maxTries = maxTriesData->Integer();
    
    Error(EL_DEBUG, "Truetalk::ConnectServer",
	  "Connecting to Truetalk server!", NULL);

    if (truetalkhost == NULL)
	if ((truetalkhost = getenv( "TRUETALKHOST" )) == NULL)
	    truetalkhost = "localhost";
    
    if(OpenTruetalk(truetalkhost, maxTries) == -1) {
	Error(EL_ERROR, "Truetalk::ConnectServer", "server connection",
	      " timed out", NULL);
	return -1;
    }
    
    Error(EL_DEBUG, "Truetalk::ConnectServer",
	  "connection to server successful", NULL);

    // Initialize Truetalk
    SetSpeechRate(180);
    SetVoice("Perfect Paul");
    
    return 0;
}

//
// return -1 if server was not started, 0 otherwise. (Will assume that
// if TRUETALKHOST environment variable is only set if the user wants to
// use a server on another machine. So essentially setting TRUETALKHOST
// to anything will prevent a server from being started.
//
int
Truetalk::StartServer()
{
    const char *host;

    // first check if an env. var. is set. I'm assuming that DECTALKHOST
    // is only set if the user wants to use a remote server
    if((host = getenv("TRUETALKHOST")) == NULL) {
	if(mercator->ExecList("trueServer")) {
	    Error(EL_DEBUG, "Truetalk::StartServer",
		  "truetalkServer not found.", NULL);
	    return -1;
	}
	return 0;
    }
    else
	Error(EL_DEBUG, "Truetalk::StartServer", "Server has been started on ",
	      host, NULL);
    return 0; // assuming the server has been started on another host
}

const Speech::Capabilities&
Truetalk::GetCapabilities() const
{
    return Truetalk::capabilities;
}

int Truetalk::Notify(const char *type){
    
    if (_client != NULL) {
	// Do whatever this is supposed to do!?
    } else {
	Speech::Notify(type);
    }
    
    return 0;
}

int
Truetalk::Speak(const char *string){ 
    
    char buffer[2050];
    char *s;
    short toDel = 1;

    Error(EL_DEBUG, "Truetalk::Speak", "|", string, "|", NULL);

    if (_client != NULL){

	// Special case for character mode. This is a hack for now.
	if(strlen(string) == 1) {
	    switch (string[0]) {
	    case ' ':
		s = Strdup("space");
		break;
	    case '\t':
		s = Strdup("tab");
		break;
	    case '\n':
		s = Strdup("new line");
		break;
	    case '\"':
		s = Strdup("quote");
		break;
	    case '."':
		s = Strdup("period");
		break;
	    case ',"':
		s = Strdup("comma");
		break;
	    case ';':
		s = Strdup("semi colon");
		break;
	    case ':':
		s = Strdup("colon");
		break;
	    case '\'':
	    case '`':
		s = Strdup("apostrophe");
		break;
	    case '?':
		s = Strdup("question mark");
		break;
	    case '!':
		s = Strdup("exclamation mark");
		break;
	    case '@':
		s = Strdup("at");
		break;
	    case '#':
		s = Strdup("pound");
		break;
	    case '^':
		s = Strdup("carret");
		break;
	    case '*':
		s = Strdup("asterisk");
		break;
	    case '(':
		s = Strdup("left paren");
		break;
	    case ')':
		s = Strdup("right paren");
		break;
	    case '[':
	    case '{':
		s = Strdup("left bracket");
		break;
	    case ']':
	    case '}':
		s = Strdup("right bracket");
		break;
	    case '>':
		s = Strdup("greater than");
		break;
	    case '<':
		s = Strdup("less than");
		break;
	    case '_':
		s = Strdup("underscore");
		break;
	    case '\\':
		s = Strdup("back slash");
		break;
	    case '|':
		s = Strdup("pipe");
		break;
	    case '~':
		s = Strdup("tilda");
		break;
	    default:
		s = (char *) string;
		toDel = 0;  // do not delete s
	    }
	    sprintf(buffer, "%s", s);
	
	    if(toDel) delete [] s;
	}
	else
	    sprintf(buffer, "%s", string);

	int ret = TT_SendText(_client, buffer, (unsigned)strlen(buffer));
	if (ret == -1)
	    Error(EL_DEBUG, "Truetalk::Speak", "TT_SendText FAILED", NULL);
	else{
	    // It may not be necessary to sync truetalk if its output is
	    // set to an audio device. Since we're sending to output to an
	    // audio device via a file stream, we have to wait until the
	    // truetalk server finished sending all its output before
	    // we can flush the file stream, or we will only get whatever
	    // the server has produced when we call fflush. 
#ifdef SOUND1
	    SyncSpeech(ret);
#endif	    
	}
    }
    else
	Speech::Speak (string);

    return 0;
}


// This function stips out unix shell special chars. before sending a
// string to the server.
int Truetalk::SpeakUnix(const char *string){
    
    Error(EL_DEBUG, "Truetalk::SpeakUnix", "|", string, "|", NULL);

    if (_client != NULL) {
	int size = strlen(string);
	int i = 0, o = 0;
	char buffer[1024];
   
	while ((string[i] == ' ') && (i < size)) i++;
   
	for (o = 0; i < size; i++) { 
	    if (iscntrl(string[i]))
		continue;
	    if (isalnum(string[i])) { 
		buffer[o++] = string[i];
		continue;
            }
	    if (string[i] == ' ') { 
		buffer[o++] = string[i];
		continue;
            }
	    if (is_member(string[i], SPECIAL_CHARS)) { 
		buffer[o++] = ' ';
		buffer[o++] = string[i];
		buffer[o++] = ' ';
            } 
        }
 
	if (o == 0) 
	    return 0;
	o--;
	
	while (o && (buffer[o] == ' '))
	    o--;
	
	o++;
	
	Error(EL_DEBUG, "Truetalk::SpeakUnix", "|", buffer, "|", NULL);

	int ret = TT_SendText(_client, buffer, (unsigned)strlen(buffer));
	if (ret == -1)
	    Error(EL_DEBUG, "Truetalk::Speak", "TT_SendText FAILED", NULL);
	else{
	    SyncSpeech(ret);  // see comments in ::Speak
	}
    } else {
	Speech::SpeakUnix(string);
    }
}

//
// Truetalk expects rate as a real. This method expects rate as an integer
// in the range of 80 to 355, in words per minute. I am converting words
// per minute to a real, with respect to the default speech rate, 1.0
// (approximately 190 words per minute).
//
int Truetalk::SetSpeechRate(int rate){
    
    float tt_rate;
    
    tt_rate = 1.0 + (190.0 - (float) rate) / 190.0;
    
    Error(EL_DEBUG, "Truetalk::SetSpeechRate", "setting speech rate", NULL);

    if (_client != NULL) {
	char buffer[256];
	sprintf(buffer, "\\!R%f", tt_rate);
	Speak((const char *) buffer);
    } else {
	Speech::SetSpeechRate(rate);
    }
    return 0;
}

int Truetalk::StopSpeaking(){
    
    Error(EL_DEBUG, "Truetalk::StopSpeaking", NULL);
    
    if (_client != NULL) {
	int ret = TT_StopOutput(_client, Graceful, Word);
	if (ret == NULL)
	    Error(EL_ERROR, "Truetalk::StopSpeaking",
		  "TT_StopOutput FAILED", NULL);
    } else {
	Speech::StopSpeaking();
    }

    return 0;
}


// Note these parameters are different than Dectalk's. They can be changed
// to follow Dectalk's parameters if necessary.
//
// "Perfect Paul": standard male voice
// "Beautiful Betty": standard female voice
// "Huge Harry": deep male voice
// "Frail Frank": older male voice
// "Kit the Kid": child's voice
// "Rough Rita": deep female voice
// "Uppity Ursula": light female voice
// "Doctor Dennis": whispery male voice
// "Whispery Wendy": whispery female voice
//
int Truetalk::SetVoice (const char *voice_name){

    char buffer[256], cmd;

    if (!strcmp(voice_name, "Perfect Paul"))
	cmd = 'p';
    else if (!strcmp(voice_name, "Beautiful Betty"))
	cmd = 'b';
    else if (!strcmp(voice_name, "Huge Harry"))
	cmd = 'h';
    else if (!strcmp(voice_name, "Frail Frank"))
	cmd = 'f';
    else if (!strcmp(voice_name, "Kit the Kid"))
	cmd = 'k';
    else if (!strcmp(voice_name, "Rough Rita"))
	cmd = 'r';
    else if (!strcmp(voice_name, "Uppity Ursula"))
	cmd = 'u';
    else if (!strcmp(voice_name, "Doctor Dennis"))
	cmd = 'd';
    else if (!strcmp(voice_name, "Whispering Wendy"))
	cmd = 'w';
    else {
	Error(EL_ERROR, "Truetalk::SetVoice", "Invalid voice name \"",
	      voice_name, "\"", NULL);
	return -1;
    }
    
    switch (cmd){
    case 'p':
	strcpy(buffer, DEF_M);
	break;
    case 'b':
	strcpy(buffer, DEF_F);
	break;
    case 'h':
	strcpy(buffer, DEEP_M);
	break;
    case 'f':
	strcpy(buffer, OLD_M);
	break;
    case 'k':
	strcpy(buffer, CHILD);
	break;
    case 'r':
	strcpy(buffer, DEEP_F);
	break;
    case 'u':
	strcpy(buffer, LIGHT_F);
	break;
    case 'd':
	strcpy(buffer, WHIS_M);
	break;
    case 'w':
	strcpy(buffer, WHIS_F);
	break;
    }

    Speak((const char *) buffer);
    return 0;
}

int
Truetalk::SetLanguage(const char *lang)
{
    mString dummy;
    
    if (!Truetalk::capabilities.languages.search(lang, dummy)) {
	Error(EL_ERROR, "Truetalk::SetLanguage", "Unsupported language: ",
	      lang, NULL);
    }

    // Truetalk actually only supports english, so don't do anything.
    return 0;
}

int
Truetalk::SetGain(int)
{
    Error(EL_ERROR, "Truetalk::SetGain",
	  "Gain control not supported by Truetalk directly", NULL);
    return 0;
}

int
Truetalk::OpenTruetalk(const char *hostname, int timeout)
{
    char *truetalkhost=NULL;
    struct hostent *host;
    int i=0;
    
    host = gethostbyname(hostname);

    if (host == NULL)  {
	Error(EL_DEBUG, "Truetalk::Opentalk",
	      "gethostbyname failed for ", hostname, ": ", Perror(), NULL);
    }
    
    EinitTtbase(NULL);   // initialize truetalk api module

    // first try to bind server
    while (i<timeout) {
	if(!(_server = TT_BindServer(host->h_name, _port))) {
	    Error(EL_DEBUG, "Truetalk::Truetalk", "CANNOT GET SERVER HANDLE",
		  NULL);

	    i++;
	    sleep(1);
	    continue;
	}
    }

    if(_server == NULL) {
	Error(EL_ERROR, "Truetalk::OpenTruetalk", "cannot bind server", NULL);
	return -1;
    }

    // Open the server. It should be up once we bound the server
    if(!(_client = TT_Open(_server))) {
	Error(EL_ERROR, "Truetalk::OpenTruetalk", "cannot open server", NULL);
	return -1;
    }

#ifdef SOUND1    
    // direct output to /dev/sound/1, the batphone audio device
    if ((_outArg = fopen("/dev/sound/1", "a")) == NULL) {
	Error(EL_DEBUG, "Truetalk::Truetalk", "CANNOT OPEN /dev/sound/1", NULL);
	return -1;
    }
    
    if (TT_SetServAudio(_client, SOUND_SPEAKER, 150) == 0) {
	Error(EL_DEBUG, "Truetalk::Truetalk", "TT_SetServAudio FAILED", NULL);
	return -1;
    }
    
    if (TT_SetOutput(_client, _who, _where, _encoding, _freq, _outArg) == 0) {
	Error(EL_DEBUG, "Truetalk::Truetalk", "TT_SetOutput FAILED", NULL);
	return -1;
    }	
#else
    // direct output to /dev/audio
    if (_who == OUTPUT_SERVER){
	if (TT_SetServAudio(_client, SOUND_SPEAKER, 150) == 0) {
	    Error(EL_DEBUG, "Truetalk::Truetalk", "TT_SetServAudio FAILED",
		  NULL);
	    return -1;
	}	    
    }
    else{
	if (TT_SetClientAudio(SOUND_SPEAKER, 150) == 0) {
	    Error(EL_DEBUG, "Truetalk::Truetalk", "TT_SetClientAudio FAILED",
		  NULL);
	    return -1;
	}
    }
    
    if (TT_SetOutput(_client, _who, _where, _encoding, _freq, NULL) == 0) {
	Error(EL_DEBUG, "Truetalk::Truetalk", "TT_SetOutput FAILED", NULL);
	return -1;
    }
#endif

    return 0;
}


int Truetalk::SyncSpeech(unsigned long ret){

    
    if (_client == NULL) {
	Error(EL_ERROR, "Truetalk::SyncSpeech",
	      "No connection to server", NULL);
	return 0;
    }

    TT_Sync(_client, ret);

#ifdef SOUND1
    if (fflush(_outArg) != 0) 
	Error(EL_ERROR, "Truetalk::SyncSpeech",
	      "Error flushing output: ", Perror(), NULL);
#endif
    
    return 0;
}

//
// This method should be called just before mercator kills all its
// subprocesses
//
void
Truetalk::ShutDown()
{
    Error(EL_DEBUG, "Truetalk::ShutDown", "Shutting down Truetalk object",
	  NULL);

    // just need to prevent Truetalk object from trying to contact the
    // server
    _client = NULL;
}
