//
//
//
//	Note:  The use of this program shall be restricted to educational,
//	non-commercial use only.  All inquiries regarding licensing this
//	software for commercial use may be addressed to:
//	
//	Georgia Tech Research Corporation
//	Technology Licensing
//	Centennial Research Building - Rm. 275
//	400 Tenth Street, N.W.
//	Atlanta, Georgia  30332-0415
//	
//	*************************************************
//	
//	       Copyright 1996
//	Georgia Tech Research Corporation
//	    All Rights Reserved
//
//
//
//
#include <rw/tphdict.h>
#include "mDictValPtr.h"

// $Author: selbie $
// $Source: /net/hm10/keith/proj/mercator/6.0/RCS/mDictValPtr.cc,v $
// $Revision: 1.6 $
// $Date: 1995/08/03 22:32:39 $

//
// $Log: mDictValPtr.cc,v $
// Revision 1.6  1995/08/03  22:32:39  selbie
// added the "kill_all" method so that dictionaries can free
// dynamicly allocated entries
//
// Revision 1.5  1995/07/28  21:21:22  spine
// #include <rw/tphdict.h>
//
// Revision 1.4  1995/07/25  22:05:34  keith
// Cleaned up some usage of strings, dicts, and lists.  Ran with -pta to
// fix some previously unseen bugs in the parameterized classes.
//
// Revision 1.3  1995/07/25  20:15:08  keith
// Cleaned up the code a little, fixed a memory leak in iter::current().
//
// Revision 1.2  1995/07/25  17:24:11  keith
// Added def of "nil", fixed but in del_item().
//
// Revision 1.1  1995/06/05  22:24:57  iansmith
// Initial revision
//
//

//
// mDictValPtr Implementation
//

//
// The BIG TRICK:
// This dictionary is implemented with a rouge wave pointer dictionary.
// The trick we use is to make OUR OWN copies of keys as they come
// in. Thus, we know that the we can delete the copies when they are
// removed.

//
// constructor... build the implementation and make it empty
//
template <class K, class V>
mDictValPtr<K, V>::mDictValPtr(unsigned (*hashKey)(const K&)) :
    impl(new RWTPtrHashDictionary<K, V>(hashKey))
{
}

//
// copy constructor
//
template <class K, class V>
mDictValPtr<K, V>::mDictValPtr(const mDictValPtr<K,V> &x)
{
    RWTPtrHashDictionaryIterator<K,V> iter(*impl);
    K *key, *keycopy;

    //
    // ok, we know that all the keys in the x.impl are our copies. Since
    // we will delete the copies when they come out, this copy of the
    // data needs its own copy of all the keys.
    //
    while ((key = iter()) != nil) {
	keycopy = new K(*key);
	impl->insertKeyAndValue(keycopy, x.impl->findValue(key));
    }
}

//
// destructor
//
template <class K, class V>
mDictValPtr<K, V>::~mDictValPtr()
{
    RWTPtrHashDictionaryIterator<K,V> iter(*impl);
    K *key = nil;
    
    // need to walk over each of the keys and delete it
    // but really can't do that easily since any modification
    // invalidates the iterator
    while ((key = iter()) != nil) {
	// impl->remove() is going to return the key so 
	// we just delete it
	delete (impl->remove(key));
	// have to reset the iterator, since we just modified
	// the collection
	iter.reset();
    }
    // blow away collection
    delete impl;
}

// operator =
template <class K, class V> mDictValPtr<K, V>&
mDictValPtr<K, V>::operator=(const mDictValPtr<K, V>& x)
{
    RWTPtrHashDictionaryIterator<K, V> iter(*impl);
    K *key, *keycopy;

    //
    // Blow away our current data.
    //
    while ((key = iter()) != nil) {
	delete (impl->remove(key));
	iter.reset();
    }
    
    // ok, we know that all the keys in the x.impl are our copies. Since
    // we will delete the copies when they come out, this copy of the
    // data needs its own copy of all the keys.
    while ((key = iter()) != nil) {
	keycopy = new K(*key);
	impl->insertKeyAndValue(keycopy, x.impl->findValue(key));
    }
    
    return *this;
}

// insert
template <class K, class V> int
mDictValPtr<K, V>::insert(const K& k, V *v)
{
    K *key, *newk;

    // check to see if its already there
    V *result = impl->findValue(&k);
    
    if (result) {
	// it is, so we need to get out the old one
	key = impl->remove(&k);
	// blow it away
	delete key;
    }
    // make a new copy of k
    newk = new K(k);
    // insert the new copy and the value
    impl->insertKeyAndValue(newk, v);
    
    // compute return result
    if (result)
	return 1;
    else
	return 0;
}

// lookup
template <class K, class V> V *
mDictValPtr<K, V>::lookup(const K& k) const
{
    return impl->findValue(&k);
}

// del... but no destructor calls
template <class K, class V> void
mDictValPtr<K, V>::del(const K& k)
{
    K *key;
    // ok here's we need to get the storage out 
    key = impl->remove(&k);
    // if key is present delete the copy
    if (key)
	delete key;
}

// del_item ... includes destructor called on V
template <class K, class V> void
mDictValPtr<K, V>::del_item(const K& k)
{
    V *v = impl->findValue(&k);
    
    if (v) {
	// we know it is there, so delete the storage that's in the dict
	delete (impl->remove(&k));
	// user wants v gone...
	delete v;
    } 
    // if k isn't in the dictionary, don't bother trying to remove it
}

template <class K, class V> void 
mDictValPtr<K,V>::walk(mDictValPtrClosure<K, V>& closure,
		       void *clientData) const
{
    RWTPtrHashDictionaryIterator<K, V> iter(*impl);
    K * current;
    
    // startup info
    closure.startup(clientData);
    // iterate until current is nil
    while ((current = iter()) != nil) {
	// if they return 0, bail out
	if (closure.action(*current, impl->findValue(current)) == 0)
	    break;
    }
    // give them the hook for ending
    closure.closedown();
}


// deletes all the Enties (not the keys) in the dictionary
// use this before calling delete on the dictionary
// use only if you are sure you know what you are doing
template <class K, class V> void
mDictValPtr<K, V>::kill_all()
{
  mDictValPtrIterator<K, V> iter(*this);
  K key;
  V *valuePtr;

  while (iter.next())
  {
    key = iter.current();
    valuePtr = this->lookup(key);
    if (valuePtr) delete valuePtr;
  }

}


///////////////////////////////////////////////////////////////////////////
// iterator
template <class K, class V> 
mDictValPtrIterator<K, V>::mDictValPtrIterator(const mDictValPtr<K, V>& dict)
{
    impl = new RWTPtrHashDictionaryIterator<K, V>(*(dict.impl));
}

// destructor
template <class K, class V>
mDictValPtrIterator<K, V>::~mDictValPtrIterator()
{
    delete impl;
}

// move up to the next value
template <class K, class V> int
mDictValPtrIterator<K, V>::next()
{
    K *key;
    
    if ((key = ((*impl)())) == nil)
	return 0;
    
    return 1;
}

// get the  current value
template <class K, class V> K
mDictValPtrIterator<K, V>::current() const
{
#ifdef BEFORE_KEITH
    K *copy;
    K *current=impl->key();

    // need to make a copy so they don't end up deleting ours
    copy=new K(*current);
    return *copy;
#endif

    return *impl->key();
}

// get the current size
template <class K, class V> unsigned
mDictValPtr<K, V>::size() const 
{
    return (impl->entries());
}

