// This may look like C code, but it is really -*- C++ -*-
//
//
//
//	Note:  The use of this program shall be restricted to educational,
//	non-commercial use only.  All inquiries regarding licensing this
//	software for commercial use may be addressed to:
//	
//	Georgia Tech Research Corporation
//	Technology Licensing
//	Centennial Research Building - Rm. 275
//	400 Tenth Street, N.W.
//	Atlanta, Georgia  30332-0415
//	
//	*************************************************
//	
//	       Copyright 1996
//	Georgia Tech Research Corporation
//	    All Rights Reserved
//
//
//
//
#ifndef MERCATOR_LIST_H
#define MERCATOR_LIST_H

// $Author: spine $
// $Source: /net/hm10/keith/proj/mercator/6.0/RCS/mList.h,v $
// $Revision: 1.4 $
// $Date: 1995/07/28 21:23:35 $

// $Log: mList.h,v $
//Revision 1.4  1995/07/28  21:23:35  spine
//Added copyright notice.
//
//Revision 1.3  1995/07/24  17:44:26  beth
//Ian's Version
//
//Revision 1.1  1995/05/25  14:48:43  iansmith
//Initial revision
//

// We implement the mList with a Rogue Wave List right now
#include <rw/tpdlist.h>

//
// This is the closure object that you pass to the iterate function.
// Subclass the action function to implement your behavior;
// if you want startup and closedown help you can implement those
// as well.
//
template <class T> class mListClosure {
public:
    // use this to setup your closure
    virtual void startup(void *clientData) {};
    // you'll still get a closedown if you short circuit
    //  out of the action routine
    virtual void closedown() {};
    // if the action routine returns 0 the closure stops
    // processing
    // if you don't subclass, we end up only calling the action routine once
    virtual int action(T *) {return 0;};
};

// 
// Linked List for Mercator
// You can ONLY put pointers in this data structure!
//
// You MUST define the following:
// operator==
// operator<
//
template <class T> class mList {
    // The implementation
    RWTPtrDlist<T> *impl;
    // friend classes are really bogus, but since they
    //both must know about the implementation using
    // Rogue Wave... sigh...
    friend class mListIterator<T>;
public:
    // add to the end
    void append(const T *v);
    // search for a value... returns 0 if not found
    T * search(const T* v) const;
    // remove an item and call its destructor
    // This is noop if its not present
    void destroy_item(T *v);
    // remove the item but return a pointer to it...
    // comparison is with the == operator on T
    T* del_item(const T *v);
    // if you want to iterate, subclass the mListClosure and
    // pass it in here... the client data is passed to the startup routine
    void walk(mListClosure<T> *closure,void *clientData) const;
    // constructor (empty list)
    mList();
    // copy constructor.. this will result in *shallow* copy with pointers
    // shared between lists
    mList(const mList &);
    // destructor... this just deletes the data structure, doesn't touch the objects
    virtual ~mList();
    // assignment is a shallow copy of the lists... you get two differents lists
    // but the nodes in them are the same
    mList<T> & operator=(const mList<T> &);
    // sometimes you want information about ordering...
    // get the head of the list
    T * head() const;
    // get the tail of the list
    T * tail() const;
    // get the entry right before or after... 
    // return 0 if they don't have the entry asked for or
    // succ/pred would be off the beginning/start of the list
    // both of these test for equality using op == on T
    T * predecessor(const T *) const;
    T * successor(const T *) const;
    // how big is the list?
    unsigned entries() const;
    // is it empty?
    int empty() { return (entries==0);};
};

// At first the iterator is NOT pointing to anything valid,
// so you must perform some operation on it to make it set
// itself to the first element.  You should not  remove elements
// from the list during iteration.  
template <class T> class mListIterator {
    // this is the implementation using rogue wave
    RWTPtrDlistIterator<T> *impl;
public:
    // construct this from an mList
    mListIterator(const mList<T> &);
    // use next to both setup the iterator the first time and
    // to iterate during use. Returns 0 when the end is reached
    // and is in an undefined state at that point
    T * next();
    // reset the iterator to the undefined state it had at creation time
    void reset();
    // destructor
    virtual ~mListIterator();
    // copy constructor
    mListIterator(const mListIterator &);
    // operator = 
    mListIterator<T> & operator=(const mListIterator<T> &);
    
};

#ifdef BEAM
#include "mList.cc"
#endif

#endif /* MERCATOR_LIST_H */
