##
##
##
##	Note:  The use of this program shall be restricted to educational,
##	non-commercial use only.  All inquiries regarding licensing this
##	software for commercial use may be addressed to:
##	
##	Georgia Tech Research Corporation
##	Technology Licensing
##	Centennial Research Building - Rm. 275
##	400 Tenth Street, N.W.
##	Atlanta, Georgia  30332-0415
##	
##	*************************************************
##	
##	       Copyright 1996
##	Georgia Tech Research Corporation
##	    All Rights Reserved
##
##
##
##
## mercator.tcl contains most of the "top level" interface functions used
## by Mercator.  It sources any other tcl files required, and installs a
## bunch of actions.
##


############################################################################
###									 ###
###	Initial setup stuff.						 ###
###									 ###
############################################################################

Error EL_DEBUG mercator Sourcing mercator.tcl

#
# user level can be Novice, Intermediate and Expert
#
set userLevel "Intermediate"

#
# These tcl files are sourced automatically when mercator is started.
# They contains all the global rules and generic functions.
# 
source "text.tcl"
source "navigate.tcl"
source "audio.tcl"
source "templateprocs.tcl"

###
### Only source log.tcl if you need to do user performance timing.
###
### source "log.tcl"
###

#
# Put up an introductory message.
#
speak "Welcome to the Mercator Application Manager."
braille display "Welcome to the Mercator Application Manager."

############################################################################
###									 ###
###	Definition of "helper" functions.  These are routines that are	 ###
###	not visible to the C++ world via addaction, and they are not	 ###
###	invoked in response to user input via bindkey/bindevent.	 ###
###									 ###
###	These helpers are primarily (1) used by other functions in	 ###
###	this file, (2) useful on the Mercator command line for		 ###
###	debugging, or (3) meant to be invoked from the console to	 ###
###	effect some change in the state of the system (the user level	 ###
###	routines below for example).					 ###
###									 ###
############################################################################

#
# Select generates a button press/release pair on the specified node, if it
# is sensitive.  Not an action.
#
proc Select {node} {
    if { [resource $node get sensitive] == "TRUE"} {
        SelectAudio $node
        button $node press 1
        button $node release 1
        Error EL_DEBUG mercator "Select called"
    }
}

#
# Return the number of children.  Not an action.
#
proc numChildren {node} {
    set count 0
    set children [modelManager children $node]
    if {$children == "NULL" || $children == "" } {
        return $count 
    } else {
        foreach child $children {
            set count [expr $count + 1]
        }
        return $count
    } 
}

#
# Raise the current object.  Not an action.
#
proc raise args {
    xwin raise [currentObject]
}


############################################################################
#
# These routines set the user level.  They are intended to be invoked from
# the console.
#
proc Novice args {
    global userLevel
    set userLevel "Novice"
    speak "Novice user mode"
}

#
#
# 
proc Intermediate args {
    global userLevel
    set userLevel "Intermediate"
    speak "Intermediate user mode"
}

#
#
# 
proc Expert args {
    global userLevel
    set userLevel "Expert"
    speak "Expert user mode"
}

############################################################################
#
# These are some helper functions for debugging - they print out a complete
# dump of the widget hierarchy to a file.  Usage:  go anywhere in a tree you
# want to dump, and say "widgetTree filename"
#
proc walkIt {obj indent fileId} {

    # use indentation to indicate parent-child hierarchy
    set ind $indent
    set spaces {}

    while {$ind > 0} {
	lappend spaces { }
	set ind [expr $ind - 1]
    }
    puts $fileId "[join $spaces {}][modelManager class $obj]:[modelManager name $obj]:[modelManager managed $obj]:[modelManager mapped $obj]:[resource $obj get x],[resource $obj get y]"

    if {[modelManager children $obj] != {}} {
	foreach child [modelManager children $obj] {
	    walkIt $child [expr $indent+3] $fileId
	}
    }
}

proc widgetTree {outfile} {

    set topWidget [currentObject]

    # first get to the top
    while {[modelManager parent $topWidget] != "NULL"} {
	set topWidget [modelManager parent $topWidget]
    }
    
    # go print out the tree, starting with an indent of 0
    set fileId [open $outfile a]
    walkIt $topWidget 0 $fileId
    close $fileId
}

############################################################################
###									 ###
###	Definition of "action" procedures.  Actions are TCL functions	 ###
###	that are callable from the C++ side of Mercator via the		 ###
###	Action class interface.  Any routine that is an action must	 ###
###	be made visible to C++ by calling addaction.			 ###
###									 ###
###	Actions are typically routines that perform interface work	 ###
###	whenever the underlying model changes.  Mercator calls out to	 ###
###	actions at particular points whenever "interesting" events	 ###
###	occur in the model.						 ###
###									 ###
############################################################################

#
# Called whenever the current client changes.  This function is an action.
#
proc ClientChange {old new location} {
    global textMode
    set textMode "FALSE"

    speak "$new is the current application"
}

#
# PopupReturn is installed as an action, and is responsible for invoking
# the PopDownAudio routine in audio.tcl.  This function is an action.
#
proc PopupReturn { } {
    PopDownAudio
}

#
# Halt speech synthesis (currently this is hacky - we generate an empty
# speech request, halting whatever's running).  This routine is an action.
#
proc StopSpeaking args {
    speak ""
}

#
# SelCurrent calls Select (above) with the current object as the argument.
# This routine is an action.
#
proc SelCurrent args {
    set current [currentObject]
    Select $current
}

#
# Special behavior for top-level shells being mapped.  This is an action.
#
proc ShellMapped {obj client} {
    PopUpAudio

    Error EL_DEBUG mercator "ShellMapped In ShellMapped"
    Error EL_DEBUG mercator "ShellMapped object $obj, client $client"
}

#
# This action gets called whenever a new client instance is created, but
# before the client is "ready."
#
proc ClientCreated {name uniqueName} {
    ApplicationStartUpAudio
    speak "A new client has been created, but is not yet ready."
}

#
# This action is called when the client is ready for use.
#
proc ClientReady {uniqueName} {
   Error EL_DEBUG mercator "ClientReady $uniqueName"
   set name [client name $uniqueName]
   speak "New client $name is now ready for use."
}

#
# This action is invoked whenever a client goes away.  We play the shutdown
# sound and reset text mode to FALSE.
#
proc ClientShutdown {uniqueName} {
    global textMode

    set textMode "FALSE"
    ApplicationShutdownAudio
}

############################################################################
###									 ###
### 	User input handlers.  These are functions that are bound to	 ###
###	events that users may generate.  Whereas actions produce the	 ###
###	Mercator-to-user interface, input handlers are responsible for	 ###
###	the user-to-Mercator interface.					 ###
###									 ###
###	Input handlers are associated with events such as keypresses,	 ###
###	and do things like perform selection and navigation over the	 ###
###	model.								 ###
###									 ###
############################################################################
#
# Returns the X, Y coordination locaiton of the current object.
# ??? WHY IS THIS TEXT POSITION?
# Bound to user input.
#
proc TextPosition args {
    set current [currentObject]
    Error EL_DEBUG mercator "TextPosition x = " \
	[resource $current get x] " y = " [resource $current get y]
}

#
# Perform an auditory preview of the current object's children.
# Bound to user input via ReaderDispatch.
#
proc Preview args {
    set current [currentObject]
    set children [modelManager children $current]
    if { $children == "NULL" || $children == "" } {
	speak "No children found"
    } else {
	set x 0
	foreach child $children {
	    if { [wtemplate $child navigable] == "TRUE"} {
		Audio $child 150
		set x [expr $x + 1]
	    }
	}
	if {$x == 0} { # no navigable children found
	    speak "No navigable children found"
	}
    }
}

#
# Perform the "info" operation on the current object.
# Bound to user input via ReaderDispatch.
#
proc Info args {
    set current [currentObject]
    Audio $current 0

    speak [wtemplate $current speakOnInfo] 
    speak [wtemplate $current speakOnEnter] nointerupt

    if {[modelManager parent $current] == "NULL" } {
	speak [modelManager class $current]
    }

    Error EL_INFORMATIONAL mercator "Info The current object is "\
	[modelManager longname $current] $current
}

#
# Play out the sounds of all ancester widgets from top to the current object
#
proc HearPath args {
    
    set current [currentObject]
    set l {}

    # build a list of ancester widgets
    while {$current != "NULL"} {
	set l [concat $current $l]
	set current [modelManager parent $current]
    }

    foreach item $l {
	Audio $item 150
    }
}

############################################################################
# 
# These routines are bound to the basic navigation keys.  They determine
# whether text mode is active or not, and call the appropriate functions
# to either navigate objects or text.
#
proc UpPressed args {
    global textMode
 
    if {$textMode == "FALSE"} {
        NavUp
    } else {
        TextUp
    }
}

proc DownPressed args {
    global textMode
 
    if {$textMode == "FALSE"} {
      NavDown
    } else {
      TextDown
    }
}

proc RightPressed args {
    global textMode
 
    if {$textMode == "FALSE"} {
      NavRight
    } else {
      TextRight
    }
}

proc LeftPressed args {
    global textMode
 
    if {$textMode == "FALSE"} {
      NavLeft
    } else {
      TextLeft
    }
}

############################################################################
#
# Inter-client navigation keybindings.
#
proc ClientNavRightContext args {
    SwitchAppAudio
    NavAppRight

    puts "***GOING RIGHT APP***"
}

#
#
#
proc ClientNavLeftContext args {
    SwitchAppAudio
    NavAppLeft

    puts "***GOING LEFT APP***"
}

proc ClientNavRight args {
    SwitchAppAudio
    NavAppRight

    speak "Context-free app navigation is not supported, doing a with-context navigation."
}

proc ClientNavLeft args {
    SwitchAppAudio
    NavAppRight

    speak "Context-free app navigation is not supported, doing a with-context navigation."
}


############################################################################
#
# Text-mode keybindings.
#
proc OnePressed args { ReaderDispatch 1 {} }
proc ShiftOnePressed args { ReaderDispatch 1 {shift} }
proc ControlOnePressed args { ReaderDispatch 1 {control} }
proc MetaOnePressed args { ReaderDispatch 1 {meta} }
proc ShiftMetaOnePressed args { ReaderDispatch 1 {shift meta} }
proc ControlMetaOnePressed args { ReaderDispatch 1 {control meta} }

proc ThreePressed args { ReaderDispatch 3 {} }
proc ShiftThreePressed args { ReaderDispatch 3 {shift} }
proc ControlThreePressed args { ReaderDispatch 3 {control} }
proc MetaThreePressed args { ReaderDispatch 3 {meta} }
proc ShiftMetaThreePressed args { ReaderDispatch 3 {shift meta} }
proc ControlMetaThreePressed args { ReaderDispatch 3 {control meta} }

proc FivePressed args { ReaderDispatch 5 {} }
proc ShiftFivePressed args { ReaderDispatch 5 {shift} }
proc AltFivePressed args { ReaderDispatch 5 {alt} }
proc ControlFivePressed args { ReaderDispatch 5 {control} }
proc MetaFivePressed args { ReaderDispatch 5 {meta} }
proc ShiftMetaFivePressed args { ReaderDispatch 5 {shift meta} }
proc ControlMetaFivePressed args { ReaderDispatch 5 {control meta} }

proc SevenPressed args { ReaderDispatch 7 {} }
proc ShiftSevenPressed args { ReaderDispatch 7 {shift} }
proc ControlSevenPressed args { ReaderDispatch 7 {control} }
proc MetaSevenPressed args { ReaderDispatch 7 {meta} }
proc ShiftMetaSevenPressed args { ReaderDispatch 7 {shift meta} }
proc ControlMetaSevenPressed args { ReaderDispatch 7 {control meta} }

proc NinePressed args { ReaderDispatch 9 {} }
proc ShiftNinePressed args { ReaderDispatch 9 {shift} }
proc ControlNinePressed args { ReaderDispatch 9 {control} }
proc MetaNinePressed args { ReaderDispatch 9 {meta} }
proc ShiftMetaNinePressed args { ReaderDispatch 9 {shift meta} }
proc ControlMetaNinePressed args { ReaderDispatch 9 {control meta} }

proc ReaderDispatch {key mods} {
    global textMode

    set current [currentObject]
    set key [expr {$key | (([lsearch $mods meta] == -1 ? 0 : 1) << 4)}]
    set key [expr {$key | (([lsearch $mods control] == -1 ? 0 : 1) << 5)}]
    set key [expr {$key | (([lsearch $mods shift] == -1 ? 0 : 1) << 6)}]
    set key [expr {$key | (([lsearch $mods alt] == -1 ? 0 : 1) << 7)}]

    if {$textMode == "FALSE"} {
	switch [expr ($key & 0x0F)] {
	    1 { switch [expr ($key & 0xF0)] {
		0 { JumpBack }
	    }}
	    3 { switch [expr ($key & 0xF0)] {
		0 { JumpForward }
	    }}
	    5 { switch [expr ($key & 0xF0)] {
		0 { Info }
		64 { Preview }
		128 { HearPath }
	    }}
	}
    } else {
	set mode [sreader $current cursor mode]
	set dir [expr {($key & 0x40) ? 
		       (($key & 0x20) ? "Error" : "Next") :
		       (($key & 0x20) ? "Prev" : "This") }]
	switch [expr {($key & 0x0F)}] {
	    1 {set unit Char}
	    3 {set unit Word}
	    5 {set unit Line}
	    7 {set unit Sentence}
	    9 {set unit Paragraph}
	}
	set op [expr {($dir != "This" &&
		       (($mode == "EDIT" && ($key & 0x10)) ||
			($mode == "REVIEW" && !($key & 0x10)))) ?
		      "Move" : "Read" }]
	if {$dir != "Error"} {
	    set action [join [list $op $dir $unit] {}]
	    Error EL_DEBUG "ReaderDispatch executing $action"
	    [expr {$action}]
	}
    }
}

#
#
#
proc SlashPressed args {
    global textMode

    if {$textMode == "TRUE"} {
      AnnounceCursorPosition
    }
}

#
#
#
proc ShiftandSlashPressed args {
    global textMode

    if {$textMode == "TRUE"} {
        ToggleCursorMode
    }
}

#
#
#
proc ControlandSlashPressed args {
    global textMode

    if {$textMode == "TRUE"} {
        ToggleFollowMode
    }
}

#
#
#
proc MetaandSlashPressed args {
    global textMode

    if {$textMode == "TRUE"} {
	Error EL_INFORMATIONAL mercator "Synchronizing edit cursor to application cursor"
        sreader [currentObject] cursor app
    }
}
 
#
#
#
proc PlusPressed args {
    global textMode filterMode

    Error EL_DEBUG PlusPressed $filterMode
    if {$textMode == "TRUE"} {
	FilterModeOn
    } else {
	NotImplemented
    }
}

#
#
#
proc MinusPressed args {
    global textMode filterMode
    
    Error EL_DEBUG MinusPressed $filterMode
    if {$textMode == "TRUE"} {
	FilterModeOff
    }
}

#
#
#
proc AsteriskPressed args {
    global textMode

    if {$textMode == "TRUE"} {
        AnnounceFilterMode
    }
}

#
#
#
proc ShiftandAsteriskPressed args {
    global textMode filterMode

    if {$textMode == "TRUE"} {
        ToggleFilterMode
    }
}

############################################################################
#
# This is a bindevent proc used for unimplemented features
#
proc NotImplemented args {
    speak "Sorry, this feature is not yet implemented."
}

############################################################################
#
# Install our actions into the model manager space
#
addaction ClientCreation
addaction ClientChange
addaction PopupReturn
addaction StopSpeaking
addaction SelCurrent
addaction ChangeTextMode

#
# These are actions so that Genovations can call to them directly.
# Genovations should really call the higher-level routines instead.
#
addaction ReadThisChar
addaction ReadThisWord
addaction ReadThisSentence
addaction ReadThisLine
addaction ReadThisParagraph
addaction ReadPrevChar
addaction ReadPrevWord
addaction ReadPrevSentence
addaction ReadPrevLine
addaction ReadPrevParagraph
addaction ReadNextChar
addaction ReadNextWord
addaction ReadNextSentence
addaction ReadNextLine
addaction ReadNextParagraph
addaction MoveNextChar
addaction MoveNextWord
addaction MoveNextSentence
addaction MoveNextLine
addaction MoveNextParagraph

#
# These actions are used by the Braille objects to call out to TCL code.
#
addaction BrailleProg
addaction BrailleHome
addaction BrailleCursor
addaction BrailleUp
addaction BrailleUpBounds
addaction BrailleLeft
addaction BrailleLeftBounds
addaction BrailleLeftJumpScroll
addaction BrailleRight
addaction BrailleRightBounds
addaction BrailleRightJumpScroll
addaction BrailleDown
addaction BrailleDownBounds

#
# I put in the addaction for GoTo *just* so that the popup/popdown
# code (in XtHook::HandleStateChangeNotify) can call it to update the
# focus/current position.  THIS IS A HACK.  We should call out to a
# new proc, called Popup/Popdown or something, and have it call GoTo.
# Take this out as soon as its fixed.			-k
#
addaction GoTo

addaction ShellMapped
addaction ClientCreated
addaction ClientReady
addaction ClientShutdown

############################################################################
#
# make the global keybindings
#
# note that we use both keypad and non-keypad keysyms...  -keith
#
bindevent press 			Up 	UpPressed
bindevent press 			KP_8 	UpPressed

bindevent press 			Down 	DownPressed
bindevent press 			KP_2 	DownPressed

bindevent press 			Right 	RightPressed
bindevent press 			KP_6 	RightPressed

bindevent press 			Left 	LeftPressed
bindevent press 			KP_4 	LeftPressed

bindevent press				R13	OnePressed
bindevent press	shift			R13	ShiftOnePressed
bindevent press		control		R13	ControlOnePressed

bindevent press			meta	R13	MetaThreePressed
bindevent press	shift		meta	R13	ShiftMetaThreePressed
bindevent press		control	meta	R13	ControlMetaThreePressed

bindevent press				F35	ThreePressed
bindevent press	shift			F35	ShiftThreePressed
bindevent press		control		F35	ControlThreePressed
bindevent press			meta	F35	MetaThreePressed
bindevent press	shift		meta	F35	ShiftMetaThreePressed
bindevent press		control	meta	F35	ControlMetaThreePressed

bindevent press				F31	FivePressed
bindevent press	shift			F31	ShiftFivePressed
bindevent press		control		F31	ControlFivePressed
bindevent press			meta	F31	MetaFivePressed
bindevent press	shift		meta	F31	ShiftMetaFivePressed
bindevent press		control	meta	F31	ControlMetaFivePressed
bindevent press			alt	F31	AltFivePressed

bindevent press				F27	SevenPressed
bindevent press	shift			F27	ShiftSevenPressed
bindevent press		control		F27	ControlSevenPressed
bindevent press			meta	F27	MetaSevenPressed
bindevent press	shift		meta	F27	ShiftMetaSevenPressed
bindevent press		control	meta	F27	ControlMetaSevenPressed

bindevent press				F29	NinePressed
bindevent press	shift			F29	ShiftNinePressed
bindevent press		control		F29	ControlNinePressed
bindevent press			meta	F29	MetaNinePressed
bindevent press	shift		meta	F29	ShiftMetaNinePressed
bindevent press		control	meta	F29	ControlMetaNinePressed

# Keypad slash: "/"
bindevent press 			F25 	SlashPressed       
bindevent press shift 			F25 	ShiftandSlashPressed
bindevent press 	control 	F25 	ControlandSlashPressed;
bindevent press			meta	F25	MetaandSlashPressed;

# Keypad asterisk: "*"
bindevent press 			F26 	AsteriskPressed
bindevent press shift 			F26 	ShiftandAsteriskPressed 

bindevent press 			KP_Insert StopSpeaking
bindevent press 			Delete 	ChangeTextMode
bindevent press 			KP_Enter SelCurrent

bindevent press 	control 	s 	StopSpeaking
bindevent press 	control 	t 	ChangeTextMode

bindevent press shift 			Right 	ClientNavRightContext
bindevent press shift 			KP_6 	ClientNavRightContext
bindevent press shift 			Left 	ClientNavLeftContext
bindevent press shift 			KP_4 	ClientNavLeftContext

bindevent press alt 			Right 	ClientNavRight
bindevent press alt 			KP_6 	ClientNavRight
bindevent press alt 			Left 	ClientNavLeft
bindevent press alt 			KP_4 	ClientNavLeft

bindevent press 			KP_Add 	PlusPressed
bindevent press 			KP_Subtract MinusPressed
bindevent press 			F24 	MinusPressed
 
bindevent press 			F1   	ReadThisChar
bindevent press 			F2   	ReadThisWord
bindevent press 			F3   	ReadThisLine
bindevent press 			F4   	ReadThisSentence
bindevent press 			F5   	ReadThisParagraph

bindevent press 			F6   	ReadNextChar
bindevent press 			F7   	ReadNextWord
bindevent press 			F8   	ReadNextLine
bindevent press 			F9   	ReadNextSentence
bindevent press 			F10  	TextPosition

